
#ifndef GAME_HXX_INCLUDED
#define GAME_HXX_INCLUDED 

/************************************************************************

    JSWED 2.3.0 - Editor for Jet Set Willy and derivatives

    Copyright (C) 2000-2001, 2004  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*************************************************************************/

#include "snapio.hxx"
#include "jnotebook.hxx"

class HexEditPage;
class Room;
class JswRoomList;

struct RefCount
{
	int id;
	int count;
	RefCount *next;
};

struct ImportedSprite
{
	int spriteID;
	jswByte bits[32];
};


#define CELL(x)     (0x43000000 | (x))
#define GUARDIAN(x) (0x47000000 | (x))
#define SPRITE(x)   (0x53000000 | (x))

class Game : public KeyListener
{
public:
	Game(SpectrumMemory *mem); // v2.3.0: Now public, because a generic
				   // Game can be edited.
	virtual ~Game();
	virtual char *getID();	// 40-character ID string
	inline void setFilename(char *s) { strcpy(m_filename, s); }
	inline char *getFilename()       { return m_filename; }
	virtual int editMain();

	virtual int decompress();
	virtual int compress();

	virtual void addPages(VideoNotebook *n);
	virtual void deletePages(void);
	inline SpectrumMemory *getMem() { return m_mem; }
	virtual int afterLoadFix(void);
protected:
	int onSave(void);
	int onTest(void);
	
	KeyListener *m_keyFilter;
	unsigned *m_spriteMap;
	int m_spriteMapLen;
public:
	// Overrides
	virtual int onKeyDown(int keysym);
	virtual int onKeyUp(int keysym);
	virtual jswByte *getFont(int *len, int *firstc = NULL);

	inline void setDirty()   { m_mem->setDirty(); }
	inline void clearDirty() { m_mem->clearDirty(); }
	inline bool isDirty()    { return m_mem->isDirty(); }

	Game *m_upgrade;

	virtual int addSprite(int fileid, int w, int h, 
			jswByte *bits);
	void clearSpriteMap();
	void addSpriteRef(int fileid, int memid);
	void clearRefCounts();
	void addRef(int id);
	void unRef(int id);
	int lookupSpriteRef(int fileid);
	int getRefCount(int id);
	void dumpRefs();

	void exportReferencedSprites(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent, int w, int h);

	virtual bool importSprites(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent);
	virtual Room *getRoomClass(int room);
	virtual JswRoomList *getRoomList(void);
	virtual void addFixedRefs();
	virtual bool importSprites(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent, int *roomMap);
	virtual jswByte *getSprite(int id);
	RefCount *m_rc;
	inline void setCurrentRoom(int room) { m_room = room; }
	inline int getCurrentRoom() { return m_room; }
protected:
	HexEditPage *m_hexPage;

	SpectrumMemory *m_mem;
	char m_filename[PATH_MAX];
	char m_namebuf[(VIDEO_W / CHAR_W) + 1];
	int m_room;
};

int saveFile(Game **gp);
int saveFile(char *filename, Game **gp);
int loadFile(Game **gp);
int loadFile(char *filename, Game **gp);

#endif // ndef GAME_HXX_INCLUDED
