/************************************************************************

    JSWED 2.3.0 - Editor for Jet Set Willy and derivatives

    Copyright (C) 2000-2001,2004,2005  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*************************************************************************/

#include "jswed.hxx"
#include "room.hxx"
#include "j48room.hxx"
#include "jswgame.hxx"

#define CELL_AIR 	0
#define CELL_EARTH	1
#define CELL_WATER	2
#define CELL_FIRE	3
#define CELL_RAMP	4
#define CELL_CONVEYOR	5
#define CELL_ITEM	6

Jsw48Room::Jsw48Room(JswGame *g, jswByte *data, int roomNumber) : 
			Room(6, roomNumber), m_data(data), m_game(g) 
{

}

/* Note: Can contain embedded NUL characters - see JSW:LOTR */
int Jsw48Room::getTitle(char *buf)
{
	memcpy(buf, m_data + 0x80, 32);
	buf[32] = 0;
        for (int n = 0; n < 32; n++)
        {
                if (buf[n] == ZXCH_COPY ) buf[n] = ZXK_COPY;
                if (buf[n] == ZXCH_POUND) buf[n] = ZXK_POUND;
        }
	return 32;
}

void Jsw48Room::setTitle(const char *buf)
{
	char fmt[33];

	sprintf(fmt, "%-32.32s", buf);
	memcpy(m_data + 0x80, fmt, 32);
}


jswByte *Jsw48Room::getCellPattern(int type)
{
	switch(type)
	{
		case CELL_AIR:		return m_data + 0xA0; 
		case CELL_EARTH:	return m_data + 0xA9; 
		case CELL_WATER:	return m_data + 0xB2; 
		case CELL_FIRE:		return m_data + 0xBB; 
		case CELL_RAMP:		return m_data + 0xC4; 
		case CELL_CONVEYOR:	return m_data + 0xCD; 
		case CELL_ITEM:
			m_itemPattern[0] = ZX_WHITE;
			memcpy(m_itemPattern + 1, m_data + 0xE1, 8);
			return m_itemPattern;
	}
	return NULL;
}


int Jsw48Room::getCellAt(int x, int y)
{
	if (y < 0 || y > 16 || x < 0 || x > 31) return CELL_AIR;

        jswByte cell = m_data[y * 8 + (x / 4)];

        cell = (cell << ((x % 4) * 2)) & 0xC0;
	return (cell >> 6);
}

int Jsw48Room::getDrawnCellAt(int cx, int cy)
{
	if (cy < 0 || cy > 16 || cx < 0 || cx > 31) return CELL_AIR;
	int cell = getCellAt(cx,cy);

// See if this is on the conveyor route
	int len;
	int w = m_data[0xD7] + 256 * m_data[0xD8];
	int x = (w & 0x1F);
	int y = (w >> 5) & 0x0F;
	int n, step;

	if (m_data[0xD9] > 32) m_data[0xD9] = 0;	// Damaged room
	len = m_data[0xD9];
	if (y == cy && x <= cx && (x + len > cx)) cell = CELL_CONVEYOR;
		
        w = m_data[0xDB] + (256*m_data[0xDC]);
        x = (w & 0x1F);
        y = ((w >> 1) & 0xF0);

        if (m_data[0xDD] > 32) /* Damaged room */
	{
		fprintf(stderr, "Room %d invalid ramp, dropped\n", m_roomNumber);
		m_data[0xDD]=0;
	}

        len = m_data[0xDD];
        if (m_data[0xDA] != 1) len = (-len);
        if (len < 0) { len = (-len); step = (-1); } else step = 1;

	y = (y >> 4);

	for (n = 0; n < len; n++)
	{
		if (y == cy && x == cx) cell = CELL_RAMP;
		x += step; --y;
	}
	return cell;
}


int Jsw48Room::getConveyorDir()
{
	return (m_data[0xD6] & 1);
}

int Jsw48Room::getConveyorY()
{
	return 0;
}

void Jsw48Room::drawOneOffs(VideoBitmap *bm, bool flashonly, bool draw)
{
	drawConveyor(bm, flashonly, draw);
	drawRamp(bm, flashonly, draw);
	Room::drawOneOffs(bm, flashonly, draw);
}

void Jsw48Room::drawConveyor(VideoBitmap *bm, bool flashonly, bool draw)
{
	int len;
	int w = m_data[0xD7] + 256 * m_data[0xD8];
	int x = (w & 0x1F);
	int y = (w >> 5) & 0x0F;
	int n;

/* Draw the conveyor */
	jswByte *pattern = getCellPattern(CELL_CONVEYOR);

	if (m_data[0xD9] > 32) m_data[0xD9] = 0;	// Damaged room
	len = m_data[0xD9];

	for (n = 0; n < len; n++)
	{
		if (y >= 0 && y < 16 && x >= 0 && x < 32)
		{
	        	if (!flashonly || (pattern[0] & 0x80))
		        {
				if (draw) m_cellBitmaps[CELL_CONVEYOR * 2 + 1]->toSurface(bm, x * 16, y * 16);
				else drawCell(bm, x, y, flashonly, false);
			}
			x++;
		}

	}	
}


void Jsw48Room::drawRamp(VideoBitmap *bm, bool flashonly, bool draw)
{
        int len;
	int n, step;
        int w = m_data[0xDB] + (256*m_data[0xDC]);
        int x = (w & 0x1F);
        int y = ((w >> 1) & 0xF0);

        if (m_data[0xDD] > 32) /* Damaged room */
	{
		fprintf(stderr, "Room %d invalid ramp, dropped\n", m_roomNumber);
		m_data[0xDD]=0;
	}

        len = m_data[0xDD];
        if (m_data[0xDA] != 1) len = (-len);
        if (len < 0) { len = (-len); step = (-1); } else step = 1;

	y = (y >> 4);

	for (n = 0; n < len; n++)
	{
		if (y >= 0 && y < 16 && x >= 0 && x < 32)
	        {
			if (draw) m_cellBitmaps[CELL_RAMP * 2]->toSurface(bm, x * 16, y * 16);
			else drawCell(bm, x, y, flashonly, false);
		}
		x += step; --y;
	}	
}



void Jsw48Room::setCell(int x, int y, int type)
{
	if (x < 0 || x > 31 || y < 0 || y > 15) return;
	type &= 3;
	int dx = (x % 4) * 2;
	jswByte andmask = 0x03;
	jswByte ormask  = type;
	jswByte *cell = m_data + (y * 8 + (x / 4));

	andmask = andmask << (6-dx);
	ormask  = ormask <<  (6-dx);

	*cell &= ~andmask;
	*cell |= ormask;
}


int Jsw48Room::getBackground()
{
	return (m_data[0xA0] >> 3) & 0x0F;
}


int Jsw48Room::getRestartX()
{
	if (!m_game->imcMode()) return -1;
	return m_data[0xEE] & 0x1F;
}       


int Jsw48Room::getRestartY()
{
	if (!m_game->imcMode()) return -1;
	return m_data[0xED];
}


void Jsw48Room::setRestartX(int x)
{       
	if (m_game->imcMode() && x >= 0 && x <= 31)
	{
		m_data[0xEE] = (m_data[0xEE] & 0xE0) | x;
	}
}       


void Jsw48Room::setRestartY(int y)
{
	if (m_game->imcMode() && y >= 0)
	{
		m_data[0xED] = y & 0xFE;
	}
}

int Jsw48Room::getItemCoords(int *px, int *py, int *counts)
{
	int count = 0;
	int n, m, x, y, found;
	int rm;

	for (n = m_game->getItemCount() ; n <= 0xFF; n++)
	{
		m_game->itemPos(n, &rm, &px[count], &py[count]);
		if (rm == m_roomNumber) 
		{
			y = (py[count] / 8) & 0x0F;
			x =  px[count] & 0x1F;
			found = 0;
			for (m = 0; m < count; m++)
			{
				if (px[m] == x && py[m] == y)
				{
					++counts[m];
					found = 1;
					break;
				}
			}
			if (!found)
			{
				py[count] = y;
				px[count] = x;
				counts[count] = 1;
				++count;
			}
		}
	}
	return count;
}


jswByte *Jsw48Room::getGuardianBuffer()
{
	int bg = getBackground();
	int msk = m_game->getGuardianMask();
	jswByte *b = m_data + 0xf0;
	jswByte *g = m_game->getGuards(m_roomNumber);

	memset(m_guardianBuffer, 0xFF, sizeof(m_guardianBuffer));
	for (int n = 0; n < 8; n++)
	{
		int ng = b[0] & msk;
		if (ng == msk) break;
		memcpy(m_guardianBuffer + 8 * n, g + 8 * ng, 8);
		m_guardianBuffer[8 * n + 2] = b[1];
		b += 2;
	}
	return m_guardianBuffer;
}


int Jsw48Room::getCellBehaviour(int idx)
{
	switch(idx)
	{
		case 0: return CB_AIR;
		case 1: return CB_WATER;
		case 2: return CB_EARTH;
		case 3: return CB_FIRE;
		case 4: if (m_data[0xDA] & 1) return CB_RRAMP;
			return CB_LRAMP;
		case 5: if (getConveyorDir()) return CB_RCONV;
			return CB_LCONV;
	}
	return CB_WATER;
}

int Jsw48Room::getCellTypeCount(bool includeSingles) 
{
	return includeSingles ? 6 : 4;
}

int Jsw48Room::getRoomLength() const
{
	return 256;
}


void Jsw48Room::getCells(Cell *c)
{
	int n;
	jswByte *b = m_data + 0xA0;
	for (n = 0; n < 6; n++)
	{
		memcpy(c[n].pattern, b, 9);
		b += 9;
		c[n].behaviour = getCellBehaviour(n);
	}
}


void Jsw48Room::setCells(Cell *c)
{
	int n;
	jswByte *b = m_data + 0xA0;
	for (n = 0; n < 6; n++)
	{
		memcpy(b, c[n].pattern, 9);
		b += 9;
	}
}



void Jsw48Room::clear()
{
	memset(m_data,        0, 160);
	memset(m_data + 0xF0, 0,  16);
	m_data[0xF0] = 0xFF;
	setTitle("");
}


static jswByte flip4(jswByte b)
{
	return ((b &  3) << 6) | ((b &  12) << 2) |
		((b & 48) >> 2) | ((b & 192) >> 6);
}


void Jsw48Room::mirrorH()
{
	int x, y;
	for (y = 0; y < 16; y++) for (x = 0; x < 4; x++)
	{
		jswByte *b = m_data + (y*8) + x;
		jswByte *c = m_data + (y*8) + (7-x);
		jswByte  a = *b;
		*b = flip4(*c);
		*c = flip4(a);
	}
}



void Jsw48Room::mirrorV()
{
	int x,y;
	for (y = 0; y < 8; y++) for (x = 0; x < 8; x++)
	{
		jswByte *b = m_data +     y  * 8 + x;
		jswByte *c = m_data + (15-y) * 8 + x;
		jswByte a  = *b;
		*b = *c;
		*c = a;
	}
}

void Jsw48Room::getRoomMemMap(char *buf)
{
	strcpy(buf,
			"bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb"	// 32
			"bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb"	// 64
			"bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb"	// 96
			"bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb"	// 128
			"tttttttttttttttttttttttttttttttt"	// 160
			"cccccccccccccccccccccccccccccccc"	// 192
			"ccccccccccccccccccccccxxxxxxxxxx"	// 224
			"xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"	// 256
		);
}

jswByte *Jsw48Room::getExits()
{
	return m_data + 0xE9;
}


/* Given a list of cell definitions, set their numberInRoom elements to
 * the cell type within the room. This can also modify their behaviour
 * if it wants.
 *
 * If numberInRoom is set to a +ve value, the corresponding cell pattern 
 * in the room is set from the pattern of this cell.
 *
 * If numberInRoom is set to a -ve value, the cell pattern is ignored and 
 * pattern (-numberInRoom) is substituted. 
 */
void Jsw48Room::normaliseCells(ImportCell *p, int count)
{
	int air = -1, water = -1, earth = -1, fire = -1, 
	    lramp = -1, rramp = -1, lconv = -1, rconv = -1;
	int n;

	for (n = 0; n < count; n++)
	{
		if (p[n].count == 0) continue;
		switch(p[n].behaviour)
		{
		    case CB_AIR:   if (air == -1)   air   = n; break;
		    case CB_WATER: if (water == -1) water = n; break;
		    case CB_EARTH: if (earth == -1) earth = n; break;
		    case CB_FIRE:  if (fire == -1)  fire  = n; break;
		    case CB_LRAMP: if (lramp == -1) lramp  = n; break;
		    case CB_RRAMP: if (rramp == -1) rramp  = n; break;
		    case CB_LCONV: if (lconv == -1) lconv  = n; break;
		    case CB_RCONV: if (rconv == -1) rconv  = n; break;
		}
	}
	// Set all cells to map to their equivalent elemental types,
	// or - failing that - water.
	for (n = 0; n < count; n++)
	{
		switch(p[n].behaviour)
		{
			case CB_AIR:	p[n].numberInRoom =  0; break;
			case CB_WATER:	p[n].numberInRoom = -1; break;
			case CB_EARTH:	p[n].numberInRoom = -2; break;
			case CB_FIRE:	p[n].numberInRoom = -3; break;
			default:	p[n].behaviour = CB_WATER;
					p[n].numberInRoom = -1;
					break;
		}
	}
	if (air != -1) 
	{ 
		p[air].behaviour = CB_AIR; 
		p[air].numberInRoom = 0; 
	}
	if (water != -1) 
	{ 
		p[water].behaviour = CB_WATER; 
		p[water].numberInRoom = 1; 
	}
	if (earth != -1) 
	{ 
		p[earth].behaviour = CB_EARTH; 
		p[earth].numberInRoom = 2; 
	}
	if (fire != -1) 
	{
	       	p[fire].behaviour = CB_FIRE; 
		p[fire].numberInRoom = 3; 
	}
	if (lramp != -1) 
	{ 
		p[lramp].behaviour = CB_LRAMP; 
		p[lramp].numberInRoom = 4; 
	}
	else if (rramp != -1)
	{
		p[rramp].behaviour = CB_RRAMP; 
		p[rramp].numberInRoom = 4; 
	}
	if (lconv != -1) 
	{ 
		p[lconv].behaviour = CB_LCONV; 
		p[lconv].numberInRoom = 5; 
	}
	else if (rconv != -1)
	{
		p[rconv].behaviour = CB_RCONV; 
		p[rconv].numberInRoom = 5; 
	}
}


jswByte Jsw48Room::getBorder()
{
	return m_data[0xDE] & 7;
}


void Jsw48Room::setBorder(int bdr)
{
	m_data[0xDE] &= ~7;
	m_data[0xDE] |= (bdr & 7);
}


void Jsw48Room::initImportBits()
{
	// Clear ramp & conveyor definitions
	memset(m_data + 0xD6, 0, 8);
}

void Jsw48Room::initImport()
{
	m_data[0xDE] &= 7; // Remove Superjump flag.
}

void Jsw48Room::initImportCells()
{
	m_data[0xA9] = 0xFF;	// Flag the water cell as unused.
}



// This is where, if we're importing a conveyor or ramp cell, we work out
// if it can be added to the existing conveyor/ramp. If so, we do this and
// set the cell "behind" it to air.
void Jsw48Room::importCell(int x, int y, int attr, int type, int behaviour)
{	
	int x0, y0;

//
// For the ramp, we rely on the fact that rooms are imported from the
// top down, so the first ramp cell imported will be the topmost. 
//
	// If ramp shares its attribute with another cell, draw that cell 
	// instead.
	if (type == 4)
	{
		if (m_data[0xC4] == m_data[0xA0]) 	// Ramp = air
		{	
			type = 0; 
			behaviour=CB_AIR;
		}
		if (m_data[0xC4] == m_data[0xA9]) 	// Ramp = water
		{	
			type = 1; 
			behaviour=CB_WATER;
		}
		if (m_data[0xC4] == m_data[0xB2]) 	// Ramp = earth
		{	
			type = 2; 
			behaviour=CB_EARTH;
		}
		if (m_data[0xC4] == m_data[0xBB]) 	// Ramp = fire
		{	
			type = 3; 
			behaviour=CB_FIRE;
		}
	}	
	if (type == 5)	// Same for conveyors
	{
		if (m_data[0xCD] == m_data[0xA0]) 	// Ramp = air
		{	
			type = 0; 
			behaviour=CB_AIR;
		}
		if (m_data[0xCD] == m_data[0xA9]) 	// Ramp = water
		{	
			type = 1; 
			behaviour=CB_WATER;
		}
		if (m_data[0xCD] == m_data[0xB2]) 	// Ramp = earth
		{	
			type = 2; 
			behaviour=CB_EARTH;
		}
		if (m_data[0xCD] == m_data[0xBB]) 	// Ramp = fire
		{	
			type = 3; 
			behaviour=CB_FIRE;
		}
		if (m_data[0xCD] == m_data[0xC4]) 	// Ramp = conveyor
		{	
			type = 4; 
			if (behaviour == CB_LCONV)
				behaviour = CB_LRAMP;
			else	behaviour = CB_RRAMP;
		}
	}
	if (type == 4) // Ramp
	{
		if (m_data[0xDD] == 0) // First ramp cell. Make a little ramp
					// containing just this cell.
		{
			m_data[0xDA] = (behaviour == CB_RRAMP);
			m_data[0xDB] = ((y * 32) + x) & 0xFF;
			m_data[0xDC] = (y >> 3) + 0x5E;
			m_data[0xDD] = 1;
			Room::importCell(x, y, m_data[0xA0], 0, CB_AIR);
			return;
		}
		else
		{
// Not the first ramp cell. See if the cell which is above this one and 
// either one to the left or one to the right is a ramp cell. If so, add
// this cell to the ramp.
			if (m_data[0xDA]) x0 = x + 1;
			else		  x0 = x - 1;
			y0 = y - 1;
			int candidate = getDrawnCellAt(x0, y0);

			if (candidate == CELL_RAMP)
			{
				m_data[0xDB] = ((y * 32) + x) & 0xFF;
				m_data[0xDC] = (y >> 3) + 0x5E;
				m_data[0xDD]++;
				Room::importCell(x, y, m_data[0xA0], 0, CB_AIR);
				return;
			}
			// We seem to have more than one ramp. If the room
			// hasn't used the water cell for anything, then 
			// take it over.
			if (m_data[0xA9] == 0xFF)
			{
				memcpy(m_data + 0xA9, m_data + 0xC4, 9);
				Room::importCell(x, y, m_data[0xA9], 1, CB_WATER);
				return;
			}
		}
	}
//
// For the ramp, we rely on the fact that rooms are imported left-to-right.
//
	if (type == 5) // Conveyor
	{
// If no previous conveyor, create a new one.
		if (m_data[0xD9] == 0) // Start of conveyor
		{
			m_data[0xD6] = (behaviour == CB_RCONV);
			m_data[0xD7] = ((y * 32) + x) & 0xFF;
			m_data[0xD8] = (y >> 3) + 0x5E;
			m_data[0xD9] = 1;
			Room::importCell(x, y, m_data[0xA0], 0, CB_AIR);
			return;
		}
		else
		{
// Otherwise, work out where the right-hand end of the conveyor is. If this 
// cell would go on the RH end of the conveyor, add it in.
			x0 = m_data[0xD7] & 31;
			y0 = (m_data[0xD7] >> 5) + ((m_data[0xD8] & 1) << 3);

			if (y == y0 && (x0 + m_data[0xD9] == x))
			{
				++m_data[0xD9];
				Room::importCell(x, y, m_data[0xA0], 0, CB_AIR);
				return;
			}
			// We seem to have more than one conveyor. If the room
			// hasn't used the water cell for anything, then 
			// take it over.
			if (m_data[0xA9] == 0xFF)
			{
				memcpy(m_data + 0xA9, m_data + 0xCD, 9);
				Room::importCell(x, y, m_data[0xA9], 0, CB_WATER);
				return;
			}
		}
	}
	Room::importCell(x, y, attr, type, behaviour);
}



bool Jsw48Room::exportHints(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent)
{
	char buf[30];
	int override, sprite;

	xmlNodePtr hint;

// Export conveyor animation
	if (m_data[0xD9])
	{
		hint = xmlNewNode(ns, (xmlChar *)"hint");
		xmlSetProp(hint, (xmlChar *)"type", (xmlChar *)"conveyor");
		sprintf(buf, "%d", m_data[0xD6]);
		xmlSetProp(hint, (xmlChar *)"direction", (xmlChar *)buf);
		sprintf(buf, "%d", m_data[0xD7] & 31);
		xmlSetProp(hint, (xmlChar *)"x", (xmlChar *)buf);
		sprintf(buf, "%d", (m_data[0xD7] >> 5) + ((m_data[0xD8] & 1) << 3));
		xmlSetProp(hint, (xmlChar *)"y", (xmlChar *)buf);
		sprintf(buf, "%d", m_data[0xD9]);
		xmlSetProp(hint, (xmlChar *)"length", (xmlChar *)buf);
		xmlAddChild(parent, hint);
	}
// Export ramp info
	if (m_data[0xDD])
	{
		hint = xmlNewNode(ns, (xmlChar *)"hint");
		xmlSetProp(hint, (xmlChar *)"type", (xmlChar *)"ramp");
		sprintf(buf, "%d", m_data[0xDA]);
		xmlSetProp(hint, (xmlChar *)"direction", (xmlChar *)buf);
		sprintf(buf, "%d", m_data[0xDB] & 31);
		xmlSetProp(hint, (xmlChar *)"x", (xmlChar *)buf);
		sprintf(buf, "%d", (m_data[0xDB] >> 5) + ((m_data[0xDC] & 1) << 3));
		xmlSetProp(hint, (xmlChar *)"y", (xmlChar *)buf);
		sprintf(buf, "%d", m_data[0xDD]);
		xmlSetProp(hint, (xmlChar *)"length", (xmlChar *)buf);
		xmlAddChild(parent, hint);
	}
	// Escalator
	if (m_data[0xC4] == m_data[0xCD] && m_data[0xDD])
	{
		hint = xmlNewNode(ns, (xmlChar *)"hint");
		xmlSetProp(hint, (xmlChar *)"type", (xmlChar *)"escalator");
		if (m_data[0xD6] == m_data[0xDA])
			xmlSetProp(hint, (xmlChar *)"direction", (xmlChar *)"up");
		else	xmlSetProp(hint, (xmlChar *)"direction", (xmlChar *)"down");
		xmlAddChild(parent, hint);	
	}
	if (m_game->allowSuperJump() && (m_data[0xDE] & 0x80))
	{
		hint = xmlNewNode(ns, (xmlChar *)"hint");
		xmlSetProp(hint, (xmlChar *)"type", (xmlChar *)"superjump");
		xmlSetProp(hint, (xmlChar *)"enabled", (xmlChar *)"1");
		xmlAddChild(parent, hint);
	}
	override = m_game->allowSpriteOverride();
	sprite   = m_game->decodeWillySprite(m_roomNumber, m_data[override]);
	if (spriteOverridden())
	{
		int frame;

		hint = xmlNewNode(ns, (xmlChar *)"hint");
		xmlSetProp(hint, (xmlChar *)"type", (xmlChar *)"sprite");
		for (frame = 0; frame < 8; frame++)
		{
			xmlNodePtr spr = xmlNewNode(ns, (xmlChar *)"spriteref");
			sprintf(buf, "%d", frame);
			xmlSetProp(spr, (xmlChar *)"frame", (xmlChar *)buf);
			sprintf(buf, "%04x", sprite*256 + 32*frame);
			xmlSetProp(spr, (xmlChar *)"xref", (xmlChar *)buf);
			xmlAddChild(hint, spr);
		}
		xmlAddChild(parent, hint);
	}
	return Room::exportHints(doc, ns, parent);
}




bool Jsw48Room::importHints(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent)
{
	xmlNodePtr cur;
	bool haveSuper = false, haveConveyor = false, haveRamp = false;

	for (cur = parent->xmlChildrenNode; cur; cur = cur->next)
	{
		if (cur->ns != ns) continue;
		if (strcmp((char *)cur->name, "hint")) continue;

		char *hintName = (char *)xmlGetProp(cur, (xmlChar *)"type");
		if (!hintName) continue;

		if (!strcmp(hintName, "superjump"))
		{
			char *attr;
			attr = (char *)xmlGetProp(cur, (xmlChar *)"enabled");
			if (attr && atoi(attr) && m_game->allowSuperJump())
			{
				m_data[0xDE] |= 0x80;
			}
			if (attr) xmlFree(attr);
			haveSuper = true;
		}

		if (!strcmp(hintName, "conveyor"))
		{
			char *attr;
		       
			attr = (char *)xmlGetProp(cur, (xmlChar *)"direction");
			if (attr) 
			{ 
				m_data[0xD6] = atoi(attr); 
				xmlFree(attr); 
			}
			attr = (char *)xmlGetProp(cur, (xmlChar *)"x");
			if (attr) 
			{
				m_data[0xD7] = (m_data[0xD7] & 0xE0) | 
							(atoi(attr) & 0x1F);
				xmlFree(attr);
			}
			attr = (char *)xmlGetProp(cur, (xmlChar *)"y");
			if (attr) 
			{
				m_data[0xD7] = (m_data[0xD7] & 0x1F) | 
					((atoi(attr) & 7) << 5);
				m_data[0xD8] = (m_data[0xD8] & 0xFE) | 
					((atoi(attr) & 8) >> 3);
				xmlFree(attr);
			}
			attr = (char *)xmlGetProp(cur, (xmlChar *)"length");
			if (attr) 
			{
				m_data[0xD9] = atoi(attr);
				xmlFree(attr);
			}
			haveConveyor = true;
		}
		if (!strcmp(hintName, "ramp"))
		{
			char *attr;
		       
			attr = (char *)xmlGetProp(cur, (xmlChar *)"direction");
			if (attr) 
			{
				m_data[0xDA] = atoi(attr);
				xmlFree(attr);
			}
			attr = (char *)xmlGetProp(cur, (xmlChar *)"x");
			if (attr) 
			{
				m_data[0xDB] = (m_data[0xDB] & 0xE0) | 
							(atoi(attr) & 0x1F);
				xmlFree(attr);
			}
			attr = (char *)xmlGetProp(cur, (xmlChar *)"y");
			if (attr) 
			{
				m_data[0xDB] = (m_data[0xDB] & 0x1F) | 
					((atoi(attr) & 7) << 5);
				m_data[0xDC] = (m_data[0xDC] & 0xFE) | 
					((atoi(attr) & 8) >> 3);
				xmlFree(attr);
			}
			attr = (char *)xmlGetProp(cur, (xmlChar *)"length");
			if (attr) 
			{
				m_data[0xDD] = atoi(attr);
				xmlFree(attr);
			}
			haveRamp = true;
		}
		if (!strcmp(hintName, "escalator"))
		{
			
		}
		xmlFree(hintName);
	}
	if (!haveSuper)    m_data[0xDE] &= 0x7F;
/* We can't rely on there being no conveyor/ramp just because there were
 * no hints. This might be a JSW64 import, for example.
	if (!haveConveyor) m_data[0xD9] = m_data[0xD6] = 0;
	if (!haveRamp)     m_data[0xDD] = 0;
*/
	return Room::importHints(doc, ns, parent);
}

void Jsw48Room::clearHints()
{
	m_data[0xDE] &= 0x7F;
//	m_data[0xD9] = m_data[0xD6] = 0;
//	m_data[0xDD] = 0;
}


void Jsw48Room::setItemBitmap(jswByte *data)
{
	memcpy(m_data + 0xE1, data, 8);
}


void Jsw48Room::clearItems()
{
	int n, rm, x, y;

	for (n = m_game->getItemCount(); n <= 255; n++)
	{
		if (!m_game->itemPos(n, &rm, &x, &y))
		{
			if (rm == m_roomNumber) 
			{
				m_game->itemDelete(n);
				--n;
			}
		}
	}
}


void Jsw48Room::addItem(int x, int y)
{
	m_game->itemInsert(m_roomNumber, x, y * 8);
}



bool Jsw48Room::exportGuardians(xmlDocPtr doc, xmlNsPtr ns, 
		xmlNodePtr grandparent)
{
	int n;
	char data[40];
	jswByte *buf = getGuardianBuffer();

	xmlNodePtr parent = xmlNewNode(ns, (xmlChar *)"guardians");
	xmlSetProp(parent, (xmlChar *)"type", (xmlChar *)"jsw");

	while(*buf != 0xFF)
	{
		if (!exportJswGuardian(doc, ns, parent, buf)) return false;
		buf += 8;
	}
	xmlAddChild(grandparent, parent);
	return true;
}

bool Jsw48Room::importGuardians(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent)
{
	xmlNodePtr cur;
	jswByte *buf;
	int *ids;
	int n, ng, slot, gc; 

/* Clear guardians. */
	slot = 0xF0;
	memset(m_data + 0xF0, 0, 16);
	m_data[0xF0] = 0xFF;
	char *type = (char *)xmlGetProp(parent, (xmlChar *)"type");

	if (type && !strcmp(type, "jsw"))
	{
/* Straight import of JSW guardians. We leave the other types for 
 * mature consideration. */

/* Start by counting the number of guardians */
		gc = 0;
		for (cur = parent->xmlChildrenNode; cur; cur = cur->next)
		{
			if (cur->ns != ns) continue;
			++gc;
		}
		if (!gc) return true;
		buf = new jswByte[gc * 8];
		ids = new int[gc];
		n = 0;

		for (cur = parent->xmlChildrenNode; cur; cur = cur->next)
		{
			if (cur->ns != ns) continue;
			importJswGuardian(doc, ns, cur, buf + 8*n, m_game);
			++n;
		}
		for (n = 0; n < gc; n++)
		{
			ng = m_game->addGuardians(buf, m_roomNumber, gc, ids);
			if (ng >= 0)
			{
				m_data[slot++] = ids[n];
				m_data[slot++] = buf[8 * n + 2];
				if (slot < 0x100) m_data[slot] = 0xFF;
			}
			if (slot >= 0x100) break;
		}
		delete ids;
		delete buf;
	}
	if (type) xmlFree(type);
	return true;
}



void Jsw48Room::addRefs()
{
	jswByte *buf    = m_data + 0xF0;
	jswByte *g;
	int n, ns, gt, frame;
	int guardBase = m_game->getGuardBase(m_roomNumber);
	int sprite, override;

	for (n = 0; n < 8; n++)
	{
		if (buf[n*2] == 0xFF) break;

		m_game->addRef(GUARDIAN(guardBase + 8 * buf[n*2]));
	}
	buf = getGuardianBuffer();
	for (n = 0; n < 8; n++)
	{
		g = buf + n*8;
		if (g[0] == 0xFF) break;
		switch(g[0] & 0x0F)
		{
			case 1: case 2:  case 5:  case 6:  case 7:
			case 9: case 10: case 13: case 14: case 15:
				for (frame = 0; frame < 8; frame++)
				{
					ns = (frame << 5) & 0xE0;

					ns &= (g[1] & 0xE0);
					ns |= (m_data[0xF1 + 2*n]) & 0xE0;
					ns += 256 * g[5];

//printf("Guardian %02x: frame=%02x ns=%04x\n", m_data[

					m_game->addRef(SPRITE(ns));
				}

				break;
		}
	}
// If sprite overridden, pick that up.
	override = m_game->allowSpriteOverride();
	sprite   = m_game->decodeWillySprite(m_roomNumber, m_data[override]);
	if (spriteOverridden())
	{
		int frame;

		for (frame = 0; frame < 8; frame++)
		{
			m_game->addRef(SPRITE(sprite*256 + 32*frame));
		}
	}
}





bool Jsw48Room::spriteOverridden()
{
	int j = m_game->allowSpriteOverride();
	if (j) return (m_data[j] != 0);

	return (m_roomNumber == m_game->getMem()->peek(0x9657));
}


void Jsw48Room::createCellPatterns()
{
        Room::createCellPatterns();
        m_cellBitmaps[CELL_ITEM*2]->getSurface()->fillReplace(
                        0, 0, 16, 16, ZX_BLACK, ZX_TRANS);
        m_cellBitmaps[CELL_ITEM*2]->makeTransparent();
}

const char *Jsw48Room::supportedGuardianTypes()
{
	return "jsw";
}
