/************************************************************************

    JSWED 2.3.0 - Editor for Jet Set Willy and derivatives

    Copyright (C) 2000-2001, 2004-6  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*************************************************************************/

#include "jswed.hxx"
#include "jswgame.hxx"
#include "room.hxx"
#include "j48roomnav.hxx"
#include "roomrange.hxx"
#include "jumproom.hxx"

#define TOP (VIDEO_H - 224)

Jsw48RoomNav::Jsw48RoomNav(JswGame *game) : Jsw48RoomDraw(game)
{
	int x,y,w,h;

	x = 9 * CHAR_W;
	y = TOP + CHAR_H + CHAR_SH + (CHAR_SH / 4);
	w =   5 * CHAR_W;
	h =       CHAR_H + (CHAR_SH / 2);

	addHotspot(new Hotspot(7,  x,    y,    w,  h,ZXK_UP    | ZXK_CONTROL));
	addHotspot(new Hotspot(6,  x,    y+2*h,w,  h,ZXK_DOWN  | ZXK_CONTROL));
	addHotspot(new Hotspot(5,  x,    y+h,  w/2,h,ZXK_LEFT  | ZXK_CONTROL));
	addHotspot(new Hotspot(8,  x+w/2,y+h,  w/2,h,ZXK_RIGHT | ZXK_CONTROL)); 
	addHotspot(new Hotspot('J',x,    y+3*h,w,  h, 'J', 'j'));
        addHotspot(new Hotspot('<',x,    y+4*h,w/2,h, '<'));
        addHotspot(new Hotspot('>',x+w/2,y+4*h,w/2,h, '>'));
	m_prevRoom = -1;
}


Jsw48RoomNav::~Jsw48RoomNav()
{
}



void Jsw48RoomNav::drawFixed(void)
{
	int x,y,h;
	int left = m_x;
        videoScreen->fillBox(left + CHAR_SW, TOP + CHAR_H + CHAR_SH,
                                6 * CHAR_W, 7 * CHAR_H, ZX_BRWHITE);
        videoScreen->box(left + CHAR_SW, TOP + CHAR_H + CHAR_SH,
                                6 * CHAR_W, 7 * CHAR_H, ZX_BLACK);

        videoScreen->drawSmallText(left + CHAR_SW, TOP + CHAR_H, "Go to       ",
                                ZX_BRWHITE, ZX_BLACK);
	videoScreen->smallZxLogo(left + 7 * CHAR_SW, TOP + CHAR_H); 
	x = left + CHAR_W;
	y = TOP + CHAR_H + CHAR_SH + (CHAR_SH / 2);
	h = CHAR_H + (CHAR_SH / 2);
        videoScreen->drawText(x, y, "  ^  ", ZX_BLACK, ZX_BRWHITE);
        videoScreen->drawText(x, y+h, " \013 \014 ", ZX_BLACK, ZX_BRWHITE);
        videoScreen->drawText(x, y+2*h, "  \011  ", ZX_BLACK, ZX_BRWHITE);
	videoScreen->drawSmallText(x, y+3*h + (CHAR_SH / 2), 
				"Enter no.", ZX_BLACK, ZX_BRWHITE);
	videoScreen->drawText(x, y+4*h, " < > ", ZX_BLACK, ZX_BRWHITE);
	y = TOP + CHAR_H + CHAR_SH + (CHAR_SH / 4);
        videoScreen->box(left + CHAR_SW, y+h,   6 * CHAR_W, h, ZX_BLACK);
	videoScreen->box(left + CHAR_SW, y+h,   3 * CHAR_W, h, ZX_BLACK);
	videoScreen->box(left + CHAR_SW, y+3*h, 6 * CHAR_W, 1, ZX_BLACK);
        videoScreen->box(left + CHAR_SW, y+4*h, 6 * CHAR_W, h, ZX_BLACK);
        videoScreen->box(left + CHAR_SW, y+4*h, 3 * CHAR_W, h, ZX_BLACK);
	
	Jsw48RoomDraw::drawFixed();
}

int Jsw48RoomNav::onHotspot(int id)
{
	jswByte *exits = getRoom()->getExits();

	switch(id)
	{
		case 5: gotoRoom(exits[0]); return ZXE_CONTINUE;
		case 6: gotoRoom(exits[3]); return ZXE_CONTINUE;
		case 7: gotoRoom(exits[2]); return ZXE_CONTINUE;
		case 8: gotoRoom(exits[1]); return ZXE_CONTINUE;
		case 'J': return jumpRoom();
		case '<': gotoRoom(m_game->prevRoom(m_room)); return ZXE_CONTINUE;
		case '>': gotoRoom(m_game->nextRoom(m_room)); return ZXE_CONTINUE;
	}
	return ZXE_CONTINUE;
}


int Jsw48RoomNav::onKeyDown(int keysym)
{
	int rv, rm;
	
	rv = HotspotManager::onKey(keysym); 
	if (rv != -1) return rv;
	switch(keysym)
	{
		case ZXK_F8:	return screenshotAll();
		case ZXK_CTRLA:
		case ZXK_CTRLJ:	intPreContextMenu();
				rv = jumpRoomNumber(&rm);
				postContextMenu();
				if (rv >= ZXE_QUIT) return rv;
				if (rm >= 0) gotoRoom(rm);
				return ZXE_CONTINUE;
		case ZXK_CTRLL: gotoRoom(m_room); return ZXE_CONTINUE;
		case ZXK_CTRLP:	if (m_prevRoom >= 0)
				{
					gotoRoom(m_prevRoom);
					return ZXE_CONTINUE;
				}
				break;
		case ZXK_PLUS:	faster(); drawBorder(); return ZXE_CONTINUE;
		case ZXK_MINUS: slower(); drawBorder(); return ZXE_CONTINUE;
				
	}
	return Jsw48RoomDraw::onKeyDown(keysym);
}

int Jsw48RoomNav::onButtonDown(int x, int y, int button)
{
        int rv;

	rv = HotspotManager::onButton(x, y, button);
	if (rv != -1) return rv;
	return Jsw48RoomDraw::onButtonDown(x, y, button);
}

int Jsw48RoomNav::jumpRoom(void)
{
	int rv = ZXE_CONTINUE;
	int rm;

	intPreContextMenu();

	JswRoomList *s = m_game->getRoomList();
	if (s)
	{
		s->setTitle("Jump to room");
		s->setSelectedRoom(m_room);
		rv = s->doModal();
		if (rv != ZXE_QUIT && rv != ZXE_OK) rv = ZXE_CONTINUE;
		else if (rv == ZXE_OK)
		{
			rm = s->getSelectedRoom();
			gotoRoom(rm);
		}
		delete s;
	}
	postContextMenu();
	return rv;
}

int Jsw48RoomNav::onReveal(int rect)
{
	if (m_room != m_game->getCurrentRoom())
	{
		gotoRoom(m_game->getCurrentRoom());
	}
        /* Room we are working on has become unavailable? */
        if (m_game->getRoom(m_room) == NULL)
        {
                int l = m_game->prevRoom(m_room);
		if (l == m_room) l = m_game->nextRoom(m_room);

		gotoRoom(l);
        }
	else	gotoRoom(m_room);
	return Jsw48RoomDraw::onReveal(rect);
}

int Jsw48RoomNav::doExport()
{
	return RoomImport::doExport(m_game);
#if 0
	int room, from, upto;

	JswRoomList *s = m_game->getRoomList();
	RoomRangeForm rf(1, s);

	int rv = rf.doModal();
	if (rv >= ZXE_QUIT) 
	{
		delete s;
		return rv;
	}
	if (rv != ZXE_OK) 
	{
		delete s;
		return ZXE_CONTINUE;
	}

	xmlDocPtr doc = xmlNewDoc((xmlChar *)"1.0");
	if (!doc)
	{
		delete s;
		return alert("Error", "Cancel", "Failed to create XML document object.");
	}
	xmlNsPtr ns; 
	xmlNodePtr root;
	root = xmlNewNode(NULL, (xmlChar *)"rooms");
	xmlDocSetRootElement(doc, root);
	ns = xmlNewNs(root, NAMESPACE, NULL);

	from = rf.getFrom();
	upto = rf.getUpto();
	m_game->clearRefCounts();
	delete s;
	for (room = from; room <= upto; room++)
	{
		Room *rm = m_game->getRoomClass(room);
		if (!rm) continue;
		rm->exportXML(doc, ns, root);
		rm->addRefs();
	}
	m_game->exportReferencedSprites(doc, ns, root, 16, 16);
        char filename[PATH_MAX];
        JswFileForm ff;

	filename[0] = 0;
	rv = ff.doModal(filename);
	if (rv != ZXE_OK)
	{
		xmlFreeDoc(doc);
		if (rv >= ZXE_QUIT) return rv;
		return ZXE_CONTINUE;
	}
	if (xmlSaveFile(filename, doc) < 0)
	{
		xmlFreeDoc(doc);
		return alert("File was not saved.", "Cancel");
	}
	xmlFreeDoc(doc);
	return ZXE_CONTINUE;
#endif
}


int Jsw48RoomNav::doImport()
{
	int rv = RoomImport::doImport(m_game);
#if 0
        char filename[PATH_MAX];
        JswFileForm ff;
	xmlDocPtr doc;
	xmlNsPtr ns;
	xmlNodePtr cur, child;
	int rv;
	int from, upto, room;
	int map[256];

	filename[0] = 0;
	rv = ff.doModal(filename);
	if (rv != ZXE_OK)
	{
		if (rv >= ZXE_QUIT) return rv;
		return ZXE_CONTINUE;
	}
	doc = xmlParseFile(filename);
	if (doc == NULL)
	{
		return alert("Couldn't load XML", "Cancel",
				"Unable to load and parse the file:\n%s",
				filename);
	}
	cur = xmlDocGetRootElement(doc);
	if (cur == NULL)
	{
		xmlFreeDoc(doc);
		return alert("Couldn't load XML", "Cancel",
				"XML document is empty");
	}
	ns = xmlSearchNsByHref(doc, cur, NAMESPACE);
	if (ns == NULL)         // No namespace
	{
		xmlFreeDoc(doc);
		return alert("Couldn't load XML", "Cancel",
				"No JSWED namespace found in the file.");
	}
	if (strcmp((char *)cur->name, "rooms")) // Wrong root
	{
		xmlFreeDoc(doc);
		return alert("Couldn't load XML", "Cancel",
				"First node is not <rooms>.");
	}
	int roomCount = 0;
	// OK. File loaded. See how many rooms there are.
	for (child = cur->xmlChildrenNode; child; child = child->next)
	{
		if (child->ns != ns) continue;
		if (!strcmp((char *)child->name, "room")) ++roomCount;
	}
	if (roomCount == 0)
	{
		xmlFreeDoc(doc);
		return alert("Couldn't load XML", "Cancel",
				"No rooms in the file.");
	}
	JswRoomList *s = m_game->getRoomList();
	RoomRangeForm rf(0, s, roomCount);
	rv = rf.doModal();

	if (rv != ZXE_OK)
	{
		delete s;
		xmlFreeDoc(doc);
		if (rv >= ZXE_QUIT) return rv;
		return ZXE_CONTINUE;
	}
	from = rf.getFrom();
	upto = rf.getUpto();
	delete s;
	child = cur->xmlChildrenNode; 
/* Populate the map of rooms */
	for (room = 0; room < 256; room++) map[room] = -1;
	for (room = from; child != NULL; child = child->next)
	{
		if (child->ns != ns) continue;
		if (strcmp((char *)child->name, "room")) continue;
		char *num = (char *)xmlGetProp(child, (xmlChar *)"number");

		if (num) 
		{
			map[room++] = atoi(num);
			xmlFree(num);
		}
	}
/* Build a reference count of guardians and sprites, in the rooms that are
 * not being used. */
	m_game->clearRefCounts();
	for (room = 0; room < 256; room++) if (map[room] == -1)
	{
		Room *rm = m_game->getRoomClass(room);
		if (rm) rm->addRefs();
	}
	m_game->addFixedRefs();
/* Import sprites and map them */
	m_game->importSprites(doc, ns, cur, map);

//	m_game->dumpRefs();
	child = cur->xmlChildrenNode; 
	for (room = from; child != NULL; child = child->next)
	{
		if (child->ns != ns) continue;
		if (strcmp((char *)child->name, "room")) continue;

		Room *rm = m_game->getRoomClass(room);
		if (!rm) continue;
		rm->importXML(doc, ns, child, map);
		child = child->next;
		++room;
		if (room > upto) break;
	}
	xmlFreeDoc(doc);
#endif
	gotoRoom(m_room);

	return rv;
}



int Jsw48RoomNav::screenshotAll()
{
	/* Start at the beginning... */
	int rm, rm0, rv;

	rm0 = m_room;
	do
	{
		rm = m_room;
		m_room = m_game->prevRoom(m_room);
	}
	while (rm != m_room);

	/* Then continue to the end */
	do
	{
		char buf[20];

		gotoRoom(m_room);
		sprintf(buf, "%03d.png", m_room);
		rv = m_bmAnim->getSurface()->savePng(0, 0, 256, 128, buf, 2);
		if (rv >= ZXE_QUIT) return rv;
		rm = m_room;
		m_room = m_game->nextRoom(m_room);

	} while (rm != m_room);
	gotoRoom(rm0);
	return ZXE_CONTINUE;
}

