/************************************************************************

    JSWED 2.3.0 - Editor for Jet Set Willy and derivatives

    Copyright (C) 2000-2001,2004  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*************************************************************************/

#include "jswed.hxx"
#include "jswtitle.hxx"
#include "j48title.hxx"
#include "jswgame.hxx"


Jsw48TitlePage::Jsw48TitlePage(JswGame *g) : 
	JswTitlePage(g->getTitleAttrs(), 512)
{
	int x, y, w, h;

	m_mem = g->getMem();

	x = VIDEO_W - 26 * CHAR_W;
	y = 24 * CHAR_H + CHAR_SH;
	w = 4 * CHAR_W;
	h = 2 * CHAR_H;

	addHotspot(new Hotspot('I', x, y, w, h, 'I', 'i')); x += w;
	addHotspot(new Hotspot('P', x, y, w, h, 'P', 'p')); x += w;
	addHotspot(new Hotspot('B', x, y, w, h, 'B', 'b')); x += w;
	addHotspot(new Hotspot('F', x, y, w, h, 'F', 'f'));

	setup();
}


Jsw48TitlePage::~Jsw48TitlePage()
{
}

int Jsw48TitlePage::getCell(int x, int y, jswByte *attr, jswByte **pStripe)
{
	int offset = y * 32 + x;
	int value, udg;

	value = m_attrs[offset];	
	if (attr) *attr = value;
	if (pStripe)
	{
		// "plain" colours
		if (value == 0 ||
		    value == m_mem->peek(0x8849) ||
		    value == m_mem->peek(0x884D) ||
		    value == m_mem->peek(0x8851) ||
		    value == m_mem->peek(0x8855))
		{	
			*pStripe = NULL;
		}
		else
		{
			// "left slope" colours
			if (value == m_mem->peek(0x8863))
			{
				*attr = value = m_mem->peek(0x886F);
				udg = 0;
			}
			else if (value == m_mem->peek(0x885B) ||
				 value == m_mem->peek(0x885F) ||
				 value == m_mem->peek(0x8867)) 
				udg = 0;
			else 	udg = 16;
			if (x & 1) udg += 8;
			*pStripe = m_mem->memoryAt(udg + 0x8431);
		}	
	}
	return value;
}

jswByte *Jsw48TitlePage::getUdgs()
{
	return m_mem->memoryAt(0x8431);
}



void Jsw48TitlePage::putCell(int x, int y, jswByte attr)
{
	int offset = y * 32 + x;

	if (offset >= 0 && offset < 512) m_attrs[offset] = attr;	
}


static int st_colourAddrs[] =
{
	0x8849, 0x884D, 0x8851, 0x8855,
	0x8863, 0x885b, 0x885f, 0x8867,
	0x886F
};

class Jsw48ColoursForm : public VideoForm
{
	VideoColourChooser *m_cc[18];
	VideoCheckbox      *m_cb[18];	
	VideoLabel	   *m_lbl[18];
	VideoButton	   *m_ok, *m_cancel;
	SpectrumMemory     *m_mem;
public:
	Jsw48ColoursForm(SpectrumMemory *m) : VideoForm(
			3 * CHAR_SW, 7 * CHAR_SH, 37 * CHAR_W, 25 * CHAR_H)
	{
		m_mem = m;
		int n, x, y;
		for (n = 0; n < 9; n++)
		{
			x = n % 4;
			y = n / 4;		
			m_cc[2*n] = new VideoColourChooser(
				m_x + ((9  * x) +1) * CHAR_W,
				m_y + ((14 * y) +5) * CHAR_SH);
			m_cc[2*n+1] = new VideoColourChooser(
				m_x + ((9  * x) +1) * CHAR_W,
				m_y + ((14 * y) +8) * CHAR_SH);
			m_cb[2*n] = new VideoCheckbox(
				m_x + ((9  * x) +8) * CHAR_W,
				m_y + ((14 * y) +11) * CHAR_SH);
			m_cb[2*n+1] = new VideoCheckbox(
				m_x + ((9  * x) +8) * CHAR_W,
				m_y + ((14 * y) +14) * CHAR_SH);
			m_lbl[2*n] = new VideoLabel(
				m_x + ((9  * x) +1) * CHAR_W,
				m_y + ((14 * y) +11) * CHAR_SH,
				"Bright");
			m_lbl[2*n+1] = new VideoLabel(
				m_x + ((9  * x) +1) * CHAR_W,
				m_y + ((14 * y) +14) * CHAR_SH,
				"Flash");
			m_cc[2*n]->setColour(m->peek(st_colourAddrs[n]) & 7);
			m_cc[2*n+1]->setColour((m->peek(st_colourAddrs[n]) >> 3) & 7);
			m_cb[2*n]->setCheck(m->peek(st_colourAddrs[n]) & 0x40);
			m_cb[2*n+1]->setCheck(m->peek(st_colourAddrs[n]) & 0x80);
		}
		m_cancel = new VideoButton(ZXE_CANCEL, m_x + CHAR_W, m_y + m_h - 3 * CHAR_SH, "Cancel");
		m_ok = new VideoButton(ZXE_OK, m_x + m_w - 7 * CHAR_W, m_y + m_h - 3 * CHAR_SH, "  OK  ");

	}

	~Jsw48ColoursForm()
	{
		int n;
		for (n = 0; n < 18; n++)
		{
			delete m_cc[n];
			delete m_cb[n];
			delete m_lbl[n];
		}
		delete m_ok;
		delete m_cancel;
	}

	virtual void redraw()
	{
		VideoForm::redraw();
		redrawBorder("Colours");

		videoScreen->drawText(m_x + CHAR_W, m_y + CHAR_H + 4,
				"Colours for blank cells:", ZX_BRBLUE, ZX_WHITE);
		videoScreen->drawText(m_x + CHAR_W, m_y + 8 * CHAR_H + 4,
				"Colours for \\ sloping cells:", ZX_BRBLUE, ZX_WHITE);
		videoScreen->drawText(m_x + CHAR_W, m_y + 15 * CHAR_H + 4,
				"Substitute colour:", ZX_BRBLUE, ZX_WHITE);
	}

	virtual void showChildren(int redraw)
	{
		for (int n = 0; n < 18; n+= 2)
		{
			addChild(m_cc[n]);
			addChild(m_cc[n+1]);
			addChild(m_cb[n]);
			addChild(m_cb[n+1]);
			addChild(m_lbl[n]);
			addChild(m_lbl[n+1]);
		}
		addChild(m_cancel);
		addChild(m_ok);
		VideoForm::showChildren(redraw);
	}


	virtual void hideChildren()
	{
		for (int n = 0; n < 18; n++)
		{
			removeChild(m_cc[n]);
			removeChild(m_cb[n]);
			removeChild(m_lbl[n]);
		}
		removeChild(m_cancel);
		removeChild(m_ok);
	}

	virtual int doModal()
	{
		int rv = VideoForm::doModal();
		if (rv == ZXE_OK)
		{
			int n;
			for (n = 0; n < 9; n++)
			{
				int ink    = m_cc[2*n  ]->getColour();
				int paper  = (m_cc[2*n+1]->getColour()) << 3;
				int bright = m_cb[2*n  ]->getCheck() ? 0x40 : 0;
				int flash  = m_cb[2*n+1]->getCheck() ? 0x80 : 0;
				m_mem->poke(st_colourAddrs[n], ink|paper|bright|flash);
			}
		}
		return rv;
	}

};




int Jsw48TitlePage::setupColours()
{
	Jsw48ColoursForm cf(m_mem);
	return cf.doModal();
}
