/************************************************************************

    JSWED 2.3.2 - Editor for Jet Set Willy and derivatives

    Copyright (C) 2000-1,2004-7  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*************************************************************************/

#include "jswed.hxx"
#include "room.hxx"
#include "j64room.hxx"
#include "jswgame.hxx"
#include "j64game.hxx"



Jsw64Room::Jsw64Room(Jsw64Game *g, jswByte *data, int roomNumber) : 
	Room(16, roomNumber), m_game(g), m_data(data) 
{
	prepareForSpriteImport();
}

int Jsw64Room::getTitle(char *buf)
{
	memcpy(buf, m_data + 0xB6, 32);
	buf[32] = 0;
        for (int n = 0; n < 32; n++)
        {
                if (buf[n] == ZXCH_COPY ) buf[n] = ZXK_COPY;
                if (buf[n] == ZXCH_POUND) buf[n] = ZXK_POUND;
        }
	return 32;
}

void Jsw64Room::setTitle(const char *buf)
{
	char fmt[33];

	sprintf(fmt, "%-32.32s", buf);
	memcpy(m_data + 0xB6, fmt, 32);
}

jswByte *Jsw64Room::getCellPattern(int cell)
{
	if (cell < 0) return NULL;

	char var = m_game->getVariant();
	int max;
	jswByte *base;

	switch(var)
	{
		case 'V': case 'X':
			base = m_data + 0x6E;
			max = 8;
			break;

		case 'W': case 'Y': case 'Z':
			base = m_data + 0x41;
			max = 13;
			if (cell >= 13 && cell < 16)	// Global cell patterns
			{
				base = m_game->getMem()->memoryAt(0xf50f, 7);
				cell -= 13;
				max = 3;
			}
			break;

		case '[':
			base = m_data + 0x26;
			max = 16;
			break;
	}
	if (cell >= 0x100)
	{
		memcpy(m_queryPattern + 1, m_data + 0x4B, 8);
		m_queryPattern[0] = cell & 0xFF;
		return m_queryPattern;
	}
	else if (cell >= max)
	{
		m_itemPattern[0] = ZX_WHITE;
		memcpy(m_itemPattern + 1, m_data + 0xE1, 8);
		return m_itemPattern;
	}
	return base + 9 * cell;
}

jswByte *Jsw64Room::getBits()
{
	switch(m_game->getVariant())
	{
		case 'V': return m_data + 0x140;
		case 'W': return m_data + 0x100;
		case 'X': return m_data + 0x340;
		case 'Y': return m_data + 0x300;
		case 'Z': return m_data + 0x200;
		case '[': return m_data + 0x300;
	}
	return NULL;
}

int Jsw64Room::getCellSize()
{
	switch(m_game->getVariant())
	{
		case 'V': case 'X': return 3;
		case 'Z': return 8;
	}
	return 4;
}

int Jsw64Room::getDrawnCellAt(int x, int y)
{
	if (getCellSize() == 8)
	{
		jswByte *bits = getBits();
		return bits[y * 32 + x];
	}
	return getCellAt(x, y);
}

int Jsw64Room::getCellAt(int x, int y)
{
	int n;
	jswByte attrib;
	jswByte *bits = getBits();
	int nBits     = getCellSize();
	unsigned bitsOffset;
	jswByte smask, dmask, value;

	if (nBits == 8)
	{
		attrib = bits[y * 32 + x];
		for (n = 0; n < 13; n++)
		{
			if (m_data[0x41 + 9 * n] == attrib) return n;
		}
		jswByte *base = m_game->getMem()->memoryAt(0xf50f, 7);
		for (n = 0; n < 3; n++)
		{
			if (base[n * 9] == attrib)
			{
				return n + 13;
			}
		}
		m_queryPattern[0] = attrib;
		return 0x100 | attrib;
	}
	bitsOffset = ((y * 32) + x) * nBits;
	bits += (bitsOffset / 8);
	smask = 0x80 >> (bitsOffset & 7);
	dmask = 1 << (nBits - 1);
	for (n = 0, value = 0; n < nBits; n++)
	{
		if (bits[0] & smask) value |= dmask;
		smask = smask >> 1;
		dmask = dmask >> 1;
		if (!smask)
		{
			smask = 0x80;
			++bits;
		}
	}
	return value;	
}


int Jsw64Room::getConveyorDir()
{
	return (m_data[0xD6] & 1);
}

int Jsw64Room::getConveyorY()
{
	return 0;
}

void Jsw64Room::setCell(int x, int y, int type)
{
	int n;
	jswByte attrib;
	jswByte *bits = getBits();
	int nBits     = getCellSize();
	unsigned bitsOffset;
	jswByte smask, dmask;

	if (nBits == 8)
	{
		if (type < 0x100) 
		{
			if (type < 13)
			{
				bits[y * 32 + x] = m_data[0x41 + 9 * type];
			}
			else // Global type
			{
				jswByte *base;
				base = m_game->getMem()->memoryAt(0xf50f, 7);
				bits[y * 32 + x] = base[(type - 13) * 9];
			}
		}
		else
		{
			bits[y * 32 + x] = type & 0xFF;
		}
		return;
	}
	bitsOffset = ((y * 32) + x) * nBits;
	bits += (bitsOffset / 8);
	smask = 1 << (nBits - 1);
	dmask = 0x80 >> (bitsOffset & 7);
	for (n = 0; n < nBits; n++)
	{
		if (type & smask) bits[0] |= dmask;
		else		  bits[0] &= ~dmask;
		smask = smask >> 1;
		dmask = dmask >> 1;
		if (!dmask)
		{
			dmask = 0x80;
			++bits;
		}
	}
}



int Jsw64Room::getBackground()
{
	int n, max;
	switch(m_game->getVariant())
	{
		case 'V': case 'X': max = 8; break;
		default: max = 16; break;
	}	
	for (n = 0; n < max; n++)
	{
		// If we find an 'air' cell, return its colour
		if (getCellBehaviour(n) == CB_AIR)
		{
			jswByte *pattern = getCellPattern(n);
			return ((*pattern) >> 3)  & 0x0F;
		}
	}
	return 0;	// Default to black
}


int Jsw64Room::getItemCoords(int *px, int *py, int *counts)
{
	int count = 0;
	int n, m, x, y;
	int rm, found = 0;

	for (n = m_game->getItemCount() ; n <= 0xFF; n++)
	{
		m_game->itemPos(n, &rm, &px[count], &py[count]);
		found = 0;
		if (rm == m_roomNumber)
		{
			y = (py[count] / 8) & 0x0F;
			x =  px[count] & 0x1F;
			for (m = 0; m < count; m++)
			{
				if (py[m] == y && px[m] == x)
				{
					found = 1;
					++counts[m];
					break;
				}
			}
			if (!found)
			{
				py[count] = y;
				px[count] = x;
				counts[count] = 1;
				++count;
			}
		}
	}
	return count;
}


jswByte *Jsw64Room::getGuardianBuffer()
{
	unsigned offset = m_data[0xDB] *256 + m_data[0xDA];

	return addressOf(offset);
}


jswByte *Jsw64Room::addressOf(unsigned offset)
{
	unsigned upperLimit;

	switch(m_game->getVariant())
	{
		case 'V': case 'W':	upperLimit = 0x8200; break;
		default:		upperLimit = 0x8400; break;
	}	
	if (offset >= 0x8000 && offset < upperLimit)
	{
		return m_data + (offset - 0x8000);
	}
 	return m_game->getMem()->memoryAt(offset);
}


void Jsw64Room::deletePortal(void)
{
	m_data[0xEE] = m_data[0xEF] = 0;
}

/* Populate a portal structure */
bool Jsw64Room::getPortal(Portal &p)
{
	unsigned offset = m_data[0xEE] + 256 * m_data[0xEF];

	if (offset == 0) return false;

	p.sprite    = addressOf(offset);
	p.x	 =   m_data[0xF0] & 0x1F;
	p.y	 = ((m_data[0xF0] & 0xE0) >> 5) | ((m_data[0xF1] & 1) << 3);
	p.attribute = m_data[0xF4];

/* Generate local sprite ID */
	if (offset >= 0x8000 && offset < 0x8400)
	       offset += (m_roomNumber * 65536);	
	{
		p.spriteID = offset;
	}
	p.targetRoom = m_data[0xF5];
	p.targetX = m_data[0xF6] & 0x1F;
	p.targetY = m_data[0xF8] >> 4;
	p.spriteID = encodeSpriteID(offset) & 0xFFFFFF;

	return true;
}



void Jsw64Room::setPortal(Portal &p)
{
	int sprite = m_game->lookupSpriteRef(p.spriteID);
	m_data[0xEF] = (sprite >> 8);
	m_data[0xEE] = (sprite & 0xFF);
	m_data[0xF4] = p.attribute;
	m_data[0xF5] = p.targetRoom;
	m_data[0xF6] = (p.targetX & 0x1F) | ((p.targetY & 7) << 5);
	m_data[0xF7] = 0x5C | ((p.targetY >> 3) & 1);
	m_data[0xF8] = (p.targetY & 0x0F) << 4;
	setPortalPos(p.x, p.y);
}




int Jsw64Room::getMaxGuards()
{
	unsigned offset = m_data[0xDA] + 256 * m_data[0xDB];
	switch(m_game->getVariant())
	{
		case 'V':	return 13;
		case 'W':	return 8;
		case 'X':	if (offset >= 0x8000 && offset < 0x8069)
				{
					return (0x8068-offset) / 8;
				}
				if (offset >= 0x8100 && offset < 0x8340)
				{
					return (0x8340-offset) / 8;
				}
				return 16;	/* Random! */
		case 'Y':	if (offset >= 0x8000 && offset < 0x8041)
				{
					return (0x8040-offset) / 8;
				}
				if (offset >= 0x8100 && offset < 0x8300)
				{
					return (0x82FF-offset) / 8;
				}
				return 16;	/* Random! */
		case 'Z':	if (offset >= 0x8000 && offset < 0x8041)
				{
					return (0x8040-offset) / 8;
				}
				if (offset >= 0x8100 && offset < 0x8200)
				{
					return (0x81FF-offset) / 8;
				}
				return 16;	/* Random! */
		case '[':	if (offset >= 0x8000 && offset < 0x8021)
				{
					return (0x8020-offset) / 8;
				}
				if (offset >= 0x8100 && offset < 0x8300)
				{
					return (0x82FF-offset) / 8;
				}
				return 16;	/* Random! */
	}
	return 0;
}



int Jsw64Room::getCellBehaviour(int index) 
{
	jswByte mask, val;
	jswByte *map;

	switch(m_game->getVariant())
	{
		case 'V': case 'X':
			if (index < 0 || index > 7) return CB_WATER;
			if (index & 1)	mask = 0xF0;
			else		mask = 0x0F; 
			val = m_data[0x69 + (index/2)] & mask;
			if (index & 1) val = val >> 4;
			return val;
		case 'W': case 'Y': case 'Z':
			if (index < 0 || index > 15) return CB_WATER;
			map = m_game->getMem()->memoryAt(0xF4FF, 7);
			return map[index];	
		case '[':
			if (index < 0 || index > 15) return CB_WATER;
			map = m_game->getMem()->memoryAt(0xF519, 7);
			return map[index];
	}
	return CB_WATER;
}


/* Given a list of cell definitions, set their numberInRoom elements to
 * the cell type within the room. This can also modify their behaviour
 * if it wants.
 *
 * If numberInRoom is set to a +ve value, the corresponding cell pattern 
 * in the room is set from the pattern of this cell.
 *
 * If numberInRoom is set to a -ve value, the cell pattern is ignored and 
 * pattern (-numberInRoom) is substituted. 
 *
 * For those JSW64 games with 8 cell types, loads the cell types into the room
 * definition. For those with 16, try to match the cell types with the global
 * list. */
void Jsw64Room::normaliseCells(ImportCell *p, int count)
{
	int max, n, c, cur, waterCell = 1;
	char variant = m_game->getVariant();

	if (variant == 'V' || variant == 'X')
	{
		for (n = 0; n < 8; n++)
		{
			if (getCellBehaviour(n) == CB_WATER)
			{
				waterCell = n;
				break;
			}
		}
		cur = 0;
		for (n = 0; n < count; n++)
		{
			if (p[n].count == 0 || cur > 7) 
			{
				p[n].behaviour = CB_WATER;
				p[n].numberInRoom = -waterCell;
			}
			else
			{
				setCellBehaviour(cur, p[n].behaviour);
				p[n].numberInRoom = cur;
				cur++;
			}					
		}
		return;
	}
	/* OK. Now we need to match up the cell types being imported with 
	 * the cell behaviours present in the game. j48room.cxx has 
	 * something similar but it's hardcoded to the fixed JSW48 cell types.
	 */
	for (n = 0; n < 16; n++)
	{
		if (getCellBehaviour(n) == CB_WATER)
		{
			waterCell = n;
			break;
		}
	}
	bool cellUsed[16];
	memset(cellUsed, 0, sizeof(cellUsed));
	/* First pass: Find unique mappings for the cells if possible */
	for (n = 0; n < count; n++)
	{
		p[n].numberInRoom = -99;
		if (!p[n].count) continue;
		for (c = 0; c < 16; c++)
		{	
			if (cellUsed[c]) continue;
			if (getCellBehaviour(c) == p[n].behaviour)
			{
				cellUsed[c] = true;
				p[n].numberInRoom = c;
				break;
			}	
		}
	}
	/* Second pass: Try to map the rest to types that do exist */
	for (n = 0; n < count; n++)
	{
		if (!p[n].count) continue;
		if (p[n].numberInRoom >= 0) continue;
		for (c = 0; c < 16; c++)
		{	
			if (getCellBehaviour(c) == p[n].behaviour)
			{
				p[n].numberInRoom = -c;
				break;	
			}
		}
	} 
	/* Any that are left, set to have CB_WATER behaviour, so they 
	 * don't trigger any ramp/conveyor/whatever detection */
	for (n = 0; n < count; n++)
	{
		if (p[n].numberInRoom == -99) 
			p[n].behaviour = -waterCell;
	}
}


void Jsw64Room::setCellBehaviour(int index, int cb)
{
	jswByte mask, val;
	jswByte *map;

	switch(m_game->getVariant())
	{
		case 'V': case 'X':
			if (index < 0 || index > 7) return;
			if (index & 1)	{ mask = 0xF0; val = cb << 4; }
			else		{ mask = 0x0F; val = cb; }
			m_data[0x69 + (index/2)] &= ~mask;
			m_data[0x69 + (index/2)] |= (val & mask);
			break;
		case 'W': case 'Y': case 'Z':
			if (index < 0 || index > 15) return;
			map = m_game->getMem()->memoryAt(0xF4FF, 7);
			map[index] = cb;
			break;
		case '[':
			if (index < 0 || index > 15) return;
			map = m_game->getMem()->memoryAt(0xF519, 7);
			map[index] = cb;
			break;
	}
}


void Jsw64Room::drawOneOffs(VideoBitmap *bm, bool flashonly, bool draw)
{
	int len;
	int w = m_data[0xD7]+(256*m_data[0xD8]);
	int x =  (w & 0x1F);
	int y =  (w >> 5) & 0x0F;

	if (m_data[0xD9] > 32) // Damaged room 
		m_data[0xD9]=0;

	len = m_data[0xD9];

//	y = y >> 4;
	for (int n = 0; n < len; n++)
	{
		if (y >= 0 && y < 16 && x >= 0 && x < 32)
		{
			drawCell(bm, x, y, flashonly, draw);
		}
		x++;
	}
	Room::drawOneOffs(bm, flashonly, draw);
}



int Jsw64Room::getCellTypeCount(bool includeSingles)
{
	switch(m_game->getVariant())
	{
		case 'V': case 'X': return 8;
	}
	return 16;
}


int Jsw64Room::getRoomLength() const
{
	switch(m_game->getVariant())
	{
		case 'V': case 'W': return 512;
	}
	return 1024;
}


jswByte Jsw64Room::getSolarAttr()
{
	return m_data[0xDD];
}

void Jsw64Room::getSolarPos(int *x, int *y)
{
	if (x) *x = m_data[0xDC] & 0x1F;
	if (y) *y = m_data[0xDC] >> 5;
}

void Jsw64Room::setSolarPos(int x, int y)
{
	if (x >  31) x = 31;
	if (x <= 0)  x = 0;
	if (y >  7)  y = 7;
	if (y <= 0)  y = 0;;

	m_data[0xDC] = (y << 5) | x;
}


void Jsw64Room::getCells(Cell *c)
{
	int n, max;
	jswByte *base;
	bool globals = false;

	switch(m_game->getVariant())
	{
		case 'V': case 'X':
			base = m_data + 0x6E;
			max = 8;
			break;

		case 'W': case 'Y': case 'Z':
			base = m_data + 0x41;
			max = 13;
			globals = true;
			break;

		case '[':
			base = m_data + 0x26;
			max = 16;
			break;
	}
	for (n = 0; n < max; n++)
	{
		memcpy(c[n].pattern, base, 9);
		base += 9;
		c[n].behaviour = getCellBehaviour(n);
	}

	if (globals)	// Global cell patterns
	{
		base = m_game->getMem()->memoryAt(0xf50f, 7);
		max = 3;
		for (n = 0; n < max; n++)
		{
			memcpy(c[n+13].pattern, base, 9);
			base += 9;
			c[n+13].behaviour = getCellBehaviour(n+13);
		}
	}
}


void Jsw64Room::setCells(Cell *c)
{
	int n, max;
	jswByte *base;
	bool globals = false;

	switch(m_game->getVariant())
	{
		case 'V': case 'X':
			base = m_data + 0x6E;
			max = 8;
			break;

		case 'W': case 'Y': case 'Z':
			base = m_data + 0x41;
			max = 13;
			globals = true;
			break;

		case '[':
			base = m_data + 0x26;
			max = 16;
			break;
	}
	for (n = 0; n < max; n++)
	{
		memcpy(base, c[n].pattern, 9);
		base += 9;
		setCellBehaviour(n, c[n].behaviour);
	}

	if (globals)	// Global cell patterns
	{
		base = m_game->getMem()->memoryAt(0xf50f, 7);
		max = 3;
		for (n = 0; n < max; n++)
		{
			memcpy(base, c[n+13].pattern, 9);
			base += 9;
			setCellBehaviour(n + 13, c[n+13].behaviour);
		}
	}
}


static char no_patch[] =
{
	0xA0, 0x86, 0x9F, 0x86, 0x9F, 0x86
};

void Jsw64Room::clear()
{
	int len;

	switch(m_game->getVariant())
	{
		case 'V':
		case 'W':	len = 512; break;
		case 'X':
		case 'Y':	
		case 'Z':
		case '[':	len = 1024; break;
	}
	memset(m_data,	  0, len);
	memset(m_data + 0xB6, ' ', 32);
	m_data[0xDB] = 0x80;
	m_data[0xDE] = 0x38;	/* White Willy */
	m_data[0xDF] = 0xFF;
	m_data[0xE0] = 0xFF;
	memcpy(m_data + 0xF9, no_patch, 6);
	m_data[0] = 0xFF;	// 2.1.6 End of grauniads
}


void Jsw64Room::mirrorH()
{
	int y, x;
	jswByte *bits = getBits();

	for (y = 0; y < 16; y++)
	{
		for (x = 0; x < 16; x++)
		{
			int c1 = getCellAt(x, y);
			int c2 = getCellAt(31 - x, y);
			setCell(x, y, c2);
			setCell(31 - x, y, c1);
		}
	}
}


void Jsw64Room::mirrorV()
{
	int y, x;
	jswByte *bits = getBits();

	for (y = 0; y < 8; y++)
	{
		for (x = 0; x < 32; x++)
		{
			int c1 = getCellAt(x, y);
			int c2 = getCellAt(x, 15 - y);
			setCell(x, y, c2);
			setCell(x, 15 - y, c1);
		}

	}
}


void Jsw64Room::getRoomMemMap(char *buf)
{
	int len, boffs, blen, maxspr;
	int portoff, guardoff, gtoff;
	jswByte *gtable;
	int n;

	switch(m_game->getVariant())
	{
		case 'V': 
			len   = 512;
			boffs = 0x140;
			blen  = 0xc0;
			memset(buf, RMM_FREE, len);
			memset(buf + 0x69, RMM_INUSE, 5);
			memset(buf + 0x6E, RMM_CELL, 72);
			break;	
		case 'W': 
			len = 512;
			boffs = 0x100;
			blen  = 0x100;
			memset(buf, RMM_FREE, len);
			memset(buf + 0x41, RMM_CELL, 117);
			break;
		case 'X': 
			len   = 1024;
			boffs = 0x340;
			blen  = 0xc0;
			memset(buf, RMM_FREE, len);
			memset(buf + 0x69, RMM_INUSE, 5);
			memset(buf + 0x6E, RMM_CELL, 72);
			break;	
		case 'Y': 
			len = 1024;
			boffs = 0x300;
			blen  = 0x100;
			memset(buf, RMM_FREE, len);
			memset(buf + 0x41, RMM_CELL, 117);
			break;
		case 'Z': 
			len = 1024;
			boffs = 0x200;
			blen  = 0x200;
			memset(buf, RMM_FREE, len);
			memset(buf + 0x41, RMM_CELL, 117);
			break;
		case '[': 
			len = 1024;
			boffs = 0x300;
			blen  = 0x100;
			memset(buf, RMM_FREE, len);
			memset(buf + 0x21, RMM_INUSE, 5);
			memset(buf + 0x26, RMM_CELL, 144);
			break;
		default:
			buf[0] = 0;
			return;
	}
	maxspr = 0x80 + (len / 256);
	portoff  = m_data[0xEE] + 256 * m_data[0xEF];
	guardoff = m_data[0xDA] + 256 * m_data[0xDB];
	buf[len] = 0;
	memset(buf + boffs, RMM_BITMAP, blen);
	memset(buf + 0xB6, RMM_TITLE, 32);
	memset(buf + 0xD6, RMM_INUSE, 42);
	if (portoff >= 0x8000 && portoff <= (0x8000 + len - 32))
	{
		memset(buf + (portoff - 0x8000), 's', 32);
	}
	if (guardoff < 0x8000 || guardoff >= (0x8000 + len)) 
	{
		gtable = m_game->getMem()->memoryAt(guardoff);
	}
	else	
	{
		gtoff = guardoff - 0x8000;
		gtable = m_data + gtoff;
		n = 8 * getMaxGuards();
		while (n >= 0)
		{
			buf[n + gtoff] = RMM_GUARDIAN;
			--n;
		}
	}
	n = 0;
	while (gtable[n] != 0xFF)
	{
		int type = gtable[n] & 0x0F;
		if (type == 8) type = gtable[n];
		switch (type)
		{
			case 1: case 2: case 5: case 6: case 7:
			case 9: case 10: case 13: case 14: case 15:
			case 0x08: case 0x18: case 0x28: case 0x38:
			case 0x48: case 0x98:

			if (gtable[n + 5] >= 0x80 && gtable[n + 5] < maxspr)
			{
				memset(buf + 256 * (gtable[n+5] - 0x80),
					RMM_SPRITE, 256);
			}
		}
		n += 8;
	}
}


void Jsw64Room::setGuardianTable(int offset)
{
	m_data[0xDA] = offset & 0xFF;
	m_data[0xDB] = (offset >> 8) & 0xFF;
}


void Jsw64Room::getSprites(int *list, int granularity)
{
	int n, offset = 0;
	bool okay;

	char buf[1025];
	getRoomMemMap(buf);

	for (offset = 0; buf[offset]; offset++)
	{
		okay = true;
		for (n = 0; n < granularity; n++) 
		{
			if (buf[offset + n] != RMM_SPRITE && 
			    buf[offset + n] != RMM_FREE)
			{
				okay = false;
				break;
			}
		}
		if (!okay) continue;
		// We have a buffer of at least 'granularity' bytes
		list[0] = offset + 0x8000;
		++list;
		offset += (granularity - 1);
	}
	list[0] = 0;
}


void Jsw64Room::setPortalPos(int x, int y)
{
	m_data[0xF0] = x + ((y << 5) & 0xE0);
	m_data[0xF1] = 0x5C | ((y >> 3) & 1);
	m_data[0xF2] = x + ((y << 5) & 0xE0);
	m_data[0xF3] = 0x60 | (y & 8);
	Room::setPortalPos(x, y);
}

jswByte *Jsw64Room::getExits()
{
	return m_data + 0xE9;
}


jswByte Jsw64Room::getBorder()
{
	return m_data[0xDE] & 7;
}


void Jsw64Room::setBorder(int bdr)
{
	m_data[0xDE] &= ~7;
	m_data[0xDE] |= (bdr & 7);
}


void Jsw64Room::setItemBitmap(jswByte *data)
{
	memcpy(m_data + 0xE1, data, 8);
}


void Jsw64Room::clearItems()
{
	int n, max, rm, x, y;

	for (n = m_game->getItemCount(); n <= 0xFF; n++)
	{
		if (!m_game->itemPos(n, &rm, &x, &y))
		{
			if (rm == m_roomNumber) 
			{
				m_game->itemDelete(n);
				--n;
			}
		}
	}
}

bool Jsw64Room::importPortal(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent,
		                int roomMap[256])
{
//	char buf[33];

//	getTitle(buf);
//	fprintf(stderr, "Importing portal for room: %s\n", buf);

	if (!Room::importPortal(doc, ns, parent, roomMap)) return false;
	return true;
}



void Jsw64Room::addItem(int x, int y)
{
	m_game->itemInsert(m_roomNumber, x, y * 8);
}


int Jsw64Room::roomLen()
{
	switch(m_game->getVariant())
	{
		case 'V': case 'W': return 0x200;
		default:	    return 0x400;
	}
}

int Jsw64Room::encodeSpriteID(int id)
{
	int rv;

	if (id >= 0x8000 && id <= (0x8000 + roomLen()))
	{
		rv = SPRITE((m_roomNumber << 16) | id);
	}
	else rv = SPRITE(id);

	return rv;
}


void Jsw64Room::addRefs()
{
	jswByte *buf, *g;
	int n, ns, gt, frame;
	int sprite, override;
	Portal p;

	if (getPortal(p))
	{
		m_game->addRef(SPRITE(p.spriteID));	
	}

	buf = getGuardianBuffer();
	for (n = 0; n < 8; n++)
	{
		g = buf + n*8;
		if (g[0] == 0xFF) break;
		int type = g[0] & 0x0F;
		if (type == 8) type = g[0];
		switch(type)
		{
			case 1: case 2:  case 5:  case 6:  case 7:
			case 9: case 10: case 13: case 14: case 15:
			case 0x18: case 0x28: case 0x38: case 0x98:
				for (frame = 0; frame < 8; frame++)
				{
					ns = (frame << 5) & 0xE0;

					ns &= (g[1] & 0xE0);
					ns |= (g[2] & 0xE0);
					ns += 256 * g[5];
					m_game->addRef(encodeSpriteID(ns));
				}

				break;
			case 0x08: 
				for (frame = 0; frame < 8; frame++)
				{
					ns = (frame << 5) & 0xE0;

					ns += 256 * g[5];
					m_game->addRef(encodeSpriteID(ns));
				}
				break;

		}
	}
// If sprite overridden, pick that up.
	override = m_game->allowSpriteOverride();
	sprite   = m_game->decodeWillySprite(m_roomNumber, m_data[override]);
	if (m_data[override] & 0x80) 
	{
		int frame;

		for (frame = 0; frame < 8; frame++)
		{
			m_game->addRef(encodeSpriteID(sprite*256 + 32*frame));
		}
	}

}



bool Jsw64Room::exportGuardians(xmlDocPtr doc, xmlNsPtr ns, 
		xmlNodePtr grandparent)
{
	int n;
	char data[40];
	int tableBase = m_data[0xDA] + 256 * m_data[0xDB];
	jswByte *buf = getGuardianBuffer();

	xmlNodePtr parent = xmlNewNode(ns, (xmlChar *)"guardians");
	xmlSetProp(parent, (xmlChar *)"type", (xmlChar *)"jsw");

	while(*buf != 0xFF)
	{
		if (!exportJswGuardian(doc, ns, parent, buf, tableBase)) 
			return false;
		buf += 8;
	}
	xmlAddChild(grandparent, parent);
	return true;
}

bool Jsw64Room::importGuardians(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent)
{
	xmlNodePtr cur;
	jswByte *buf;
	int n, ng, slot, gc; 
	char variant;
	int bufmax, bufmax2, bufbase, tableBase;

	char *type = (char *)xmlGetProp(parent, (xmlChar *)"type");

	bufbase = 0;
	variant = m_game->getVariant();
	if (type && !strcmp(type, "jsw"))
	{
/* Straight import of JSW guardians. We leave the other types for 
 * mature consideration. */

/* Start by counting the number of guardians */
		gc = 0;
		for (cur = parent->xmlChildrenNode; cur; cur = cur->next)
		{
			if (cur->ns != ns) continue;
			++gc;
		}
// We can't omit this. Even if there are no guardians we need
// to import a blank to zap any existing data using the buffer.
// if (!gc) return true;
		buf = new jswByte[gc * 8 + 1];

// Will these guardians fit in the room's standard guardian buffer?
		switch(variant)
		{
			case 'V': bufmax = 0x69; bufmax2 = 0x40; break;
			case 'W': bufmax = 0x41; bufmax2 = 0; break;
			case 'X': bufmax = 0x69; bufmax2 = 0x240; break;
			case 'Y': bufmax = 0x41; bufmax2 = 0x200; break;
			case 'Z': bufmax = 0x41; bufmax2 = 0x100; break;
			case '[': bufmax = 0x21; bufmax2 = 0x200; break;
		}
		if ((gc*8+1) <= bufmax) 	bufbase = 0;
		else if ((gc*8+1) <= bufmax2)	bufbase = 0x100;
		else if (bufmax > bufmax2)	bufbase = 0;
		else 				bufbase = 0x100;

		tableBase = bufbase + 0x8000;
		n = 0;
// Import the guardians.
		for (cur = parent->xmlChildrenNode; cur; cur = cur->next)
		{
			if (cur->ns != ns) continue;
			importJswGuardian(doc, ns, cur, buf + 8*n, m_game,
					tableBase);
			++n;
//			printf("Buffer: ");
//			for (int q = 0; q < 8*n; q++) printf("%02x ", buf[q]);
//			printf("\n");
		}
		buf[8*n] = 0xFF;	// End of guardians

//		printf("Buffer: ");
//		for (int q = 0; q <= 8*n; q++) printf("%02x ", buf[q]);
//		printf("\n");

		if (bufbase == 0x100) bufmax = bufmax2;
		while ((gc*8+1) > bufmax) --gc;
		memcpy(m_data + bufbase, buf, (gc*8+1));
		delete [] buf;
	}
	xmlFree(type);
	return true;
}


bool Jsw64Room::exportHints(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent)
{
	char buf[30];
	int override, sprite;

	xmlNodePtr hint;

// Willy colour
	if ((m_data[0xDE] & 0x38) != 0x38)
	{
		hint = xmlNewNode(ns, (xmlChar *)"hint");
		xmlSetProp(hint, (xmlChar *)"type", (xmlChar *)"willy");
		sprintf(buf, "%d", (m_data[0xDE] >> 3) & 7);
		xmlSetProp(hint, (xmlChar *)"ink", (xmlChar *)buf);
		xmlAddChild(parent, hint);
	}
// Final barrier
	if (m_data[0xD6] & 0xFC)
	{
		hint = xmlNewNode(ns, (xmlChar *)"hint");
		xmlSetProp(hint, (xmlChar *)"type", (xmlChar *)"finalbarrier");

		if ((m_data[0xD6] >> 2) & 7)
		{
			sprintf(buf, "%d", (m_data[0xD6] >> 2) & 7);
			xmlSetProp(hint, (xmlChar *)"top", (xmlChar *)buf);
		}
		if ((m_data[0xD6] >> 5) & 7)
		{
			sprintf(buf, "%d", (m_data[0xD6] >> 5) & 7);
			xmlSetProp(hint, (xmlChar *)"bottom", (xmlChar *)buf);
		}
		xmlAddChild(parent, hint);
	}

// Export conveyor animation
	if (m_data[0xD9])
	{
		hint = xmlNewNode(ns, (xmlChar *)"hint");
		xmlSetProp(hint, (xmlChar *)"type", (xmlChar *)"conveyor");
		sprintf(buf, "%d", m_data[0xD6] & 3);
		xmlSetProp(hint, (xmlChar *)"direction", (xmlChar *)buf);
		sprintf(buf, "%d", m_data[0xD7] & 31);
		xmlSetProp(hint, (xmlChar *)"x", (xmlChar *)buf);
		sprintf(buf, "%d", (m_data[0xD7] >> 5) + ((m_data[0xD8] & 1) << 3));
		xmlSetProp(hint, (xmlChar *)"y", (xmlChar *)buf);
		sprintf(buf, "%d", m_data[0xD9]);
		xmlSetProp(hint, (xmlChar *)"length", (xmlChar *)buf);
		xmlAddChild(parent, hint);
	}
// Export solar power
	if (m_data[0xDD])
	{
		hint = xmlNewNode(ns, (xmlChar *)"hint");
		xmlSetProp(hint, (xmlChar *)"type", (xmlChar *)"solar");
		sprintf(buf, "%d", m_data[0xDC] & 31);
		xmlSetProp(hint, (xmlChar *)"x", (xmlChar *)buf);
		sprintf(buf, "%d", (m_data[0xDC] >> 5) );
		xmlSetProp(hint, (xmlChar *)"y", (xmlChar *)buf);
		sprintf(buf, "%d", (m_data[0xDD] & 7) );
		xmlSetProp(hint, (xmlChar *)"ink", (xmlChar *)buf);
		sprintf(buf, "%d", (m_data[0xDD] >> 3) & 7 );
		xmlSetProp(hint, (xmlChar *)"paper", (xmlChar *)buf);
		if (m_data[0xDD] & 0x40)
			xmlSetProp(hint, (xmlChar *)"bright", (xmlChar *)"1");
		if (m_data[0xDD] & 0x80)
			xmlSetProp(hint, (xmlChar *)"flash", (xmlChar *)"1");
		xmlAddChild(parent, hint);
	}
	// Escalator
	if (m_data[0xFF] & 4)
	{
		hint = xmlNewNode(ns, (xmlChar *)"hint");
		xmlSetProp(hint, (xmlChar *)"type", (xmlChar *)"escalator");
		if (m_data[0xFF] & 2)
			xmlSetProp(hint, (xmlChar *)"direction", (xmlChar *)"down");
		else	xmlSetProp(hint, (xmlChar *)"direction", (xmlChar *)"up");
		xmlAddChild(parent, hint);	
	}
	if (m_data[0xFF] & 1)
	{
		hint = xmlNewNode(ns, (xmlChar *)"hint");
		xmlSetProp(hint, (xmlChar *)"type", (xmlChar *)"nokamikaze");
		xmlSetProp(hint, (xmlChar *)"enabled", (xmlChar *)"1");
		xmlAddChild(parent, hint);
	}
	if (m_data[0xFF] & 8)
	{
		hint = xmlNewNode(ns, (xmlChar *)"hint");
		xmlSetProp(hint, (xmlChar *)"type", (xmlChar *)"fallanyheight");
		xmlSetProp(hint, (xmlChar *)"enabled", (xmlChar *)"1");
		xmlAddChild(parent, hint);
	}
	if (m_data[0xFF] & 0x80)
	{
		hint = xmlNewNode(ns, (xmlChar *)"hint");
		xmlSetProp(hint, (xmlChar *)"type", (xmlChar *)"bonusroom");
		xmlSetProp(hint, (xmlChar *)"enabled", (xmlChar *)"1");
		xmlAddChild(parent, hint);
	}
	if (m_data[0xDE] & 0x80)
	{
		hint = xmlNewNode(ns, (xmlChar *)"hint");
		xmlSetProp(hint, (xmlChar *)"type", (xmlChar *)"superjump");
		xmlSetProp(hint, (xmlChar *)"enabled", (xmlChar *)"1");
		xmlAddChild(parent, hint);
	}
	if (m_data[0xE0] != 0xFF)
	{
		int air = (m_data[0xDF] - 37) * 6;
		if      (m_data[0xE0] &  4) air += 5;
		else if (m_data[0xE0] &  8) air += 4;
		else if (m_data[0xE0] & 16) air += 3;
		else if (m_data[0xE0] & 32) air += 2;
		else if (m_data[0xE0] & 64) air += 1;
		hint = xmlNewNode(ns, (xmlChar *)"hint");
		xmlSetProp(hint, (xmlChar *)"type", (xmlChar *)"air");
		sprintf(buf, "%d", air );
		xmlSetProp(hint, (xmlChar *)"amount", (xmlChar *)buf);
		xmlAddChild(parent, hint);
	}
	override = m_game->allowSpriteOverride();
	sprite   = m_game->decodeWillySprite(m_roomNumber, m_data[override]);
	if (m_data[0xED] & 0x80)
	{
		int frame;

		hint = xmlNewNode(ns, (xmlChar *)"hint");
		xmlSetProp(hint, (xmlChar *)"type", (xmlChar *)"sprite");
		for (frame = 0; frame < 8; frame++)
		{
			xmlNodePtr spr = xmlNewNode(ns, (xmlChar *)"spriteref");
			sprintf(buf, "%d", frame);
			xmlSetProp(spr, (xmlChar *)"frame", (xmlChar *)buf);
			sprintf(buf, "%04x", sprite*256 + 32*frame);
			xmlSetProp(spr, (xmlChar *)"xref", (xmlChar *)buf);
			xmlAddChild(hint, spr);
		}
		xmlAddChild(parent, hint);
	}
	return Room::exportHints(doc, ns, parent);
}




bool Jsw64Room::importHints(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent)
{
	xmlNodePtr cur;
	bool haveConveyor, haveSuper, haveSolar, haveEscalator, haveNoKami,
	haveFallAny, haveBonus, haveAir, haveColour, haveSprite, haveFinal;
//	char buf[33];

//	getTitle(buf);
//	fprintf(stderr, "Importing hints for room: %s\n", buf);

	haveConveyor = haveSuper = haveSolar = haveEscalator = haveNoKami = 
	haveFallAny = haveBonus = haveAir = haveColour = haveSprite = 
	haveFinal = false;

	for (cur = parent->xmlChildrenNode; cur; cur = cur->next)
	{
		if (cur->ns != ns) continue;
		if (strcmp((char *)cur->name, "hint")) continue;

		char *hintName = (char *)xmlGetProp(cur, (xmlChar *)"type");
		if (!hintName) continue;

		if (!strcmp(hintName, "superjump"))
		{
			char *attr;
			attr = (char *)xmlGetProp(cur, (xmlChar *)"enabled");
			if (attr && atoi(attr) && m_game->allowSuperJump())
			{
				m_data[0xDE] |= 0x80;
				haveSuper = true;
			}
			if (attr) xmlFree(attr);
		}

		if (!strcmp(hintName, "conveyor"))
		{
			char *attr;

			haveConveyor = true;
			attr = (char *)xmlGetProp(cur, (xmlChar *)"direction");
			if (attr) 
			{
				m_data[0xD6] &= ~3;
				m_data[0xD6] |= atoi(attr) & 3;
				xmlFree(attr);
			}
			attr = (char *)xmlGetProp(cur, (xmlChar *)"x");
			if (attr) 
			{
				m_data[0xD7] = (m_data[0xD7] & 0xE0) | 
							(atoi(attr) & 0x1F);
				xmlFree(attr);
			}
			attr = (char *)xmlGetProp(cur, (xmlChar *)"y");
			if (attr) 
			{
				m_data[0xD7] = (m_data[0xD7] & 0x1F) | 
					((atoi(attr) & 7) << 5);
				m_data[0xD8] = (m_data[0xD8] & 0xFE) | 
					((atoi(attr) & 8) >> 3);
				xmlFree(attr);
			}
			attr = (char *)xmlGetProp(cur, (xmlChar *)"length");
			if (attr) 
			{
				m_data[0xD9] = atoi(attr);
				xmlFree(attr);
			}
		}
		if (!strcmp(hintName, "escalator"))
		{
			haveEscalator = true;
			char * attr;

			attr = (char *)xmlGetProp(cur, (xmlChar *)"direction");
			if (attr)
			{
				if (!strcmp(attr, "up") || 
				    !strcmp(attr, "Up") ||
				    !strcmp(attr, "UP")) m_data[0xFF] &= ~2;
				else			 m_data[0xFF] |=  2;
				m_data[0xFF] |= 4;
				xmlFree(attr);
			}
		}
		if (!strcmp(hintName, "solar"))
		{
			haveSolar = true;

			char *attr;
			int x = 0, y = 0, i = 6, p = 7, f = 0, b = 1;

			attr = (char *)xmlGetProp(cur, (xmlChar *)"x");
			if (attr) { x = atoi(attr); xmlFree(attr); }
			attr = (char *)xmlGetProp(cur, (xmlChar *)"y");
			if (attr) { y = atoi(attr); xmlFree(attr); }
			attr = (char *)xmlGetProp(cur, (xmlChar *)"ink");
			if (attr) { i = atoi(attr); xmlFree(attr); }
			attr = (char *)xmlGetProp(cur, (xmlChar *)"paper");
			if (attr) { p = atoi(attr); xmlFree(attr); }
			attr = (char *)xmlGetProp(cur, (xmlChar *)"bright");
			if (attr) { b = atoi(attr); xmlFree(attr); }
			attr = (char *)xmlGetProp(cur, (xmlChar *)"flash");
			if (attr) { f = atoi(attr); xmlFree(attr); }
			m_data[0xDC] = (x & 31) + 32 * (y & 7);
			m_data[0xDD] = (i & 7) + 8 * (p & 7) + 
					(b ? 64 : 0) + (f ? 128 : 0);
		}
		if (!strcmp(hintName, "nokamikaze"))
		{
			char *attr;
		
			haveNoKami = true;	
			attr = (char *)xmlGetProp(cur, (xmlChar *)"enabled");
			if (attr && atoi(attr)) m_data[0xFF] |= 1;
			else			m_data[0xFF] &= ~1;
			if (attr) xmlFree(attr);
		}
		if (!strcmp(hintName, "fallanyheight"))
		{
			char *attr;
		
			haveFallAny = true;	
			attr = (char *)xmlGetProp(cur, (xmlChar *)"enabled");
			if (attr && atoi(attr)) m_data[0xFF] |= 8;
			else			m_data[0xFF] &= ~8;
			if (attr) xmlFree(attr);
		}
		if (!strcmp(hintName, "bonusroom"))
		{
			char *attr;
		
			haveBonus = true;	
			attr = (char *)xmlGetProp(cur, (xmlChar *)"enabled");
			if (attr && atoi(attr)) m_data[0xFF] |= 0x80;
			else			m_data[0xFF] &= ~0x80;
			if (attr) xmlFree(attr);
		}
		if (!strcmp(hintName, "air"))
		{	
			char *attr;
		
			haveAir = true;	
			attr = (char *)xmlGetProp(cur, (xmlChar *)"amount");
			if (attr && atoi(attr) < 162)
			{
				m_data[0xDF] = (atoi(attr) / 6) + 37;
				switch(atoi(attr) % 6)
				{
					case 5: m_data[0xE0] = 0xFC; break;
					case 4: m_data[0xE0] = 0xF8; break;
					case 3: m_data[0xE0] = 0xF0; break;
					case 2: m_data[0xE0] = 0xE0; break;
					case 1: m_data[0xE0] = 0xC0; break;
					case 0: m_data[0xE0] = 0x80; break;
				}
			}
			else	m_data[0xE0] = m_data[0xDF] = 0xFF;
			if (attr) xmlFree(attr);
		}
		/* Willy's ink */
		if (!strcmp(hintName, "willy"))
		{	
			char *attr;
	
			haveColour = true;
			attr = (char *)xmlGetProp(cur, (xmlChar *)"ink");
			if (attr)
			{
				m_data[0xDE] &= ~0x38;
				m_data[0xDE] |= ((atoi(attr) & 7) << 3);
				xmlFree(attr);
			}
		}
		/* Willy's sprite */
		if (!strcmp(hintName, "sprite"))
		{
			char *prop;
			haveSprite = true;
			xmlNodePtr spriteref;
			for (spriteref = cur->xmlChildrenNode; 
					spriteref != NULL; 
					spriteref = spriteref->next)
			{
				if (spriteref->ns != ns) continue;
				if (strcmp((char *)spriteref->name, 
						"spriteref")) continue;
				prop = (char *)xmlGetProp(spriteref, (xmlChar *)"xref");
				if (prop)
				{
					int sprPage;
					sscanf(prop, "%x", &sprPage);
					sprPage = m_game->lookupSpriteRef(sprPage);	
					m_data[0xED] = sprPage / 256;
					xmlFree(prop);
				}
			}
		}
		/* Final Barrier options */
		if (!strcmp(hintName, "finalbarrier"))
		{
			haveFinal = true;
			char *attr = (char *)xmlGetProp(cur, (xmlChar *)"top");
			if (attr)
			{
				m_data[0xD6] &= ~0x1C;
				m_data[0xD6] |= ((atoi(attr) & 7) << 2);
				xmlFree(attr);
			}
			attr = (char *)xmlGetProp(cur, (xmlChar *)"bottom");
			if (attr)
			{
				m_data[0xD6] &= ~0xE0;
				m_data[0xD6] |= ((atoi(attr) & 7) << 5);
				xmlFree(attr);
			}
		}
		xmlFree(hintName);
	}
	/* For those hints not supplied, reset to defaults */
	if (!haveSuper)     m_data[0xDE] &= 0x7F;
	if (!haveColour)    m_data[0xDE] |= 0x38;
	if (!haveSprite)    m_data[0xED] = 0;
	if (!haveConveyor)  
	{
	//	fprintf(stderr, "No conveyor, deleting it\n");
		m_data[0xD6] &= ~3;
		m_data[0xD9] = 0;
	}
	if (!haveFinal)     m_data[0xD6] &= 3;
	if (!haveSolar)     m_data[0xDD] = 0;
	if (!haveEscalator) m_data[0xFF] &= ~4;
	if (!haveNoKami)    m_data[0xFF] &= ~1;
	if (!haveFallAny)   m_data[0xFF] &= ~8;
	if (!haveBonus)     m_data[0xFF] &= ~0x80;
	if (!haveAir)	    m_data[0xE0] = m_data[0xDF] = 0xFF;
	return Room::importHints(doc, ns, parent);
}

void Jsw64Room::clearHints()
{
	m_data[0xD6] = 0;
	m_data[0xD9] = 0;
	m_data[0xDD] = 0;
	m_data[0xFF] &= ~(0x8D);
	m_data[0xDE] &= 0x7F;
	m_data[0xDE] |= 0x38;
	m_data[0xDF] = m_data[0xE0] = 0xFF;
	m_data[0xED] = 0;
}

void Jsw64Room::createCellPatterns()
{
        char var = m_game->getVariant();
        int max;

        Room::createCellPatterns();

	max = cellCount() * 2;
        m_cellBitmaps[max]->getSurface()->fillReplace(
                        0, 0, 16, 16, ZX_BLACK, ZX_TRANS);
        m_cellBitmaps[max]->makeTransparent();
}


int Jsw64Room::encodeCellNumber(int c)
{
	if (getCellSize() == 8)
	{
		jswByte *pat = getCellPattern(c);
		if (pat) return pat[0];
	}
	return c;
}

bool Jsw64Room::exportExtraCells(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent)
{
	jswByte *cp[16];
	jswByte pattern[9];
	int n, attr, used;

	if (getCellSize() == 8)
	{
		/* Get the 16 cell patterns for this room */
		for (n = 0; n < 16; n++) cp[n] = getCellPattern(n);
		/* And base our extra cells on type 1 */
		memcpy(pattern, cp[1], 9);
		/* See which attributes are (a) nonstandard and (b) 
		 * used in this screen */
		for (attr = 0; attr < 256; attr++)
		{
			used = 0;
			/* Compare against the 16 we've already done */
			for (n = 0; n < 16; n++) if (cp[n][0] == attr) 
			{
				used = -1;
				break;
			}
			/* If used < 0, it's one of the ones that's been done */
			if (used < 0)  continue;
			/* Is it actually used onscreen? */
			for (n = 0; n < 512; n++)
			{
				if (m_data[0x200 + n] == attr) used++;
			}
			if (used == 0) continue;
			/* Yes. So export it. */
			pattern[0] = attr;
			if (!exportCell(doc, ns, parent, attr, CB_WATER, pattern)) 
				return false;
		}
	}
	return true;
}


void Jsw64Room::importCell(int x, int y, int attr, int type, int behaviour)
{
	if (getCellSize() == 8)
	{
		m_data[y*32 + x + 0x200] = attr;
	}
	else	Room::importCell(x,y,attr,type,behaviour);
}



int Jsw64Room::addLocalSpritePage(jswByte *bits)
{
	int max, id;
	int page = 0x82;
	int n, m, type;
	char buf[1025];

	getRoomMemMap(buf);
	max = getRoomLength();
	// Scan the room memory map to see if there's a sprite page 
	// available (ie, a complete run of 256 bytes marked RMM_FREE, on 
	// a page boundary).
	for (n = 0; n < max; n += 256)
	{
		for (m = 0; m < 256; m++)
		{
			if (m_spriteUsage[n + m]) break;
			if (buf[n + m] != RMM_FREE) break;
		}
		if (m == 256) 
		{
			memcpy(m_data + n, bits, 256);
			for (m = 0; m < 256; m++)
				m_spriteUsage[m + n] = true;
			id = (m_roomNumber << 16) | 0x8000 | n;
			m_game->addRef(SPRITE(id));
			return id;
		}
	}
	return -1;
}


int Jsw64Room::addLocalSprite(jswByte *bits)
{
	int page = 0x82;
	int n, m, type, max;
	int id;
	char buf[1025];

	getRoomMemMap(buf);
	max = getRoomLength() - 31;
	// Scan for a 32-byte long unused area
	for (n = 0; n < max; n++)
	{
		for (m = 0; m < 32; m++)
		{
			if (m_spriteUsage[n + m]) break;
			if (buf[n + m] != RMM_FREE) break;
		}
		if (m == 32) 
		{
			memcpy(m_data + n, bits, 32);
			for (m = 0; m < 32; m++)
				m_spriteUsage[m + n] = true;
			id = (m_roomNumber << 16) | 0x8000 | n;
			m_game->addRef(SPRITE(id));
			return id;
		}
	}
	return -1;
}

void Jsw64Room::prepareForSpriteImport(void)
{
	memset(m_spriteUsage, 0, sizeof(m_spriteUsage));
}

const char *Jsw64Room::supportedGuardianTypes()
{
	return "jsw";
}

