/************************************************************************

    JSWED 2.3.0 - Editor for Jet Set Willy and derivatives

    Copyright (C) 2000-2001,2004  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*************************************************************************/

#include "jswed.hxx"
#include "jsw2game.hxx"
#include "jsw2guard.hxx"

Jsw2Arrow::Jsw2Arrow(jswByte *a, SpectrumMemory *mem)
{
	m_mem = mem;
	m_arrow = a;
	memcpy(m_orig,  a, 2);
        m_bmp      = newVideoBitmap(32, 32);
        m_bmpSmall = newVideoBitmap(16, 16);
	reset();
}



Jsw2Arrow::~Jsw2Arrow()
{
	delete m_bmp;
	delete m_bmpSmall;
}

void Jsw2Arrow::constructSprites()
{
	VideoBitmap *spb, *spbs;
	int fr, fg;
	jswByte *spriteBase;

	fg = ZX_WHITE;	
	if (m_arrow[1] & 0x80)	spriteBase = m_mem->memoryAt(0xEB01);
	else			spriteBase = m_mem->memoryAt(0xEB21);

	getBounds(spriteBase, &m_w, &m_h);

	delete m_bmp;
	delete m_bmpSmall;	
	m_bmp      = bitmapFromSprite(16, 16, spriteBase, 1, fg, 0);
	m_bmpSmall = bitmapFromSprite(16, 16, spriteBase, 0, fg, 0);
}


void Jsw2Arrow::draw(VideoSurface *s)
{
	int x;

	x = 8 * m_x;

	if (x >= 0 && x < 256)  
	{
		m_bmp->toSurface(s, 2 * x, 2 * m_y, 0, 0, 2*m_w, 2*m_h);
	}
}


void Jsw2Arrow::move()
{
	m_x += m_xstep;
}


void Jsw2Arrow::drawFrames(VideoSurface *s, int x, int y)
{
	int n;

	for (n = 0; n < 8; n++)
	{
		m_bmp->toSurface(s, x + 64 * n, y, 0, 0, 32, 32);
	}
}

void Jsw2Arrow::drawThumb(VideoSurface *s, int cx, int cy)
{
	if (m_w <= 8 && m_h <= 8)
		m_bmp->toSurface(s, cx, cy, 0, 0, 16, 16);
	else	m_bmpSmall->toSurface(s, cx, cy, 0, 0, 16, 16);
}


void Jsw2Arrow::drawBounds(VideoSurface *s, int ink, int cx, int cy)
{
	int acy = (cy - 1)  * 2;
	int ach = 6;
	int y   = (m_y * 2) - 2;
	int h   = 6;

	if (y   < 0) {   h +=   y; y   = 0; }
	if (acy < 0) { ach += acy; acy = 0; }

	s->box(0, y, 512, h, ZX_WHITE);
	if (ink >= 0) s->box(0, acy, 512, ach, ink);
}	


void Jsw2Arrow::setPos(int x, int y)
{
	m_y = y;
	m_arrow[1] = (m_arrow[1] & 0x80) | (y & 0x7F);
}



void Jsw2Arrow::reset()
{
	constructSprites();

	m_x = m_arrow[0];
	m_y = m_arrow[1] & 0x7F;
	if (m_arrow[1] & 0x80)	m_xstep = -1;
	else			m_xstep = 1;
}

jswByte *Jsw2Arrow::getBase()
{
        return m_arrow;
}

