/************************************************************************

    JSWED 2.3.0 - Editor for Jet Set Willy and derivatives

    Copyright (C) 2000-2001,2004  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*************************************************************************/

#include "jswed.hxx"
#include "jsw2game.hxx"
#include "jsw2guard.hxx"
#include "jsw2arrowform.hxx"
#include "jswdirbox.hxx"

Jsw2ArrowForm::Jsw2ArrowForm(int roomNo, Jsw2Game *g,jswByte *guard,VideoBitmap *bg)
	 : VideoForm(CHAR_W * 3, CHAR_H * 4, CHAR_W * 34, CHAR_H * 22)
{
	m_roomNo = roomNo;
	m_bmBg   = bg;
	m_bmAnim = newVideoBitmap(512, 256); 
	m_game = g;
	m_adata = guard;
	m_ot = NULL;
	m_haveTick = 0;
	m_arrow = new Jsw2Arrow(guard, g->getMem());

        m_bnOk = new VideoButton(ZXE_OK,  m_x +      CHAR_W, 
					  m_y + m_h - 2 * CHAR_H, "  OK  ");
        m_bnCancel = new VideoButton(ZXE_CANCEL, m_x + m_w - 7 * CHAR_W, 
					m_y + m_h - 2 * CHAR_H, "Cancel");

	m_labelX = new VideoLabel(m_x + 13 * CHAR_W, m_y + 18 * CHAR_H,
					"X:");
	m_spinX = new VideoSpinControl(m_x + 15 * CHAR_W, 
				       m_y + 18 * CHAR_H, 
				       7 * CHAR_W, CHAR_H + 4);
	m_spinX->setValue(m_adata[0]);
	m_spinX->setLimits(0, 255);
	m_spinX->setListener(this);

	m_labelY = new VideoLabel(m_x + 24 * CHAR_W, m_y + 18 * CHAR_H,
					"Y:");
	m_spinY = new VideoSpinControl(m_x + 26 * CHAR_W, 
				       m_y + 18 * CHAR_H, 
				       7 * CHAR_W, CHAR_H + 4);
	m_spinY->setValue(m_adata[1]);
	m_spinY->setLimits(0, 127);
	m_spinY->setListener(this);

	m_cbDirection    = new JswDirCheckbox(m_x + 6*CHAR_W, m_y + 18 * CHAR_W);
	m_cbDirection->setListener(this);
	m_cbDirection->setCheck(!(m_adata[1] & 0x80));
	m_labelDirection = new VideoLabel(m_x + CHAR_W, m_y + 18 * CHAR_W, "L/R?");
}



Jsw2ArrowForm::~Jsw2ArrowForm()
{
	delete m_labelDirection;
	delete m_bmAnim;
	delete m_labelX;
	delete m_spinX;
	delete m_labelY;
	delete m_spinY;
	// nb: Delete guard *after* deleting controls; otherwise the
	// control destructors may cause callbacks that try to access
	// the guard.
	if (m_arrow) delete m_arrow;
}


void Jsw2ArrowForm::drawFixed(void)
{
        videoScreen->fillBox(m_x, m_y, m_w, CHAR_H, ZX_WHITE);
        videoScreen->fillBox(m_x, m_y, CHAR_W, m_h, ZX_WHITE);
        videoScreen->fillBox(m_x + m_w - CHAR_W, m_y, CHAR_W, m_h, ZX_WHITE);
        videoScreen->fillBox(m_x, m_y + 17 * CHAR_H, m_w, m_h - 17 * CHAR_H, ZX_WHITE); 
        videoScreen->box(m_x, m_y, m_w, m_h, ZX_BLACK);
        for (int n = 0; n < m_ccount; n++)
        {
                m_controls[n]->redraw();
                m_controls[n]->drawFocus(n == m_focus);
        }
}

void Jsw2ArrowForm::drawData(int tickonly)
{
	char hexBytes[80];

	if (m_bmBg) m_bmBg->toSurface(m_bmAnim->getSurface(), 0, 0);
	else	    m_bmAnim->getSurface()->fillBox(0, 0, 512, 256, ZX_BLACK);

	m_arrow->draw(m_bmAnim->getSurface());
	m_arrow->drawBounds(m_bmAnim->getSurface());

	m_bmAnim->toScreen(m_x + CHAR_W, m_y + CHAR_H);	

	strcpy(hexBytes, "As Hex:");
	for (int n = 0; n < 2; n++) 
	{
		sprintf(hexBytes + strlen(hexBytes), " %02x", m_adata[n]);
	}
	strcat(hexBytes, ".");
	
	videoScreen->drawSmallText(m_x +      CHAR_W,
                                    m_y + 17 * CHAR_H + (CHAR_SH / 2),
				    hexBytes, ZX_BLACK, ZX_WHITE);
	if (!tickonly)
	{	
		drawTsData();
	}
}


void Jsw2ArrowForm::redraw(void)
{
	// Don't call base class. 
	//        videoScreen->fillBox(m_x, m_y, m_w, m_h, ZX_WHITE);

	drawFixed();
	drawData(0);
}



void Jsw2ArrowForm::showChildren(int redraw)
{
	postContextMenu();
	addChild(m_cbDirection);
	addChild(m_labelX);
	addChild(m_spinX);
	addChild(m_labelY);
	addChild(m_spinY);
	addChild(m_labelDirection);
	addChild(m_bnOk);
	addChild(m_bnCancel);
	VideoForm::showChildren(redraw);
}

void Jsw2ArrowForm::hideChildren(void)
{
	m_spinY->setListener(NULL);
	m_spinX->setListener(NULL);
	removeChild(m_spinY);
	removeChild(m_labelY);
	removeChild(m_spinX);
	removeChild(m_labelX);
	removeChild(m_cbDirection);
	removeChild(m_labelDirection);
	removeChild(m_bnOk);
	removeChild(m_bnCancel);
	preContextMenu();
	VideoForm::hideChildren();
}



void Jsw2ArrowForm::preContextMenu()
{
	if (m_haveTick)
	{
		setTickListener(m_ot);
		m_haveTick = 0;
	}
}


void Jsw2ArrowForm::postContextMenu()
{
        if (!m_haveTick)
        {
                m_ot = setTickListener(this);
                m_haveTick = 1;
        }
}





void Jsw2ArrowForm::onTick(void)
{
	++m_tick;
	if ((m_tick % 5) == 0)
	{
		m_arrow->move();
		drawData(1);
	}
}






int  Jsw2ArrowForm::onCheckboxSelect(VideoCheckbox *c, int value)
{
	if (c == m_cbDirection)
	{
		m_adata[1] &= 0x7F;
		if (!value) m_adata[1] |= 0x80;
		m_arrow->reset();
		m_cbDirection->setCheck(!(m_adata[1] & 0x80));
		return ZXE_OK;	
	}
	return ZXE_CONTINUE;
}


void Jsw2ArrowForm::addChild(VideoControl *c)
{
	VideoForm::addChild(c);
	c->setVisible(1, 0);
}

void Jsw2ArrowForm::removeChild(VideoControl *c)
{
	c->setVisible(0, 0);	
	VideoForm::removeChild(c);
}

void Jsw2ArrowForm::drawTsData()
{
	// Memo to self: Give functions more self-documenting names.
	m_spinX->setValue(m_adata[0]);
	m_spinY->setValue(m_adata[1] & 0x7F);
	m_cbDirection->setCheck(!(m_adata[1] & 0x80));
}

void Jsw2ArrowForm::onEditChange(VideoControl *v)
{
	if (v == m_spinY || v == m_spinX)
	{
		m_adata[0] = m_spinX->getValue();
		m_adata[1] &= 0x80;
		m_adata[1] |= m_spinY->getValue();
		m_arrow->reset();
	}
}




