/************************************************************************

    JSWED 2.3.0 - Editor for Jet Set Willy and derivatives

    Copyright (C) 2000-2001,2004  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*************************************************************************/

#include "jswed.hxx"
#include "jsw2cellform.hxx"
#include "room.hxx"
#include "jsw2room.hxx"
#include "jsw2game.hxx"
#include "celllist.hxx"


static char *st_captions[] =
{
	"Air",
	"Water",
	"Earth",
	"Fire",
	"Ramp /",
	"Conveyor <<",
	"Item",
	"Ramp \\",
	"Conveyor >>"
};

Jsw2CellForm::Jsw2CellForm(Jsw2Room *r, Jsw2Game *g) : 
VideoForm(13 * CHAR_W + CHAR_SW, 15 * CHAR_H, CHAR_W * 24, CHAR_H * 17)
{
	int n, x, y;

	m_room = r; 
	m_game  = g;

	m_cancel = new VideoButton(ZXE_CANCEL, m_x + CHAR_W, m_y + 15 * CHAR_H, "Cancel");
	m_ok     = new VideoButton(ZXE_OK,     m_x + 16 * CHAR_W, m_y + 15 * CHAR_H, "  OK  ");

	x = m_x + 15 * CHAR_W;	
	y = m_y +  2 * CHAR_H;

	for (n = 0; n < 8; n++)
	{
		m_select[n] = new VideoButton(ZXE_CONTINUE, x, y, "Change");
		m_select[n]->setListener(this);
		y += CHAR_H + CHAR_SH;
	}
}


Jsw2CellForm::~Jsw2CellForm()
{
	int n;
	for (n = 0; n < 8; n++)
	{
		delete m_select[n];
	}
	delete m_ok;
	delete m_cancel;
}

void Jsw2CellForm::redraw(void)
{
	int n, x, y, fg, bg;
	VideoBitmap *bmp;

	VideoForm::redraw();

	x = m_x +  3 * CHAR_W;	
	y = m_y +  2 * CHAR_H;

	for (n = 1; n < 9; n++)
	{
		unsigned cpat = m_room->getCellNumber(n);

		jswByte *mem = m_game->getMem()->memoryAt(JSW2CELLS + 9 * cpat);
		if (mem[0] & 0x80)
		{
			bg = ZX_BRIGHT | (mem[0] & 7);
			fg = ZX_BRIGHT | ((mem[0] >> 3) & 7);
		}
		else
		{
			fg = ZX_BRIGHT | (mem[0] & 7);
			bg = ZX_BRIGHT | ((mem[0] >> 3) & 7);
		}
		bmp = bitmapFromSprite(8, 8, mem + 1, 1, fg, bg);
		bmp->toScreen(x - CHAR_W - CHAR_SW, y);
		delete bmp;
		videoScreen->drawText(x, y, st_captions[n],
				ZX_BLACK, ZX_WHITE);
		y += CHAR_H + CHAR_SH;
	}
	redrawBorder("Room graphics");
}



int Jsw2CellForm::doModal()
{
	return VideoForm::doModal();
}


void Jsw2CellForm::showChildren(int redraw)
{
        int n;
	for (n = 0; n < 8; n++)
	{
		addChild(m_select[n]);
	}
        addChild(m_ok);
        addChild(m_cancel);
	VideoForm::showChildren(redraw);
}


void Jsw2CellForm::hideChildren(void)
{
        int n;
	VideoForm::hideChildren();
	for (n = 0; n < 8; n++)
	{
		removeChild(m_select[n]);
	}
        removeChild(m_ok);
        removeChild(m_cancel);
}


int Jsw2CellForm::onButtonSelect(VideoButton *b)
{
	int n, rv;

	for (n = 0; n < 8; n++)
	{
		if (m_select[n] == b) break;
	}
	++n;

	int y = b->getY() + b->getH();

	CellList list(m_game->getMem(), 2, y, VIDEO_W - 4, CHAR_H + CHAR_SH + 40);
	m_game->fillCells(-1, &list);
	list.setTitle("Choose cell");
	list.setSelected(m_room->getCellNumber(n) * 9 + JSW2CELLS);
	rv = list.doModal();
	if (rv >= ZXE_QUIT) return rv;
	if (rv != ZXE_OK) return ZXE_CONTINUE;
	m_room->setCellNumber(n, (list.getSelected(NULL) - JSW2CELLS) / 9);

	redraw();
	return ZXE_CONTINUE;
}

