/************************************************************************

    JSWED 2.3.0 - Editor for Jet Set Willy and derivatives

    Copyright (C) 2000-2001,2004  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*************************************************************************/

#include "jswed.hxx"
#include "jsw2game.hxx"
#include "jsw2cellpage.hxx"

Jsw2CellPage::Jsw2CellPage(Jsw2Game *g)
{
	m_list = new CellList(g->getMem(), 2, m_y + 20 * CHAR_H, VIDEO_W - 4,
				               CHAR_H + CHAR_SH + 40);
	g->fillCells(-1, m_list);
	m_list->setTitle("Cell patterns in this game");
	addChild(m_list);
	m_editor = new BitmapEditor(9 * CHAR_W, CHAR_H, 8, 8);
	m_cc[0]  = new VideoColourChooser(9 * CHAR_W, 10 * CHAR_H);
	m_cc[1]  = new VideoColourChooser(9 * CHAR_W, 12 * CHAR_H);
	m_cb[0]  = new VideoCheckbox(9 * CHAR_W, 16 * CHAR_H);
	m_cb[1]  = new VideoCheckbox(9 * CHAR_W, 14 * CHAR_H);
	m_lbl[0] = new VideoLabel(19 * CHAR_W, 10 * CHAR_H, "Ink");
	m_lbl[1] = new VideoLabel(19 * CHAR_W, 12 * CHAR_H, "Paper");
	m_lbl[2] = new VideoLabel(19 * CHAR_W, 16 * CHAR_H, "Bright?");
	m_lbl[3] = new VideoLabel(19 * CHAR_W, 14 * CHAR_H, "Inverse");
	addChild(m_editor);
	addChild(m_cc[0]);
	addChild(m_cc[1]);
	addChild(m_cb[1]);
	//addChild(m_cb[0]);
	addChild(m_lbl[0]);
	addChild(m_lbl[1]);
	addChild(m_lbl[3]);
	//addChild(m_lbl[2]);
	m_list->setListener(this);
	m_cb[0]->setListener(this);
	m_cb[1]->setListener(this);
	m_cc[0]->setListener(this);
	m_cc[1]->setListener(this);
	setTitle("Cells");
	setFocus(m_list);
	m_reload = NULL;
}

Jsw2CellPage::~Jsw2CellPage()
{
	delete m_editor;
	delete m_list;
	delete m_cc[0];
	delete m_cc[1];
	delete m_cb[0];
	delete m_cb[1];
	delete m_lbl[0];
	delete m_lbl[1];
	delete m_lbl[2];
	delete m_lbl[3];
}


int Jsw2CellPage::redraw(int whichRectangle)
{
	int rv = VideoNotePage::redraw(whichRectangle);

	if (whichRectangle & VPR_RIGHT)
	{
		videoScreen->bottomBar("Cell definitions");

		videoScreen->drawSmallText(0, VIDEO_H - 4 * CHAR_H,
			"Keyboard hints: Select a cell using cu"
			"rsor keys and ENTER; then press TAB to", 
			ZX_BLACK, ZX_WHITE);
		videoScreen->drawSmallText(0, VIDEO_H - 4 * CHAR_H + CHAR_SH,
			"switch to the editor. ^Z or ^U to undo", 
			ZX_BLACK, ZX_WHITE);
	}
	return rv;
}

void Jsw2CellPage::reload(Jsw2Game *g)
{	
	m_reload = g;
}

int Jsw2CellPage::onReveal(int rect)
{
	int r = VideoNotePage::onReveal(rect);

	if (m_reload)
	{
		m_list->emptyBuffer();
		m_reload->fillCells(-1, m_list);
		m_list->redraw();
		m_reload = 0;
	}
	return r;
}

void Jsw2CellPage::onChangeSprite(jswByte *newBitmap)
{
	m_editor->setBitmap(newBitmap + 1);
//	m_cb[0]->setCheck(newBitmap[0] & 0x40);
	m_cb[1]->setCheck(newBitmap[0] & 0x80);
	m_cc[0]->setColour(newBitmap[0] & 7);
	m_cc[1]->setColour((newBitmap[0] >> 3) & 7);
	memcpy(m_undoBuf, newBitmap, 9);
}

void Jsw2CellPage::onChooseColour(VideoColourChooser *c)
{
	if (!m_editor->getBitmap()) return;

	jswByte *bm = m_editor->getBitmap() - 1;
	if (c == m_cc[0])
	{
		bm[0] &= 0xF8;
		bm[0] |= c->getColour();	
	}
	if (c == m_cc[1])
	{
		bm[0] &= 0xC7;
		bm[0] |= (c->getColour() << 3);	
	}
}


int Jsw2CellPage::onCheckboxSelect(VideoCheckbox *b, int value)
{
	if (!m_editor->getBitmap()) return ZXE_CONTINUE;

	jswByte *bm = m_editor->getBitmap() - 1;
	if (b == m_cb[0])
	{
		if (value) bm[0] |= 0x40;
		else	   bm[0] &= ~0x40;
		m_cb[0]->setCheck(value);
	}
	if (b == m_cb[1])
	{
		if (value) bm[0] |= 0x80;
		else	   bm[0] &= ~0x80;
		m_cb[1]->setCheck(value);
	}
	return ZXE_CONTINUE;
}


void Jsw2CellPage::undo(void)
{
	jswByte buf[9];

	if (!m_editor->getBitmap()) return;

	jswByte *bm = m_editor->getBitmap() - 1;
	
	memcpy(buf, bm, 9);
	memcpy(bm, m_undoBuf, 9);
	memcpy(m_undoBuf, buf, 9);
	m_editor->setBitmap(bm + 1);	// Make it redraw
//	m_cb[0]->setCheck(bm[0] & 0x40);
	m_cb[1]->setCheck(bm[0] & 0x80);
	m_cc[0]->setColour(bm[0] & 7);
	m_cc[1]->setColour((bm[0] >> 3) & 7);
}

int Jsw2CellPage::preContextMenu(void)
{
	return m_editor->doContextMenu();
}


int Jsw2CellPage::onKeyDown(int keysym)
{
	switch(keysym)
	{
		case 'Z' - '@':
		case 'U' - '@':
			undo();
			return ZXE_CONTINUE;
	}
	return VideoNotePage::onKeyDown(keysym);
}
