/************************************************************************

    JSWED 2.3.0 - Editor for Jet Set Willy and derivatives

    Copyright (C) 2000-2001,2004  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*************************************************************************/

#include "jswed.hxx"
#include "game.hxx"
#include "room.hxx"
#include "snapio.hxx"
#include "jsw2game.hxx"
#include "jsw2room.hxx"
#include "jsw2patchpage.hxx"
#include "jsw2msgpage.hxx"
#include "jsw2sprite.hxx"
#include "jsw2roomdraw.hxx"
#include "jsw2cellpage.hxx"
#include "jsw2roompage.hxx"
#include "jsw2splpage.hxx"
#include "jswtitle.hxx"
#include "jsw2title.hxx"
#include "progressform.hxx"
#include "vspritelist.hxx"
#include "celllist.hxx"

// JSW2 initialisation code. This is overwritten at startup; so if you 
// snapshot a JSW2 game and later want it back, this has to be copied
// into the snapshot
jswByte initFragment[] =
{
	0x31,0xFF,0x5F,0x01,0x85,0x01,0x21,0x78,
	0x8C,0x7E,0xCB,0x7F,0x28,0x11,0xCB,0xBF,
	0x77,0xE5,0x23,0xC5,0x06,0x08,0x7E,0xEE,
	0xFF,0x77,0x23,0x10,0xF9,0xC1,0xE1,0x11,
	0x09,0x00,0x19,0x0B,0x78,0xB1,0x20,0xE1,
	0x21,0x00,0x5E,0x11,0x00,0x60,0x73,0x23,
	0x72,0x23,0x14,0x7A,0xE6,0x07,0x20,0xF6,
	0x7A,0xD6,0x08,0x57,0x7B,0xC6,0x20,0x5F,
	0x20,0xEC,0x7A,0xC6,0x08,0x57,0xFE,0x70,
	0x20,0xE4,0x01,0x1F,0x00,0xAF,0xED,0x58,
	0xB3,0x10,0xFB,0x2F,0xE6,0xC0,0x32,0x00,
	0x70,0x3A,0xB2,0x61,0xB7,0x28,0xFF,0x21,
	0x01,0x70,0x11,0x02,0x70,0x36,0x00,0x01,
	0x87,0x00,0xC3,0x4D,0x74,0x00,0x00,0x00,
	0x00
};

Jsw2Game::Jsw2Game(SpectrumMemory *mem) : Game(mem)
{
	m_patchPage = NULL;
	m_messagePage = NULL;
	m_spritePage = NULL;
	m_cellPage = NULL;
	m_roomPage = NULL;
	m_splPage = NULL;
	m_ttlPage = NULL;

        m_mem->poke(0x5BFE, 0x00);      // Initial PC for JSW2
        m_mem->poke(0x5BFF, 0x70);


        m_roomCache = NULL;
        m_roomCacheSize = 0;
	m_wordHead = NULL;
	m_spriteGroups = NULL;
	m_sgCount = 0;
	m_sgMax = 0;
}


Jsw2Game::~Jsw2Game()
{
	clearSpriteGroups();
	deletePages();
	clearWords();
}

void Jsw2Game::addPages(VideoNotebook *n)
{
	if (m_patchPage == NULL)   m_patchPage   = new Jsw2PatchPage(this);
	if (m_messagePage == NULL) m_messagePage = new Jsw2MessagePage(m_mem);
	if (m_spritePage == NULL)  m_spritePage  = new Jsw2SpritePage(this);
	if (m_cellPage   == NULL)  m_cellPage    = new Jsw2CellPage(this);
	if (m_roomPage   == NULL)  m_roomPage    = new Jsw2RoomPage(this);
	if (m_splPage    == NULL)  m_splPage     = new Jsw2SpecialPage(this);
	if (m_ttlPage    == NULL)  m_ttlPage     = new Jsw2TitlePage(m_mem);
	n->addPage(m_patchPage);
	n->addPage(m_messagePage);
	n->addPage(m_spritePage);
	n->addPage(m_cellPage);
	n->addPage(m_roomPage);
	n->addPage(m_splPage);
	n->addPage(m_ttlPage);
	Game::addPages(n);
}


void Jsw2Game::deletePages(void)
{
	if (m_patchPage)   delete m_patchPage;
	if (m_messagePage) delete m_messagePage;
	if (m_spritePage)  delete m_spritePage;
	if (m_cellPage)    delete m_cellPage;
	if (m_roomPage)    delete m_roomPage;
	if (m_splPage)     delete m_splPage;
	if (m_ttlPage)     delete m_ttlPage;
	m_patchPage   = NULL;
	m_messagePage = NULL;
	m_spritePage  = NULL;
	m_cellPage    = NULL;
	m_roomPage    = NULL;
	m_splPage     = NULL;
	m_ttlPage     = NULL;
	Game::deletePages();
}

void Jsw2Game::fillSprites(int room, SpriteList *s, int every)
{
	for (int n = JSW2SPRITES; n < JSW2SPRITESEND; n += 32)
	{
		s->addSprite(n);
	}
}

void Jsw2Game::fillSprites(int room, VSpriteList *s)
{
	for (int n = JSW2SPRITES; n < JSW2SPRITESEND; n += 32)
	{
		s->addSprite(n);
	}
}


void Jsw2Game::fillCells(int room, CellList *s)
{
	for (int n = JSW2CELLS; n < JSW2CELLSEND; n += 9)
	{
		s->addSprite(n);
	}
}


void Jsw2Game::clearSpriteGroups()
{
	if (m_spriteGroups) delete [] m_spriteGroups;
	m_spriteGroups = NULL;
	m_sgCount = 0;
	m_sgMax = 0;
}


void Jsw2Game::addSpriteGroup(int fileid, int baseid)
{
	SpriteGroup *sg;
	int oldBase;
     
	// Does this sprite ID already exist?	
	sg = lookupSpriteGroup(fileid);
	if (sg) return;
	// No. Does the group exist?
	sg = lookupSpriteGroup(baseid);
	// If it exists, and its baseid is higher than the new ID, then
	// the new ID becomes the baseid for all members of the group.
	if (sg && (sg->baseid > fileid))
	{
		oldBase = sg->baseid;
		for (int n = 0; n < m_sgCount; n++)
			if (m_spriteGroups[n].baseid == oldBase)
				m_spriteGroups[n].baseid = fileid;
		baseid = fileid;
	}
	if (!m_spriteGroups)
	{
		m_spriteGroups = new SpriteGroup[32];
		m_sgMax = 32;
		m_sgCount = 0;
	}
	if (m_sgMax == m_sgCount)
	{
		sg = new SpriteGroup[2 * m_sgMax];
		memcpy(sg, m_spriteGroups, m_sgMax * sizeof(SpriteGroup));
		delete [] m_spriteGroups;
		m_spriteGroups = sg;
		m_sgMax *= 2;
	}
	m_spriteGroups[m_sgCount].fileid = fileid;
	m_spriteGroups[m_sgCount].memid  = -1;
	m_spriteGroups[m_sgCount].baseid = baseid;
	++m_sgCount;
}

SpriteGroup *Jsw2Game::lookupSpriteGroup(int id)
{
	for (int n = 0; n < m_sgCount; n++)
		if (m_spriteGroups[n].fileid == id)
			return &m_spriteGroups[n];
	return NULL;
}


jswByte *Jsw2Game::getRoom(int room)
{
	return NULL;
}

int  Jsw2Game::getRoomCount(void)
{
	return m_roomCount;
}

int Jsw2Game::roomReserved(int room)
{
	return 0;
}

void Jsw2Game::deleteSprites(void)
{

}

Room *Jsw2Game::newRoom(int room)
{
	if (room >= 0 && room < m_roomCount)
	{
		return new Jsw2Room(room, this);
	}
	return NULL;
}


jswByte *Jsw2Game::getRoomData(int rmid)
{
	if (rmid >= m_originalRoomCount) rmid = 0;

	jswByte *base = m_mem->memoryAt(0x7E69);
	unsigned roomTable = base[0] + 256 * base[1];
        base = m_mem->memoryAt(roomTable + 2 * rmid);
	return m_mem->memoryAt(base[0] + 256 * base[1]);
}


int Jsw2Game::decompress()
{
	int n;

	/* Is it encrypted? */
	if (m_mem->peek(0x7089) == 0xEC)
	{
		// Replicate the builtin decryptor. Firstly, copy to final
		// position.
		int hl = 0xF4B0;
		int de = 0xFFB0;
		for (n = 0; n < 0x8FB3; n++)
		{
			m_mem->poke(de, m_mem->peek(hl));
			--de;
			--hl;
		}
		/* Decrypt */
		jswByte *key  = m_mem->memoryAt(0x6480);
		jswByte *data = m_mem->memoryAt(0x7000);
		for (n = 0x0; n < 0x8F00; n++)
		{
			data[n] ^= key[n % 0x22];	
		}
		FILE *fp = fopen("dec.dbg", "wb");
		int n;
		for (n = 0x3fe5; n <= 0xffff; n++)
			fputc(m_mem->peek(n), fp);
		fclose(fp);
		if (m_mem->peek(0x7000) != 0x31)
		{
			int rv = alert("Load JSW2 game", "Cancel", 
				"This does not appear to be a valid JSW2 game.");
			if (rv == ZXE_QUIT) return rv;
			return ZXE_CANCEL;
		}
	}
// After decrypt: Copy in the initialisation fragment, which overwrites 
// itself.
	memcpy(m_mem->memoryAt(0x7000), initFragment, sizeof(initFragment));
	// Necessary otherwise initcode will trash the data.
	m_mem->poke(0x61b2, 1);	
// Move stack pointer to 5FFE
	m_mem->pokeSnaHeader(0x17, 0xFE);
	m_mem->pokeSnaHeader(0x18, 0x5F);
	m_mem->poke(0x5FFE, 0x00);
	m_mem->poke(0x5FFF, 0x70);

        jswByte *base = m_mem->memoryAt(0x7E69);
        unsigned roomTable = base[0] + 256 * base[1];
        base = m_mem->memoryAt(roomTable);
	unsigned room0     = base[0] + 256 * base[1];

	m_roomCount = m_originalRoomCount = (room0 - roomTable) / 2;

	for (n = 0; n < m_roomCount; n++)
	{
		getRoomClass(n);	// Decompress all rooms
	}	
	return ZXE_CONTINUE;
}


// Max room size:
//
// 512 bytes shape
//  32 bytes title
//  17 bytes cell types, exits, flags etc.
//  56 bytes guardians
//   2 bytes directory entry
//
#define MAXROOMSIZE	(512 + 32 + 17 + 56 + 2)
#define COMPBASE	0x9A2E

void Jsw2Game::clearWords()
{
	WordBuf *wp;

	while (m_wordHead)
	{
		wp = m_wordHead;
		m_wordHead = m_wordHead->next;
		delete wp;
	}
}

// Add a token to the list of all tokens for analysis.
void Jsw2Game::addWord(const char *s, bool atEnd)
{
	WordBuf *wp;
	int saving = strlen(s);

	// If a token is at the end of a title, the number of bytes saved 
	// is reduced by 2. One because the following space isn't removed 
	// (there not being one) and one because a following space must be
	// added (a token can't come last).
	//
	// Since we don't in fact make substitutions that would increase
	// stored size, we don't allow saving to go negative.
	//
	if (atEnd) saving -= 2;
	if (saving < 0) saving = 0;

	for (wp = m_wordHead; wp != NULL; wp = wp->next)
	{
		if (!strcmp(s, wp->word))
		{
			++wp->count;
			wp->weight += saving;
			return;
		}		
	}
	wp = new WordBuf;
	strcpy(wp->word, s);
	wp->count = 1;
	wp->weight = saving;
	wp->next = m_wordHead;
	m_wordHead = wp;
}

static void ltrim(char *s)
{
	int n, m, l;
	l = strlen(s);

	for (n = 0; n < l; n++)
	{
		if (s[n] != ' ') break;
	}
	for (m = n; m <= l; m++)
	{
		s[m-n] = s[m];
	}
}

static void rtrim(char *s)
{
	for (int n = strlen(s) - 1; n >= 0; n--)
	{
		if (s[n] != ' ') break;
		s[n] = 0;
	}
	// If string ends with a token, stick a space on the end.
	char *p = s + strlen(s);
	if (p[-1] < ' ') strcat(s, " ");
}
/*
static void showWord(char *s)
{
	putchar('\'');
	while (*s)
	{
		if (*s < ' ') printf("<%02x>", *s);
		else putchar(*s);
		++s;
	}
	putchar('\'');
}*/

// Generate a dictionary of tokens for use in room titles, and compress room
// titles using it.
//
// JSW2 itself was compressed using an algorithm that replaced only whole 
// words or pairs of words with tokens. JSWED is rather more aggressive; 
// any sequence of characters ending with a space is a candidate for 
// compression. 
//
// The compression algorithm used is:
// * Remove leading and trailing spaces from room names.
// * Scan all room names. Store all possible tokens; a token is any sequence
//  of characters that ends with either a space or the end of the room name.
//  XXX Currently, this does not include tokens with a space in them; so the
//  compression of 'MAIN LIFT' that appears in the original dictionary does
//  not appear.
// * For each token, give it a 'weight' which is the number of bytes saved by
//  using it. Each time the token appears at the end of a title, the weight 
//  increases by (token length - 1) (since it must be followed by a space; a 
//  compressed line can't end with a token). Each time it occurs elsewhere 
//  in a line, its weight increases by (token length).
// * Take the token with the greatest weight.
//  Add it to the dictionary, and replace all instances of that token with the
//  single-byte token ID (1-31).
// * Repeat until the dictionary's full or the token ID reaches 32.
//
// This method appears to give better compression than JSW2 itself 
// contains, while still being compatible with the JSW2 expander. 
//
void Jsw2Game::compressTitles(ProgressForm *pf)
{
	char titleBuf[33];
	char tokenBuf[33];
	char *s;
	int found, saving;
	WordBuf *wp, *candidate;
	int n, m, m2;
	int dictSpace = 0x6E; 	// Buffer size minus start mark 
	jswByte dictBuf[0x70];
	int dictEnd = 0;
	int dictToken = 1;

	dictBuf[dictEnd++] = 0x80;	// Start marker
	for (n = 0; n < m_roomCount; n++)
	{
		getRoomClass(n)->getTitle(m_roomTitles[n]);
		// Strip leading & trailing spaces
		ltrim(m_roomTitles[n]);
		rtrim(m_roomTitles[n]);
		//printf("%3d '%s'\n", n, m_roomTitles[n]);
	}
	while (dictSpace)
	{
		if (dictToken == 31) break;	// All tokens used
		// Empty the word table
		clearWords();
		for (n = 0; n < m_roomCount; n++)
		{
			strcpy(titleBuf, m_roomTitles[n]);
			// A 'word' is any sequence of characters terminated
			// by a space.
			for (m = 0; titleBuf[m]; m++)
			{
				if (titleBuf[m] == ' ') continue;
				strcpy(tokenBuf, titleBuf + m);
				char *sp = strchr(tokenBuf, ' ');
				if (sp) 
				{
					*sp = 0;
					addWord(tokenBuf, false);
				}
			//	printf("Token=%s\n", tokenBuf);
				else	addWord(tokenBuf, true);
			}
		}
		saving = 0;
		candidate = NULL;
		for (wp = m_wordHead; wp; wp = wp->next)
		{
			// If the candidate word won't fit in the 
			// dictionary, skip it.
			if (strlen(wp->word) > dictSpace) continue;
			// Don't try to compress a token that's already
			// compressed.
			if (wp->word[0] < ' ' && wp->word[1] == 0) continue;
			int sav = wp->weight;
			if (sav > saving)
			{
				saving = sav;
				candidate = wp;
			}
		}
		// Add the word to the dictionary. If no words found, then
		// abandon compression.
		if (!candidate) break;
		// Add candidate to the dictionary
		memcpy(dictBuf + dictEnd, candidate->word, 
				strlen(candidate->word));
		dictEnd   += strlen(candidate->word);
		dictSpace -= strlen(candidate->word);
		dictBuf[dictEnd - 1] |= 0x80;	// End of word

		// Compression includes the space at the end.
		strcat(candidate->word, " ");
//		printf("Compression candidate: ");
//		showWord(candidate->word);
//		printf("\n");
		for (n = 0; n < m_roomCount; n++)
		{
			char *s;
			int ns;

			//showWord(m_roomTitles[n]);
			// Stick a space at the end, so the last word can be
			// compressed if need be. Note that the tokenised
			// last word must be followed by a space in this case;
			// rtrim() handles this.
			// (but, only do the strcat() if the token is longer
			// than 2 characters. There's no point in replacing 
			// a single 'e' (frex) with a tokenized "e " which 
			// has to be encoded as "<xx><20>". Likewise
			// replacing "er" by "<xx><20>" doesn't save any
			// memory.
			if (strlen(candidate->word) > 2)
			{
				strcat(m_roomTitles[n], " ");
			}
			//printf(" -> ");
			while (s = strstr(m_roomTitles[n], candidate->word))
			{
				ns = s - m_roomTitles[n];
				if (ns) memcpy(titleBuf, m_roomTitles[n], ns);
				titleBuf[ns] = dictToken;
				s += strlen(candidate->word);
				strcpy(titleBuf + ns + 1, s);
				strcpy(m_roomTitles[n], titleBuf);
			}
			// Remove trailing space (if any)
			rtrim(m_roomTitles[n]);
			//showWord(m_roomTitles[n]);
			//printf("\n");
		}
		++dictToken;
		pf->inc();
	}
	//dictBuf[dictEnd++] = 0x80;	// End marker

	memcpy(m_mem->memoryAt(JSW2DICT), dictBuf, 0x6F);
//	for (wp = m_wordHead; wp; wp = wp->next)
//	{
//		showWord(wp->word);
//		printf(" count=%d\n", wp->count);
//	}

}



int Jsw2Game::compress()
{
	int n, m, p;
	jswByte compData[MAXROOMSIZE * 256];
	int bufSize = 0;
	int roomTable = 0;
	ProgressForm pf(2 * m_roomCount + 30);
	unsigned shapeOffset[256];

	pf.bar("Compressing room titles", 0);
	compressTitles(&pf);
/* 9A2E-BAFC: Room shapes
 * BAFD-BC08: Room directory
 * BC09-D4A0: Room definitions 
 *
 * 7E69 contains address of room directory, so we can shuffle these as required
 */
	pf.bar("Compressing room shapes", 30);
	for (n = 0; n < m_roomCount; n++)
	{
		int obs = bufSize;
		shapeOffset[n] = bufSize + COMPBASE;
		bufSize += ((Jsw2Room *)getRoomClass(n))->compressShape(compData, bufSize);
//
// See if this compressed pattern has already been generated. If so, this room
// can share it.
//
		for (m = 0; m < n; m++)
		{
			jswByte *match = compData + (shapeOffset[m] - COMPBASE);
			if (!memcmp(compData + obs, match, bufSize - obs))
			{
				shapeOffset[n] = shapeOffset[m];
				bufSize = obs;
				break;
			}
		}

		pf.inc();
	}
/*	for (n = 0; n < bufSize; n++)
	{
		printf("%02x ", compData[n]);
		if ((n & 0x0F) == 0x0F) putchar('\n');
	}
*/
// Skip over the room directory
	roomTable = bufSize;
	bufSize += 2 * m_roomCount;

	pf.bar("Compressing rooms", m_roomCount + 30);
	for (n = 0; n < m_roomCount; n++)
	{
		compData[roomTable + 2*n  ] = (bufSize + COMPBASE) & 0xFF;
		compData[roomTable + 2*n+1] = ((bufSize + COMPBASE) >> 8) & 0xFF;
		bufSize += ((Jsw2Room *)getRoomClass(n))->compress(compData, roomTable, bufSize, shapeOffset[n], m_roomTitles[n]);
		pf.inc();
	}
	if (bufSize + COMPBASE > JSW2COMPEND)
	{
                int rv = alert("Compression error", "Cancel",
		"The rooms will not compress enough to fit in the "
		"space available.\n"
		"Space available=%5d\n"
		"Space needed   =%5d", JSW2COMPEND - COMPBASE, bufSize);
		if (rv >= ZXE_QUIT) return rv;
                return ZXE_CANCEL;
	}
	memcpy(m_mem->memoryAt(COMPBASE), compData, bufSize);
	m_mem->pokeW(0x7E69, (roomTable + COMPBASE));
//	printf("Buffer top at %04x\n", bufSize + COMPBASE);
	return ZXE_CONTINUE;
}



char *Jsw2Game::getID()
{
        char *s = Game::getID();

        strcpy(s + ((VIDEO_W / CHAR_W) - 7), " [JSW2]");
        return s;
}

int Jsw2Game::prevRoom(int rm)
{
	if (rm > 0) return rm - 1;
	return rm;
}

int Jsw2Game::nextRoom(int rm)
{
	if (rm < m_roomCount) return rm + 1;
	return rm;
}


int Jsw2Game::getStartRoom()
{
	return m_mem->peek(0x752E);
}


void Jsw2Game::setStartRoom(int rm)
{
	if (rm >= 0 && rm < 134)
	{
		m_mem->poke(0x752E, rm);
	}
}



typedef Room *RoomPtr;

Room *Jsw2Game::getRoomClass(int room)
{
	Room *r, **rc;
	int newsize, n;

	if (room < 0 || room >= getRoomCount()) return NULL;

	if (room < m_roomCacheSize)
	{
		if (m_roomCache[room]) return m_roomCache[room];
		r = newRoom(room);
		m_roomCache[room] = r;
		return r;
	}

	r = newRoom(room);
	if (m_roomCacheSize == 0) newsize = 64;
	else			  newsize = 2 * m_roomCacheSize;

	while (newsize < room) newsize *= 2;

	rc = new RoomPtr[newsize];
	for (n = 0; n < newsize; n++) rc[n] = NULL;
	if (m_roomCache)
	{
		for (n = 0; n < m_roomCacheSize; n++) rc[n] = m_roomCache[n];
		delete m_roomCache;
	}
	m_roomCache = rc;
	m_roomCacheSize = newsize;
	rc[room] = r;
	return r;
}


JswRoomList *Jsw2Game::getRoomList(void)
{
        int n;  
	Room *rm;
	char rname[40];

	JswRoomList *s = new JswRoomList();

	for (n = 0; n < 256; n++) 
	{
		if (roomReserved(n)) continue;
		rm = getRoomClass(n);
		if (!rm) continue;
		sprintf(rname, "%03d ", n);
		rm->getTitle(rname + 4);
		s->addString(rname);
	}
	return s;
}

bool Jsw2Game::importSprites(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent)
{
	return importSprites(doc, ns, parent, NULL);
}

bool Jsw2Game::importSprites(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent,
		int *roomMap)
{
	xmlNodePtr room, guards, guard, spriteref;
	int baseid, fileid, sr;
	int n;

	clearSpriteGroups();
	/* In JSW2, sprites don't go around in pages like in JSW1; but 
	 * groups must be kept together. */
	for (room = parent->xmlChildrenNode; room; room = room->next)
	{
		if (room->ns != ns) continue;
		if (strcmp((char *)room->name, "room")) continue;
		for (guards = room->xmlChildrenNode; guards;
				guards = guards->next)
		{
			if (guards->ns != ns) continue;
			if (strcmp((char *)guards->name, "guardians")) continue;
			for (guard = guards->xmlChildrenNode; guard; 
					guard = guard->next)
			{
				if (guard->ns != ns) continue;
				baseid = -1;
				for (spriteref = guard->xmlChildrenNode;  
					spriteref; spriteref = spriteref->next)
				{
					if (spriteref->ns != ns) continue;
					if (strcmp((char *)spriteref->name, "spriteref")) continue;
					char *prop = (char *)xmlGetProp(
						spriteref, (xmlChar *)"xref");
					sr = 0;
					if (prop)
					{
						sr =sscanf(prop, "%x", &fileid);
						xmlFree(prop);
					}
					if (sr)
					{
						if (baseid == -1)
							baseid = fileid;
						addSpriteGroup(fileid,baseid);
					}
				}
			}
		}
	}
	Game::importSprites(doc, ns, parent);
	// All sprites have now been read into m_spriteGroups.
	// So for each group that hasn't been mapped...
	for (n = 0; n < m_sgCount; n++) if (m_spriteGroups[n].memid == -1)
	{
		int basememid = -1;
		int baseid = m_spriteGroups[n].baseid;
		int count = 0, cur = 0;
		int m;
		for (m = 0; m < m_sgCount; m++)
			if (m_spriteGroups[m].baseid == baseid) ++count;
		// OK, we have a group with a given baseid, length count.
		// Try to match it in the sprite memory.
		jswByte *match = new jswByte[count * 32];
		for (m = 0; m < m_sgCount; m++)
			if (m_spriteGroups[m].baseid == baseid) 
		{
			memcpy(match + 32*cur, m_spriteGroups[m].data, 32);
			++cur;
			if (cur >= count) break; // Should never happen
		}
		// match is the imported group of sprites
		for (m = JSW2SPRITES; m < JSW2SPRITESEND; m += 32)
		{
			// Got a match?
			if (!memcmp(match, m_mem->memoryAt(m), 32*count))
			{
				basememid = m;
				break;
			}
		}
		if (basememid == -1) // That didn't work
		{
			// Try to find 'count' consecutive unreferenced
			// sprites.
			for (m = JSW2SPRITES; m < JSW2SPRITESEND; m += 32)
			{
				bool possible = true;
				for (cur = 0; cur < count; cur++)
				{
					if (getRefCount(SPRITE(m + 32*cur)))
					{
						possible = false;
						break;
					}
				}
				if (possible) 
				{
					basememid = m;
					break;
				}
			}
		}
		if (basememid != -1) // OK, we have a match!
		{
			memcpy(m_mem->memoryAt(basememid), match, 32*count);
			cur = 0;
			for (m = 0; m < m_sgCount; m++)
				if (m_spriteGroups[m].baseid == baseid) 
			{
				m_spriteGroups[m].memid = basememid + 32 * cur;
				addRef(SPRITE(m_spriteGroups[m].memid));
				addSpriteRef(m_spriteGroups[m].fileid,
					     m_spriteGroups[m].memid);
				++cur;
				if (cur >= count) break; // Should never happen
			}
		}
		delete [] match;
	}
	clearSpriteGroups();
	return true;
}


// This queues all the imported sprites; it does not map any of them into
// memory.
int Jsw2Game::addSprite(int fileid, int w, int h, jswByte *bits)
{
	int n, x, y;
	int memid;
	int baseid = -1;
	bool found = false;
	int count;
	int wb = (w + 7) / 8;

	for (n = 0; n < m_sgCount; n++)
	{
		if (m_spriteGroups[n].fileid == fileid)
		{
			for (y = 0; y < h; y++)
			{
				for (x = 0; (8*x) < w; x++)
				{
					if (y < 16 && x < 2)
						m_spriteGroups[n].data[y*2+x]
							= bits[y*wb+x];
				}
			}
		}
	}
	// Don't add to mapping table
	return -1;
}


void Jsw2Game::addFixedRefs()
{
	int n;
	// Put sprite-xref locks on Willy and the toilet, so that 
	// sprite import doesn't try to overwrite them.
	for (n = 0; n < 12; n++)
	{
		addRef(SPRITE(0xD561 + 32 * n));
	}
	// Same for the foot and the barrel.
	addRef(SPRITE(0xD4A1));
	addRef(SPRITE(0xD4C1));
	// and the arrows 
	addRef(SPRITE(0xEB01));
	addRef(SPRITE(0xEB21));
}

