
#ifndef JSW2GAME_HXX_INCLUDED
#define JSW2GAME_HXX_INCLUDED 

/************************************************************************

    JSWED 2.3.0 - Editor for Jet Set Willy and derivatives

    Copyright (C) 2000-2001,2004  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*************************************************************************/

#include "snapio.hxx"
#include "jnotebook.hxx"
#include "jswroomlist.hxx"

class Room;
class Jsw2Room;
class CellList;
class SpriteList;
class VSpriteList;
class Jsw2SpritePage;
class Jsw2CellPage;
class Jsw2RoomPage;
class Jsw2SpecialPage;
class Jsw2MessagePage;
class Jsw2PatchPage;
class Jsw2TitlePage;
class ProgressForm;

class WordBuf
{
public:
	WordBuf *next;
	char word[33];
	int count;
	int weight;
};

class SpriteGroup
{
public:
	int fileid;
	int memid;
	int baseid;
	jswByte data[32];
};

class Jsw2Game : public Game
{
public:
	Jsw2Game(SpectrumMemory *mem);
	virtual ~Jsw2Game();
        virtual int decompress();
	virtual int compress();
	virtual char *getID();

	virtual void addPages(VideoNotebook *n);
	virtual void deletePages(void);
	virtual void fillCells(int room, CellList *s);
	virtual void fillSprites(int room, SpriteList *s, int every = 1);
	virtual void fillSprites(int room, VSpriteList *s);
	virtual bool importSprites(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent, int *roomMap);
	virtual bool importSprites(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent);
	virtual int addSprite(int fileid, int w, int h, jswByte *bits);
	virtual void addFixedRefs(void);

	virtual jswByte *getRoom(int room);
	Room *getRoomClass(int room);

	virtual int  getRoomCount(void);
	inline void setRoomCount(int n) { m_roomCount = n; }

	inline int  getItemCount(void) { return m_mem->peek(0x877E); }
	inline void setItemCount(int n) { m_mem->poke(0x877E, n); }

	virtual JswRoomList *getRoomList(void);
	virtual int roomReserved(int room);

	int getStartRoom();
	void setStartRoom(int rm);

	int nextRoom(int rm);
	int prevRoom(int rm);

	void clearWords();
	void addWord(const char *s, bool atEnd);

	jswByte *getRoomData(int rmid);

protected:
	void compressTitles(ProgressForm *pf);
	virtual void deleteSprites(void);
	virtual Room *newRoom(int room);

	void clearSpriteGroups();
	void addSpriteGroup(int fileid, int baseid);
	SpriteGroup *lookupSpriteGroup(int id);

	Jsw2PatchPage   *m_patchPage;
	Jsw2MessagePage *m_messagePage;
	Jsw2SpritePage  *m_spritePage;
	Jsw2CellPage    *m_cellPage;
	Jsw2RoomPage    *m_roomPage;
	Jsw2SpecialPage *m_splPage;
	Jsw2TitlePage   *m_ttlPage;

	Room **m_roomCache;
	int  m_roomCacheSize;

	WordBuf * m_wordHead;

	int m_roomCount;	
	int m_originalRoomCount;	
	char m_roomTitles[256][36];

	SpriteGroup *m_spriteGroups;
	int m_sgCount, m_sgMax;
};

#define JSW2GUARDINKS  0x70A9
#define JSW2CELLS      0x8C78
#define JSW2CELLSEND   0x9A2E
#define JSW2SPRITES    0xD4A1
#define JSW2DICT       0xFA81
#define JSW2SPRITESEND JSW2DICT
#define JSW2COMPEND    JSW2SPRITES

#endif // ndef JSW2GAME_HXX_INCLUDED
