/************************************************************************

    JSWED 2.3.0 - Editor for Jet Set Willy and derivatives

    Copyright (C) 2000-2001,2004  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*************************************************************************/

#include "jswed.hxx"
#include "jsw2game.hxx"
#include "jsw2guard.hxx"


Jsw2HVGuardian::Jsw2HVGuardian(jswByte *g, SpectrumMemory *mem)
{
	m_mem = mem;
	m_guard = g;
	memcpy(m_orig,  g, 7);
        m_bmp      = newVideoBitmap(256, 32);
        m_bmpSmall = newVideoBitmap(128, 16);
	constructSprites();
	reset();
}


Jsw2HVGuardian::~Jsw2HVGuardian()
{
        delete m_bmp;
        delete m_bmpSmall;
}


void Jsw2Guardian::getBounds(jswByte *frame, unsigned *x, unsigned *y)
{
	int r,c;
	jswByte mask;
	
	*x = 0;
	*y = 0;

	for (r = 15; r >= 0; r--)
	{
		if ((frame[r*2] || frame[r*2+1]) && (*y == 0))
		{
			*y = (r + 1);
		}
		mask = 1;
		for (c = 15; c >= 0; c--)
		{
			if (frame[r*2+(c/8)] & mask)
			{
				if (c >= (*x))
				{
					*x = c + 1;
				}
				break;
			}
			mask = mask << 1;
			if (mask == 0) mask = 1;
		}
	}
}



void Jsw2HVGuardian::constructSprites()
{
	VideoBitmap *spb, *spbs;
	int fr, fg, ink;
	jswByte *spriteBase, *frame; 

	ink = (m_guard[6] & 0x0C) >> 2;	
	fg = m_mem->peek(JSW2GUARDINKS + ink);
	fg = (fg & 7) | ((fg & 0x40) >> 3);
/*
	switch(m_guard[6] & 0x0C)
	{
		case 0:  fg = ZX_WHITE; break;
		case 4:  fg = ZX_YELLOW | ZX_BRIGHT; break;
		case 8:  fg = ZX_CYAN   | ZX_BRIGHT; break;
		case 12: fg = ZX_GREEN  | ZX_BRIGHT; break;
	} */
	if (m_guard[5] & 0x80)	// Unidirectional guardian
	{
		fg = ZX_WHITE;	// Don't know why, but making a guardian
				// unidirectional also makes it white.
	}
	spriteBase = m_mem->memoryAt(JSW2SPRITES);
	unsigned nSprite = getSpritePage();

	frame = spriteBase + 32 * nSprite;
	for (fr = 0; fr < 8; fr++)
	{
		spb = bitmapFromSprite(16, 16, frame + 32 * fr, 1, fg, ZX_TRANS);
		spbs = bitmapFromSprite(16,16, frame + 32 * fr, 0, fg, ZX_BLACK);
		spb ->toSurface(m_bmp,      (32 * fr), 0, 0, 0, 32, 32);
		spbs->toSurface(m_bmpSmall, (16 * fr), 0, 0, 0, 16, 16);
		m_bmp ->makeTransparent();
		getBounds(frame + 32 * fr, &m_w[fr], &m_h[fr]);
		delete spbs;
		delete spb;
	}	
}


void Jsw2HVGuardian::draw(VideoSurface *s)
{
	int x, frame;

	x = m_x & 0xFC;

	if (m_guard[6] & 0x80)	frame = m_x    & m_guard[6] & 3;
	else			frame = m_tick & m_guard[6] & 3;

	if (m_guard[6] & 0x40)	// Direction-dependent frames
	{
		if (m_guard[6] & 0x80)	
		{
			if ((~m_xstep) & 0x80) frame += 4;
		}
		else if (m_ystep & 0x80) frame += 4;
	}
	if (m_guard[5] & 0x80)	// Unidirectional guardian
	{
		if (m_xstep > 0 || m_ystep < 0) return;
	}

	m_bmp->toSurface(s, 4 * x, 2 * m_y, frame * 32, 0, 2 * m_w[frame], 
			2 * m_h[frame]);
}


void Jsw2HVGuardian::drawThumb(VideoSurface *s, int x, int y)
{
	int frame;

	if (m_guard[6] & 0x80)	frame = m_x    & m_guard[6] & 3;
	else			frame = m_tick & m_guard[6] & 3;

	if (m_guard[6] & 0x40)	// Direction-dependent frames
	{
		if (m_guard[6] & 0x80)	
		{
			if ((~m_xstep) & 0x80) frame += 4;
		}
		else if (m_ystep & 0x80) frame += 4;
	}
	// Unidirectional guardian
	if (m_guard[5] & 0x80 && (m_xstep > 0 || m_ystep < 0))
	{
		s->fillBox(x, y, 16, 16, ZX_BLACK);
	}
	else
	{
		m_bmpSmall->toSurface(s, x, y, frame * 16, 0, 16, 16);
	}
}

void Jsw2HVGuardian::drawBounds(VideoSurface *s, int ink, int cx, int cy)
{
	int xmin, xmax, ymin, ymax;
	// Draw a box in every position...
	int gx, gy;
	int n;
	int dx, dy;
	int x1 = cx / 2;
	int y1 = cy;
	int x0 = m_guard[4] & 0x7F;
	int y0 = m_guard[5] & 0x7F;
	int diag  = (m_guard[6] & 0x30);
	int horiz = (m_guard[6] & 0x80);
	int fg, revs;

	switch(m_guard[6] & 0x0C)
	{
		case 0:  fg = ZX_WHITE; break;
		case 4:  fg = ZX_YELLOW | ZX_BRIGHT; break;
		case 8:  fg = ZX_CYAN   | ZX_BRIGHT; break;
		case 12: fg = ZX_GREEN  | ZX_BRIGHT; break;
	}
	getLimits(m_guard, &xmin, &xmax, &ymin, &ymax);

	/* If it's horizontal, draw a nice bounded box */	
	if (ymin == ymax)
	{
		s->box(xmin * 4, ymin * 2, (xmax-xmin) * 4+20, 32, fg);
		s->box(x0   * 4, ymin * 2, 32, 32, fg);
		if (ink > -1)
		{
			s->box((xmin+(cx/2)-x0) * 4, 
			       (ymin+cy-y0) * 2, (xmax-xmin) * 4+20, 32, ink);
			s->box(cx*2, cy*2, 32, 32, ink);

		}
		return;
	}
	/* If it's vertical, draw a nice bounded box */	
	else if (xmin == xmax)
	{
		s->box(xmin * 4, ymin * 2, 32, (ymax-ymin) * 2+30, fg);
		s->box(xmin * 4, y0 * 2, 32, 32, fg);
		if (ink > -1)
		{
			s->box( cx * 2, 
			       (ymin+cy-y0) * 2, 32, (ymax-ymin) * 2+30, ink);
			s->box(cx*2, cy*2, 32, 32, ink);
		}
		return;
	}
	/* It's diagonal. We have no option but to draw out all the positions
	 * in detail. */

	if (m_guard[6] & 0x80)	
	{
		dx = (signed char)(m_guard[3]);
		dy = (m_guard[6] >> 4) & 3;	
	}	
	else
	{
		dx = (m_guard[6] >> 4) & 3;		
		dy = (signed char)(m_guard[3]);
	}

	for (revs = 0, n = (unsigned char)m_guard[0]; revs < 2;)
	{
//		if (diag || (n == m_guard[0] && revs == 0))
//		{
			boundBox(s, 2*x0, y0, fg);
			if (ink > -1) boundBox(s, 2*x1, y1, ink);
//		}
//		else if (horiz)
//		{
//			boundHoriz(s, 2*x0, y0, fg);
//			if (ink > -1) boundHoriz(s, 2*x1, y1, ink);
//		}
//		else 
//		{
//			boundVert(s, 2*x0, y0, fg);
//			if (ink > -1) boundVert(s, 2*x1, y1, ink);
//		}
		--n;
		if (n <= 0)
		{
			++revs;
			dx = -dx;
			dy = -dy;
			n = (unsigned char)m_guard[1];
		}
		x0 += dx;
		x1 += dx;
		y0 += dy;
		y1 += dy;
	}
}

void Jsw2HVGuardian::boundBox(VideoSurface *s, int x, int y, int ink)
{
	int x0, y0, x1, y1, w, h;

	x0 = x;    if (x0 < 0) x0 = 0; if (x0 >= 255) x0 = 255;
	x1 = x+16; if (x1 < 0) x1 = 0; if (x1 >= 255) x1 = 255;
	y0 = y;    if (y0 < 0) y0 = 0; if (y0 >= 127) y0 = 127;
	y1 = y+16; if (y1 < 0) y1 = 0; if (y1 >= 127) y1 = 127;

	w = x1 - x0;
	h = y1 - y0;

	if (w <= 0 || h <= 0) return;
	s->box(2*x0, 2*y0, 2*w, 2*h, ink);
}


void Jsw2HVGuardian::boundHoriz(VideoSurface *s, int x, int y, int ink)
{
	int x0, y0, x1, y1, w, h;

	x0 = x;    if (x0 < 0) x0 = 0; if (x0 >= 255) x0 = 255;
	x1 = x+16; if (x1 < 0) x1 = 0; if (x1 >= 255) x1 = 255;
	y0 = y;    if (y0 < 0) y0 = 0; if (y0 >= 127) y0 = 127;
	y1 = y+16; if (y1 < 0) y1 = 0; if (y1 >= 127) y1 = 127;

	w = x1 - x0;
	h = y1 - y0;

	if (w <= 0 || h <= 0) return;
	s->box(2*x0, 2*y0, 2*w, 1, ink);
	s->box(2*x0, 2*y1, 2*w, 1, ink);
}



void Jsw2HVGuardian::boundVert(VideoSurface *s, int x, int y, int ink)
{
	int x0, y0, x1, y1, w, h;

	x0 = x;    if (x0 < 0) x0 = 0; if (x0 >= 255) x0 = 255;
	x1 = x+16; if (x1 < 0) x1 = 0; if (x1 >= 255) x1 = 255;
	y0 = y;    if (y0 < 0) y0 = 0; if (y0 >= 127) y0 = 127;
	y1 = y+16; if (y1 < 0) y1 = 0; if (y1 >= 127) y1 = 127;

	w = x1 - x0;
	h = y1 - y0;

	if (w <= 0 || h <= 0) return;
	s->box(2*x0, 2*y0, 1, 2*h, ink);
	s->box(2*x1, 2*y0, 1, 2*h, ink);
}



void Jsw2HVGuardian::drawFrames(VideoSurface *s, int x, int y)
{
	int frame, n, x1, xs, ys;

	x1 = m_guard[4] & 0x7F;
	if (m_guard[6] & 0x80)	
	{
		xs = (signed char)(m_guard[3]);
		ys = (m_guard[6] >> 4) & 3;	
	}	
	else
	{
		xs = (m_guard[6] >> 4) & 3;		
		ys = (signed char)(m_guard[3]);
	}
	for (n = 0; n < 8; n++)
	{
		if (m_guard[6] & 0x80)	frame = x1    & m_guard[6] & 3;
		else			frame = (8-n) & m_guard[6] & 3;

		x1 += xs;
		if (n == 4) 
		{
			xs = -xs;
			ys = -ys;
		}

		if (m_guard[6] & 0x40)	// Direction-dependent frames
		{
			if (m_guard[6] & 0x80)	
			{
				if ((~xs) & 0x80) frame += 4;
			}
			else if (ys & 0x80) frame += 4;
		}
		// Unidirectional guardian
		if (m_guard[5] & 0x80 && (xs > 0 || ys < 0))
		{
			s->fillBox(x + 64 * n, y, 32, 32, ZX_BLACK);
		}
		else
		{
			VideoBitmap *b2 = newVideoBitmap(32,32);
			b2->getSurface()->fillBox(0,0,32,32,ZX_BLACK);
			m_bmp->toSurface(b2->getSurface(), 0, 0, frame*32, 0,
					32,32);
			b2->toSurface(s, x + 64 * n, y, 0, 0, 32, 32);
			delete b2;

			//m_bmp->toSurface(s, x + 64 * n, y, frame * 32, 0, 32, 32);
		}
	}
}



void Jsw2HVGuardian::move()
{
	--m_tick;
	if (m_tick == 0)
	{
		m_xstep = -m_xstep;
		m_ystep = -m_ystep;
		m_tick = (unsigned char)m_guard[1];
	}
	m_x += m_xstep;
	m_y += m_ystep;

}

void Jsw2HVGuardian::setPos(int x, int y)
{
	m_guard[4] &= 0x80;
	m_guard[4] |= (x/2);
	m_guard[5] &= 0x80;
	m_guard[5] |= y;

	reset();
} 



void Jsw2HVGuardian::reset()
{
	m_x = m_guard[4] & 0x7F;
	m_y = m_guard[5] & 0x7F;
	m_tick = (unsigned char)m_guard[0];
	if (m_guard[6] & 0x80)
	{
		m_xstep = (signed char)(m_guard[3]);
		m_ystep = (m_guard[6] >> 4) & 3;
	}
	else
	{
		m_ystep = (signed char)(m_guard[3]);
		m_xstep = (m_guard[6] >> 4) & 3;
	}
}

unsigned Jsw2HVGuardian::getSpritePage()
{
	unsigned nSprite = m_guard[2];
	if (m_guard[4] & 0x80) nSprite += 256;
	return nSprite;
}


void Jsw2HVGuardian::setSpritePage(unsigned spriteNo)
{
	m_guard[2] = spriteNo & 0xFF;
	m_guard[4] &= 0x7F;
	m_guard[4] |= (spriteNo >> 1) & 0x80;
	constructSprites();
}



unsigned Jsw2HVGuardian::getInk()
{
	return (m_guard[6] >> 2) & 3;
}


void Jsw2HVGuardian::setInk(unsigned i)
{
	m_guard[6] &= 0xF3;
	m_guard[6] |= ((i & 3) << 2);
	constructSprites();
}

jswByte *Jsw2HVGuardian::getBase()
{
	return m_guard;
}


void Jsw2HVGuardian::getLimits(jswByte *guardian, int *x1, int *x2,
			                          int *y1, int *y2)
{
	int dx, dy;
	int n;
	int count = guardian[0];
	int x = guardian[4] & 0x7F;
	int y = guardian[5] & 0x7F;

	if (guardian[6] & 0x80)
	{
		dx = (signed char)(guardian[3]);
		dy = (guardian[6] >> 4) & 3;
	}
	else
	{
		dy = (signed char)(guardian[3]);
		dx = (guardian[6] >> 4) & 3;
	}
	*x1 = *x2 = x;
	*y1 = *y2 = y;
	// Wait until the guardian has reversed twice, so we can be sure
	// it reached both ends of its travel.
	for (n = 0; n < 2;)
	{
		--count;
		if (count <= 0)
		{
			count = guardian[1];
			dx = -dx;
			dy = -dy;
			++n;
		}
		x += dx;
		y += dy;
		if (x < *x1) *x1 = x;
		if (x > *x2) *x2 = x;
		if (y < *y1) *y1 = y;
		if (y > *y2) *y2 = y;
	}
}


/*
        virtual void drawFrames(VideoSurface *s, int x, int y) = 0;
	void drawRope(VideoSurface *s);
	virtual void move(void) = 0;

        virtual void setFrame(jswByte b) = 0;
	virtual void constructSprites(void) = 0;

// Accessor methods

	jswByte getSpritePage(void);
	void setSpritePage(jswByte page);
	jswByte getInk(void);
	virtual void setInk (jswByte b);	
        jswByte getY(void);
	jswByte getStep(void);
        jswByte getAnim(void);
	void    setAnim(jswByte b);
	virtual void setY(jswByte y) = 0; // setY() is very different
					  // in the methods it uses to 
					  // compensate for diagonal movement
					  // in JSW128 and Geoff Mode
	virtual void setStep(jswByte s)=0; // Ditto for setStep()
        virtual void setDirection(jswByte right) = 0;
        virtual jswByte getDirection(void) = 0;

// For the rope
	void setRopeLen(jswByte len);
	jswByte getRopeLen(void);
	void setRopeSwing(jswByte swing);
	jswByte getRopeSwing(void);

protected:
	jswByte m_guard[7];
	jswByte m_orig[7];
	jswByte m_counter;
//	int m_x, m_y, m_frame;	
	VideoBitmap *m_bmp, *m_bmpSmall;
	SpectrumMemory *m_mem;	

//	void moveRope(void);
};
*/
