#ifndef JSW2GUARD_HXX_INCLUDED
#define JSW2GUARD_HXX_INCLUDED

/************************************************************************

    JSWED 2.3.0 - Editor for Jet Set Willy and derivatives

    Copyright (C) 2000-2001,2004  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*************************************************************************/

class Jsw2Game;

class Jsw2Guardian
{
public:
        inline void draw(VideoBitmap *b) { draw(b->getSurface()); }
	virtual void draw(VideoSurface *s) = 0;
	virtual void move() = 0;
        virtual void drawFrames(VideoSurface *s, int x, int y) = 0;
        virtual void drawThumb(VideoSurface *s, int cx, int cy) = 0;
        virtual void drawBounds(VideoSurface *s, int ink = -1, int cx = 0, int cy = 0) = 0; 
	virtual void setPos(int cx, int cy) = 0;
	virtual jswByte *getBase() = 0;
	void getBounds(jswByte *frame, unsigned *x, unsigned *y);
};


class Jsw2HVGuardian : public Jsw2Guardian
{
public:
        Jsw2HVGuardian(jswByte *g, SpectrumMemory *mem);
        virtual ~Jsw2HVGuardian();

	virtual void draw(VideoSurface *s) ;
        virtual void drawFrames(VideoSurface *s, int x, int y);
        virtual void drawThumb(VideoSurface *s, int cx, int cy);
        virtual void drawBounds(VideoSurface *s, int ink = -1, int cx = 0, int cy = 0); 
	virtual void move(void);
	virtual void setPos(int cx, int cy);
	virtual void constructSprites(void);

	unsigned getSpritePage(void);
	void setSpritePage(unsigned page);

	unsigned getInk(void);
	virtual void setInk (unsigned value);	
	virtual jswByte *getBase();

	void reset();
	static void getLimits(jswByte *guardian, int *x1, int *x2,
					int *y1, int *y2);

protected:
	void boundBox(VideoSurface *s, int x, int y, int ink);
	void boundVert(VideoSurface *s, int x, int y, int ink);
	void boundHoriz(VideoSurface *s, int x, int y, int ink);
	void boundHL(VideoSurface *s, int x, int y, int ink);
	void boundHR(VideoSurface *s, int x, int y, int ink);
	void boundVT(VideoSurface *s, int x, int y, int ink);
	void boundVB(VideoSurface *s, int x, int y, int ink);

	jswByte *m_guard;
	jswByte m_orig[7];
	jswByte m_counter;
	VideoBitmap *m_bmp, *m_bmpSmall;
	SpectrumMemory *m_mem;	

	int m_x, m_y, m_tick, m_ystep, m_xstep;
	unsigned m_w[8], m_h[8];
};



class Jsw2Arrow : public Jsw2Guardian
{
public:
        Jsw2Arrow(jswByte *arr, SpectrumMemory *mem);
        virtual ~Jsw2Arrow();

	virtual void draw(VideoSurface *s) ;
        virtual void drawFrames(VideoSurface *s, int x, int y);
        virtual void drawThumb(VideoSurface *s, int cx, int cy);
        virtual void drawBounds(VideoSurface *s, int ink = -1, int cx = 0, int cy = 0);
	virtual void move(void);
	virtual void setPos(int cx, int cy);
	virtual void constructSprites(void);
	virtual jswByte *getBase();

	void reset();

protected:
	jswByte *m_arrow;
	jswByte m_orig[2];
	jswByte m_counter;
	VideoBitmap *m_bmp, *m_bmpSmall;
	SpectrumMemory *m_mem;	

	int m_x, m_y, m_xstep;
	unsigned m_w, m_h;
};


#endif //ndef JSW2GUARD_HXX_INCLUDED
