/************************************************************************

    JSWED 2.3.0 - Editor for Jet Set Willy and derivatives

    Copyright (C) 2000-2001,2004  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*************************************************************************/

#include "jswed.hxx"
#include "jsw2game.hxx"
#include "jsw2guard.hxx"
#include "jsw2guardform.hxx"
#include "jswdirbox.hxx"
#include "hexedit.hxx"
#include "vspritelist.hxx"

Jsw2GuardEditor::Jsw2GuardEditor(int roomNo, Jsw2Game *g,jswByte *guard,VideoBitmap *bg)
	 : VideoForm(CHAR_W * 3, CHAR_H, CHAR_W * 34, CHAR_H * 28)
{
	m_roomNo = roomNo;
	m_bmBg   = bg;
	m_bmAnim = newVideoBitmap(512, 256); 
	m_game = g;
	m_gdata = guard;
	m_ot = NULL;
	m_haveTick = 0;
	m_guard = new Jsw2HVGuardian(guard, g->getMem());

	m_bnPling  = new VideoButton('!', m_x +      CHAR_W,
					  m_y + 24 * CHAR_H,"  !   ");
	m_bnSprite = new VideoButton('S', m_x +      CHAR_W,
					  m_y + 18 * CHAR_H, " Sprite ");
        m_bnOk = new VideoButton(ZXE_OK,  m_x +      CHAR_W, 
					  m_y + m_h - 2 * CHAR_H, "  OK  ");
        m_bnCancel = new VideoButton(ZXE_CANCEL, m_x + m_w - 7 * CHAR_W, 
					m_y + m_h - 2 * CHAR_H, "Cancel");
	m_bnPling->setListener(this);	
	m_bnSprite->setListener(this);

	// Move this so we can have an X: as well.
	m_labelX = new VideoLabel(m_x + 13 * CHAR_W, m_y + 18 * CHAR_H,
					"X:");
	m_spinX = new VideoSpinControl(m_x + 15 * CHAR_W, 
				       m_y + 18 * CHAR_H, 
				       7 * CHAR_W, CHAR_H + 4);
	m_spinX->setValue(2 * (m_gdata[4] & 0x7F));
	m_spinX->setDelta(2);
	m_spinX->setLimits(0, 254);
	m_spinX->setListener(this);

	m_labelY = new VideoLabel(m_x + 24 * CHAR_W, m_y + 18 * CHAR_H,
					"Y:");
	m_spinY = new VideoSpinControl(m_x + 26 * CHAR_W, 
				       m_y + 18 * CHAR_H, 
				       7 * CHAR_W, CHAR_H + 4);
	m_spinY->setValue(m_gdata[5] & 0x7F);
	m_spinY->setLimits(0, 127);
	m_spinY->setListener(this);

	m_labelStep = new VideoLabel(m_x + 8 * CHAR_W, m_y + 24 * CHAR_H,
			"Movements");
	m_spinStep1  = new VideoSpinControl(m_x + 18 * CHAR_W, 
			m_y + 24 * CHAR_H, 7 * CHAR_W, 4 + CHAR_H);
	m_spinStep2  = new VideoSpinControl(m_x + 26 * CHAR_W, 
			m_y + 24 * CHAR_H, 7 * CHAR_W, 4 + CHAR_H);
	m_spinStep1->setSigned(1);
	//m_spinStep1->setLimits(-128, 127);
	m_spinStep2->setLimits(0, 3);
	m_spinStep1->setListener(this);
	m_spinStep2->setListener(this);
	
	m_labelTick = new VideoLabel(m_x + 8 * CHAR_W, m_y + 20 * CHAR_H,
			"Counters");
	m_spinTick1  = new VideoSpinControl(m_x + 18 * CHAR_W, 
			m_y + 20 * CHAR_H, 7 * CHAR_W, 4 + CHAR_H);
	m_spinTick2  = new VideoSpinControl(m_x + 26 * CHAR_W, 
			m_y + 20 * CHAR_H, 7 * CHAR_W, 4 + CHAR_H);
	m_spinTick1->setValue(m_gdata[0]);
	m_spinTick2->setValue(m_gdata[1]);
	m_spinTick1->setListener(this);
	m_spinTick2->setListener(this);

	m_chInk = new VideoColourChooser(m_x + CHAR_W, m_y + 22 * CHAR_H,
			m_game->getMem()->memoryAt(JSW2GUARDINKS), 4);
	m_chInk->setListener(this);

	m_cbDirection    = new VideoCheckbox(m_x + 6*CHAR_W, m_y + 20 * CHAR_W);
	m_cbDirection->setListener(this);
	m_cbDirection->setCheck(m_gdata[6] & 0x80);
	m_labelDirection = new VideoLabel(m_x + CHAR_W, m_y + 20 * CHAR_W, "L/R?");

	m_labelAnim = new VideoLabel(m_x + 8 * CHAR_W, m_y + 22 * CHAR_H, "Animation"); 
	m_spinAnim  = new VideoSpinControl(m_x + 18 * CHAR_W, m_y + 22 * CHAR_H,
					   7 * CHAR_W, CHAR_H + 4);
	m_spinAnim->setLimits(0, 7);
	m_spinAnim->setListener(this);

	m_cbUniDir = new VideoCheckbox(m_x + 26 * CHAR_W, m_y + 22 * CHAR_W);
	m_cbUniDir->setListener(this);
	m_cbUniDir->setCheck(m_gdata[5] & 0x80);

	m_labelUniDir = new VideoLabel(m_x + 27 * CHAR_W, m_y + 22 * CHAR_W,
				"Oneway");
}



Jsw2GuardEditor::~Jsw2GuardEditor()
{
	delete m_labelUniDir;
	delete m_cbUniDir;
	delete m_labelDirection;
	delete m_bmAnim;
	delete m_bnPling;
	delete m_bnSprite;
	delete m_labelX;
	delete m_spinX;
	delete m_labelY;
	delete m_spinY;
	delete m_labelAnim;
	delete m_labelStep;
	delete m_spinStep1;
	delete m_spinStep2;
	delete m_labelTick;
	delete m_spinTick1;
	delete m_spinTick2;
	m_spinAnim->bind(NULL);
	delete m_spinAnim;
	delete m_chInk;
	// nb: Delete guard *after* deleting controls; otherwise the
	// control destructors may cause callbacks that try to access
	// the guard.
	if (m_guard) delete m_guard;
}


void Jsw2GuardEditor::drawFixed(void)
{
        videoScreen->fillBox(m_x, m_y, m_w, CHAR_H, ZX_WHITE);
        videoScreen->fillBox(m_x, m_y, CHAR_W, m_h, ZX_WHITE);
        videoScreen->fillBox(m_x + m_w - CHAR_W, m_y, CHAR_W, m_h, ZX_WHITE);
        videoScreen->fillBox(m_x, m_y + 17 * CHAR_H, m_w, m_h - 17 * CHAR_H, ZX_WHITE); 
        videoScreen->box(m_x, m_y, m_w, m_h, ZX_BLACK);
        for (int n = 0; n < m_ccount; n++)
        {
                m_controls[n]->redraw();
                m_controls[n]->drawFocus(n == m_focus);
        }
}

void Jsw2GuardEditor::drawData(int tickonly)
{
	char hexBytes[80];

	if (m_bmBg) m_bmBg->toSurface(m_bmAnim->getSurface(), 0, 0);
	else	    m_bmAnim->getSurface()->fillBox(0, 0, 512, 256, ZX_BLACK);

	m_guard->draw(m_bmAnim->getSurface());
	m_guard->drawBounds(m_bmAnim->getSurface());

	m_bmAnim->toScreen(m_x + CHAR_W, m_y + CHAR_H);	

	strcpy(hexBytes, "As Hex:");
	for (int n = 0; n < 7; n++) 
	{
		sprintf(hexBytes + strlen(hexBytes), " %02x", m_gdata[n]);
	}
	strcat(hexBytes, ". Click \"!\" for manual override");
	
	videoScreen->drawSmallText(m_x +      CHAR_W,
                                    m_y + 17 * CHAR_H + (CHAR_SH / 2),
				    hexBytes, ZX_BLACK, ZX_WHITE);
	if (!tickonly)
	{	
		drawTsData();
	}
}


void Jsw2GuardEditor::redraw(void)
{
	// Don't call base class. 
	//        videoScreen->fillBox(m_x, m_y, m_w, m_h, ZX_WHITE);

	drawFixed();
	drawData(0);
}



void Jsw2GuardEditor::showChildren(int redraw)
{
	postContextMenu();
	addChild(m_bnSprite);
	addChild(m_labelX);
	addChild(m_spinX);
	addChild(m_labelY);
	addChild(m_spinY);
	addChild(m_cbDirection);
	addChild(m_labelTick);
	addChild(m_spinTick1);
	addChild(m_spinTick2);
	addChild(m_chInk);
	addChild(m_labelAnim);
	addChild(m_spinAnim);
	addChild(m_labelUniDir);
	addChild(m_cbUniDir);
	addChild(m_labelDirection);
	addChild(m_bnPling);
	addChild(m_labelStep);
	addChild(m_spinStep1);
	addChild(m_spinStep2);
	addChild(m_bnOk);
	addChild(m_bnCancel);
	VideoForm::showChildren(redraw);
}

void Jsw2GuardEditor::hideChildren(void)
{
        m_bnPling->setListener(NULL);
        m_bnSprite->setListener(NULL);
        m_spinX->setListener(NULL);
        m_spinY->setListener(NULL);
        m_spinStep1->setListener(NULL);
        m_spinStep2->setListener(NULL);
        m_spinTick1->setListener(NULL);
        m_spinTick2->setListener(NULL);
        m_chInk->setListener(NULL);
        m_cbDirection->setListener(NULL);
        m_spinAnim->setListener(NULL);
        m_cbUniDir->setListener(NULL);


	removeChild(m_cbUniDir);
	removeChild(m_bnSprite);
	removeChild(m_spinY);
	removeChild(m_labelY);
	removeChild(m_spinX);
	removeChild(m_labelX);
	removeChild(m_labelTick);
	removeChild(m_spinTick1);
	removeChild(m_spinTick2);
	removeChild(m_labelStep);
	removeChild(m_cbUniDir);
	removeChild(m_labelUniDir);
	removeChild(m_spinStep1);
	removeChild(m_spinStep2);
	removeChild(m_bnPling);
	removeChild(m_cbDirection);
	removeChild(m_labelDirection);
	removeChild(m_chInk);
	removeChild(m_labelAnim);
	removeChild(m_spinAnim);
	removeChild(m_bnOk);
	removeChild(m_bnCancel);
	preContextMenu();
	VideoForm::hideChildren();
}



void Jsw2GuardEditor::preContextMenu()
{
	if (m_haveTick)
	{
		setTickListener(m_ot);
		m_haveTick = 0;
	}
}


void Jsw2GuardEditor::postContextMenu()
{
        if (!m_haveTick)
        {
                m_ot = setTickListener(this);
                m_haveTick = 1;
        }
}





void Jsw2GuardEditor::onTick(void)
{
	++m_tick;
	if ((m_tick % 5) == 0)
	{
		m_guard->move();
		drawData(1);
	}
}

int Jsw2GuardEditor::setSprite(void)
{
	unsigned spriteNo;

	preContextMenu();

        VSpriteList list(m_game->getMem());
        m_game->fillSprites(m_roomNo, &list);
        list.setTitle("Choose sprite");

	spriteNo = m_gdata[2] | ((m_gdata[4] & 0x80) << 1);
        list.setSelected(spriteNo * 32 + JSW2SPRITES);

        int rv = list.doModal();

        if (rv >= ZXE_QUIT) return rv;
        if (rv != ZXE_OK)   
	{
		postContextMenu();
		return ZXE_CONTINUE;
	}
	spriteNo = (list.getSelected() - JSW2SPRITES) / 32;
	m_gdata[2] = spriteNo & 0xFF;
	m_gdata[4] &= 0x7F;
	m_gdata[4] |= (spriteNo >> 1) & 0x80;

	postContextMenu();
	m_guard->setSpritePage(spriteNo);
	drawData(0);
	return ZXE_CONTINUE;
}





int Jsw2GuardEditor::onButtonSelect(VideoButton *b)
{
	switch(b->getId())
	{
		case 'S': return setSprite();
		case '!': return onPling();

	}
	return b->getId();
}



void Jsw2GuardEditor::onChooseColour(VideoColourChooser *c)
{
	if (c == m_chInk)
	{
		m_gdata[6] &= 0xF3;
		m_gdata[6] |= ((c->getColour() & 3) << 2);
		m_guard->setInk(c->getColour());
	}
}


int  Jsw2GuardEditor::onCheckboxSelect(VideoCheckbox *c, int value)
{
	if (c == m_cbDirection)
	{
		m_gdata[6] &= 0x7F;
		if (value) m_gdata[6] |= 0x80;
		m_guard->reset();
		m_cbDirection->setCheck(m_gdata[6] & 0x80);
		return ZXE_OK;	
	}
	if (c == m_cbUniDir)
	{
		m_gdata[5] &= 0x7F;
		if (value) m_gdata[5] |= 0x80;
		m_guard->setInk(m_chInk->getColour());
		m_cbUniDir->setCheck(m_gdata[5] & 0x80);
		return ZXE_OK;	
	}
	return ZXE_CONTINUE;
}


void Jsw2GuardEditor::addChild(VideoControl *c)
{
	VideoForm::addChild(c);
	c->setVisible(1, 0);
}

void Jsw2GuardEditor::removeChild(VideoControl *c)
{
	c->setVisible(0, 0);	
	VideoForm::removeChild(c);
}

void Jsw2GuardEditor::drawTsData()
{
	// Memo to self: Give functions more self-documenting names.
	m_chInk->setColour((m_gdata[6] >> 2) & 3);

	m_spinAnim->setValue((m_gdata[6] & 3) | (m_gdata[6] >> 4) & 4);
	m_spinStep1->setValue((signed char)m_gdata[3]);
	m_spinStep2->setValue((m_gdata[6] >> 4) & 3);
	m_spinX->setValue(2 * (m_gdata[4] & 0x7F));
	m_spinY->setValue(m_gdata[5] & 0x7F);
	m_spinTick1->setValue(m_gdata[0]);
	m_spinTick2->setValue(m_gdata[1]);
}

void Jsw2GuardEditor::subEditChange(VideoControl *v)
{
	if (v == m_spinAnim)
	{
		m_gdata[6] &= ~0x43;
		m_gdata[6] |=  (m_spinAnim->getValue() & 3);
		m_gdata[6] |= ((m_spinAnim->getValue() & 4) << 4);
	}
	if (v == m_spinStep1)
	{
		m_gdata[3] = m_spinStep1->getValue();
		m_guard->reset();
	}
	if (v == m_spinStep2)
	{
		m_gdata[6] &= 0xCF;
		m_gdata[6] |= ((m_spinStep2->getValue() & 3) << 4);
		m_guard->reset();
	}
	if (v == m_spinTick1)
	{
		m_gdata[0] = m_spinTick1->getValue();
		m_guard->reset();
	}
	if (v == m_spinTick2)
	{
		m_gdata[1] = m_spinTick2->getValue();
		m_guard->reset();
	}
	if (v == m_spinX)
	{
		m_gdata[4] &= 0x80;
		m_gdata[4] |= m_spinX->getValue() / 2;
		m_guard->setPos((m_gdata[4] & 0x7F) * 2, (m_gdata[5] & 0x7F));
	}
	if (v == m_spinY)
	{
		m_gdata[5] &= 0x80;
		m_gdata[5] |= m_spinY->getValue() & 0x7F;
		m_guard->setPos((m_gdata[4] & 0x7F) * 2, (m_gdata[5] & 0x7F));
	}
}





int Jsw2GuardEditor::onPling(void)
{
	HexEditor hex(m_gdata, 7);

	preContextMenu();
	int rv = hex.doModal();
	postContextMenu();
	if (rv >= ZXE_QUIT) return rv;
	if (rv != ZXE_OK) return ZXE_CONTINUE;

        delete m_guard;
        m_guard = new Jsw2HVGuardian(m_gdata, m_game->getMem());
        drawData(0);                                                            
	return ZXE_CONTINUE;
}

void Jsw2GuardEditor::onEditChange(VideoControl *v)
{
	subEditChange(v);
}
