/************************************************************************

    JSWED 2.3.0 - Editor for Jet Set Willy and derivatives

    Copyright (C) 2000-2001,2004  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*************************************************************************/

#include "jswed.hxx"
#include "jsw2game.hxx"
#include "jsw2patchpage.hxx"
#include "filetypelist.hxx"
#include "patch.hxx"

Jsw2PatchPage::Jsw2PatchPage(Jsw2Game *g)
{
	int x, y;
	m_game = g;

	x = 9 * CHAR_W;
	y = CHAR_H;

	m_fflabel = new VideoLabel(x, y, "File format:");

	m_fflist = new FileTypeList((jswByte *)&m_game->getMem()->m_type, 
				x + 13 * CHAR_W, y, 8 * CHAR_W, 6*CHAR_H);
	m_fflist->setTitle("");
	m_fflist->requireClick(1);

	setTitle("Game");
	addChild(m_fflabel);
	addChild(m_fflist);
	setFocus(m_fflist);

	m_roomCountLabel = new VideoLabel(x, 7 * CHAR_H + 4, "Rooms in game");
	m_roomCount = new VideoSpinControl(x + 15 * CHAR_W, 
			7 * CHAR_H + 4, 7 * CHAR_W, CHAR_H + 4);
	m_roomCount->setListener(this);
	m_roomCount->setLimits(1,255);
	addChild(m_roomCountLabel);
	addChild(m_roomCount);

	m_itemCountLabel = new VideoLabel(x, 9 * CHAR_H + 4, "Items to win");
	m_itemCount = new VideoSpinControl(x + 15 * CHAR_W, 
			9 * CHAR_H + 4, 7 * CHAR_W, CHAR_H + 4);
	m_itemCount->setListener(this);
	m_itemCount->setLimits(0,255);
	addChild(m_itemCountLabel);
	addChild(m_itemCount);

	y += 8 * CHAR_H + 2;

	m_patchData = new Patch[20];
	m_patchCount = 0;
	m_patchMax = 20;

	/* Add the patches */
//	patchSpecialsAnywhereBB.label    = new VideoLabel(x, y, "Specials by room");
//	patchSpecialsAnywhereBB.checkbox = new VideoCheckbox(x + 21 * CHAR_W, y);
//	addPatch(&patchSpecialsAnywhereBB);
	y += CHAR_H + 4;
	updatePatches();

}

Jsw2PatchPage::~Jsw2PatchPage()
{	
	int n;
	delete m_fflabel;
	delete m_fflist;
	delete m_roomCount;
	delete m_roomCountLabel;
	delete m_itemCount;
	delete m_itemCountLabel;

	for (n = 0; n < m_patchCount; n++)	
	{
		delete m_patchData[n].label;
		delete m_patchData[n].checkbox;
	}	
	delete m_patchData;
}


void Jsw2PatchPage::updatePatches(void)
{
	int check;
	int n, m, value;
	SpectrumMemory *mem = m_game->getMem();
	
	for (n = 0; n < m_patchCount; n++)
	{
		check = -1;
		for (m = 0; m < m_patchData[n].lineCount; m++)
                {

                	value = mem->peek(m_patchData[n].lines[m].address);

			if (value == m_patchData[n].lines[m].oldVal &&
			    (check == -1 || check == 0)) check = 0;
			else if (value == m_patchData[n].lines[m].newVal && 
                            (check == -1 || check == 1)) check = 1;
			else check = 2;
// for debugging	
// 			printf("%04x: %02x %02x %02x  %d\n", 
//				m_patchData[n].lines[m].address, 
//				m_patchData[n].lines[m].oldVal, 
//				m_patchData[n].lines[m].newVal, 
//				value, check);
		}
		if (check == -1) check = 0;
		if (check == 2) m_patchData[n].checkbox->setStates(3);
		else		m_patchData[n].checkbox->setStates(2);
		m_patchData[n].checkbox->setCheck(check);
		m_patchData[n].checkbox->setListener(this);	
	}
}


void Jsw2PatchPage::addPatch(Patch *p)
{
	if (m_patchCount == m_patchMax)
	{
		Patch *q = new Patch[2 * m_patchMax];
		memcpy(q, m_patchData, m_patchMax * sizeof(Patch));
		delete m_patchData;
		m_patchData = q;
		m_patchMax *= 2;
	}
	memcpy(&m_patchData[m_patchCount], p, sizeof(Patch));
	++m_patchCount;
	addChild(p->label);
	addChild(p->checkbox);
}


int Jsw2PatchPage::redraw(int whichRectangle)
{
	int rv = VideoNotePage::redraw(whichRectangle);

	m_roomCount->setValue(m_game->getRoomCount());
	m_itemCount->setValue(m_game->getItemCount());

	videoScreen->bottomBar("Game properties");
	return rv;
}

void Jsw2PatchPage::onEditChange(VideoControl *c)
{
	if (c == m_roomCount)
		m_game->setRoomCount(m_roomCount->getValue());
	if (c == m_itemCount)
		m_game->setItemCount(m_itemCount->getValue());
}



int Jsw2PatchPage::onCheckboxSelect(VideoCheckbox *b, int value)
{
	int n, m;
	jswByte from, to;
	int addr;
	SpectrumMemory *mem = m_game->getMem();

	if (value == 2) return ZXE_CANCEL;
	if (b->getCheck() == 2) return ZXE_CANCEL;	

	for (n = 0; n < m_patchCount; n++)
	{
		if (m_patchData[n].checkbox == b)
		{
			for (m = 0; m < m_patchData[n].lineCount; m++)
			{
				addr = m_patchData[n].lines[m].address;
				if (value)
				{ 
					from = m_patchData[n].lines[m].oldVal;
					to   = m_patchData[n].lines[m].newVal;
				}
				else
				{
                                        to   = m_patchData[n].lines[m].oldVal;
                                        from = m_patchData[n].lines[m].newVal;
				}
				if (mem->peek(addr) == from) mem->poke(addr,to);
			}
			updatePatches();
			return ZXE_OK;
		}
	}
	return ZXE_CANCEL;
}

int Jsw2PatchPage::onReveal(int rect)
{
	m_roomCount->setValue(m_game->getRoomCount());
	m_itemCount->setValue(m_game->getItemCount());
	return VideoNotePage::onReveal(rect);
}

/*
int Jsw2PatchPage::preContextMenu(void)
{
	VideoMenu vm("Game options", "Cancel", "Clear rooms", 
			"Clear guardians", "Clear sprites", "Exit", NULL);
	int rv = vm.doModal();

	if (rv >= ZXE_QUIT) return rv;
	if (rv != ZXE_OK) return ZXE_CONTINUE;
	switch(vm.getSelected())
	{
		case 0: return ZXE_CONTINUE;
		case 1: return m_game->clearRooms();
		case 2: return m_game->clearGuardians();
		case 3: return m_game->clearSprites();
		case 4: return ZXE_OK;
	}
	return ZXE_CONTINUE;
}
*/
