/************************************************************************

    JSWED 2.3.2 - Editor for Jet Set Willy and derivatives

    Copyright (C) 2000-2001,2004-6  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*************************************************************************/

#include "jswed.hxx"
#include "room.hxx"
#include "jsw2room.hxx"
#include "jsw2game.hxx"
#include "jsw2guard.hxx"
//
// A note on JSW2 items.
//
// In JSW2, there is no overall item table. Instead, each room has its own
// bitmap of items collected. This bitmap occupies two bytes; therefore, a
// JSW2 room can have at most 16 items.
//

Jsw2Room::Jsw2Room(int rmid, Jsw2Game *g) : Room(9, rmid)
{
	m_animMask = 0x0F;
	m_bmConvey0 = NULL;
	m_bmConvey2 = NULL;
	m_mem = g->getMem();

	m_data = g->getRoomData(rmid);
	m_game = g;

	unsigned shape = m_data[0] + 256 * m_data[1];
	m_shape = m_mem->memoryAt(shape);
	m_arrowCount = 0;

	decompress();
}

Jsw2Room::Jsw2Room(Jsw2Room &other) : Room(other)
{
	m_animMask = 0x0F;
	m_bmConvey0 = NULL;
	m_bmConvey2 = NULL;
	Jsw2Room::copyFrom(&other);
}
/*
Jsw2Room::Jsw2Room() : Room(9, -1)
{
	m_mem = NULL;
	m_arrowCount = 0;
	m_shape = NULL;
	m_T4 = 0;
	m_T5 = 0;
}*/

Jsw2Room::~Jsw2Room()
{
	if (m_bmConvey0) delete m_bmConvey0;
	if (m_bmConvey2) delete m_bmConvey2;
}

void Jsw2Room::decompress()
{
	int n, src, dst, rep, cel;

	// Decompress room shape.
	for (src = dst = 0; dst < 512; ++src)
	{
		if (m_shape[src] < 0x90)
		{
			rep = (m_shape[src] & 0x0F) + 1;
			cel = (m_shape[src] & 0xF0) >> 4;
		}
		else
		{
			rep = m_shape[src] - 0x7F;
			cel = 0;
		}
		while (rep)
		{
			m_cells[dst++] = cel;
			if (dst >= 512) break;
			--rep;
		}
	}
	// Cell shapes
	memset(m_cellShapes, 0, sizeof(m_cellShapes));
	memset(m_cellNumber, 0, sizeof(m_cellNumber));
	jswByte overflow = m_data[2];
	jswByte mask = 0x80;
	for (n = 0; n < 8; n++)
	{
		unsigned cellNo = m_data[3 + n];
		if (overflow & mask) cellNo += 256;
		m_cellNumber[n+1] = cellNo;

		jswByte *cellShape = m_cellShapes + 9 * (n+1);
		if (n == 5)
		{
			memcpy(m_itemBits,
			       m_mem->memoryAt(JSW2CELLS + 9 * cellNo), 9); 
			memcpy(m_cellShapes + 81, m_itemBits, 9);
			if (m_itemBits[0] & 0x80)
			{
				for (int j = 1; j < 9; j++) 
					m_itemBits[j] = ~m_itemBits[j];
				m_cellShapes[81] = 0x38;
			}
			else	m_cellShapes[81] = 0x07;
		}
		memcpy(cellShape, m_mem->memoryAt(JSW2CELLS + 9 * cellNo), 9); 	
		// Top bit of this byte = Invert.
		if (cellShape[0] & 0x80)
		{
			for (int j = 1; j < 9; j++) 
				cellShape[j] = ~cellShape[j];
		}	
		cellShape[0] &= 0x7F;
		cellShape[0] |= 0x40;	// All colours bright, or so it seems
		mask = mask >> 1;
	}
	// Room title
	sprintf((char *)m_title, "%-32.32s", "");

	jswByte *tDest = m_title + (m_data[0x0B] & 0x1F);
	jswByte *tSrc  = m_data + 0x0C;

	tSrc += expandString(0, tSrc, &tDest, m_title + 33);
	memcpy(m_exits, tSrc, 4);
//	printf("Room %02x exits: %02x %02x %02x %02x\n", 
//			m_roomNumber, m_exits[0], m_exits[1], 
//				m_exits[2], m_exits[3]);
	tSrc += 4;
	m_T4 = *tSrc++;
	if (m_T4 & 0x10)
	{
		m_T5 = *tSrc++;
//		if (m_T5 & 0x7F) printf("Room %03d (%-32.32s): Patch vector %d\n",
//				m_roomNumber, m_title, m_T5 & 0x7F); 
	}
	else
	{
		m_T5 = 0;
	}
	// You can only have 8 guardians. Any more => assume the room is 
	// invalid.
	if ((m_T4 & 0x0F) > 8) m_T4 &= 0xF0;

	memcpy(m_guardDefs, tSrc, 7 * (m_T4 & 0x0F));
	tSrc += 7 * (m_T4 & 0x0F);

	if (m_T5 & 0x80)
	{
		m_arrowCount = *tSrc++;
		if (m_arrowCount > 8) m_arrowCount = 8;
		memcpy(m_arrowDefs, tSrc, 2 * m_arrowCount);
	}		
}


void Jsw2Room::putRle(int lct, int reps, jswByte *buffer, unsigned *offset)
{
	int nr;
	jswByte c;

	while (reps)
	{
		if (lct == 0)
		{
			if (reps > 128) nr = 128;
			else	        nr = reps;

			if (nr > 16) 	c = 0x7F + nr;
			else		c = nr - 1;
		}
		else
		{
			if (reps > 16)	nr = 16;
			else		nr = reps;
			
			c = (nr - 1) | (lct << 4); 
		}
		buffer[*offset] = c;
		++(*offset);
		reps -= nr;	
	}
}

unsigned Jsw2Room::compressShape(jswByte *buffer, unsigned offset)
{
	int n;
	int lct  = -1;
	int reps = 0;
	unsigned off0 = offset;

	for (n = 0; n < 512; n++)
	{
		if (m_cells[n] != lct)
		{
			putRle(lct, reps, buffer, &offset);
			lct = m_cells[n];
			reps = 1;
		}
		else
		{
			++reps;
		}
	}
	putRle(lct, reps, buffer, &offset);
	return offset - off0;
}


unsigned Jsw2Room::compress(jswByte *buffer, unsigned table, unsigned offset,
		unsigned shapeOffset, char *title)
{
	int n, m;
	jswByte bt;

	unsigned off0 = offset;
// 
// First few bytes of the room are fixed...
//
	buffer[offset++] = shapeOffset & 0xFF;
	buffer[offset++] = (shapeOffset >> 8) & 0xFF;

	bt = 0;
	for (n = 1; n < 9; n++)
	{
		bt = bt << 1;
		if (m_cellNumber[n] > 255) bt |= 1;	
	}
	buffer[offset++] = bt;
	for (n = 1; n < 9; n++)
	{
		buffer[offset++] = m_cellNumber[n] & 0xFF;
	}
//
// X position of room title
//
	bt = 0;
	for (n = 0; n < 32; n++)
	{
		if (m_title[n] != ' ') break;
		++bt;
	}
	buffer[offset++] = bt;
//
// Room title. This was compressed when the dictionary was generated.
//
	for (n = 0; n < strlen(title); n++)
	{
		buffer[offset++] = title[n];
	}	
	--offset;	// Set bit 7 on last character
	buffer[offset++] |= 0x80;
//
// Exits, T4 and T5
//
	memcpy(buffer + offset, m_exits, 4);
	offset += 4;

	if (m_arrowCount) m_T5 |= 0x80;
	else		  m_T5 &= 0x7F;
	if (m_T5)         m_T4 |= 0x10;
	else		  m_T4 &= 0xEF;

	buffer[offset++] = m_T4;
	if (m_T4 & 0x10)
	{
		buffer[offset++] = m_T5;
	}
//
// Guardians
//
	memcpy(buffer + offset, m_guardDefs, 7 * getGuardianCount());
	offset += 7 * getGuardianCount();
	if (m_arrowCount)
	{
		buffer[offset++] = m_arrowCount;
		memcpy(buffer + offset, m_arrowDefs, 2 * m_arrowCount);
		offset += 2 * m_arrowCount;
	}
	return offset - off0;
}




int Jsw2Room::expandString(int addr, jswByte *tSrc, jswByte **tDest, 
		jswByte *last)
{
	jswByte *base = tSrc;
	do
	{
		if (tSrc[0] < 0x1F) 
		{
			expandToken(tSrc[0], tDest, last);
		}
		else
		{
			if ((*tDest) < last)
			{
				(*tDest)[0] = tSrc[0] & 0x7F;
			}
			++(*tDest);
		}
		++tSrc;
		++addr;
		if (addr > 0xFFFF) return tSrc - base;
	} while (!(tSrc[-1] & 0x80));

	return tSrc - base;
}


int Jsw2Room::expandToken(jswByte tokenID, jswByte **tDest, jswByte *last)
{
	int addr = 0xFA81;
	jswByte *dict = m_mem->memoryAt(0xFA81);
	while (tokenID)
	{
		while (!(dict[0] & 0x80)) ++dict;
		--tokenID;
		++dict;
		++addr;
		if (addr > 0xFFFF) return 0;
	}
	expandString(addr, dict,  tDest, last);
	if ((*tDest) < last) 
	{
		(*tDest)[0] = ' ';
	}
	++(*tDest);
	return 1;
}

jswByte *Jsw2Room::getExits()
{
	return m_exits;
}


jswByte *Jsw2Room::getCellPattern(int celltype)
{
	return m_cellShapes + 9 * celltype;
}

int Jsw2Room::getConveyorDir()
{
	return 0;
}

int Jsw2Room::getConveyorY()
{
	return 0;
}

int Jsw2Room::getItemCoords(int *px, int *py, int *counts)
{
	int n, x, y;

	for (n = y = 0; y < 16; y++)
		for (x = 0; x < 32; x++)
	{
		if (m_cells[y*32+x] == 6) 
		{
			px[n] = x;
			py[n] = y;
			counts[n] = 1;
			++n;
		}

	}
	return n;
}

int Jsw2Room::getTitle(char *buf)
{
	sprintf(buf, "%-32.32s", m_title);
	return 32;
}

void Jsw2Room::setTitle(const char *buf)
{
	sprintf((char *)m_title, "%-32.32s", buf);
}


int Jsw2Room::getBackground()
{
	return 0;
}

int Jsw2Room::getCellTypeCount(bool)
{
	return 9;
}

int Jsw2Room::getCellBehaviour(int index)
{
	switch(index)
	{
		case 0: return CB_AIR;
		case 1: return CB_WATER;
		case 2: return CB_EARTH;
		case 3: return CB_FIRE;
		case 4: return CB_RRAMP;
		case 5: return CB_LCONV;
		case 6: return CB_AIR;	// Item on air return CT_ITEM;
		case 7: return CB_LRAMP;
		case 8: return CB_RCONV;
	}
	return CB_WATER;
}

void Jsw2Room::setCell(int x, int y, int type)
{
	if (x < 0 || x > 31 || y < 0 || y > 15) return;
	m_cells[y*32+x] = type;
}

int Jsw2Room::getCellAt(int x, int y)
{
	if (x < 0 || x > 31 || y < 0 || y > 15) return 0;
	return m_cells[y*32+x];
}


int Jsw2Room::getDrawnCellAt(int x, int y)
{
	int cell = getCellAt(x, y);
	if (cell == 6) return 0;	// Convert items to air
	return cell;
}

int Jsw2Room::getRoomLength() const
{
	return 640;	// Arbitrary :-)
}

void Jsw2Room::getCells(Cell *c)
{

}

void Jsw2Room::setCells(Cell *c)
{

}

void Jsw2Room::clear()
{
	memset(m_cells, 0, sizeof(m_cells));
	setTitle("");
	m_T4 = 0;
	m_T5 = 0;
	m_arrowCount = 0;
}

void Jsw2Room::mirrorH()
{
	int x, y;
	for (y = 0; y < 16; y++)
		for (x = 0; x < 16; x++)
		{
			jswByte c = m_cells[y*32+x];
			m_cells[y*32+x] = m_cells[y*32+(31-x)];
			m_cells[y*32+(31-x)] = c;
		}
}


void Jsw2Room::mirrorV()
{
	int x, y;
	for (y = 0; y < 8; y++)
		for (x = 0; x < 32; x++)
		{
			jswByte c = m_cells[y*32+x];
			m_cells[y*32+x] = m_cells[(15-y)*32+x];
			m_cells[(15-y)*32+x] = c;
		}
}



void Jsw2Room::getRoomMemMap(char *buf)
{

}



void Jsw2Room::drawItems(VideoBitmap *bm, int inkidx)
{
	int xs[256];
        int ys[256];
	int counts[256];
        int n, count;
//
// This is an approximation to the real item flash algorithm; but it 
// doesn't seem to do too bad a job.
//
	jswByte ink = (inkidx + m_animMask) & 7;

        count = getItemCoords(xs, ys, counts);
        for (n = 0; n < count; n++)
        {
		ink = (ink + 1) & 7;
		if (!ink) ++ink;
		drawItem(bm, xs[n], ys[n], ink | ZX_BRIGHT);
	}
}


jswByte *Jsw2Room::getGuardianBuffer()
{
	return m_guardDefs;
}


int Jsw2Room::getItemCount()
{
	int n, c;

	for (n = c = 0; n < 512; n++)
	{
		if (m_cells[n] == 6) ++c;
	}
	return c;
}


void Jsw2Room::copyFrom(Jsw2Room *other)
{
	memcpy(m_cellNumber, other->m_cellNumber, sizeof(m_cellNumber));
	memcpy(m_cellShapes, other->m_cellShapes, sizeof(m_cellShapes));
	memcpy(m_cells,      other->m_cells,      sizeof(m_cells));
	memcpy(m_title,      other->m_title,      sizeof(m_title));
	memcpy(m_exits,      other->m_exits,      sizeof(m_exits));
	m_T4 = other->m_T4;
	m_T5 = other->m_T5;
	m_data = other->m_data;
	m_shape = other->m_shape;
	memcpy(m_guardDefs,  other->m_guardDefs,      136);
	memcpy(m_arrowDefs,  other->m_arrowDefs,      16);
	m_arrowCount = other->m_arrowCount;
	m_mem = other->m_mem;
}

void Jsw2Room::createCellPatterns()
{
	jswByte *bm;
	int n;

	// Regenerate cell shapes from cell numbers
	for (n = 0; n < 8; n++)
	{
		unsigned cellNo = m_cellNumber[n+1];

		jswByte *cellShape = m_cellShapes + 9 * (n+1);
		if (n == 5)
		{
			memcpy(m_cellShapes + 81, 
			       m_mem->memoryAt(JSW2CELLS + 9 * cellNo), 9); 
			if (m_cellShapes[81] & 0x80)
				m_cellShapes[81] = 0x38;
			else	m_cellShapes[81] = 0x07;
		}
		memcpy(cellShape, m_mem->memoryAt(JSW2CELLS + 9 * cellNo), 9); 	
		// Top bit of this byte = Invert.
		if (cellShape[0] & 0x80)
		{
			for (int j = 1; j < 9; j++) 
				cellShape[j] = ~cellShape[j];
		}	
		cellShape[0] &= 0x7F;
		cellShape[0] |= 0x40;	// All colours bright, or so it seems
	}
//
// If either conveyor cell definition has changed, recreate the 
// conveyor bitmaps.
//
	for (int j = 5; j <= 8; j += 3)
	{
		bm = m_zxCellBitmaps + 9 * j;
		jswByte *buf = getCellPattern(j/2);
		if (m_cellBitmaps[j])
		{
			if (memcmp(bm, buf, 9))
			{
				if (m_bmConvey0) delete m_bmConvey0;
				if (m_bmConvey2) delete m_bmConvey2;
				m_bmConvey0 = NULL;
				m_bmConvey2 = NULL;
			}
		}
	}
	Room::createCellPatterns();
}


#define CONVEYORANIM 2

void Jsw2Room::animateConveyor()
{
	int n, m, c, ct, fg, bg, x, y;

	for (n = 0; n < 512; n++)
	{
		if (m_cells[n] == 5 || m_cells[n] == 8) break;
	}
	if (n == 512) return;	// No conveyors in this room.
	ct = m_cells[n];

	if (!m_bmConvey0)
	{
		fg = m_cellShapes[9 * ct] & 7;
		bg = (m_cellShapes[9 * ct] >> 3) & 7;
		m_bmConvey0 = newVideoBitmap(16, 2);
		m_bmConvey0->getSurface()->fillBox(0, 0, 8, 2, bg | ZX_BRIGHT);
		m_bmConvey0->getSurface()->fillBox(8, 0, 8, 2, fg | ZX_BRIGHT);
	}
	if (!m_bmConvey2)
	{
		fg = m_cellShapes[9 * ct] & 7;
		bg = (m_cellShapes[9 * ct] >> 3) & 7;
		m_bmConvey2 = newVideoBitmap(16, 2);
		m_bmConvey2->getSurface()->fillBox(0, 0, 8, 2, fg | ZX_BRIGHT);
		m_bmConvey2->getSurface()->fillBox(8, 0, 8, 2, bg | ZX_BRIGHT);
	}
	if (ct == 5) // Conveyor goes left
	{
		m_bmConvey0->getSurface()->rotateL(0, 0, 16, 2, CONVEYORANIM);
		m_bmConvey2->getSurface()->rotateR(0, 0, 16, 2, CONVEYORANIM);
		m_animMask = ((m_animMask & 0x80) >> 7) | (m_animMask << 1);
	}
	else	// Conveyor goes right
	{
		m_bmConvey0->getSurface()->rotateR(0, 0, 16, 2, CONVEYORANIM);
		m_bmConvey2->getSurface()->rotateL(0, 0, 16, 2, CONVEYORANIM);
		m_animMask = ((m_animMask & 1) << 7) | (m_animMask >> 1);
	}
	if (m_T4 & 0x40)
	{
		m_bmConvey0->toSurface(m_cellBitmaps[2*ct+1], 0, 0);
		if (m_T4 & 0x20)
		{
			m_bmConvey2->toSurface(m_cellBitmaps[2*ct+1], 0, 4);
		}
	}
}


void Jsw2Room::drawOneOffs(VideoBitmap *bm, bool flashonly, bool draw)
{
	int n, m, c, ct, fg, bg, x, y;

	// Animate conveyors
	for (n = 0; n < 512; n++)
	{
		if (m_cells[n] == 5 || m_cells[n] == 8) 
		{
			drawCell(bm, (n % 32), (n / 32), flashonly, draw);	
		}
	}
	Room::drawOneOffs(bm, flashonly, draw);
}



jswByte Jsw2Room::getBorder()
{
	int x = 0;
	while (m_title[x] == ' ' && x < 32) ++x;
	return (x & 7);
}


void Jsw2Room::setBorder(int bdr)
{
}


void Jsw2Room::clearItems()
{
	int x, y;

	for (y = 0; y < 16; y++) for (x = 0; x < 32; x++)
	{
		if (getCellAt(x, y) == 6) setCell(x, y, 0);
	}
}

void Jsw2Room::addItem(int x, int y)
{
	if (getItemCount() >= 16) return;
	setCell(x, y, 6);
}

jswByte *Jsw2Room::getItemBitmap()
{
	return m_itemBits + 1;
}

void Jsw2Room::setItemBitmap(jswByte *bm)
{
	jswByte bits[9];

	// Copy to our own item bitmap, 
	// and then map it as a cell in the normal way.
	memcpy(bits + 1, bm, 8);
	bits[0] = 7;
	int ncur = mapCellGraphic(bits, false);
	memcpy(m_cellShapes + 56, bits, 9);
	m_cellNumber[6] = ncur;
	m_data[8] = (ncur & 0xFF);
	if (ncur > 256) m_data[2] |= 4;
	else		m_data[2] &= ~4;	
	createCellPatterns();
}


void Jsw2Room::addRefs()
{
	int n;
	for (n = 0; n < 9; n++) 
	{
		m_game->addRef(CELL(m_cellNumber[n]));
	}
	jswByte *buf = getGuardianBuffer();
	for (n = 0; n < (m_T4 & 0x0F); n++)
	{
		jswByte *guard = buf + 7*n;
		unsigned spriteNo = guard[2];
	       	if (guard[4] & 0x80) spriteNo += 256;
		int anim = guard[6] & 3;
		if (guard[6] & 0x40) anim |= 4;
		for (int frame = 0; frame < 8; frame++)
		{
			m_game->addRef(SPRITE((spriteNo + (frame & anim)) * 32 
						+ JSW2SPRITES));
		}
	}
}

static const char *st_exnames[4] =
{
	"left", "up", "right", "down"
};

bool Jsw2Room::exportExits(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr node)
{
	char buf[30];
	jswByte *ex = getExits();
	int n;

	for (n = 0; n < 4; n++)
	{
		sprintf(buf, "%d", ex[n] - 1);
		xmlSetProp(node, (xmlChar *)st_exnames[n], (xmlChar *)buf);	
	}
	return true;
}

bool Jsw2Room::importExits(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr node, 
			int roomMap[256])
{
	int n, r;
	char *attr;
	jswByte *ex = getExits();
	if (!ex) return true;

	for (n = 0; n < 4; n++)
	{
		attr = (char *)xmlGetProp(node, (xmlChar *)st_exnames[n]);
		if (attr) 
		{
			for (r = 0; r < 256; r++) 
			{
				if (roomMap[r] == atoi(attr))
				{
					ex[n] = 1 + r;
				}
			}
			xmlFree(attr);
		}	
	}	
	return true;
}


// We export guardians in two formats: Native and downlevel JSW.
// XXX Do this for other non-JSW games, such as Manic Miner and Geoff Mode
// as well?
bool Jsw2Room::exportGuardians(xmlDocPtr doc, xmlNsPtr ns, 
		xmlNodePtr grandparent)
{
	int n;
	char data[40];
	jswByte *buf = getGuardianBuffer();
	xmlNodePtr node      = xmlNewNode(ns, (xmlChar *)"guardians");
	xmlNodePtr downlevel = xmlNewNode(ns, (xmlChar *)"guardians");
	xmlSetProp(node,      (xmlChar *)"type", (xmlChar *)"jsw2");
	xmlSetProp(downlevel, (xmlChar *)"type", (xmlChar *)"jsw");

	for (n = 0; n < (m_T4 & 0x0F); n++)
	{
		if (!exportJsw2Guardian(doc, ns, node, downlevel, buf + 7 * n))
			return false;
	}
	// Export ropes. The JSW2 rope has no attributes, because JSW2 only
	// records its existence. Whereas the JSW1 rope has a set of standard
	// values (copied, in this case, from the rope in the Beach).
	if (m_T4 & 0x80)
	{
		xmlNodePtr r = xmlNewNode(ns, (xmlChar *)"rope");
		xmlAddChild(node, r);
		r = xmlNewNode(ns, (xmlChar *)"rope");
		xmlSetProp(r, (xmlChar *)"direction", (xmlChar *)"L");
		xmlSetProp(r, (xmlChar *)"pos", (xmlChar *)"34");
		xmlSetProp(r, (xmlChar *)"x", (xmlChar *)"14");
		xmlSetProp(r, (xmlChar *)"len", (xmlChar *)"32");
		xmlSetProp(r, (xmlChar *)"swing", (xmlChar *)"54");
		xmlAddChild(downlevel, r);
	}
	if (m_T5 & 0x80)	// Arrows present?
	{
		// Export them
		for (n = 0; n < m_arrowCount; n++)
		{
			if (!exportJsw2Arrow(doc, ns, node, downlevel, 
				m_arrowDefs + 2 * n)) return false;
		}
	}
	xmlAddChild(grandparent, node);
	xmlAddChild(grandparent, downlevel);
	return true;	
}

bool Jsw2Room::exportJsw2Arrow(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent,
		xmlNodePtr downlevel, jswByte *guard)
{
	char txt[20];
	
	xmlNodePtr node = xmlNewNode(ns, (xmlChar *)"arrow");
	sprintf(txt, "%d", guard[0]);
	xmlSetProp(node, (xmlChar *)"x", (xmlChar *)txt);
	sprintf(txt, "%d", guard[1] & 0x7F);
	xmlSetProp(node, (xmlChar *)"y", (xmlChar *)txt);
	if (guard[1] & 0x80)
		xmlSetProp(node, (xmlChar *)"direction", (xmlChar *)"L");
	else	xmlSetProp(node, (xmlChar *)"direction", (xmlChar *)"R");
	xmlAddChild(parent, node);
	// So much for JSW2. Now for the downlevel arrow.
	node = xmlNewNode(ns, (xmlChar *)"arrow");
	sprintf(txt, "%d", guard[0]);
	xmlSetProp(node, (xmlChar *)"x", (xmlChar *)txt);
	sprintf(txt, "%d", guard[1] & 0x7F);
	xmlSetProp(node, (xmlChar *)"y", (xmlChar *)txt);
	if (guard[1] & 0x80)
	{
		xmlSetProp(node, (xmlChar *)"direction", (xmlChar *)"L");
		sprintf(txt, "%02x", m_game->getMem()->peek(0xEB01));
		xmlSetProp(node, (xmlChar *)"edges", (xmlChar *)txt);
		sprintf(txt, "%02x", m_game->getMem()->peek(0xEB03));
		xmlSetProp(node, (xmlChar *)"middle", (xmlChar *)txt);
	}
	else	
	{
		xmlSetProp(node, (xmlChar *)"direction", (xmlChar *)"R");
		sprintf(txt, "%02x", m_game->getMem()->peek(0xEB21));
		xmlSetProp(node, (xmlChar *)"edges", (xmlChar *)txt);
		sprintf(txt, "%02x", m_game->getMem()->peek(0xEB23));
		xmlSetProp(node, (xmlChar *)"middle", (xmlChar *)txt);
	}
	xmlSetProp(node, (xmlChar *)"ink", (xmlChar *)"7");
	xmlAddChild(downlevel, node);
	return true;
}	

bool Jsw2Room::exportJsw2Guardian(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent,
		xmlNodePtr downlevel, jswByte *guard)
{
	char txt[20];
	int n, majorstep, minorstep;
	int xmin, xmax, ymin, ymax;

	xmlNodePtr node = xmlNewNode(ns, (xmlChar *)"gg");
	sprintf(txt, "%d", guard[0]);
	xmlSetProp(node, (xmlChar *)"tick", (xmlChar *)txt);
	sprintf(txt, "%d", guard[1]);
	xmlSetProp(node, (xmlChar *)"reset", (xmlChar *)txt);
	majorstep = (signed char)guard[3];
	sprintf(txt, "%d", majorstep);
	xmlSetProp(node, (xmlChar *)"majorstep", (xmlChar *)txt);
	sprintf(txt, "%d", guard[4] & 0x7F);
	xmlSetProp(node, (xmlChar *)"x", (xmlChar *)txt);
	sprintf(txt, "%d", guard[5] & 0x7F);
	xmlSetProp(node, (xmlChar *)"y", (xmlChar *)txt);
	if (guard[5] & 0x80)
		xmlSetProp(node, (xmlChar *)"unidirectional", (xmlChar *)"1");
	int anim = guard[6] & 3;
	if (guard[6] & 0x40) anim |= 4;
	sprintf(txt, "%d", anim);
	xmlSetProp(node, (xmlChar *)"animation", (xmlChar *)txt);
	int ink = (guard[6] >> 2) & 3;
	sprintf(txt, "%d", 7 & m_game->getMem()->peek(JSW2GUARDINKS + ink));
	xmlSetProp(node, (xmlChar *)"ink", (xmlChar *)txt);
	minorstep = (guard[6] >> 4) & 3;
	sprintf(txt, "%d", minorstep);
	xmlSetProp(node, (xmlChar *)"minorstep", (xmlChar *)txt);
	if (guard[6] & 0x80)
		xmlSetProp(node, (xmlChar *)"direction", (xmlChar *)"h");
	else	xmlSetProp(node, (xmlChar *)"direction", (xmlChar *)"v");

	xmlAddChild(parent, node);
	xmlNodePtr dnode;

	// OK, that was the JSW2 bit. Now for the JSW bit.
	Jsw2HVGuardian::getLimits(guard, &xmin, &xmax, &ymin, &ymax);
	if (minorstep == 0)
	{
		if (guard[6] & 0x80) 
		{
			// Export downlevel HG
			dnode = xmlNewNode(ns, (xmlChar *)"hg");
			sprintf(txt, "%d", xmin / 4);
			xmlSetProp(dnode, (xmlChar *)"limit1", (xmlChar *)txt);
			sprintf(txt, "%d", xmax / 4);
			xmlSetProp(dnode, (xmlChar *)"limit2", (xmlChar *)txt);
			if (majorstep > 0) 
				xmlSetProp(dnode, (xmlChar *)"direction", (xmlChar *)"1");
			else	xmlSetProp(dnode, (xmlChar *)"direction", (xmlChar *)"0");
		}
		else
		{
			dnode = xmlNewNode(ns, (xmlChar *)"vg");
			sprintf(txt, "%d", 2 * ymin);
			xmlSetProp(dnode, (xmlChar *)"limit1", (xmlChar *)txt);
			sprintf(txt, "%d", 2 * ymax);
			xmlSetProp(dnode, (xmlChar *)"limit2", (xmlChar *)txt);
			sprintf(txt, "%d", majorstep * 2);
			xmlSetProp(dnode, (xmlChar *)"speed", (xmlChar *)txt);
		}
	}	
	else
	{
		if (guard[6] & 0x80) 
		{
			// Export downlevel HG
			dnode = xmlNewNode(ns, (xmlChar *)"dgnw");
			sprintf(txt, "%d", xmin / 4);
			xmlSetProp(dnode, (xmlChar *)"limit1", (xmlChar *)txt);
			sprintf(txt, "%d", xmax / 4);
			xmlSetProp(dnode, (xmlChar *)"limit2", (xmlChar *)txt);
			if (majorstep > 0) 
				xmlSetProp(dnode, (xmlChar *)"direction", (xmlChar *)"1");
			else	xmlSetProp(dnode, (xmlChar *)"direction", (xmlChar *)"0");
			if (majorstep > 0)
				sprintf(txt, "%d", minorstep * -2);
			else	sprintf(txt, "%d", minorstep * 2);
			xmlSetProp(dnode, (xmlChar *)"speed", (xmlChar *)txt);
		}
		else
		{
			dnode = xmlNewNode(ns, (xmlChar *)"dgnw");
			sprintf(txt, "%d", xmin / 4);
			xmlSetProp(dnode, (xmlChar *)"limit1", (xmlChar *)txt);
			sprintf(txt, "%d", xmax / 4);
			xmlSetProp(dnode, (xmlChar *)"limit2", (xmlChar *)txt);
			xmlSetProp(dnode, (xmlChar *)"direction", (xmlChar *)"1");
			sprintf(txt, "%d", majorstep * 2);
			xmlSetProp(dnode, (xmlChar *)"speed", (xmlChar *)txt);
		}
	}
	sprintf(txt, "%d", (guard[4] & 0x7F) / 4);
	xmlSetProp(dnode, (xmlChar *)"x", (xmlChar *)txt);
	sprintf(txt, "%d", (guard[4] & 3));
	xmlSetProp(dnode, (xmlChar *)"frame", (xmlChar *)txt);
	sprintf(txt, "%d", guard[5] & 0x7F);
	xmlSetProp(dnode, (xmlChar *)"y", (xmlChar *)txt);
	sprintf(txt, "%d", 7 & m_game->getMem()->peek(JSW2GUARDINKS + ink));
	xmlSetProp(dnode, (xmlChar *)"ink", (xmlChar *)txt);
	xmlSetProp(dnode, (xmlChar *)"bright", (xmlChar *)"1");
	sprintf(txt, "%d", anim);
	xmlSetProp(dnode, (xmlChar *)"framemask", (xmlChar *)txt);
	xmlAddChild(downlevel, dnode);
//
// Now the sprite refs for both
//	
	unsigned spriteNo = guard[2];
	if (guard[4] & 0x80) spriteNo += 256;
	for (int frame = 0; frame < 8; frame++)
	{
		xmlNodePtr ref  = xmlNewNode(ns, (xmlChar *)"spriteref");
		xmlNodePtr dref = xmlNewNode(ns, (xmlChar *)"spriteref");
		sprintf(txt, "%04x", (spriteNo + (frame & anim)) * 32 
				+ JSW2SPRITES);
		xmlSetProp(ref, (xmlChar *)"xref", (xmlChar *)txt);
		xmlSetProp(dref, (xmlChar *)"xref", (xmlChar *)txt);
		sprintf(txt, "%d", frame);
		xmlSetProp(ref, (xmlChar *)"frame", (xmlChar *)txt);
		xmlSetProp(dref, (xmlChar *)"frame", (xmlChar *)txt);
		xmlAddChild(node, ref); 
		xmlAddChild(dnode, dref); 
	}
	return true;
}


bool Jsw2Room::exportHints(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr node)
{
	char buf[30];
	xmlNodePtr hint;
	int cpos, clen;

	if (m_T4 & 0x40)	// Conveyor is animated
	{
		for (cpos = 0; cpos < 512; cpos++)
		{
			if (m_cells[cpos] == 5 || m_cells[cpos] == 8) break;
		}
		if (cpos >= 512) clen = 0;
		else while (m_cells[cpos + clen] == m_cells[cpos])
		{
			++clen;
		}
		if (clen > 0)
		{

			hint = xmlNewNode(ns, (xmlChar *)"hint");
			xmlSetProp(hint, (xmlChar *)"type", 
					(xmlChar *)"conveyor");
			sprintf(buf, "%d", (m_T4 & 0x20) ? 0 : 1);
			xmlSetProp(hint, (xmlChar *)"animation", 
					(xmlChar *)buf);
			sprintf(buf, "%d", cpos % 32);
			xmlSetProp(hint, (xmlChar *)"x", (xmlChar *)buf);
			sprintf(buf, "%d", cpos / 32);
			xmlSetProp(hint, (xmlChar *)"y", (xmlChar *)buf);
			sprintf(buf, "%d", clen);
			xmlSetProp(hint, (xmlChar *)"length", (xmlChar *)buf);
			sprintf(buf, "%d", (m_cells[cpos] == 8));
			xmlSetProp(hint, (xmlChar *)"direction", 
					(xmlChar *)buf);
			xmlAddChild(node, hint);
		}
	}
	if (m_T5 & 31)
	{
		hint = xmlNewNode(ns, (xmlChar *)"hint");
		xmlSetProp(hint, (xmlChar *)"type", (xmlChar *)"jsw2");
		sprintf(buf, "%d", m_T5 & 31);
		xmlSetProp(hint, (xmlChar *)"vector", (xmlChar *)buf);
		xmlAddChild(node, hint);
	}
	return Room::exportHints(doc, ns, node);
}


/* Given a list of cell definitions, set their numberInRoom elements to
 * the cell type within the room. This can also modify their behaviour
 * if it wants.
 *
 * If numberInRoom is set to a +ve value, the corresponding cell pattern 
 * in the room is set from the pattern of this cell.
 *
 * If numberInRoom is set to a -ve value, the cell pattern is ignored and 
 * pattern (-numberInRoom) is substituted. 
 */
void Jsw2Room::normaliseCells(ImportCell *p, int count)
{
	int air = -1, water = -1, earth = -1, fire = -1, 
	    lramp = -1, rramp = -1, lconv = -1, rconv = -1;
	int n;

	// First pass: Map only 1:1
	for (n = 0; n < count; n++)
	{
		if (p[n].count == 0) continue;
		switch(p[n].behaviour)
		{
		    case CB_AIR:   if (air == -1)   air   = n; break;
		    case CB_WATER: if (water == -1) water = n; break;
		    case CB_EARTH: if (earth == -1) earth = n; break;
		    case CB_FIRE:  if (fire == -1)  fire  = n; break;
		    case CB_LRAMP: if (lramp == -1) lramp  = n; break;
		    case CB_RRAMP: if (rramp == -1) rramp  = n; break;
		    case CB_LCONV: if (lconv == -1) lconv  = n; break;
		    case CB_RCONV: if (rconv == -1) rconv  = n; break;
		}
	}
	// Second pass: If there are no good candidates for ramps and
	// there are for crumbly, map crumbly to both ramps. 
	for (n = 0; n < count; n++)
	{
		if (p[n].count == 0) continue;
		if (p[n].behaviour == CB_CRUMBLY) 
		{
			if (lramp == -1) lramp = n;
			if (rramp == -1) rramp = n;
		}
	}
	// Set all cells to map to their equivalent elemental types,
	// or - failing that - water.
	for (n = 0; n < count; n++)
	{
		switch(p[n].behaviour)
		{
			case CB_AIR:	p[n].numberInRoom =  0; break;
			case CB_WATER:	p[n].numberInRoom = -1; break;
			case CB_EARTH:	p[n].numberInRoom = -2; break;
			case CB_FIRE:	p[n].numberInRoom = -3; break;
			case CB_RRAMP:	p[n].numberInRoom = -4; break;
			case CB_LCONV:	p[n].numberInRoom = -5; break;
			case CB_LRAMP:	p[n].numberInRoom = -7; break;
			case CB_RCONV:	p[n].numberInRoom = -8; break;
			default:	p[n].behaviour = CB_WATER;
					p[n].numberInRoom = -1;
					break;
		}
	}
	if (air != -1) 
	{ 
		p[air].behaviour = CB_AIR; 
		p[air].numberInRoom = 0; 
	}
	if (water != -1) 
	{ 
		p[water].behaviour = CB_WATER; 
		p[water].numberInRoom = 1; 
	}
	if (earth != -1) 
	{ 
		p[earth].behaviour = CB_EARTH; 
		p[earth].numberInRoom = 2; 
	}
	if (fire != -1) 
	{
	       	p[fire].behaviour = CB_FIRE; 
		p[fire].numberInRoom = 3; 
	}
	if (lramp != -1) 
	{ 
		p[lramp].behaviour = CB_LRAMP; 
		p[lramp].numberInRoom = 7; 
	}
	if (rramp != -1)
	{
		p[rramp].behaviour = CB_RRAMP; 
		p[rramp].numberInRoom = 4; 
	}
	if (lconv != -1) 
	{ 
		p[lconv].behaviour = CB_LCONV; 
		p[lconv].numberInRoom = 5; 
	}
	if (rconv != -1)
	{
		p[rconv].behaviour = CB_RCONV; 
		p[rconv].numberInRoom = 8; 
	}
}

void Jsw2Room::initImportCells(void)
{
	memset(m_cellShapes, 0, sizeof(m_cellShapes));
}

bool Jsw2Room::importCells(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr node, 
		xmlNodePtr bits, ImportCell **parsedCells, int *ncells)
{
	int n, m;
	jswByte *ptr;
	bool b;
       
	b = Room::importCells(doc, ns, node, bits, parsedCells, ncells);
	if (!b) return false;
	memset(m_cellShapes, 0, 9);	// Air is always black
	jswByte mask = 0x80;
	for (n = 0; n < 8; n++)
	{
		ptr = m_cellShapes + 9 * (n+1);
		m = mapCellGraphic(ptr, (n != 6));
		if (m < 256) m_data[2] &= ~mask;
		else	     m_data[2] |= mask;
		m_data[3+n] = (m & 0xFF);
		m_cellNumber[n+1] = m;
		mask = mask >> 1;
	}
	createCellPatterns();
	return true;
}


// Given a cell graphic, find a home for it.
int Jsw2Room::mapCellGraphic(jswByte *ptr, bool colourMatters)
{
	int n, y, pass;
	bool match, inverse;
	jswByte *compare;

	for (pass = 0; pass < 2; pass++)
	{
		for (n = JSW2CELLS; n < JSW2CELLSEND; n += 9)
		{
			compare = m_game->getMem()->memoryAt(n);
			inverse = false;
			match = true;
			for (y = 1; y < 9; y++) 
				if (ptr[y] != compare[y])	
				{
					match = false;
					break;
				}
			if (!match)
			{
				inverse = true;
				match = true;
				for (y = 1; y < 9; y++)
					if (ptr[y] != ~compare[y])
					{
						match = false;
						break;
					}
			}
			if (!match) continue;
			// "compare" is a possible candidate. If colour is
			// important, compare colours.
			if (pass == 0 && colourMatters)
			{
				int cink, cpaper;
				int myink, mypaper;

				if (compare[0] & 0x80)
				{
					cink   = compare[0] & 7;
					cpaper = (compare[0] >> 3) & 7;
				}
				else
				{
					cpaper = compare[0] & 7;
					cink   = (compare[0] >> 3) & 7;
				}
				if (inverse)
				{
					myink   = (ptr[0] >> 3) & 7;
					mypaper = ptr[0] & 7;
				}
				else
				{
					myink   = ptr[0] & 7;
					mypaper = (ptr[0] >> 3) & 7;
				}
				if (myink != cink || mypaper != cpaper) 
					match = false;
			}
			if (match)
			{
				int no = (n - JSW2CELLS) / 9;
				m_game->addRef(CELL(no));
				return no;
			}
		}
		// End of pass, and no match found. See if there's a free 
		// space.
		for (n = JSW2CELLS; n < JSW2CELLSEND; n += 9)
		{
			int no = (n - JSW2CELLS) / 9;
			// Aha! A free space!
			if (!m_game->getRefCount(CELL(no)))
			{
				compare = m_game->getMem()->memoryAt(n);
				compare[0] = ptr[0] & 0x3F;
				memcpy(compare + 1, ptr + 1, 8);
				m_game->addRef(CELL(no));
				return no;	
			}
		}	
	}
	// Right. We've tried everything but no match was found. Default to 0.
	return 0;
}


bool Jsw2Room::importHints(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent)
{
	xmlNodePtr cur;
	bool haveSpecial = false, haveConveyor = false;

	for (cur = parent->xmlChildrenNode; cur; cur = cur->next)
	{
		if (cur->ns != ns) continue;
		if (strcmp((char *)cur->name, "hint")) continue;

		char *hintName = (char *)xmlGetProp(cur, (xmlChar *)"type");
		if (!hintName) continue;

		if (!strcmp(hintName, "conveyor"))
		{
			char *attr = (char *)xmlGetProp(cur, (xmlChar *)"animation");
			m_T4 &= ~0x20;
			if (attr)
			{
				if (atoi(attr) == 0) m_T4 |= 0x20;
				xmlFree(attr);
			}
			m_T4 |= 0x40;
			haveConveyor = true;
		}
		if (!strcmp(hintName, "jsw2"))
		{
			char *attr = (char *)xmlGetProp(cur, (xmlChar *)"vector");
			m_T5 &= ~0x1F;
			if (attr)
			{
				m_T5 |= (atoi(attr) & 0x1F);
				xmlFree(attr);
			}
			haveSpecial = true;
		}
		xmlFree(hintName);
	}
	if (!haveConveyor) m_T4 &= ~0x60;
	if (!haveSpecial)  m_T5 &= ~0x1F;
	return Room::importHints(doc, ns, parent);
}

void Jsw2Room::clearHints()
{
	m_T4 &= ~0x60;
	m_T5 &= ~0x1F;
}


bool Jsw2Room::importGuardians(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent)
{
	int ng, na;
	m_T4 &= ~0x0F;	// Remove existing guardians
	m_T4 &= ~0x80;	// Remove existing rope
	m_T5 &= ~0x80;	// Remove existing arrows
	m_arrowCount = 0;

	char *type = (char *)xmlGetProp(parent, (xmlChar *)"type");

	if (type && !strcmp(type, "jsw2"))
	{
		ng = na = 0;
		xmlNodePtr cur;
		for (cur = parent->xmlChildrenNode; cur; cur = cur->next)
		{
			if (!strcmp((char *)cur->name, "rope"))
			{
				m_T4 |= 0x80;
				continue;
			}
			if (ng < 8 && !strcmp((char *)cur->name, "gg"))
			{
				importJsw2Guardian(doc, ns, cur,
						m_guardDefs + 7 * ng);
				++ng;
			}
			if (na < 8 && !strcmp((char *)cur->name, "arrow"))
			{
				importJsw2Arrow(doc, ns, cur,
						m_arrowDefs + 2 * na);
				++na;
			}
		}
		m_T4 |= (ng & 15);
		if (na)
		{
			m_T5 |= (0x80 + na);
			m_arrowCount = na;
		}
	}
	else if (type && !strcmp(type, "jsw"))
	{
		ng = na = 0;
		xmlNodePtr cur;
		for (cur = parent->xmlChildrenNode; cur; cur = cur->next)
		{
			if (!strcmp((char *)cur->name, "rope"))
			{
				m_T4 |= 0x80;
				continue;
			}
			if (ng < 8 && 
			    (!strcmp((char *)cur->name, "hg")   ||
			     !strcmp((char *)cur->name, "vg")   ||
			     !strcmp((char *)cur->name, "dgnw") ||
			     !strcmp((char *)cur->name, "dgne")))
			{
				importJsw1Guardian(doc, ns, cur,
						m_guardDefs + 7 * ng);
				++ng;
			}
			if (na < 8 && !strcmp((char *)cur->name, "arrow"))
			{
// The JSW1 arrow is a superset of the JSW2 arrow, so the same import routine
// will handle both.
				importJsw2Arrow(doc, ns, cur,
						m_arrowDefs + 2 * na);
				++na;
			}
		}
		m_T4 |= (ng & 15);
		if (na)
		{
			m_T5 |= (0x80 + na);
			m_arrowCount = na;
		}
	}
	if (type) xmlFree(type);
	return true;
}

const char *Jsw2Room::supportedGuardianTypes()
{
	return "jsw2|jsw";
}


bool Jsw2Room::importJsw2Arrow(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr node,
		jswByte *arrow)
{
	char *prop;
      	arrow[0] = 0;
	arrow[1] = 1;	

	prop = (char *)xmlGetProp(node, (xmlChar *)"x");
	if (prop)
	{
		arrow[0] = atoi(prop);
		xmlFree(prop);
	}
	prop = (char *)xmlGetProp(node, (xmlChar *)"y");
	if (prop)
	{
		arrow[1] = atoi(prop) & 0x7F;
		xmlFree(prop);
	}
	prop = (char *)xmlGetProp(node, (xmlChar *)"direction");
	if (prop)
	{
		if (prop[0] == 'L' || prop[0] == 'l') arrow[1] |= 0x80;
		xmlFree(prop);
	}
	return true;
}



bool Jsw2Room::importJsw2Guardian(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr node,
		jswByte *guardian)
{
	char *prop;
	memset(guardian, 0, 7);

	prop = (char *)xmlGetProp(node, (xmlChar *)"tick");
	if (prop)
	{
		guardian[0] = atoi(prop);
		xmlFree(prop);
	}
	prop = (char *)xmlGetProp(node, (xmlChar *)"reset");
	if (prop)
	{
		guardian[1] = atoi(prop);
		xmlFree(prop);
	}
	// XXX guardian[2] and top bit of guardian[4]  = sprite number
	prop = (char *)xmlGetProp(node, (xmlChar *)"majorstep");
	if (prop)
	{
		guardian[3] = atoi(prop);
		xmlFree(prop);
	}
	prop = (char *)xmlGetProp(node, (xmlChar *)"x");
	if (prop)
	{
		guardian[4] &= 0x80;
		guardian[4] |= (atoi(prop) & 0x7F);
		xmlFree(prop);
	}
	prop = (char *)xmlGetProp(node, (xmlChar *)"unidirectional");
	if (prop)
	{
		if (atoi(prop)) guardian[5] |= 0x80;
		xmlFree(prop);
	}
	prop = (char *)xmlGetProp(node, (xmlChar *)"y");
	if (prop)
	{
		guardian[5] &= 0x80;
		guardian[5] |= (atoi(prop) & 0x7F);
		xmlFree(prop);
	}
	prop = (char *)xmlGetProp(node, (xmlChar *)"animation");
	if (prop)
	{
		guardian[6] = atoi(prop) & 3;
		if (atoi(prop) & 4) guardian[6] |= 0x40;
		xmlFree(prop);
	}
	jswByte *inks = m_game->getMem()->memoryAt(JSW2GUARDINKS);
	prop = (char *)xmlGetProp(node, (xmlChar *)"ink");
	if (prop)
	{
		for (int n = 0; n < 4; n++)
		{
			if ((atoi(prop) & 7) == (inks[n] & 7))
			{
				guardian[6] |= ((n << 2) & 0x0C);
			}
		}
		xmlFree(prop);
	}
	prop = (char *)xmlGetProp(node, (xmlChar *)"minorstep");
	if (prop)
	{
		guardian[6] |= (atoi(prop) << 4) & 0x30;
		xmlFree(prop);
	}
	prop = (char *)xmlGetProp(node, (xmlChar *)"direction");
	if (prop)
	{
		if (prop[0] == 'h' || prop[0] == 'H') guardian[6] |= 0x80;
		xmlFree(prop);
	}
	importGuardianSprite(doc, ns, node, guardian);
	return true;
}



bool Jsw2Room::importJsw1Guardian(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr node,
		jswByte *guardian)
{
	jswByte guard1[8];
	int type;
	bool horizontal = false;
	bool diagonal = false;
	bool vertical = false;
	bool unidirectional = false;
	int y1, x1, x, y, majorstep, minorstep, lim1, lim2, tick, reset;
	
	char *uni = (char *)xmlGetProp(node, (xmlChar *)"unidirectional");
	if (uni)
	{
		if (atoi(uni)) unidirectional = true;
		xmlFree(uni);
	}
	Room::importJswGuardian(doc, ns, node, guard1, m_game, 0x8000);
	type = guard1[0] & 0x0F;
	if (type == 8) type = guard1[0];

	switch(type)
	{
		case 1: case 9: horizontal = true; break;
		case 5: case 6: 
		case 13: case 14: diagonal = true; break;
		case 2: case 10:
		case 7: case 15:  vertical = true; break;
	}
	if (horizontal || diagonal)
	{
		majorstep = (guard1[0] & 0x80) ? 1 : -1;
		if (diagonal)
			minorstep = ((signed char)guard1[4]) / 2;
		else	minorstep = 0;
		if (guard1[7] & 0x80) majorstep *= 2;
		if (type == 6 || type == 14) minorstep = -minorstep;
		x = (guard1[2] & 0x1F) * 4;
		y = guard1[3] / 2;
		lim1 = (guard1[6] & 0x1F) * 4;
		lim2 = (guard1[7] & 0x1F) * 4;
		if (guard1[0] & 0x80)
		{
			tick  = (lim2 - x) + 1;
			reset = (lim2 - lim1) + ((guard1[1] >> 5) & 3);
		}
		else
		{	
			tick  = (x - lim1) + 1;
			reset = (lim2 - lim1) + ((guard1[1] >> 5) & 3);
		}
		// XXX we need more than this to make diagonals work properly.
		// Making the vertical step be the major step is fine, but
		// this doesn't distinguish between dgnw and dgne, let alone
		// which direction each component of the vector is heading 
		// at a given time.
		if (diagonal)
		{
			x1 = majorstep; majorstep = minorstep; minorstep = x1;
		}
	}
	else if (vertical)
	{
		majorstep = ((signed char)guard1[4]) / 2;
		minorstep = 0;
		x = (guard1[2] & 0x1F) * 4;
		y = guard1[3] / 2;
		lim1 = guard1[6] / 2;
		lim2 = guard1[7] / 2;
		// Work out tick and reset steps
		tick = 0;
		reset = 0;
		y1 = majorstep;
		if (y1 < 0) y1 = -y1;
		if (!majorstep) 
		{
			tick = reset = 0;
		}
		else if (majorstep > 0) 
		{
			tick  = ((lim2 - y) / y1) + 1;
			reset = (lim2 - lim1) / y1;
		}
		else
		{
			tick  = ((y - lim1) / y1) + 1;
			reset = (lim2 - lim1) / y1;
		}
	}
	guardian[0] = tick;
	guardian[1] = reset;
	guardian[3] = majorstep;
	guardian[4] = x & 0x7F;
	guardian[5] = y & 0x7F;
	importGuardianSprite(doc, ns, node, guardian);
	if (unidirectional) guardian[5] |= 0x80;
	// Animation (low 2 bits)
	guardian[6] = (guard1[1] >> 5) & 3;
	// Ink
        jswByte *inks = m_game->getMem()->memoryAt(JSW2GUARDINKS);

	for (int n = 0; n < 4; n++)
	{
		if ((inks[n] & 7) == (guard1[1] & 7))
		{
			guardian[6] |= ((n << 2) & 0x0C);
		}
	}
	guardian[6] |= ((minorstep << 4) & 0x30);
	if (guard1[1] & 0x80) guardian[6] |= 0x40;	// Bidirectional
	if (horizontal) guardian[6] |= 0x80;

	return true;

}


bool Jsw2Room::importGuardianSprite(xmlDocPtr doc, xmlNsPtr ns,
		xmlNodePtr cur, jswByte *guardian)
{
	xmlNodePtr spriteref;
	int sr;
	int baseid, fileid, memid;

	baseid = -1;
	for (spriteref = cur->xmlChildrenNode;  
		spriteref; spriteref = spriteref->next)
	{
		if (spriteref->ns != ns) continue;
		if (strcmp((char *)spriteref->name, "spriteref")) continue;
		char *prop = (char *)xmlGetProp(spriteref, (xmlChar *)"xref");
		sr = 0;
		if (prop)
		{
			sr = sscanf(prop, "%x", &fileid); 
			xmlFree(prop);
		}
		if (sr && (baseid > fileid || baseid < 0)) 
		{
			baseid = fileid;
		}
	}
	memid = JSW2SPRITES;
	if (baseid >= 0)
	{
		memid = m_game->lookupSpriteRef(baseid);
		if (memid < 0) memid = JSW2SPRITES;
	}
	memid -= JSW2SPRITES;
	memid /= 32;
	guardian[2] = (memid & 0xFF);
	if (memid & 0x100) guardian[4] |= 0x80;
	else		   guardian[4] &= 0x7F;
	return true;
}
