/************************************************************************

    JSWED 2.3.0 - Editor for Jet Set Willy and derivatives

    Copyright (C) 2000-2001,2004-5  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*************************************************************************/

class Jsw2Game;

class Jsw2Room : public Room
{
private:
	unsigned m_cellNumber[9];
	jswByte m_cellShapes[90];
	jswByte m_cells[512];
	jswByte m_title[33];
	jswByte m_exits[4];
	jswByte m_T4;	// Options 1
	jswByte m_T5;	// Options 2
	jswByte *m_data;
	jswByte *m_shape;
	jswByte m_guardDefs[136];
	jswByte m_arrowDefs[16];
	unsigned m_arrowCount;
	SpectrumMemory *m_mem;
	VideoBitmap *m_bmConvey0;
	VideoBitmap *m_bmConvey2;
	jswByte m_animMask;
	jswByte m_itemBits[9];
	Jsw2Game *m_game;
protected:
	void decompress();
	int expandToken(jswByte tokenID, jswByte **tDest, jswByte *last);
	int expandString(int addr, jswByte *src, jswByte **tDest, jswByte *last);
	int mapCellGraphic(jswByte *ptr, bool colourMatters);

	virtual jswByte *getCellPattern(int celltype);
	virtual int getConveyorDir();
	virtual int getConveyorY();
	virtual int getItemCoords(int *px, int *py, int *counts);
public:
	Jsw2Room(int rmid, Jsw2Game *g);
	Jsw2Room(Jsw2Room &other);
	~Jsw2Room();

	inline unsigned getCellNumber(int n)
	{
		if (n >= 1 && n <= 8) return m_cellNumber[n];
	}

	inline void setCellNumber(int n, unsigned v)
	{
		if (n >= 1 && n <= 8) m_cellNumber[n] = v;
	}

	inline jswByte getT5()             { return (m_T5 & 0x7F); }
	inline void setT5(jswByte b)       { m_T5 &= 0x80; m_T5 |= (b & 0x7F); }

	inline jswByte getT4()             { return (m_T4 & 0xE0); }
	inline void setT4(jswByte b)       { m_T4 &= 0x1F; m_T4 |= (b & 0xE0); }
	inline unsigned getArrowCount()    { return m_arrowCount; }
	inline unsigned getGuardianCount() { return (m_T4 & 0x0F); }
	inline void setGuardianCount(unsigned v) 
	{ 
		if ((v + m_arrowCount) < 9) { m_T4 &= 0xF0; m_T4 |= v; }
  	}
	inline void setArrowCount(unsigned v) 
	{ 
		if ((v + (m_T4 & 0x0F)) < 9) { m_arrowCount = v; }
  	}
	virtual int getTitle(char *buf);
	virtual void setTitle(const char *buf);
	/* Get the background colour to use for guardians */
	virtual int getBackground();
	int getItemCount();

	virtual jswByte *getExits();
	virtual int getCellTypeCount(bool includeSingles);
	virtual int getCellBehaviour(int index);

	virtual void drawOneOffs(VideoBitmap *bm, bool flashonly, bool draw = true);
	virtual void createCellPatterns();
	virtual void animateConveyor();
	virtual void setCell(int x, int y, int type);
	/* Get the cell type at position x,y */
	virtual int getCellAt(int x, int y);
	virtual int getDrawnCellAt(int x, int y);

	virtual int getRoomLength() const;

	virtual void getCells(Cell *c);
	virtual void setCells(Cell *c);

	virtual void clear();
	virtual void mirrorH();
	virtual void mirrorV();

	virtual void copyFrom(Jsw2Room *rm);
	virtual void getRoomMemMap(char *buf);
	virtual void drawItems(VideoBitmap *bm, int colour);
	virtual jswByte *getGuardianBuffer();
	inline jswByte *getArrowBuffer() { return m_arrowDefs; }

	virtual jswByte getBorder();
	virtual void setBorder(int bdr);

	virtual void clearItems();
	virtual void addItem(int x, int y);
	virtual void setItemBitmap(jswByte*);
	virtual void addRefs();

	unsigned compressShape(jswByte *buffer, unsigned offset);
	unsigned compress(jswByte *buffer, unsigned table, unsigned offset,
			unsigned shapeOffset, char *title);

	virtual void normaliseCells(ImportCell *p, int count);
	virtual void initImportCells(void);

	virtual bool importCells(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr node,
			xmlNodePtr bits, ImportCell **parsedCells, int *ncells);
	virtual bool exportHints(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr node);
	virtual bool importHints(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr node);
	virtual void clearHints(void);
	virtual bool exportExits(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr node);
	virtual bool importExits(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr node,
			int roomMap[256]);
	virtual bool exportGuardians(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr grandparent);
	virtual bool importGuardians(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent);
        virtual const char *supportedGuardianTypes(void);

	bool exportJsw2Guardian(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent,
		xmlNodePtr downlevel, jswByte *guard);
	bool importJsw2Guardian(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent,
		jswByte *guard);
	bool importJsw2Arrow(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent,
		jswByte *guard);
	bool exportJsw2Arrow(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent,
		xmlNodePtr downlevel, jswByte *arrow);
	bool importJsw1Guardian(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent,
		jswByte *guard);
	bool importGuardianSprite(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent,
		jswByte *guard);

	virtual jswByte *getItemBitmap();
private:
	void putRle(int lct, int reps, jswByte *buffer, unsigned *offset);
};

