/************************************************************************

    JSWED 2.3.0 - Editor for Jet Set Willy and derivatives

    Copyright (C) 2000-2001,2004  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*************************************************************************/

#include "jswed.hxx"
#include "jsw2game.hxx"
#include "jsw2roomdraw.hxx"
#include "room.hxx"
#include "jsw2room.hxx"


Jsw2RoomDraw::Jsw2RoomDraw(Jsw2Game *game)
{
//	setTitle("Debug");
	m_game = game;
	m_ticker = 0;
	m_room = game->getStartRoom();
	m_prevRoom = -1;
	m_ot   = NULL;
	m_haveTimer = 0;
	m_visible = 0;

	for (int n = 0; n < (GUARD_MAX + J2FG_MAX); n++) m_guards[n] = NULL;
	loadRoom(m_room);

	m_bmRoom  = newVideoBitmap(512, 256);
	m_bmAnim  = newVideoBitmap(512, 256);

	m_drawnFixed = 0;

}

Jsw2RoomDraw::~Jsw2RoomDraw()
{
	delete m_bmRoom;
	delete m_bmAnim;
	for (int n = 0; n < (GUARD_MAX + J2FG_MAX); n++) 
		if (m_guards[n]) delete m_guards[n]; 

        if (m_haveTimer) setTickListener(m_ot);
}




void Jsw2RoomDraw::drawFixed(void)
{
	m_drawnFixed = 1;
}

VideoBitmap *Jsw2RoomDraw::bitmapFromSprite(const jswByte *sprite)
{
	int ink   = sprite[0] & 7;
	int paper = (sprite[0] >> 3) & 7;

	if (sprite[0] & 64)  { ink |= ZX_BRIGHT; paper |= ZX_BRIGHT; }
	return ::bitmapFromSprite(8, 8, sprite+1, 1, ink, paper);
}

#define CONVEYORANIM 4	// 2 or even 1 look nicer, but 4 matches JSW best 

// Overridables 
void Jsw2RoomDraw::drawCellType() { }
void Jsw2RoomDraw::drawCursor(void) { }
void Jsw2RoomDraw::drawGuardList(void) { }

void Jsw2RoomDraw::flashColours(void)
{
	++m_ticker;

	if (!m_visible) return;
	
	drawCellType();

	if ((m_ticker % 20)== 0)   drawData(1);

	if ( (m_ticker % 5)== 0)
	{
		int tock = tick();
		
		if (tock) getRoom()->animateConveyor();
		getRoom()->drawOneOffs(m_bmRoom, false);
		m_bmRoom->toSurface(m_bmAnim, 0, 0);
		for (int n = 0; n < (GUARD_MAX + J2FG_MAX); n++)
		{	
			if (m_guards[n])
			{
				if (tock) m_guards[n]->move();
				m_guards[n]->draw(m_bmAnim);
			}
			else if (n < GUARD_MAX) n = GUARD_MAX - 1; 
			else break;
		}
		if (getRoom()->getT4() & 0x80)
		{
			if (tock) moveRope();
			drawRope();
		}
		drawItems();
		drawTeleports();
		drawCursor();
	        m_bmAnim->toScreen(m_x + X_OFFSET, m_y + Y_OFFSET);
		drawBorder();
	}

}







void Jsw2RoomDraw::flashSprite(VideoBitmap *bmp, jswByte attr)
{
        int ink   = attr & 7;
        int paper = (attr >> 3) & 7;
	int im;

	if (!(attr & 128)) return; 

        if (attr & 64)  { ink |= ZX_BRIGHT; paper |= ZX_BRIGHT; }

	for (im = 1; im < 15; im++)
	{
		if (im != paper && im != ink && im != 8) break;
	}
	bmp->getSurface()->fillReplace(0, 0, 16, 16, paper, im);
	bmp->getSurface()->fillReplace(0, 0, 16, 16, ink, paper);
	bmp->getSurface()->fillReplace(0, 0, 16, 16, im, ink);
}

void Jsw2RoomDraw::drawData(int flashonly)
{
	jswByte *DEBUGDATA = getRoom()->getGuardianBuffer() + 8;

	getRoom()->createCellPatterns();

	if (flashonly)
	{
		getRoom()->flashCells();
	}

	drawBasic(flashonly);
	if (!flashonly)
	{
		char sbuf[33];
		int n;

		getRoom()->drawOneOffs(m_bmRoom, false);
		//drawConveyor();
		drawGuardList();	
		getRoom()->getTitle(sbuf);
		videoScreen->drawText(m_x + X_OFFSET, m_y + Y_OFFSET + 256, 
				sbuf, ZX_WHITE, ZX_BLACK);
	        m_bmRoom->toSurface(m_bmAnim, 0, 0);

 		for (n = 0; n < (GUARD_MAX + J2FG_MAX); n++)
		{
			if (m_guards[n]) 
			{
                        	m_guards[n]->draw(m_bmAnim);
			}
			else if (n < GUARD_MAX) n = GUARD_MAX - 1; 
		} 
		drawItems();
		if (getRoom()->getT4() & 0x80)
		{
			drawRope();
		}
		drawTeleports();
		m_bmAnim->toScreen(m_x + X_OFFSET, m_y + Y_OFFSET);
		drawBorder();
	}
}


void Jsw2RoomDraw::drawBasic(int flashonly)
{
        int x, y;

	for (y = 0; y < 16; y++) for (x = 0; x < 32; x++)
	{
		getRoom()->drawCell(m_bmRoom, x, y, flashonly, false);
	}
}






int Jsw2RoomDraw::onReveal(int whichRectangle)
{
	if (!m_haveTimer)
	{
		m_ot = setTickListener(this);
		m_ticker = 0;
		m_haveTimer = 1;
	}
	return VideoNotePage::onReveal();
}







void Jsw2RoomDraw::intPreContextMenu(void)
{
        if (m_haveTimer)
        {
                setTickListener(m_ot);
                m_ot = NULL;
                m_haveTimer = 0;
        }
}


void Jsw2RoomDraw::postContextMenu(void)
{
	m_ot = setTickListener(this);	
	m_haveTimer = 1;
	VideoNotePage::postContextMenu();
}



int Jsw2RoomDraw::onConceal()
{
	if (m_haveTimer)
        {
                setTickListener(m_ot);
                m_ot = NULL;
                m_haveTimer = 0;
        }

	return VideoNotePage::onConceal();
}


int Jsw2RoomDraw::redraw(int whichRectangle)
{
	int rv = VideoNotePage::redraw(whichRectangle);
	if (whichRectangle & VPR_RIGHT)
	{
		drawFixed();
		drawData();
	}
	m_visible = 1;
	return rv;
}

void Jsw2RoomDraw::gotoRoom(int room)
{
	Room *rm = m_game->getRoomClass(room);
	if (!rm) return;

	rm->createCellPatterns();

	if (m_room != room) 
	{
		m_prevRoom = m_room;
		m_game->setCurrentRoom(room);
	}
	m_room = room;
	loadRoom(m_room);

	drawData();
}




void Jsw2RoomDraw::loadRoom(int room, int reload)
{
	int ng, n, bg;
	jswByte *buf  = getRoom()->getGuardianBuffer();
	jswByte *abuf = getRoom()->getArrowBuffer();
	m_maxGuards = getRoom()->getMaxGuards();
	bg = getRoom()->getBackground();

	for (n = 0; n < (GUARD_MAX + J2FG_MAX); n++)
	{
		if (m_guards[n]) delete m_guards[n];
		m_guards[n] = NULL;
	}
	for (n = ng = 0; n < getRoom()->getGuardianCount(); n++)
	{
		if (ng >= GUARD_MAX) break;	
		m_guards[ng] = new Jsw2HVGuardian(buf, m_game->getMem());
		buf += 7;
		++ng;
	}
	for (n = 0; n < getRoom()->getArrowCount(); n++)
	{
		if (ng >= GUARD_MAX) break;	
		m_guards[ng] = new Jsw2Arrow(abuf, m_game->getMem());
		abuf += 2;
		++ng;
	}
	checkSpecials();
	m_ropeX = 34;
	m_ropeDir = -1;
}

void Jsw2RoomDraw::moveRope()
{
	int ropeX = m_ropeX;

	if (m_ropeDir < 0)	// Going left
	{
		if (ropeX < 0)	// In left-hand half of screen
		{
			ropeX -= 2;
			if (ropeX < -108) ropeX -= 2;
			if (ropeX <= -128) ropeX = 0;
		}
		else
		{
			ropeX += 2;
			if (ropeX < 18) ropeX += 2;
		}
	}
	else	// Going right
	{
		if (ropeX < 0)
		{
			ropeX += 2;
			if (ropeX < -110) ropeX += 2;
		}
		else
		{
			ropeX -= 2;
			if (ropeX < 20) ropeX -= 2;
			if (ropeX == 0) ropeX = -128;
		}
	}
	m_ropeX = ropeX;
	if (m_ropeX == 54 || m_ropeX == -74)
	{
		m_ropeDir = -m_ropeDir;
	}
}

static jswByte ropeTable[] =
{
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	1,1,1,1,1,1,1,1,1,1,1,1,
	2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,1,2,2,1,1,2,1,1,2,2,
	3,2,3,2,3,3,3,3,3,3,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,
	6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
	6,6,6,6,6,6,6,6,6,6,6,6,6,6,4,6,6,4,6,4,6,4,6,4,4,4,6,4,4,4,4,4,4,4,
	4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
};

void Jsw2RoomDraw::drawRope()
{
	int y = 0;
	int iy = 0;
	int ropeX = m_ropeX;
	int x = 0x70;
	VideoSurface *s = m_bmAnim->getSurface();

	do
	{
		s->fillBox(x * 2, iy, 2, 2, ZX_WHITE);
		int p = (y + ropeX) & 0x7F;
		int dy = ropeTable[p + 128];
		int dx = ropeTable[p];

		iy += dy;

		if (ropeX < 0)  x -= dx;
		else		x += dx;
	}
	while ((y++) < 32);
}



void Jsw2RoomDraw::checkSpecials()
{
	int n = GUARD_MAX;
	int bg;
	Jsw2Room *r = getRoom();
	SpectrumMemory *mem = m_game->getMem();

        bg = r->getBackground();

	if (m_game->getStartRoom() == m_room)
	{
		m_guards[n++] = new Jsw2FixedGuard(FG_WILLY, mem);
	}
	switch(r->getT5())
	{
		case 9:	m_guards[n++] = new Jsw2FixedGuard(FG_TOILET, mem); 
			break;
		case 2: m_guards[n++] = new Jsw2HVGuardian(mem->memoryAt(0xFB30),mem);
			m_guards[n++] = new Jsw2HVGuardian(mem->memoryAt(0xFB37),mem);
			break;
		case 3: m_guards[n++] = new Jsw2HVGuardian(mem->memoryAt(0xFB3E),mem);
			m_guards[n++] = new Jsw2HVGuardian(mem->memoryAt(0xFB45),mem);
			break;
		case 4: m_guards[n++] = new Jsw2HVGuardian(mem->memoryAt(0xFB4C),mem);
			m_guards[n++] = new Jsw2HVGuardian(mem->memoryAt(0xFB53),mem);
			break;
		case 5: m_guards[n++] = new Jsw2HVGuardian(mem->memoryAt(0xFB5A),mem);
			m_guards[n++] = new Jsw2HVGuardian(mem->memoryAt(0xFB61),mem);
			break;
		case 24:m_guards[n++] = new Jsw2HVGuardian(mem->memoryAt(0xFB68),mem);
			m_guards[n++] = new Jsw2HVGuardian(mem->memoryAt(0xFB6F),mem);
			break;
		case 19:m_guards[n++] = new Jsw2HVGuardian(mem->memoryAt(0xFB76),mem);
			m_guards[n++] = new Jsw2HVGuardian(mem->memoryAt(0xFB7D),mem);
			break;
		case 21:m_guards[n++] = new Jsw2HVGuardian(mem->memoryAt(0xFB84),mem);
			m_guards[n++] = new Jsw2HVGuardian(mem->memoryAt(0xFB8B),mem);
			break;
	}
}


void Jsw2RoomDraw::drawItems(void)
{
	int ink = (m_ticker / 5);

	getRoom()->drawItems(m_bmAnim, ink);
}                                                     



void Jsw2RoomDraw::onTick(void)
{
	flashColours();
}


void Jsw2RoomDraw::drawTeleports()
{
	int n, x, y;
	int inks[4];
	char buf[2];
	VideoSurface *s = m_bmAnim->getSurface();
	buf[1] = 0;

	if (getRoom()->getT5() != 11 && getRoom()->getT5() != 22)
	{
		return;
	}

	inks[((m_ticker / 5)) % 4] = ZX_BRORANGE;
	inks[((m_ticker / 5)+1) % 4] = ZX_ORANGE;
	inks[((m_ticker / 5)+2) % 4] = ZX_BRYELLOW;
	inks[((m_ticker / 5)+3) % 4] = ZX_YELLOW;

	for (n = 0; n < 4; n++) 
	{
		buf[0] = n + '1';
		jswByte *t = m_game->getMem()->memoryAt(0x7435 + 6 * n);
		if (t[0] == m_room)
		{
			x = t[1] * 4;
			y = t[2] * 2;
			s->box(x,   y,   32, 32, inks[0]);
			s->box(x+1, y+1, 30, 30, inks[1]);
			s->box(x+2, y+2, 28, 28, inks[2]);
			s->box(x+4, y+4, 24,  2, inks[3]);		
			s->box(x+15, y+4, 2, 24, inks[3]);		
			s->drawSmallText(x+4, y+16, buf, ZX_BRORANGE, ZX_BLACK);
		}
	}

}


void Jsw2RoomDraw::drawBorder(void)
{
	int w = X_OFFSET * 2 + 512;
	int h = Y_OFFSET * 2 + 256 + CHAR_H;

	int bdr = getRoom()->getBorder();
	videoScreen->fillBox(m_x, m_y, w, Y_OFFSET, bdr);
	videoScreen->fillBox(m_x, m_y, X_OFFSET, h, bdr);
	videoScreen->fillBox(m_x + w - X_OFFSET, m_y, X_OFFSET, h, bdr);
	videoScreen->fillBox(m_x, m_y + h - Y_OFFSET, w, Y_OFFSET, bdr);
        drawTick(videoScreen, m_x, m_y);
}

