
/************************************************************************

    JSWED 2.3.0 - Editor for Jet Set Willy and derivatives

    Copyright (C) 2000-2001,2004  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*************************************************************************/

#include "jswed.hxx"
#include "jsw2game.hxx"
#include "room.hxx"
#include "jsw2room.hxx"
#include "jsw2roomform.hxx"

Jsw2RoomForm::Jsw2RoomForm(int roomno, Jsw2Game *g) : 
		VideoForm( 3 * CHAR_W, 8 * CHAR_H, 34 * CHAR_W, 16 * CHAR_H)
{
	m_roomno = roomno;
	m_room = (Jsw2Room *)(g->getRoomClass(roomno));
	m_game = g;

	m_room->getTitle(m_title);
	m_edit = new VideoTextEdit(m_x + CHAR_SW, m_y + 2 * CHAR_H, 
				  33 * CHAR_W + 4, CHAR_H + 4);
	m_edit->setLimit(32);
	m_edit->bindFixed(m_title);

	m_spin = new VideoSpinControl(m_x + CHAR_SW + 12 * CHAR_W, 
			m_y + 4 *CHAR_H - 2, 7 * CHAR_W, CHAR_H + 4);
	m_patch = m_room->getT5();
	m_spin->setLimits(0, 26);
	m_spin->bind(&m_patch);	
	m_spin->setListener(this);

	m_check[0] = new VideoCheckbox(30 * CHAR_W, 14 * CHAR_H);
	m_check[1] = new VideoCheckbox(30 * CHAR_W, 16 * CHAR_H);
	m_check[2] = new VideoCheckbox(30 * CHAR_W, 18 * CHAR_H);

	jswByte t4 = m_room->getT4();
	m_check[0]->setCheck(t4 & 0x80);
	m_check[1]->setCheck(t4 & 0x40);
	m_check[2]->setCheck(t4 & 0x20);

	m_lbl[0] = new VideoLabel(10 * CHAR_W + CHAR_SW, m_y + 6 * CHAR_H, "Room has a rope");
	m_lbl[1] = new VideoLabel(10 * CHAR_W + CHAR_SW, m_y + 8 * CHAR_H, "Conveyor animation");
	m_lbl[2] = new VideoLabel(10 * CHAR_W + CHAR_SW, m_y + 10 * CHAR_H, "Full conveyor anim");
	m_lbl[3] = new VideoLabel(m_x + CHAR_SW, m_y + 4 * CHAR_H, "Patch code");

	m_ok = new VideoButton(ZXE_OK, 30 * CHAR_W, 22 * CHAR_H, 
		     "  OK  ");
	m_cancel = new VideoButton(ZXE_CANCEL, 22 * CHAR_W, 22 * CHAR_H, 
                     "Cancel");

	if (roomno == 0x1C)
	{
		m_bmpSprite = bitmapFromSprite(16, 16, 
				g->getMem()->memoryAt(0xE0E1), 
				1, ZX_BRWHITE, ZX_BLACK);
	}
	else if (roomno >= 0x60 && roomno < 0x82)
	{
		m_bmpSprite = bitmapFromSprite(16, 16, 
				g->getMem()->memoryAt(0xEB81), 
				1, ZX_BRWHITE, ZX_BLACK);
	}
	else
	{
		m_bmpSprite = bitmapFromSprite(16, 16, 
				g->getMem()->memoryAt(0xD561), 
				1, ZX_BRWHITE, ZX_BLACK);
	}
}



Jsw2RoomForm::~Jsw2RoomForm()
{
	delete m_lbl[3];
	delete m_lbl[2];
	delete m_lbl[1];
	delete m_lbl[0];
	delete m_edit;
	m_spin->bind(NULL);
	delete m_spin;
	delete m_check[0];
	delete m_check[1];
	delete m_check[2];
	delete m_ok;
	delete m_cancel;
	delete m_bmpSprite;
}

void Jsw2RoomForm::redraw(void)
{
	VideoForm::redraw();

	redrawBorder("Room properties");	
	videoScreen->drawText(m_x + CHAR_SW, m_y + 12 * CHAR_H + CHAR_SH, "Willy's sprite: ", ZX_BLACK, ZX_WHITE);
	m_bmpSprite->toScreen(m_x + CHAR_SW + 15 * CHAR_W, m_y + 12 * CHAR_H);
	drawPatch();
}



int Jsw2RoomForm::doModal()
{
	int r = VideoForm::doModal();

	if (r == ZXE_OK)
	{
		jswByte t4 = 0;

		m_room->setTitle(m_title);
		if (m_check[0]->getCheck()) t4 |= 0x80;
		if (m_check[1]->getCheck()) t4 |= 0x40;
		if (m_check[2]->getCheck()) t4 |= 0x20;
		m_room->setT4(t4);
		m_room->setT5(m_patch);
	}
	return r;
}


void Jsw2RoomForm::showChildren(int redraw)
{
	// Order does matter - it dictates tabbing order 
        addChild(m_edit);
	addChild(m_lbl[3]);
        addChild(m_spin);
	addChild(m_check[0]);
	addChild(m_lbl[0]);
	addChild(m_check[1]);
	addChild(m_lbl[1]);
	addChild(m_check[2]);
	addChild(m_lbl[2]);
	addChild(m_ok);
        addChild(m_cancel);
	VideoForm::showChildren(redraw);
}

void Jsw2RoomForm::hideChildren(void)
{
	VideoForm::hideChildren();
	removeChild(m_check[0]);
	removeChild(m_lbl[0]);
	removeChild(m_check[1]);
	removeChild(m_lbl[1]);
	removeChild(m_check[2]);
	removeChild(m_lbl[2]);
	removeChild(m_lbl[3]);
	removeChild(m_spin);
	removeChild(m_edit);
	removeChild(m_ok);
        removeChild(m_cancel);
}


int Jsw2RoomForm::onButtonSelect(VideoButton *b)
{
	return ZXE_CONTINUE;
}

void Jsw2RoomForm::onEditChange(VideoControl *v)
{
	drawPatch();
}

static const char *st_patchNames[27] = 
{
	"No patch    ",	// 0
	"Central Cave",	// 1
	"Lift 1      ",	// 2
	"Lift 2      ",	// 3
	"Lift 3      ",	// 4
	"Lift 4      ",	// 5
	"Tribble trbl",	// 6
	"------------", // 7
	"Rocket Room ", // 8
	"The Bathroom", // 9
	"Master Bedrm", // 10
	"Beam me Down", // 11
	"Belfry      ", // 12
	"Eggoids     ", // 13
	"The Yacht   ", // 14
	"Trip Switch ", // 15
	"Rigor Mortis", // 16
	"Crypt Switch", // 17
	"Foot Room   ", // 18
	"Lift 5      ", // 19
	"First Lnding", // 20
	"Lift 6      ", // 21
	"Desert Isle ", // 22
	"Macaroni Ted", // 23
	"Dumb Waiter ", // 24
	"Hway to Hell", // 25
	"------------", // 26
};

void Jsw2RoomForm::drawPatch()
{
	const char *patchName;

	if (m_patch < 27) patchName = st_patchNames[m_patch];
	else		  patchName = "OUT OF RANGE";

	videoScreen->drawText(m_x + 20 * CHAR_W, m_y + 4 * CHAR_H, patchName,
			ZX_BLACK, ZX_WHITE);
		
}
