/************************************************************************

    JSWED 2.3.2 - Editor for Jet Set Willy and derivatives

    Copyright (C) 2000-2001,2004-6  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*************************************************************************/

#include "jswed.hxx"
#include "jsw2game.hxx"
#include "room.hxx"
#include "jsw2room.hxx"
#include "jsw2roompage.hxx"
#include "jsw2roomform.hxx"
#include "jsw2cellform.hxx"
#include "jsw2guard.hxx"
#include "jsw2guardform.hxx"
#include "jsw2arrowform.hxx"

static Jsw2Room *gl_clipRoom = NULL;

#define TOP (VIDEO_H - 224)

Jsw2RoomPage::Jsw2RoomPage(Jsw2Game *game) : Jsw2RoomNav(game)
{
	m_ot   = NULL;
	m_cx = m_cy = m_cyminor = m_cxminor = 0;
	m_oct = -1;
	m_oft = -1;
	m_mb = 0;
	m_ocx = m_ocy = -1;
	m_cdir = 0;
	m_undoRoom = NULL;

	// Active areas
	int x, y, w, h;
	x = VIDEO_W - 10 * CHAR_W;
	y = TOP + CHAR_H + CHAR_SH;
	w = 10 * CHAR_W;
	h = CHAR_H;

	addHotspot(new Hotspot('a', x, y, w, h, 'A', 'a')); y += CHAR_H;
	addHotspot(new Hotspot('w', x, y, w, h, 'W', 'w')); y += CHAR_H;
	addHotspot(new Hotspot('e', x, y, w, h, 'E', 'e')); y += CHAR_H;
	addHotspot(new Hotspot('f', x, y, w, h, 'F', 'f')); y += CHAR_H;
        addHotspot(new Hotspot('r', x, y, w, h, '/', 'r')); y += CHAR_H;
        addHotspot(new Hotspot('c', x, y, w, h, '<', 'c')); y += CHAR_H;
        addHotspot(new Hotspot('i', x, y, w, h, 'I', 'i')); y += CHAR_H;
        addHotspot(new Hotspot('R', x, y, w, h, '\\', 'R')); y += CHAR_H;
        addHotspot(new Hotspot('C', x, y, w, h, '>', 'C')); y += CHAR_H;

	w = 8 * CHAR_W;
	y = TOP + 2*CHAR_H;
	x = VIDEO_W - 19 * CHAR_W;

        addHotspot(new Hotspot('1', x, y, w, h, '1')); y += CHAR_H;
        addHotspot(new Hotspot('2', x, y, w, h, '2')); y += CHAR_H;
        addHotspot(new Hotspot('3', x, y, w, h, '3')); y += CHAR_H;
        addHotspot(new Hotspot('4', x, y, w, h, '4')); y += CHAR_H;
        addHotspot(new Hotspot('5', x, y, w, h, '5')); y += CHAR_H;
        addHotspot(new Hotspot('6', x, y, w, h, '6')); y += CHAR_H;
        addHotspot(new Hotspot('7', x, y, w, h, '7')); y += CHAR_H;
        addHotspot(new Hotspot('8', x, y, w, h, '8')); y += CHAR_H;
	
        x = CHAR_W * 16;
        y = TOP + (2 * CHAR_H);
	h = CHAR_H;
	w = 11 * CHAR_W;

	addHotspot(new Hotspot('X', x, y,         w, h, 'X', 'x'));
	addHotspot(new Hotspot('B', x, y +     h, w, h, 'B', 'b'));
	addHotspot(new Hotspot('O', x, y + 2 * h, w, h, 'O', 'o'));
	addHotspot(new Hotspot('G', x, y + 3 * h, w, h, 'G', 'g'));
	addHotspot(new Hotspot('C' | ZXK_ALT, x, y+4*h, w, h, 'C' | ZXK_ALT,
				'c' | ZXK_ALT));

	setTitle("Rooms");
	setCellType(0);
}

Jsw2RoomPage::~Jsw2RoomPage()
{
	if (m_undoRoom) delete m_undoRoom;
}



void Jsw2RoomPage::drawFixed(void)
{
	int x,y,h;

	videoScreen->bottomBar("Room editor");

	videoScreen->fillBox(VIDEO_W - 10 * CHAR_W, TOP + CHAR_H + CHAR_SH,
				10 * CHAR_W, 9 * CHAR_H + CHAR_SH, ZX_BRWHITE);
        videoScreen->box(VIDEO_W - 10 * CHAR_W, TOP + CHAR_H,
                                10 * CHAR_W, 10 * CHAR_H, ZX_BLACK);

	videoScreen->fillBox(VIDEO_W - 19 * CHAR_W, TOP + CHAR_H + CHAR_SH,
				8 * CHAR_W, 9 * CHAR_H, ZX_BRWHITE);
	videoScreen->box(VIDEO_W - 19 * CHAR_W, TOP + CHAR_H + CHAR_SH,
				8 * CHAR_W, 9 * CHAR_H, ZX_BLACK);

	videoScreen->drawSmallText(VIDEO_W - 10 * CHAR_W, TOP + CHAR_H,
			      "Cells                   ", 
				ZX_BRWHITE, ZX_BLACK);
	videoScreen->smallZxLogo(VIDEO_W - 6 * CHAR_SH, TOP + CHAR_H);

	x = VIDEO_W - 10 * CHAR_W + CHAR_SH;
	y = TOP + CHAR_H + CHAR_SH;
	h = CHAR_H;
	videoScreen->drawText(x,y,  "a:Air", ZX_BLACK, ZX_BRWHITE);
        videoScreen->drawText(x,y+h,"w:Water", ZX_BLACK, ZX_BRWHITE);
        videoScreen->drawText(x,y+2*h,"e:Earth", ZX_BLACK, ZX_BRWHITE);
        videoScreen->drawText(x,y+3*h,"f:Fire", ZX_BLACK, ZX_BRWHITE);
        videoScreen->drawText(x,y+4*h,"r:Ramp /", ZX_BLACK, ZX_BRWHITE);
        videoScreen->drawText(x,y+5*h,"c:Conv <", ZX_BLACK, ZX_BRWHITE);
        videoScreen->drawText(x,y+6*h,"i:Item", ZX_BLACK, ZX_BRWHITE);
        videoScreen->drawText(x,y+7*h,"R:Ramp \\", ZX_BLACK, ZX_BRWHITE);
        videoScreen->drawText(x,y+8*h,"C:Conv >", ZX_BLACK, ZX_BRWHITE);

        videoScreen->drawSmallText(VIDEO_W - 19 * CHAR_W, TOP + CHAR_H,
                              "Guards          ", 
				ZX_BRWHITE, ZX_BLACK);
        videoScreen->smallZxLogo(VIDEO_W - 11 * CHAR_H - 6 * CHAR_SH, 
				TOP + CHAR_H);
	videoScreen->fillBox(VIDEO_W - 13 * CHAR_W, TOP + CHAR_H + CHAR_SH, 1, 
				9 * CHAR_H, ZX_BLACK);

	x = CHAR_W * 16;
	y = TOP + CHAR_H;
	h = CHAR_H;

	videoScreen->fillBox(x - CHAR_SW, y + CHAR_SH, 25 * CHAR_SW, 7 * CHAR_H,
			ZX_BRWHITE);
        videoScreen->box(x - CHAR_SW, y + CHAR_SH, 25 * CHAR_SW, 7 * CHAR_H,
                        ZX_BLACK);

	videoScreen->drawSmallText(x - CHAR_SW,y,
			"Other options            ", ZX_BRWHITE, ZX_BLACK);
	videoScreen->drawText(x,y + h, "X:set eXits", ZX_BLACK, ZX_BRWHITE);
	videoScreen->drawText(x,y+2*h, "B:clipBoard", ZX_BLACK, ZX_BRWHITE);
	videoScreen->drawText(x,y+3*h, "O:Options  ", ZX_BLACK, ZX_BRWHITE);
	videoScreen->drawText(x,y+4*h, "G:Guardians", ZX_BLACK, ZX_BRWHITE);
	videoScreen->drawText(x,y+5*h, "AltC:Cells ", ZX_BLACK, ZX_BRWHITE);
	videoScreen->smallZxLogo(x + 18 * CHAR_SW, y);

	x = m_x;
	y = TOP + 9 * CHAR_H;
	videoScreen->drawSmallText(x, y, "Keyboard: ^Z or ^U to undo; CTRL+cursors", ZX_BLACK, ZX_WHITE);
	y += CHAR_SH;
	videoScreen->drawSmallText(x, y, "or J,<,> change room; cursor keys and", ZX_BLACK, ZX_WHITE);
	y += CHAR_SH;
	videoScreen->drawSmallText(x, y, "SPACE to draw", ZX_BLACK, ZX_WHITE);

	Jsw2RoomNav::drawFixed();
}


void Jsw2RoomPage::setCellType(int ct)
{
	m_ct = ct;
	drawCellType();
}

void Jsw2RoomPage::drawCellType()
{
	if ((m_oct == m_ct) || !m_visible || !m_drawnFixed) return;

	int w  = 10 * CHAR_W;
	int x  = VIDEO_W - w;
	int y  = TOP + 2 * CHAR_H;
	int ox = x;
	int ow = w;
	int oy = y;

	y  += (m_ct  % 10) * CHAR_H;
	oy += (m_oct % 10) * CHAR_H;

	if (m_ct  > 8) {  x -= 9 * CHAR_W;  w = 8 * CHAR_W; }
	else		 y -= CHAR_SH;
	if (m_oct > 8) { ox -= 9 * CHAR_W; ow = 8 * CHAR_W; } 
	else		oy -= CHAR_SH;

	if (m_oct >= 0) videoScreen->fillReplace (ox, oy, ow, CHAR_H,
                                        ZX_BRCYAN,
                                        ZX_BRWHITE);
	videoScreen->fillReplace(x, y, w, CHAR_H, ZX_BRWHITE,
                                        ZX_BRCYAN);
	m_oct = m_ct;
}


void Jsw2RoomPage::flashColours(void)
{
//	int objcount = m_game->getItemCount();

	if (!m_visible) return;


//	if (m_oft != objcount && m_drawnFixed)
//	{
//		char buf[20];
//
//		sprintf(buf, "%03d item%s", objcount, 
//				(objcount == 1) ? " free " : "s free" );
//		videoScreen->drawSmallText(VIDEO_W - 10 * CHAR_W, 
//					   VIDEO_H - 4 * CHAR_H, buf, 
//					   ZX_BLACK, ZX_WHITE);
//
//		m_oft = objcount;
//	}
        Jsw2RoomNav::flashColours();

	/* Draw guardian thumbnails */
	for (int n = 0; n < 8; n++)
	{
        	if (m_guards[n]) m_guards[n]->drawThumb(videoScreen,
                                                       VIDEO_W - 12 * CHAR_W
                                                        - CHAR_SW,
                                                        TOP + (2 +n)*CHAR_H);

                else videoScreen->fillBox(VIDEO_W - 12 * CHAR_W - CHAR_SW,
                                                     TOP + (2+n) * CHAR_H,
                                                     CHAR_W, CHAR_H * (8-n),
                                                     ZX_BRWHITE);
	}

	/* Draw animation frames */
        if (m_ct >= 10 && m_ct < 20 && m_guards[m_ct % 10])
        {
                m_guards[m_ct % 10]->drawFrames(videoScreen, 80, VIDEO_H - 32);
        }
        else    videoScreen->fillBox(80, VIDEO_H - 32, 480, 32, ZX_WHITE);

}



void Jsw2RoomPage::drawXCursor(VideoSurface *s, int x, int y, int i)
{
	s->plot(x + 8, y + 8, i);
        s->plot(x + 7, y + 7, i);
        s->plot(x + 7, y + 9, i);
        s->plot(x + 6, y + 6, i);
        s->plot(x + 6, y +10, i);
        s->plot(x + 9, y + 7, i);
        s->plot(x + 9, y + 9, i);
        s->plot(x +10, y + 6, i);
        s->plot(x +10, y +10, i);
}

/*
void Jsw2RoomPage::drawRampCursor(VideoSurface *s, int i)
{
        int x, y, p, q, l;

        if      (m_cx < m_ocx) m_cdir = -1;
        else if (m_cx > m_ocx) m_cdir = 1;

        // Note that if m-cx != m_ocx, we don't touch m_cdir 

	l = abs(m_cx - m_ocx) + 1;
        x = m_ocx;
	y = m_ocy;
        for (q = 0; q < l; q++)
        {
		s->box(x * 16, y * 16, 16, 16, i);
                if (m_cdir)
		{
			for (p = 0; p < 16; p++)
			{
                            if (m_cdir < 0)
                                 s->plot(x * 16 + p, y * 16 + p, i);
                            else s->plot(x * 16 + p, y * 16 + (16-p), i);
                        }
                }
                else drawXCursor(s, x * 16, m_ocy * 16, i);
		x += m_cdir; 
      		y--; 	
		if (y < 0) 
		{
			m_cx = x;
			break;
		}
	}

}


void Jsw2RoomPage::drawConveyorCursor(VideoSurface *s, int i)
{
	int x, w;

	if      (m_cx < m_ocx) m_cdir = -1;
	else if (m_cx > m_ocx) m_cdir = 1;

	// Note that if m-cx != m_ocx, we don't touch m_cdir 

	x = (m_cdir < 0) ? m_cx : m_ocx;
	w = (m_cdir > 0) ? m_cx : m_ocx;

	w -= (x-1);

	s->box(x * 16, m_ocy * 16, w * 16, 16, i);

	for (int x1 = x; x1 < (x + w); x1++)
	{
		if (m_cdir)
		{
			s->box(x1 * 16 + 2, m_ocy * 16 + 8, 12, 1, i);
			if (m_cdir < 0)
			{
				s->plot(x1 * 16 + 3, m_ocy * 16 + 7, i);
				s->plot(x1 * 16 + 3, m_ocy * 16 + 9, i);
				s->plot(x1 * 16 + 4, m_ocy * 16 + 6, i);
				s->plot(x1 * 16 + 4, m_ocy * 16 +10, i);
			}
			else
			{
				s->plot(x1 * 16 +13, m_ocy * 16 + 7, i);
				s->plot(x1 * 16 +13, m_ocy * 16 + 9, i);
				s->plot(x1 * 16 +12, m_ocy * 16 + 6, i);
				s->plot(x1 * 16 +12, m_ocy * 16 +10, i);
			}
		}
		else drawXCursor(s, x1 * 16, m_ocy * 16, i);
	}
}
*/


void Jsw2RoomPage::drawCursor()
{
	VideoSurface *s = m_bmAnim->getSurface();
	int i = ZX_BRWHITE;
	if ((m_ticker % 20) >= 10) i = ZX_BRBLUE;

	if (m_ocx != -1)
	{
//		if (m_ct == 5) drawConveyorCursor(s, i);
//		else if (m_ct == 4) drawRampCursor(s, i);
	}
	else 
	{
		if (m_ct >= 10)
		{
			if (m_guards[m_ct % 10])
			{
				Jsw2Guardian *g = m_guards[m_ct % 10];

				g->drawBounds(s,i, m_cx*8 + m_cxminor, 
						   m_cy*8 + m_cyminor);
			}
			else	// Non-existent guardian
			{	
				if (m_cx > 30) m_cx = 30;
                                if (m_cy > 14) m_cy = 14;              
				s->box(m_cx * 16, m_cy * 16, 32, 32, i);
			}
		}		
		else s->box(m_cx * 16, m_cy * 16, 16, 16, i);
	}
}




void Jsw2RoomPage::drawData(int flashonly)
{
	if (!flashonly)
	{
	        char s[30];

       		sprintf(s, "Room editor [%d]", m_room);
	        videoScreen->bottomBar(s);
	}
	Jsw2RoomNav::drawData(flashonly);
}



void Jsw2RoomPage::drawGuardList(void)
{
	Jsw2Room *room = getRoom();
	int x, y, n, p, stop;
	char buf[8];
	int gc = room->getGuardianCount();
	int ac = room->getArrowCount();

	y = TOP + 2 * CHAR_H;
	x = VIDEO_W - 18 * CHAR_W;
	
	stop = 0;
	for (n = 0; n < 8; n++)
	{
		if (n >= (gc + ac))
		{
			strcpy(buf, "     ");
		}
		else if (n >= gc)
		{
			sprintf(buf, "%d:Arr", n + 1);
		}
		else
		{
			sprintf(buf, "%d:Grd", n + 1);
		}

		if (n + 10 == m_ct)	p = ZX_BRCYAN;
		else			p = ZX_BRWHITE;

		videoScreen->drawText(x, y, buf, ZX_BLACK, p);
		y += CHAR_H;
	}
}





void Jsw2RoomPage::putCell()
{
	Jsw2Room *room = getRoom();
	int gc = room->getGuardianCount();
	int ac = room->getArrowCount();

	if (m_ct >= 10)
	{
		Jsw2Guardian *g = m_guards[m_ct % 10];

		if (g)
		{
			g->setPos(m_cx * 8 + m_cxminor,
				  m_cy * 8 + m_cyminor);
		}
		return;	
	}

	if (m_ct == 6)
	{
		// Can't have more than 16 items in a room.
		if (room->getItemCount() >= 16) return;
	}

	room->setCell (m_cx, m_cy, m_ct);
	room->drawCell(m_bmRoom, m_cx, m_cy, false, false);
}






int Jsw2RoomPage::roomExits(void)
{
	char str[4][48];
	char title[4][33];
	jswByte *b = NULL;
	Jsw2Room *room = getRoom();
	jswByte *exits = room->getExits();
	Jsw2Room *rl   = (Jsw2Room *)(m_game->getRoomClass(exits[0]-1)),
	         *rr   = (Jsw2Room *)(m_game->getRoomClass(exits[2]-1)),
                 *ru   = (Jsw2Room *)(m_game->getRoomClass(exits[1]-1)),
                 *rd   = (Jsw2Room *)(m_game->getRoomClass(exits[3]-1));
	Jsw2Room dummy(*room);

	dummy.setTitle(":Out of range:");
	if (!rr) rr = &dummy; rr->getTitle(title[1]);
	if (!rl) rl = &dummy; rl->getTitle(title[0]);
	if (!ru) ru = &dummy; ru->getTitle(title[2]);
	if (!rd) rd = &dummy; rd->getTitle(title[3]);
	sprintf(str[0], "Left  [%03d %-24.24s]", exits[0]-1, title[0]);
	sprintf(str[1], "Right [%03d %-24.24s]", exits[2]-1, title[1]);
	sprintf(str[2], "Up    [%03d %-24.24s]", exits[1]-1, title[2]);
	sprintf(str[3], "Down  [%03d %-24.24s]", exits[3]-1, title[3]);

	VideoMenu vm("Set exits", str[0], str[1], str[2], str[3], NULL); 

	int rv = vm.doModal();

	if (rv >= ZXE_QUIT) return rv;
	if (rv != ZXE_OK) return ZXE_CONTINUE;

	JswRoomList *s = m_game->getRoomList();
	if (s)
	{
		switch(vm.getSelected())
		{
			case 0: b = &(exits[0]); s->setTitle("Left exit");
				break;
			case 1: b = &(exits[2]); s->setTitle("Right exit");
				break;
			case 2: b = &(exits[1]); s->setTitle("Up exit");
				break;
			case 3: b = &(exits[3]); s->setTitle("Down exit");
				break;
		}
		s->setTitle("Exit destination");
		s->setSelectedRoom(*b - 1);
		rv = s->doModal();
		if (rv != ZXE_QUIT && rv != ZXE_OK) rv = ZXE_CONTINUE;
		else if (rv == ZXE_OK)
		{
			*b = s->getSelectedRoom() + 1;
			rv = ZXE_CONTINUE;
		}
		delete s;
	}
	return rv;
}



int Jsw2RoomPage::roomProperties(void)
{
	Jsw2Room cimage(*getRoom());

	Jsw2RoomForm rf(m_room, m_game);
	int rv = rf.doModal();
	if (rv >= ZXE_QUIT) return rv;
	if (rv != ZXE_OK) 
	{
		getRoom()->copyFrom(&cimage);
	}
	else
	{
		drawData();
	}
	return ZXE_CONTINUE;
}


int Jsw2RoomPage::roomUdgs(void)
{
        Jsw2Room cimage(*getRoom());
        int rv;

	Jsw2CellForm rf(&cimage, m_game);
        rv = rf.doModal();
        if (rv >= ZXE_QUIT) return rv;
        if (rv == ZXE_OK)
        {
                getRoom()->copyFrom(&cimage);
		getRoom()->createCellPatterns();

                drawData();
        }
        return ZXE_CONTINUE;
}



int Jsw2RoomPage::guardianMenu(void)
{
	Jsw2Room *room = getRoom();
	VideoMenu vmg2("Guardian", "Cancel", "Add guardian", "Add arrow", 
			"Edit", "Delete", NULL);
        VideoMenu vmg3("Guardian", "Cancel", "Edit", "Delete", "Cancel", NULL);
        VideoMenu vmg4("Guardian", "Cancel", "Add guardian", "Add arrow", NULL);

	VideoMenu *vm;

	int gc, ac;
	int x, rv, ng, gi, bg;
	jswByte *g;

	gc = room->getGuardianCount();
	ac = room->getArrowCount();

	if (m_ct < 10)		vm = &vmg4;
	else if (gc + ac >= 8)	vm = &vmg3;
	else			vm = &vmg2;

	rv = vm->doModal();
        if (rv >= ZXE_QUIT) return rv;
        if (rv == ZXE_CANCEL) return ZXE_CONTINUE;
	int ns = vm->getSelected();

	if (m_ct >= 10 && (gc + ac >= 8)) ns += 2;
	// 0 = Cancel
	// 1 = Add guardian
	// 2 = Add arrow
	// 3 = Edit
	// 4 = Delete
        switch(ns)
	{
		// New guardian
		case 1: if (m_guards[7]) return ZXE_CONTINUE;
					// All guardians taken 
			for (gi = gc + ac; gi > gc; gi--)
			{
				m_guards[gi] = m_guards[gi - 1];
			}
			// gi = index to insert new guard
			g = room->getGuardianBuffer() + 7 * gi;
			g[0] = 0x21;	// Initial count
			g[1] = 0x20;	// Step
			g[2] = 26;	// Sprite
			g[3] = 1;	// Movement step
			g[4] = 60;	// X
			g[5] = 60;	// Y
			g[6] = 0x0B;	// Flags
			m_guards[gi] = new Jsw2HVGuardian(g, m_game->getMem());	
			room->setGuardianCount(gc + 1);
    			drawData(); 
			return ZXE_CONTINUE;
		// New arrow
		case 2: if (m_guards[7]) return ZXE_CONTINUE;
					// All guardians taken 
			gi = gc + ac;
			g = room->getArrowBuffer() + 2 * ac;
			g[0] = 0x0F;	// Arrow X
			g[1] = 0x40;	// Arrow Y
			m_guards[gi] = new Jsw2Arrow(g, m_game->getMem());	
			room->setArrowCount(ac + 1);
    			drawData(); 
			return ZXE_CONTINUE;

		case 3: // Edit guardian
			if (m_ct < 10) return ZXE_CONTINUE;
			ng = m_ct % 10;
			if (ng >= gc)	// Editing arrow
			{
				ng -= gc;
				jswByte gprev[2];
				g = room->getArrowBuffer() + 2 * ng;
				memcpy(gprev, g, 2);
				Jsw2ArrowForm ed(m_room, m_game, g, m_bmRoom);
				rv = ed.doModal();
				if (rv >= ZXE_QUIT) return rv;
				ng += gc;
				if (rv == ZXE_OK)
				{
					delete m_guards[ng];
					m_guards[ng] = new Jsw2Arrow(g, m_game->getMem());
				}
				else
				{
					memcpy(g, gprev, 2);
				}
			}
			else
			{
				jswByte gprev[7];
				g = room->getGuardianBuffer() + 7 * ng;
				memcpy(gprev, g, 7);
				Jsw2GuardEditor ed(m_room, m_game, g, m_bmRoom);

				rv = ed.doModal();
				if (rv >= ZXE_QUIT) return rv;
				if (rv == ZXE_OK)
				{
					delete m_guards[ng];
					m_guards[ng] = new Jsw2HVGuardian(g, m_game->getMem());
				}
				else
				{
					memcpy(g, gprev, 7);
				}
			}
			return ZXE_CONTINUE;

		case 4:	// Delete guardian
                        if (m_ct < 10) return ZXE_CONTINUE;
			ng = m_ct % 10;

			if (ng >= gc)	// Deleting arrow
			{
				delete m_guards[ng];
				for (gi = ng; gi < 7; gi++)
				{
					m_guards[gi] = m_guards[gi+1];
				}
				m_guards[7] = NULL;
				ng -= gc;
				g = room->getArrowBuffer();
				for (gi = ng; gi < 8; gi++)
				{
					g[2*gi]   = g[2*gi+2];
					g[2*gi+1] = g[2*gi+3];
				}
				room->setArrowCount(ac - 1);
			}
			else		// Deleting guardian
			{
				delete m_guards[ng];
				for (gi = ng; gi < 7; gi++)
				{
					m_guards[gi] = m_guards[gi+1];
				}
				m_guards[7] = NULL;
				g = room->getGuardianBuffer();
				for (gi = ng; gi < 8; gi++)
				{
					memcpy(g, g + 7, 7);
					g += 7;
				}
				room->setGuardianCount(gc - 1);
			}
			drawGuardList();
			return ZXE_CONTINUE;	
	}
	return ZXE_CONTINUE;
}




int Jsw2RoomPage::preContextMenu(void)
{
	intPreContextMenu();
	VideoMenu vm("Options", "Editor", "Guardian", "Room options", 
		     "Cell patterns", "Room exits", "Clipboard etc.", 
                     "Import", "Export", "Exit editor", NULL);
	int rv = vm.doModal();

	if (rv >= ZXE_QUIT) return rv;
	if (rv == ZXE_CANCEL) 
	{
		postContextMenu();
		return ZXE_CONTINUE;
	}
	switch (vm.getSelected())
	{
		case 0: postContextMenu();
		        return ZXE_CONTINUE;
		case 1: rv = guardianMenu();
			postContextMenu();
			return rv;
		case 2: rv = roomProperties();
			postContextMenu();
			return rv;
		case 3: rv = roomUdgs();
			postContextMenu();
			return rv;
		case 4: rv = roomExits();
			postContextMenu();
			return rv;
		case 5: rv = roomClipboard();
			postContextMenu();
			return rv;
		case 6: rv = doImport();
			postContextMenu();
			return rv;
		case 7: rv = doExport();
			postContextMenu();
			return rv;
		case 8: postContextMenu();
			return ZXE_OK;
	}
	postContextMenu();
	return Jsw2RoomNav::preContextMenu();
}







void Jsw2RoomPage::undo(void)
{
	Jsw2Room *b = getRoom();
	Jsw2Room buf(*b);

	if (m_undoRoom)
	{
		b->copyFrom(m_undoRoom);
		delete m_undoRoom;
	}
	m_undoRoom = new Jsw2Room(buf);

	if (m_visible) drawData();
}


void Jsw2RoomPage::loadRoom(int room, int reload)
{
        Jsw2Room *b = (Jsw2Room *)(m_game->getRoomClass(room));

	if (!reload && b) 
	{
		if (m_undoRoom) delete m_undoRoom;
		m_undoRoom = new Jsw2Room(*b);
	}
	Jsw2RoomNav::loadRoom(room, reload);
}



int Jsw2RoomPage::onHotspot(int id)
{
	int rv;

	if (id >= '1' && id <= '8')
	{
		if (m_guards[(id-'1')]) setCellType(id - '1' + 10);
		return ZXE_CONTINUE;
	}
	switch(id)
	{
                case 'a':
                        setCellType(0); return ZXE_CONTINUE;
                case 'w':
                        setCellType(1); return ZXE_CONTINUE;
                case 'e':
                        setCellType(2); return ZXE_CONTINUE;
                case 'f':
                        setCellType(3); return ZXE_CONTINUE;
                case 'r':
                        setCellType(4); return ZXE_CONTINUE;
                case 'c':
                        setCellType(5); return ZXE_CONTINUE;
                case 'i':
                        setCellType(6); return ZXE_CONTINUE;
                case 'R':
                        setCellType(7); return ZXE_CONTINUE;
                case 'C':
                        setCellType(8); return ZXE_CONTINUE;

		case 'X': // Exits
			intPreContextMenu();
			rv = roomExits();
			postContextMenu();
			return rv;
		case 'B': // Clipboard
			intPreContextMenu();
			rv = roomClipboard();
			postContextMenu();
			return rv;	
		case 'O': // Options
			intPreContextMenu();
			rv = roomProperties();
			postContextMenu();
			return rv;
		case 'G': // Guardians
			intPreContextMenu();
			rv = guardianMenu();
			postContextMenu();
			return rv;
		case ('C' | ZXK_ALT): // Conveyor extra
			intPreContextMenu();
			rv = roomUdgs();
			postContextMenu();
			return rv;
	}
	return Jsw2RoomNav::onHotspot(id);
}


int Jsw2RoomPage::onKeyDown(int keysym)
{
	switch(keysym)
	{
		case ZXK_BREAK:
			m_ocx = m_ocy = -1;
			m_cdir  = -1;
			return ZXE_CONTINUE;
		case 'U' - '@':
		case 'Z' - '@':
			undo();
			return ZXE_CONTINUE;
		case ZXK_LEFT:
			if (m_ct >= 10)
			{
				if (m_cxminor > 1) { m_cxminor -= 2; return ZXE_CONTINUE; }
				if (m_cx > 0)      { m_cxminor = 6; --m_cx; return ZXE_CONTINUE; }
			}
			if (m_cx > 0) --m_cx;
			return ZXE_CONTINUE;
		case ZXK_RIGHT:
                        if (m_ct >= 10)
                        {
				if (m_cx == 30 && m_cxminor >= 5) return ZXE_CONTINUE;
                                if (m_cxminor < 5) { m_cxminor += 2; return ZXE_CONTINUE; }
                                if (m_cx < 30)      { m_cxminor = 0; ++m_cx; return ZXE_CONTINUE; }
                        }
			if (m_cx < 31) ++m_cx;
			return ZXE_CONTINUE;
		case ZXK_UP:
			if (m_ct >= 10)
			{
				if (m_cyminor > 0) { m_cyminor--; return ZXE_CONTINUE; }
				if (m_cy > 0)      { m_cyminor = 7; --m_cy; return ZXE_CONTINUE; }
			}
			if (m_cy > 0) --m_cy;
			return ZXE_CONTINUE;
		case ZXK_DOWN:
                        if (m_ct >= 10)
                        {
				if (m_cy == 14 && m_cyminor >= 6) return ZXE_CONTINUE;
                                if (m_cyminor < 7) { m_cyminor++; return ZXE_CONTINUE; }
                                if (m_cy < 15)      { m_cyminor = 0; ++m_cy; return ZXE_CONTINUE; }
                        }
			if (m_cy < 15) ++m_cy;
			return ZXE_CONTINUE; 
		case ZXK_ENTER:
		case ' ': 
			 putCell(); return ZXE_CONTINUE;
	}
	return Jsw2RoomNav::onKeyDown(keysym);
}

int Jsw2RoomPage::onButtonDown(int x, int y, int button)
{
	if (y >= m_y + Y_OFFSET && y < (m_y + 256 + Y_OFFSET) &&
            x >= m_x + X_OFFSET && x < (m_x + 512 + X_OFFSET))
	{
		m_mb = 1;
		return onMouseMove(x, y, 0, 0);
	}
	return Jsw2RoomNav::onButtonDown(x, y, button);
}

int Jsw2RoomPage::onButtonUp(int x, int y, int button)
{
	m_mb = 0;
	return Jsw2RoomNav::onButtonUp(x, y, button);
}

int Jsw2RoomPage::onMouseMove(int x, int y, int xrel, int yrel)
{
	if (y >= m_y + Y_OFFSET && y < (m_y + 256 + Y_OFFSET) &&
            x >= m_x + X_OFFSET && x < (m_x + 512 + X_OFFSET))
	{
		m_cx = (x - X_OFFSET - m_x)/16;
		m_cy = (y - Y_OFFSET - m_y)/16;
		m_cyminor = ((y - Y_OFFSET - m_y) % 16) / 2;
		m_cxminor = (((x - X_OFFSET - m_x) % 16) / 4) * 2;
		if (m_mb) putCell();
	}
	return ZXE_CONTINUE;
}





int Jsw2RoomPage::roomClipboard(void)
{
	jswByte fill;
	int y,x;

	if (!gl_clipRoom) gl_clipRoom = new Jsw2Room(*getRoom());

	VideoMenu vm("Clipboard etc.",
		     "Cut", "Copy", "Paste", "Clear to Air",
		    /* "Clear to Earth", "Clear to Water", "Clear to Fire", */
		     "Flip vertically", "Flip horizontally", NULL);
	int rv = vm.doModal();

	if (rv >= ZXE_QUIT) return rv;
	if (rv != ZXE_OK) return ZXE_CONTINUE;

	switch(vm.getSelected())
	{
		case 0: gl_clipRoom->copyFrom(getRoom());
			// FALL THROUGH to
		case 3:	getRoom()->clear();
			if (m_visible) gotoRoom(m_room);
			break;
		case 1: gl_clipRoom->copyFrom(getRoom());
			break;
		case 2: getRoom()->copyFrom(gl_clipRoom);
			if (m_visible) gotoRoom(m_room);
			break;
//		case 4: 
//		case 5:
//		case 6:
//			fill = (vm.getSelected() - 3); // 1 2 or 3
//			fill += (fill * 4) + (fill * 16) + (fill * 64);
//			getRoom()->clear();
//			memset(room, fill, 0x80);
//                      if (m_visible) drawData();
                        break;

		case 4: getRoom()->mirrorV();
			if (m_visible) drawData();
			break;

		case 5: getRoom()->mirrorH();
			if (m_visible) drawData();
			break;
	}
	return ZXE_CONTINUE;
}



