/************************************************************************

    JSWED 2.3.0 - Editor for Jet Set Willy and derivatives

    Copyright (C) 2000-2001,2004  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*************************************************************************/

#include "jswed.hxx"
#include "jswtitle.hxx"
#include "jsw2title.hxx"


Jsw2TitlePage::Jsw2TitlePage(SpectrumMemory *mem) : 
	JswTitlePage(findMem(mem), findLen(mem))
{
	int x, y, w, h;

	m_mem = mem;
	x = VIDEO_W - 26 * CHAR_W;
	y = 24 * CHAR_H + CHAR_SH;
	w = 4 * CHAR_W;
	h = 2 * CHAR_H;

	addHotspot(new Hotspot('I', x, y, w,   h, 'I', 'i')); x += w;
	addHotspot(new Hotspot('P', x, y, w,   h, 'P', 'p')); x += w;
	addHotspot(new Hotspot('T', x, y, w*2, h, 'T', 't')); 

	setup();
}

jswByte *Jsw2TitlePage::findMem(SpectrumMemory *m)
{
	 return m->memoryAt(m->peekW(0xFE94));
}

int Jsw2TitlePage::findLen(SpectrumMemory *m)
{
	int max;
	jswByte *attrs = findMem(m);

	for (max = 0; attrs[max] != 0xFF; max++);
	return max;
}


Jsw2TitlePage::~Jsw2TitlePage()
{
}

int Jsw2TitlePage::getCell(int x, int y, jswByte *attr, jswByte **pStripe)
{
	int base   = m_mem->peek(0xFE97);
	int offset = y * 32 + x;
	int value, udg;

	if (offset < base || offset >= (base + m_imgLen))
	{	
		return JswTitlePage::getCell(x, y, attr, pStripe);
	}
	value = m_attrs[offset-base];	
	if (attr)
	{
		if ((value & 0x3F) == m_mem->peek(0xFEC9))
		       	*attr = m_mem->peek(0xFECD) | 0x40;
		else	*attr = (value & 0x7F) | 0x40;
	}
	if (pStripe)
	{
		if (value & 0x80)
		{
			udg = 0xFE73;
			if (value & 0x40) udg += 16;
			if (x & 1) udg += 8;
			*pStripe = m_mem->memoryAt(udg);			
		}
		else	*pStripe = NULL;
	}
	return value;
}



void Jsw2TitlePage::putCell(int x, int y, jswByte attr)
{
	int base   = m_mem->peek(0xFE97);
	int offset = y * 32 + x;
	int value, udg;

	if (offset < base || offset >= (base + m_imgLen))
	{	
		return;
	}
	m_attrs[offset - base] = attr;
}


jswByte *Jsw2TitlePage::getUdgs()
{
	return m_mem->memoryAt(0xFE73);
}



int Jsw2TitlePage::onHotspot(int id)
{
	int oa = m_attrib;

	if (id != 'T') return JswTitlePage::onHotspot(id);

	m_attrib &= 0x3F;
	switch(oa & 0xC0)
	{
		case 0x00: case 0x40:
		       m_attrib	|= 0x80;
		       showAttributes();
		       break;
		case 0x80: 
		       m_attrib	|= 0xC0;
		       showAttributes();
		       break;
		case 0xC0: 
		       showAttributes();
		       break;

	}
	return ZXE_CONTINUE;
}

void Jsw2TitlePage::showAttributes()
{
	int x,y;
	int ifg, ibg, pfg, pbg;

	ibg = (m_attrib & 7) | ZX_BRIGHT;	       // Current ink
	pbg = ((m_attrib >> 3) & 7) | ZX_BRIGHT;	// Current paper

	//
	// "INK 9" - choose black/white ink for contrast
	//
	if (ibg >= ZX_BRGREEN) ifg = ZX_BLACK; else ifg = ZX_BRWHITE;
	if (pbg >= ZX_BRGREEN) pfg = ZX_BLACK; else pfg = ZX_BRWHITE;

	x = VIDEO_W - 24 * CHAR_W + CHAR_SW;
	y = 25 * CHAR_H;

	// Show colours
	videoScreen->drawText(x,          y, (ibg + '0' - ZX_BRIGHT), ifg, ibg);
	videoScreen->drawText(x+4*CHAR_W, y, (pbg + '0' - ZX_BRIGHT), pfg, pbg);

	// Show tile
	switch(m_attrib & 0xC0)
	{
		case 0: case 0x40:
		videoScreen->drawText(x+12*CHAR_W, y, 
				' ', ZX_BRBLUE, ZX_BRWHITE);
		break;
		case 0x80:
		videoScreen->drawText(x+12*CHAR_W, y, 
				'/', ZX_BRBLUE, ZX_BRWHITE);
		break;
		case 0xC0:
		videoScreen->drawText(x+12*CHAR_W, y, 
				'\\', ZX_BRBLUE, ZX_BRWHITE);
		break;
	}
}


int Jsw2TitlePage::redraw(int whichRectangle)
{
	m_visible = 1;
	int rv = VideoNotePage::redraw(whichRectangle);

	int x,y,w,h;
	x = VIDEO_W - 26 * CHAR_W;
	y = 24 * CHAR_H;
	w = 4 * CHAR_W;
	h = 2 * CHAR_H;
	videoScreen->fillBox(x, y, 16 * CHAR_W, 5*CHAR_SH, ZX_BRWHITE);
	videoScreen->box(x, y + CHAR_SH, 16 * CHAR_W, h, ZX_BLACK);
	videoScreen->drawSmallText(x, y, "Attributes                      ", ZX_BRWHITE, ZX_BLACK);
	videoScreen->drawText(x + CHAR_SW, y + CHAR_H, "I:* P:*  Tile:"
			, ZX_BLACK, ZX_BRWHITE);
	videoScreen->smallZxLogo(VIDEO_W - 26 * CHAR_SW, y);
	videoScreen->box(x, y + CHAR_SH, w, h, ZX_BLACK);
	videoScreen->box(x, y + CHAR_SH, w + 4*CHAR_W, h, ZX_BLACK);

	if (m_h < 192)
	{
		videoScreen->fillBox(m_x, m_y + 2 * m_h, 512, 384 - 2 * m_h, ZX_GREY);
	}
	if (m_w < 256)
	{
                videoScreen->fillBox(m_x + 2 * m_w, 0, 512 - 2 * m_w, 384, ZX_GREY);
	}
	drawData(0);
	return rv;
}


class Jsw2ColoursForm : public VideoForm
{
	VideoColourChooser *m_cc[4];
	VideoCheckbox      *m_cb;	
	VideoButton	   *m_ok, *m_cancel;
	SpectrumMemory     *m_mem;
public:
	Jsw2ColoursForm(SpectrumMemory *m) : VideoForm(
			10 * CHAR_W + CHAR_SW, 10 * CHAR_H, 19 * CHAR_W, 9 * CHAR_H)
	{
		m_mem = m;
		m_cc[0] = new VideoColourChooser(m_x + CHAR_W, m_y + 5 * CHAR_SH);
		m_cc[1] = new VideoColourChooser(m_x + CHAR_W, m_y + 8 * CHAR_SH);
		m_cc[2] = new VideoColourChooser(m_x + 10 * CHAR_W, m_y + 5 * CHAR_SH);
		m_cc[3] = new VideoColourChooser(m_x + 10 * CHAR_W, m_y + 8 * CHAR_SH);
		m_cb = new VideoCheckbox(m_x + 17 * CHAR_W, m_y + 11 * CHAR_SH);
		int n;
		m_cc[0]->setColour(m->peek(0xFEC9) & 7);
		m_cc[1]->setColour((m->peek(0xFEC9) >> 3) & 7);
		m_cc[2]->setColour(m->peek(0xFECD) & 7);
		m_cc[3]->setColour((m->peek(0xFECD) >> 3) & 7);
		m_cb->setCheck(m->peek(0xFECD) & 0x80);
		m_cancel = new VideoButton(ZXE_CANCEL, m_x + CHAR_W, m_y + 14 * CHAR_SH, "Cancel");
		m_ok = new VideoButton(ZXE_OK, m_x + 12 * CHAR_W, m_y + 14 * CHAR_SH, "  OK  ");

	}

	~Jsw2ColoursForm()
	{
		int n;
		for (n = 0; n < 4; n++)
		{
			delete m_cc[n];
		}
		delete m_cb;
		delete m_ok;
		delete m_cancel;
	}

	virtual void redraw()
	{
		VideoForm::redraw();
		redrawBorder("Colours");

		videoScreen->drawText(m_x + CHAR_W, m_y + CHAR_H,
				"Map this  to this", ZX_BLACK, ZX_WHITE);
		videoScreen->drawText(m_x + 10 * CHAR_W, m_y + 11 * CHAR_SH,
				"Flash?", ZX_BLACK, ZX_WHITE);
	}

	virtual void showChildren(int redraw)
	{
		for (int n = 0; n < 4; n++)
		{
			addChild(m_cc[n]);
		}
		addChild(m_cb);
		addChild(m_cancel);
		addChild(m_ok);
		VideoForm::showChildren(redraw);
	}


	virtual void hideChildren()
	{
		for (int n = 0; n < 4; n++)
		{
			removeChild(m_cc[n]);
		}
		removeChild(m_cb);
		removeChild(m_cancel);
		removeChild(m_ok);
	}

	virtual int doModal()
	{
		int rv = VideoForm::doModal();
		if (rv == ZXE_OK)
		{
			int from = m_cc[0]->getColour() + 8 * 
				   m_cc[1]->getColour();		
			int to   = m_cc[2]->getColour() + 8 * 
				   m_cc[3]->getColour();		
			if (m_cb->getCheck()) to |= 0x80;
			m_mem->poke(0xFEC9, from);
			m_mem->poke(0xFECD, to);
		}
		return rv;
	}

};


int Jsw2TitlePage::setupColours()
{
	Jsw2ColoursForm cf(m_mem);
	return cf.doModal();
}
