/************************************************************************

    JSWED 2.3.0 - Editor for Jet Set Willy and derivatives

    Copyright (C) 2000-2001,2004  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*************************************************************************/

#include "jswed.hxx"
#include "jswtitle.hxx"
#include "j48udgsform.hxx"
#include "tick.xbm"
#include "png.h"

extern "C" {

#include <SDL.h>
#include <SDL_image.h>

};


JswTitlePage::JswTitlePage(jswByte *image, int len)
{
	m_attrs  = image;
	m_imgLen = len;

	setTitle("Title");
	m_bDraw = false;
	m_bmAttr1 = m_bmAttr2 = NULL;
	m_tcount = 0;
	m_haveTimer = 0;
	m_ot = NULL;
	m_pcx = m_pcy = 0;
	m_acx = m_acy = 0;
	m_visible = 0;
	m_attrib = 0x47;
	m_buttonDown = 0;

	m_bmTick = bitmapFromXbm(tick_width, tick_height, tick_bits, 0,
				ZX_BRBLUE, ZX_BRWHITE);
}


JswTitlePage::~JswTitlePage()
{
	deleteBitmaps();	
	delete m_bmTick;
	if (m_haveTimer) setTickListener(m_ot);
}

void JswTitlePage::deleteBitmaps(bool withUndo)
{
	if (m_bmAttr1)   delete m_bmAttr1;
	if (m_bmAttr2)   delete m_bmAttr2;
        m_bmAttr1 = m_bmAttr2 = NULL;
}

void JswTitlePage::setup(bool withUndo)
{	
	int x, y, offset, ink, paper, udg;
	jswByte attr;
	bool stopped, flash;

	deleteBitmaps(withUndo);
	m_w = 256;
	m_h = 128;
	m_aw = 32;
	m_ah = 16;
	m_bmAttr1 = newVideoBitmap(m_w*2, m_h*2);
        m_bmAttr2 = newVideoBitmap(m_w*2, m_h*2);
	stopped = false;
	for (y = 0; y < 16; y++) for (x = 0; x < 32; x++)
	{
		setCellAttribute(x,y);
	}
	if (withUndo) for (int n = 0; n < m_imgLen; n++)
	{
		m_undoAttrs[n] = m_attrs[n];
	}
}


int JswTitlePage::redraw(int whichRectangle)
{
	m_visible = 1;
	int rv = VideoNotePage::redraw(whichRectangle);

	int x,y,w,h;
	x = VIDEO_W - 26 * CHAR_W;
	y = 24 * CHAR_H;
	w = 4 * CHAR_W;
	h = 2 * CHAR_H;
	videoScreen->fillBox(x, y, 16 * CHAR_W, 5*CHAR_SH, ZX_BRWHITE);
	videoScreen->box(x, y + CHAR_SH, 16 * CHAR_W, h, ZX_BLACK);
	videoScreen->drawSmallText(x, y, "Attributes                      ", ZX_BRWHITE, ZX_BLACK);
	videoScreen->drawText(x + CHAR_SW, y + CHAR_H, "I:* P:* B:* F:"
			, ZX_BLACK, ZX_BRWHITE);
	videoScreen->smallZxLogo(x + 26 * CHAR_SW, y);
	videoScreen->box(x, y + CHAR_SH, w, h, ZX_BLACK);
	videoScreen->box(x, y + CHAR_SH, w + 4*CHAR_W, h, ZX_BLACK);
	videoScreen->box(x, y + CHAR_SH, w + 8*CHAR_W, h, ZX_BLACK); 

	if (m_h < 192)
	{
		videoScreen->fillBox(m_x, m_y + 2 * m_h, 512, 384 - 2 * m_h, ZX_GREY);
	}
	if (m_w < 256)
	{
                videoScreen->fillBox(m_x + 2 * m_w, 0, 512 - 2 * m_w, 384, ZX_GREY);
	}

	drawData(0);

	return rv;
}


void JswTitlePage::drawData(int flashOnly)
{
	int t = m_tcount % 40;

	if (t >= 20 && m_bmAttr2) m_bmAttr2->toScreen(m_x, m_y);
        else if       (m_bmAttr1) m_bmAttr1->toScreen(m_x, m_y);

	drawCursor();
	
	if (!flashOnly)
	{
		showAttributes();
	}

}

void JswTitlePage::drawCursor(void)
{
	int t = m_tcount % 40;
	int ci;

	if (t >= 20) ci = ZX_BRBLUE;
	else	     ci = ZX_BRWHITE;

	videoScreen->box(m_x + 16 * m_acx, m_y + 16 * m_acy, 16, 16, ci);	
}


void JswTitlePage::drawCell(int x, int y)
{
        int t = m_tcount % 40;
	int sx = x * 16;
	int sy = y * 16;

        if (t >= 20 && m_bmAttr2) m_bmAttr2->toScreen(m_x+sx, m_y + sy,
				    sx, sy, 16, 16);
        else if       (m_bmAttr1) m_bmAttr1->toScreen(m_x+sx, m_y + sy,
				    sx, sy, 16, 16);
}




//
// Show the currently selected attributes
//
void JswTitlePage::showAttributes(void)
{
        int x,y;
	int ifg, ibg, pfg, pbg;

	ibg = (m_attrib & 7) | ZX_BRIGHT;		// Current ink
	pbg = ((m_attrib >> 3) & 7) | ZX_BRIGHT;	// Current paper

	//
	// "INK 9" - choose black/white ink for contrast
	//
	if (ibg >= ZX_BRGREEN) ifg = ZX_BLACK; else ifg = ZX_BRWHITE;
	if (pbg >= ZX_BRGREEN) pfg = ZX_BLACK; else pfg = ZX_BRWHITE;

        x = VIDEO_W - 24 * CHAR_W + CHAR_SW;
        y = 25 * CHAR_H;

	//
	// Display colours
	//
	videoScreen->drawText(x,          y, (ibg + '0' - ZX_BRIGHT), ifg, ibg);
	videoScreen->drawText(x+4*CHAR_W, y, (pbg + '0' - ZX_BRIGHT), pfg, pbg);

	//
	// Empty the bright/flash checkboxes
	//
	videoScreen->fillBox(x + 8*CHAR_W, y, CHAR_W, CHAR_H, ZX_BRWHITE);
        videoScreen->fillBox(x +12*CHAR_W, y, CHAR_W, CHAR_H, ZX_BRWHITE);

	//
	// And set them as appropriate.
	//
	if (m_attrib & 0x40) m_bmTick->toScreen(x +  8 * CHAR_W + 2, y+2);
        if (m_attrib & 0x80) m_bmTick->toScreen(x + 12 * CHAR_W + 2, y+2);
	
}

void JswTitlePage::onTick(void)
{
	if (!m_visible) return;
	++m_tcount;
	if ((m_tcount % 20) == 0)
	{
		drawData(1);
	}
	else if ((m_tcount % 20) == 10) 
	{
		drawCursor();
	}

}


void JswTitlePage::intPreContextMenu(void)
{
	if (m_haveTimer)
        {
                setTickListener(m_ot);
                m_tcount = 0;
                m_haveTimer = 0;
        }
}


int JswTitlePage::preContextMenu(void)
{
	jswByte tiles[32], *udgs;
	Jsw48UdgsForm fm(4, tiles, 8);

	intPreContextMenu();
	udgs = getUdgs();	
	if (udgs) memcpy(tiles, udgs, 32);
	VideoMenu vm("Options", "Editor", "Tiles", "Colours", "Import image", 
			"Export image", "Exit", NULL);
	int rv = vm.doModal();
	if (rv >= ZXE_QUIT) return rv;
	if (rv == ZXE_CANCEL)
	{
		postContextMenu();
		return ZXE_CONTINUE;
	}
	switch(vm.getSelected())
	{
		case 0: postContextMenu();
			return ZXE_CONTINUE;

		case 1: if (udgs) 
			{
				rv = fm.doModal();
				if (rv >= ZXE_QUIT) return rv;
				if (rv == ZXE_OK)
				{
					memcpy(udgs, tiles, 32);
					setup(false);
					drawData(0);
				}	
			}
			postContextMenu();
			return ZXE_CONTINUE;

		case 2:	rv = setupColours();
			if (rv == ZXE_OK)
			{
				setup(false);
				drawData(0);
			}
			postContextMenu();
			if (rv >= ZXE_QUIT) return rv;
			return ZXE_CONTINUE;

		case 3: rv = doImport();
			postContextMenu();
			return rv;

		case 4: rv = doExport();
			postContextMenu();
			return rv;

		case 5: postContextMenu();
			return ZXE_OK;
	}
	postContextMenu();
	return ZXE_CONTINUE;
}


void JswTitlePage::postContextMenu(void)
{
        if (!m_haveTimer)
        {
                m_ot = setTickListener(this);
                m_tcount = 0;
                m_haveTimer = 1;
        }
	VideoNotePage::postContextMenu();
}



void JswTitlePage::putCell(int x, int y, jswByte attr)
{
	
}

int JswTitlePage::getCell(int x, int y, jswByte *attr, jswByte **pStripe)
{
	if (attr)    *attr = 0;
	if (pStripe) *pStripe = NULL;
	return -1;
}

void JswTitlePage::setCellAttribute(int x, int y, int newAttr)
{
	int rv, ink, paper;
	jswByte attr;
	jswByte *stripe;
	VideoBitmap *bmCell;

	if (newAttr >= 0) putCell(x, y, newAttr);
	rv = getCell(x, y, &attr, &stripe);

	if (rv < 0)
	{
		m_bmAttr1->getSurface()->fillBox(x*16, y*16, 16, 16, ZX_GREY);
		m_bmAttr2->getSurface()->fillBox(x*16, y*16, 16, 16, ZX_GREY);
		return;
	}

	ink   =  attr & 7;
	paper = (attr >> 3) & 7;
	if (attr & 0x40)
	{
		ink   += 8;
		paper += 8;
	}
	if (stripe)
	{
		bmCell = bitmapFromSprite (8 ,8, stripe, 1, ink, paper);
		bmCell->toSurface(m_bmAttr1, x*16, y*16);
		bmCell->toSurface(m_bmAttr2, x*16, y*16);
		delete bmCell;
		if (attr & 0x80)
		{
			bmCell = bitmapFromSprite(8 ,8, stripe, 
						 1, paper, ink);
			bmCell->toSurface(m_bmAttr2, x*16, y*16);
			delete bmCell;
		}
	}
	else
	{
		m_bmAttr1->getSurface()->fillBox(x*16, y*16, 16, 16, paper);
		if (attr & 0x80)
			m_bmAttr2->getSurface()->fillBox(x*16, y*16, 16, 16, ink);
		else	m_bmAttr2->getSurface()->fillBox(x*16, y*16, 16, 16, paper);
		
	}
}




jswByte JswTitlePage::getCellAttribute(int x, int y)
{
	jswByte attr;

	getCell(x, y, &attr, NULL);

	return attr;
}



jswByte JswTitlePage::get8Pixels(int x, int y)
{
	jswByte *pStripe;

	getCell(x, y/8, NULL, &pStripe);

	if (pStripe) return pStripe[y & 7];
	return 0;
}



int JswTitlePage::onReveal(int whichRectangle)
{
        if (!m_haveTimer)
        {
                m_ot = setTickListener(this);
                m_tcount = 0;
                m_haveTimer = 1;
        }
        return VideoNotePage::onReveal();
}


int JswTitlePage::onConceal(void)
{
        if (m_haveTimer)
        {
		setTickListener(m_ot);
		m_tcount = 0;
                m_haveTimer = 0;
        }
        return VideoNotePage::onConceal();
}


int  JswTitlePage::onKeyDown(int keysym)
{
	int rv = onKey(keysym);
	if (rv != -1) return rv;

	if (keysym == 'Z' - '@' || keysym == 'U' - '@') return onUndo();

	switch(keysym)
	{
		case ZXK_UP:    moveCursor(m_acx, m_acy - 1); return ZXE_CONTINUE;
                case ZXK_DOWN:  moveCursor(m_acx, m_acy + 1); return ZXE_CONTINUE;
                case ZXK_LEFT:  moveCursor(m_acx - 1, m_acy); return ZXE_CONTINUE;
                case ZXK_RIGHT: moveCursor(m_acx + 1, m_acy); return ZXE_CONTINUE;
		case ' ':	setCellAttribute(m_acx, m_acy, m_attrib);
				drawCell(m_acx, m_acy);
			        drawCursor();
				break;

	}
	return VideoNotePage::onKeyDown(keysym);
}

int JswTitlePage::onMouseMove(int x, int y, int xrel, int yrel)
{
	if (x >= m_x && x < (m_x + (16*m_aw)) &&
	    y >= m_y && y < (m_y + (16*m_ah)))
	{
		x = (x - m_x) / 16;
		y = (y - m_y) / 16;
		if (m_buttonDown)
		{
			setCellAttribute(x, y, m_attrib);
			drawCell(x, y);		
		}
		moveCursor(x, y);
		return ZXE_CONTINUE;		
	}
	return VideoNotePage::onMouseMove(x, y, xrel, yrel);
}



int JswTitlePage::onButtonDown(int x, int y, int button)
{
        int rv = onButton(x, y, button);
        if (rv != -1) return rv;

	if (x >= m_x && y >= m_y && x < (m_x + 2 * m_w) && y < (m_y + 2*m_h))
	{
		m_buttonDown = 1;
		setCellAttribute(m_acx, m_acy, m_attrib);
		drawCell(m_acx, m_acy);
		drawCursor();
	}
	return VideoNotePage::onButtonDown(x, y, button);
}


int JswTitlePage::onButtonUp  (int x, int y, int button)
{
	m_buttonDown = 0;
	return VideoNotePage::onButtonUp(x, y, button);
}



void JswTitlePage::moveCursor(int x, int y)
{
	if (x < 0 || y < 0 || x >= m_aw || y >= m_ah) return;

	drawCell(m_acx, m_acy);
	m_acx = x;
	m_acy = y;
	drawCursor();	
}
/*
void JswTitlePage::drawZoom(void)
{
	int x, y;
	int x1, y1;
	x = m_acx;     
	y = m_acy;

	VideoBitmap *bm = newVideoBitmap(128, 128);
	for (x1 = 0; x1 < 8; x1++) for (y1 = 0; y1 < 8; y1++)	
	{
		int pixel = m_bmAttr1->getSurface()->pointZx(16 * x + 2*x1, 
							     16 * y + 2*y1);

		bm->getSurface()->fillBox(16 * x1, 16 * y1, 16, 16, pixel);
	}
	drawZoomCursor(bm->getSurface());
	bm->toScreen(m_lx, m_ly);
	delete bm;
}*/

int JswTitlePage::onHotspot(int id)
{
	int oa = m_attrib;

	switch(id)
	{
		case 'I':
			m_attrib &= (~7);
			m_attrib |= (oa + 1) & 7;
			showAttributes();
			break;
		case 'P':
			m_attrib &= (~0x38);
			m_attrib |= (oa + 8) & 0x38;
			showAttributes();
			break;
		case 'B': 
			m_attrib ^= 0x40;
			showAttributes();
			break;
		case 'F':
			m_attrib ^= 0x80;
			showAttributes();
			break;
	}
	return ZXE_CONTINUE;
}








int JswTitlePage::onUndo(void)
{
	int n;
	jswByte s;

	for (n = 0; n < m_imgLen; n++)
	{
		s = m_attrs[n];
		m_attrs[n] = m_undoAttrs[n];
		m_undoAttrs[n] = s;
	}
	setup(false);
	drawData(0);
	return ZXE_CONTINUE;
}





int JswTitlePage::doExport()
{
	VideoMenu vm("Export format ", "Screen$", "PNG", "Cancel", NULL);

	int rv = vm.doModal();
	if (rv >= ZXE_QUIT) return rv;
	if (rv != ZXE_OK) return ZXE_CONTINUE;

	switch (vm.getSelected())
	{
		case 0: return exportScreen();
		case 1: return exportPNG();
	}
	return ZXE_CONTINUE;
}

static unsigned char scrHeader[128] = 
{
	 'P',  'L',  'U',  'S',  '3',  'D',  'O',  'S',
	0x1A, 0x01, 0x00, 0x80, 0x1B, 0x00, 0x00, 0x03,
	0x00, 0x1B, 0x00, 0x40
};

int JswTitlePage::exportScreen()
{
        char filename[PATH_MAX];
        JswFileForm ff;
        int rv;
	int x, y;
	jswByte *row;
	unsigned char scrBuf[0x1B80];

        filename[0] = 0;
        rv = ff.doModal(filename);
        if (rv != ZXE_OK) return rv;

	FILE *fp = fopen(filename, "rb");
	if (fp)
	{
		fclose(fp);
		VideoMenu vm("Overwrite existing file?", "Yes", "No", NULL);

		if ( (rv = vm.doModal()) != ZXE_OK) return ZXE_CONTINUE;
		if (vm.getSelected()) return ZXE_CONTINUE;
	}
	// Generate screen bitmap
	memset(scrBuf, 0, sizeof(scrBuf));
	for (y = 0; y < 192; y++)
	{
		row = scrBuf + 0x80 + ((y & 0xC0) * 0x20) + 
			              ((y & 0x38) * 0x04) +
			              ((y & 0x07) * 0x100);
		for (x = 0; x < 32; x++)
		{
			*row++ = get8Pixels(x, y);	
		}
	} 
	// Attributes
	for (y = 0; y < 24; y++)
	{
		row = scrBuf + 0x1880 + 32 * y;
		for (x = 0; x < 32; x++)
		{
			if (y < m_ah && x < m_aw)
			{
       				row[x] = getCellAttribute(x, y);
			}
			else	row[x] = 0;
		}
	}	
	// +3DOS header
	memcpy(scrBuf, scrHeader, 128);
	for (x = 0; x < 127; x++)
	{
		scrBuf[127] += scrBuf[x];
	}
	fp = fopen(filename, "wb");
	if (fp)
	{
		if ((fwrite(scrBuf, 1, 0x1B80, fp) < 0x1B80) ||
		    (fclose(fp) == EOF))
		{
			alert("Write failed", "Cancel", 
					"Write failed on file:\n%s", filename);
		}
	}
	else
	{
		alert("Write failed", "Cancel", 
					"Could not open file:\n%s", filename);
	}
	return ZXE_CONTINUE;
}

// Define png_jumpbuf() in case we are using a pre-1.0.6 version of libpng 
#ifndef png_jmpbuf
#  define png_jmpbuf(png_ptr) png_ptr->jmpbuf
#endif

static void se_png_error(png_structp png_ptr, png_const_charp error_msg)
{
	alert("LibPNG error", "Cancel", (char *)error_msg);
	longjmp(png_ptr->jmpbuf, 1);
}

static void se_png_warn(png_structp png_ptr, png_const_charp warning_msg)
{
	fprintf(stderr, "LibPNG warning: %s\n", warning_msg);
}


extern int gl_b[], gl_r[], gl_g[];




int JswTitlePage::exportPNG()
{
        char filename[PATH_MAX];
        JswFileForm ff;
        int rv;
	int x, y, n;
	png_byte  rows[128 * 192];
	png_bytep rowPtr[192];
	jswByte mask;
	png_color pngPalette[16];

	memset(rows, 0, sizeof(rows));
	for (y = 0; y < 192; y++)
	{
		rowPtr[y] = &rows[128 * y];
	}
	for (n = 0; n < 16; n++)
	{
		pngPalette[n].red   = gl_r[n];
		pngPalette[n].green = gl_g[n];
		pngPalette[n].blue  = gl_b[n];
	}

	png_structp pPng  = NULL;
      	png_infop   pInfo = NULL; 

	pPng = png_create_write_struct(PNG_LIBPNG_VER_STRING,
		(png_voidp)this, se_png_error, se_png_warn);
	if (pPng)
	{
		pInfo = png_create_info_struct(pPng);
	}
	if (!pPng || !pInfo)
	{
		if (pPng) png_destroy_write_struct(&pPng, NULL);
		rv = alert("Out of memory", "Cancel");
		if (rv >= ZXE_QUIT) return rv;
		return ZXE_CONTINUE;	
	}
	if (setjmp(png_jmpbuf(pPng)))
	{
		png_destroy_write_struct(&pPng, &pInfo);
		return ZXE_CONTINUE;	
	}
        filename[0] = 0;
        rv = ff.doModal(filename);
        if (rv != ZXE_OK) 
	{
		png_destroy_write_struct(&pPng, &pInfo);
		return rv;
	}

	FILE *fp = fopen(filename, "rb");
	if (fp)
	{
		fclose(fp);
		VideoMenu vm("Overwrite existing file?", "Yes", "No", NULL);

		if ( (rv = vm.doModal()) != ZXE_OK) 
		{
			png_destroy_write_struct(&pPng, &pInfo);
			return ZXE_CONTINUE;
		}
		if (vm.getSelected()) 
		{
			png_destroy_write_struct(&pPng, &pInfo);
			return ZXE_CONTINUE;
		}
	}
	fp = fopen(filename, "wb");
	if (!fp)
	{
		png_destroy_write_struct(&pPng, &pInfo);
		rv = alert("File create failed", "Cancel", 
				"Could not create file:\n%s", filename);
		if (rv >= ZXE_QUIT) return rv;
		return ZXE_CONTINUE;	
	}
	// Let's go.
	png_init_io(pPng, fp);
	png_set_IHDR(pPng, pInfo, m_w, m_h, 4, PNG_COLOR_TYPE_PALETTE,
			PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_DEFAULT,
			PNG_FILTER_TYPE_DEFAULT);
	png_set_PLTE(pPng, pInfo, pngPalette, 16);

	// Generate the image.
	for (y = 0; y < 192; y++)
	{
		mask = 0x80;
		for (x = 0; x < 256; x++)
		{
			bool pix = false;
			if (x < m_w && y < m_h) 
			{
				jswByte pix8 = get8Pixels(x/8, y);
			 	pix = ((pix8 << (x & 7)) & 0x80) != 0;
			}
			jswByte attr = getCellAttribute(x/8, y/8);
			
			if (pix) attr = (attr & 7) | ((attr & 0x40) >> 3);
			else	 attr = (attr & 0x78) >> 3;

			if (x & 1) 
			{
				rowPtr[y][x/2] |= (attr & 0x0F);
			}
			else
			{
				rowPtr[y][x/2] |= ((attr & 0x0F) << 4);
			}
		}
	}
	png_set_rows(pPng, pInfo, rowPtr);
	png_write_png(pPng, pInfo, PNG_TRANSFORM_IDENTITY, NULL);
	png_destroy_write_struct(&pPng, &pInfo);
	fclose(fp);

	return ZXE_CONTINUE;
}

int JswTitlePage::doImport()
{
        char filename[PATH_MAX];
        JswFileForm ff;
	int rv, n;
	FILE *fp;
	jswByte magic[128], sum;

        filename[0] = 0;
	rv = ff.doModal(filename);

	if (rv >= ZXE_QUIT) return rv;
	if (rv != ZXE_OK)   return ZXE_CONTINUE;

	fp = fopen(filename, "rb");
	if (!fp)
	{
		rv = alert("File open failed", "Cancel", 
				"Could not open file:\n%s", filename);
		if (rv >= ZXE_QUIT) return rv;
		return ZXE_CONTINUE;
	}
	// Work out what we have here
	memset(magic, 0, sizeof(magic));
	fread(magic, 1, sizeof(magic), fp);
	if (!png_sig_cmp(magic, 0, 8))	// Importing PNG
	{
		fclose(fp);
		return importImage(filename);
	}
	if (!memcmp(magic, "BM", 2))	// Importing BMP
	{
		fclose(fp);
		return importImage(filename);
	}
	if (!memcmp(magic, "\377\330\377", 3)) // Importing JPG
	{
		fclose(fp);
		return importImage(filename);
	}
	if (!memcmp(magic, scrHeader, 14))
	{
		sum = 0;
		for (n = 0; n < 127; n++)
		{
			sum += magic[n];
		}
		if (sum == magic[n]) /* PLUS3DOS magic found */
		{
			return importScreen(fp);
		}
	}
	// No magic number. Treat as raw Spectrum screen.
	fseek(fp, 0, SEEK_SET);
	return importScreen(fp);
}



int JswTitlePage::importScreen(FILE *fp)
{
	int y,x;
	jswByte *row;
	jswByte scrBuf[6912];

	memset(scrBuf, 0, sizeof(scrBuf));
	fread(scrBuf, 1, sizeof(scrBuf), fp);
	fclose(fp);

	// Attributes
	for (y = 0; y < 24; y++)
	{
		row = scrBuf + 0x1800 + 32 * y;
		for (x = 0; x < 32; x++)
		{
			if (y < m_ah && x < m_aw)
			{
				setCellAttribute(x, y, row[x] & 0x3F);
			}
		}
	}	
	// Redraw
	setup(false);
	return ZXE_CONTINUE;
}

void JswTitlePage::importCell(int x, int y, SDL_Surface *s0)
{
	int x1, y1, n, ink, paper;
	int zxc[8][8];
	int frequency[16];
	Uint8 r,g,b;

	memset(frequency, 0, sizeof(frequency));
	// Initialise the colour array
	for (y1 = 0; y1 < 8; y1++)
		for (x1 = 0; x1 < 8; x1++)
			zxc[y1][x1] = -1;


	for (y1 = 0; y1 < 8; y1++)
	{
		if (y + y1 >= s0->h) break;
		for (x1 = 0; x1 < 8; x1++)
		{
			if (x + x1 >= s0->w) break;

			jswByte *pixel = (jswByte *)(s0->pixels);

			pixel += ((y + y1) * s0->pitch);
			pixel += ((x + x1) * s0->format->BytesPerPixel);
			Uint32 pixval = 0;
			switch(s0->format->BitsPerPixel)
			{
				case 8:  pixval = *(Uint8  *)pixel; break;
				case 16: pixval = *(Uint16 *)pixel; break;
				// XXX 24-bit pixels are nassty
				case 24: pixval = *(Uint32 *)pixel; break;
				case 32: pixval = *(Uint32 *)pixel; break;
			}
			SDL_GetRGB(pixval, s0->format, &r, &g, &b);
			zxc[y1][x1] = 0;
			if (b >= 192) zxc[y1][x1] |= 1;
			if (r >= 192) zxc[y1][x1] |= 2;
			if (g >= 192) zxc[y1][x1] |= 4;
			if (b >= 224) zxc[y1][x1] |= 8;
			if (r >= 224) zxc[y1][x1] |= 8;
			if (g >= 224) zxc[y1][x1] |= 8;
			frequency[zxc[y1][x1]]++;
		}
	}
	int max = 0;

	// Work out ink and paper
	for (n = 0; n < 16; n++)
	{
		if (frequency[n] > max) 
		{
			max = frequency[n];	
			ink = n;
		}
	}
	max = paper = 0;
	for (n = 0; n < 16; n++)
	{
		if (frequency[n] > max && n != ink) 
		{
			max = frequency[n];	
			paper = n;
		}
	}
	// OK. We have ink and paper.
	jswByte attr = (ink & 7) | ((paper & 7) << 3);
	if ((ink & 8) | (paper & 8)) attr |= 0x40;

	int ay = y / 8;
	int ax = x / 8;
	if (ay < m_ah && ax < m_aw)
	{
		setCellAttribute(ax, ay, attr & 0x3F);
	}
}

int JswTitlePage::importImage(char *filename)
{
        SDL_Surface *s0 = IMG_Load(filename);
	int rv;
	int x, y;

	if (!s0)
	{
		rv = alert("Failed to load image", "Cancel",
				"%s", IMG_GetError());
		if (rv >= ZXE_QUIT) return rv;
		return ZXE_CONTINUE;	
	}
	int imgw = s0->w;
	int imgh = s0->h;

	for (y = 0; y < imgh; y += 8)
	{
		for (x = 0; x < imgw; x += 8)
		{
			importCell(x, y, s0);
		}
	}
	SDL_FreeSurface(s0);
	// Redraw
	setup(false);
	return ZXE_CONTINUE;
}
