/************************************************************************

    JSWED 2.3.0 - Editor for Jet Set Willy and derivatives

    Copyright (C) 2000-2001,2004  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*************************************************************************/

#include "jswed.hxx"
#include "spritelist.hxx"
#include "manicgame.hxx"
#include "manicattrpage.hxx"
#include "manicpatchpage.hxx"
#include "manic128game.hxx"
#include "manic128attrpage.hxx"
#include "manic128patchpage.hxx"
#include "progressform.hxx"

#define ROOMCOUNT 20	// Hardcoded for now, but with compression you
			// could get more in.

Manic128Game::Manic128Game(SpectrumMemory *mem) : ManicGame(mem)
{
	m_attrPage  = NULL;
	m_patchPage = NULL;
}


Manic128Game::~Manic128Game()
{

}

char *Manic128Game::getID()
{
        char *s = ManicGame::getID();

        strcpy(s + ((VIDEO_W / CHAR_W) - 11), " [MANIC128]");
        return s;
}


// Do a rough imitation of the Software Projects engine for now.
unsigned short Manic128Game::xltAddress(unsigned short address)
{
        if (address < 0x846E) return address;
	// This line should get roomHasFeatures() working
	if (address < 0x87A2) return address + 48;
        if (address < 0x8C2A) return address + 6;
        if (address < 0x8C73) return address + 13;
        if (address < 0x9357) return address + 11;
        return address;
}

void Manic128Game::addPages(VideoNotebook *n)
{
	if (m_attrPage == NULL)
	{
		m_attrPage = new Manic128AttrPage(this);
	}
	if (m_patchPage == NULL)
	{
		m_patchPage = new Manic128PatchPage(this);
	}
	return ManicGame::addPages(n);
}


int Manic128Game::decompress()
{
	int n;

	for (n = 0; n < ROOMCOUNT; n++)
	{
		expandRoom(n, m_roomBuffer + n * 0x400);
	}
	expandData(m_mem->memoryAt(0x9D00), m_screenBuffer,        4096);

	// Look up attributes; JSWED's compressor can move them around.
	unsigned attrAddr = 256 * m_mem->peek(0x8617) + m_mem->peek(0x8616);
	expandData(m_mem->memoryAt(attrAddr), m_screenBuffer+0x1800,  768);

	return ZXE_CONTINUE;
}

int Manic128Game::compress()
{
	int n;
	unsigned compSize = 0;
	jswByte encodeBuf[1024 * ROOMCOUNT * 2];
	unsigned roomOffset[ROOMCOUNT];
	ProgressForm pf(4864 + 1024 * ROOMCOUNT);
	char buf[32];

	for (n = 0; n < ROOMCOUNT; n++)
	{
		sprintf(buf, "Compressing room %d", n);
		pf.bar(buf, n * 1024);
		roomOffset[n] = compSize + 0xC984;
		compSize += compressData(m_roomBuffer + 1024*n, 
				&encodeBuf[compSize], 1024, &pf);
	}
	if (compSize > 0x367C)
	{
		int rv = alert("Compression error", "Cancel",
			"The rooms will not compress enough to fit in the "
			"space available.\n"
			"Space available=%5d\n"
			"Space needed   =%5d\n", 0x367C, compSize);
		if (rv >= ZXE_QUIT) return rv;
		return ZXE_CANCEL;
	}
	memcpy(m_mem->memoryAt(0xC984), encodeBuf, 0x367C);

	pf.bar("Compressing screen", 1024 * ROOMCOUNT);
	compSize = compressData(m_screenBuffer, encodeBuf, 4096, &pf);
	memcpy(m_mem->memoryAt(0x9D00), encodeBuf, compSize);

	unsigned attrTable = 0x9D00 + compSize;
	// Set pointer to attributes
	m_mem->pokeW(0x8616, attrTable);

	pf.bar("Compressing screen", 1024 * ROOMCOUNT + 4096);
	compSize = compressData(m_screenBuffer+0x1800, encodeBuf, 768, &pf);
	memcpy(m_mem->memoryAt(attrTable), encodeBuf, compSize);

	unsigned roomTable = attrTable + compSize;

	printf("Attributes at %04x; room table at %04x\n",
			attrTable, roomTable);

	for (n = 0; n < ROOMCOUNT; n++)
	{
		m_mem->pokeW(roomTable + 3*n,     roomOffset[n]);
		m_mem->poke (roomTable + 3*n + 2, 0);
	}
	m_mem->pokeW(0x86C6, roomTable);
	return ZXE_CONTINUE;
}


void Manic128Game::expandRoom(int room, jswByte *buffer)
{
	jswByte *rTable, *data;
	unsigned offset;
	jswByte bank;

	// Get room table from LD HL,rooms_table. This allows JSWED's
	// compressor to move the rooms table around.
	data = m_mem->memoryAt(0x86C6);
       	rTable = m_mem->memoryAt(data[0] + 256 * data[1]);

	offset = rTable[3*room] + 256 * rTable[3*room+1];
	// Hypothesis: The third byte in the room address table 
	// is the bank containing the room definition.
	bank = 0; // rTable[3*room+2];

	data = m_mem->memoryAt(offset, bank);	

//	printf("Room %d:\n", room);
	expandData(data, buffer, 0x400);
}



void Manic128Game::expandData(jswByte *data, jswByte *buffer, unsigned len)
{
	jswByte rleByte, *base;
	unsigned offset;

	base = data;
	rleByte = *data++;
//	printf("\trleByte = %02x\n", rleByte);
	offset = 0;
	while (offset < len)
	{
		// Not the RLE trigger
		if (data[0] != rleByte)
		{
//			printf("Literal:    %02x\n",data[0]);
//			printf("%04x: Mundane byte %02x\n",
//				       	data - base, data[0]);
			buffer[offset] = data[0];
			++data;
			++offset;
			continue;
		}
		++data;
		// Literal byte 
		if (data[0] == 0)
		{
//			printf("Literal*:   %02x\n", rleByte);
//			printf("%04x: Literal byte %02x\n",
//				       	data - base, rleByte);
			buffer[offset++] = rleByte;
			++data;
			continue;
		}
		// Compressed sequence
		unsigned short cword = data[0] + 256 * data[1];
		unsigned short repeats = cword & 0x3F;
		unsigned short rlen    = cword >> 6;

//		printf("%04x: Repeat %04x -> %d of %d\n",
//				       	data - base, cword, repeats, rlen);
		if (repeats > rlen)
		{
			//printf("Overlap:    repeats=%2d rlen=%2d\n", repeats, rlen);
		}
		else
		{
			//printf("No overlap: repeats=%2d rlen=%2d\n", repeats, rlen);
		}
		data += 2;
		jswByte *source = &buffer[offset - rlen];
		do
		{
			buffer[offset++] = *source++;
			--repeats;
			if (offset >= len) return;
		}
		while (repeats);
	}
}


unsigned Manic128Game::compressData(jswByte *uncomp, jswByte *compressed, 
		unsigned len, ProgressForm *pf)
{
	unsigned freq[256];
	unsigned count, soFar, n, rleByte, min;
	unsigned output;

	memset(freq, 0, sizeof(freq));
	for (n = 0; n < len; n++)
	{
		++freq[uncomp[n]];
	}
	for (min = 65536, n = 0; n < 256; n++)
	{
		if (freq[n] < min) 
		{
			rleByte = n;
			min = freq[n];
		}		
	}
	count = soFar = 0;
	output = 1;
//	printf("rleByte=0x%02x\n", rleByte);

	/* */
	if (compressed) compressed[0] = rleByte;
	while (soFar < len)
	{
		unsigned sf = soFar;

		compChunk(uncomp, compressed, &soFar, &output, rleByte, 
				len - soFar);
		pf->inc(soFar - sf);
//		char buf[20];
//		sprintf(buf, "%04x", soFar);
//		videoScreen->drawText(32 * CHAR_W, 16 * CHAR_H, buf, ZX_BLUE, ZX_WHITE);
	}
	return output;
}

void Manic128Game::compChunk(jswByte *uncomp, jswByte *comp,
		unsigned *soFar, unsigned *output, jswByte rleByte,
		unsigned remaining)
{
	unsigned seqLen;
	unsigned repeats;
	jswByte *matchPattern, *candidatePattern;
	unsigned matches[64];
	unsigned offsets[64];
	unsigned n, max, matched, mmax;
//	bool trace = (soFar[0] >= 0x470);
	int anyMatches;

	memset(matches, 0, sizeof(matches));

//	if (trace) printf("** soFar=%d\n", *soFar);
	unsigned maxSeq = 63;
	if (*soFar < maxSeq) 
	{
		maxSeq = *soFar;
	}
	if (remaining < maxSeq)
	{
		maxSeq = remaining;
	}
	mmax = 0;
//	for (seqLen = maxSeq; seqLen > 0; seqLen--)
	for (seqLen = 1; seqLen <= maxSeq; seqLen++)
	{
		anyMatches = 0;
//		printf("Trying sequence len %d\n", seqLen);

		// matchPattern is the pattern to encode. We have to find if
		// it's occurred already.
		matchPattern = uncomp + (*soFar);

		for (candidatePattern = matchPattern - 1;
		     candidatePattern >= uncomp;
		     candidatePattern--)
		{
			// Compression uses a 1k window.
			if (matchPattern - candidatePattern > 1023) break;
//			printf("Checking candidate at %d\n", 
//					matchPattern - candidatePattern);

			// If we're in the area where sequences can't repeat
			// and there's already a longer pattern than what we
			// have, then bail out.
			if (candidatePattern + seqLen < matchPattern
			    && (mmax > seqLen || 4 > seqLen))
			{
				break;
			}	
			if (!memcmp(candidatePattern, matchPattern, seqLen))
			{
	// OK. Candidate matches. 
				jswByte *cp = candidatePattern + seqLen;
				repeats = 1;
				anyMatches++;

	// If the candidate is immediately before the match, then it may 
	// be the start of a repeated run.
				if (candidatePattern == matchPattern - seqLen)
				  while (!memcmp(candidatePattern, cp, seqLen))
				{
	// Don't let the match pattern count twice.
					if (cp != matchPattern)
						++repeats;
					cp += seqLen;
				}
//				printf("Matched with len=%d repeats=%d "
//				 	"cp=%d matchP=%d candidateP=%d "
//					"offset=%d\n",
//						seqLen, repeats,
//				cp-uncomp, matchPattern-uncomp, 
//				candidatePattern-uncomp, 
//				matchPattern - candidatePattern);
	// If this is a better pattern than the one already present,
	// switch to it.
				matches[seqLen] = seqLen * repeats; 
				offsets[seqLen] = matchPattern - candidatePattern;
	// A run can't be longer than 63 bytes. 
				if (matches[seqLen] > 63) matches[seqLen] = 63;
				if (matches[seqLen] > mmax) 
				{
					mmax = matches[seqLen];
				}				
			} // end if memcmp
		} // end for candidatePattern

		// There were no matches for this sequence. Looking for 
		// longer supersets is pointless.
		if (anyMatches == 0) break;
	}
	max = matched = 0;
	for (n = 1; n < 64; n++)
	{
		if (matches[n] > max)
		{
			matched = n;
			max = matches[n];
		}
	}
	// Runs are encoded as 3 bytes. So compressing something that's only
	// 1-3 bytes long isn't worth the effort.
	if (max <= 3) matched = 0;

	if (matched)
	{
//		if (trace) printf("Output run %d,%d\n", matches[matched],offsets[matched]);
		if (comp)
		{
			comp[*output  ]  = rleByte;
			comp[1+*output]  = matches[matched] & 0x3F;
			comp[1+*output] |= ((offsets[matched] << 6) & 0xC0); 
			comp[2+*output]  = (offsets[matched] >> 2);
		}
		*output += 3;
		*soFar += matches[matched];
	}
	else
	{
//		if (trace) printf("Output literal %02x\n", uncomp[*soFar]);
		if (uncomp[*soFar] == rleByte)	// Literal rleByte
		{
			if (comp)
			{
				comp[*output  ] = rleByte;
				comp[1+*output] = 0;
			}
			*output += 2;
		}
		else				// Literal normal byte
		{
			if (comp) comp[*output] = uncomp[*soFar];
			++*output;
		}
		++*soFar;	
	}
}





jswByte *Manic128Game::getRoom(int room)   
{
        if (room < 0 || room > 19) return NULL;
        return m_roomBuffer + 1024 * room;
}



void Manic128Game::fillSprites(int room, SpriteList *s, int every)
{
        int n, m, step;

        step = every ? 0x20 : 0x100;

        for (n = 33280; n < 33536; n += step)
        {
                s->addSprite(n);
        }
	for (n = 0xB000; n < 0xFFFF; n += 1024)
	{
		jswByte *addr = m_roomBuffer + (n - 0xB000);
		for (m = 768; m < 1024; m += step)
		{
			s->addSprite(n + m, addr + m);
		}
	}
}


char *Manic128Game::getTitleMessage()
{
//	FILE *fp = fopen("debug.bin", "wb");
//	jswByte *ram = m_mem->memoryAt(0x8000);
//	fwrite(ram, 1, 0x8000, fp);
//	fclose(fp);
	return (char *)m_mem->memoryAt(0xA3CE);
}


jswByte *Manic128Game::getRoomTitleAttrs(void)
{
	return getScreenBuffer() + 0x1A00;
}
