/************************************************************************

    JSWED 2.3.0 - Editor for Jet Set Willy and derivatives

    Copyright (C) 2000-2001,2004  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*************************************************************************/

#include "jswed.hxx"
#include "room.hxx"
#include "spritelist.hxx"
#include "manicgame.hxx"
#include "manicpatchpage.hxx"
#include "manicmsg.hxx"
#include "manicsprite.hxx"
#include "manicattrpage.hxx"
#include "manic48attrpage.hxx"
#include "manicroompage.hxx"
#include "manicguardpage.hxx"
#include "manicportalpage.hxx"
#include "manicroom.hxx"

ManicGame::ManicGame(SpectrumMemory *mem) : Game(mem)
{
	m_patchPage  = NULL;
	m_msgPage    = NULL;
	m_attrPage   = NULL;
	m_roomPage   = NULL;
	m_sprPage    = NULL;
	m_portalPage = NULL;
	m_guardPage  = NULL;

// Set initial SNA stack 
        mem->pokeSnaHeader(23, 0xFE);
        mem->pokeSnaHeader(24, 0x5B); 
	mem->poke(0x5BFE, 0x00);
	mem->poke(0x5BFF, 0x84);
	m_roomCache = NULL;
	m_roomCacheSize = 0;
	m_sprites = NULL;
}	

ManicGame::~ManicGame()
{
	deletePages();
        if (m_roomCache)
        {
                for (int n = 0; n < m_roomCacheSize; n++)
                        if (m_roomCache[n]) delete m_roomCache[n];
                delete m_roomCache;
        }
	if (m_sprites) delete [] m_sprites;
}


void ManicGame::addPages(VideoNotebook *n)
{
	if (!m_patchPage)  m_patchPage  = new ManicPatchPage(this);
	if (!m_msgPage)    m_msgPage    = new ManicMessagePage(this);
	if (!m_sprPage)    m_sprPage    = new ManicSpritePage(this);
	if (!m_attrPage)   m_attrPage   = new Manic48AttrPage(this);
	if (!m_roomPage)   m_roomPage   = new ManicRoomPage(this); 
	if (!m_guardPage)  m_guardPage  = new ManicGuardPage(this);
	if (!m_portalPage) m_portalPage = new ManicPortalPage(this); 
	n->addPage(m_patchPage);
	n->addPage(m_msgPage);
	n->addPage(m_sprPage);
	n->addPage(m_roomPage);
	n->addPage(m_guardPage);
	n->addPage(m_portalPage);
	n->addPage(m_attrPage);
	Game::addPages(n);
}


void ManicGame::deletePages(void)
{
	if (m_attrPage)   delete m_attrPage;
	if (m_sprPage)    delete m_sprPage;
	if (m_patchPage)  delete m_patchPage;
	if (m_msgPage)    delete m_msgPage;
	if (m_roomPage)   delete m_roomPage;
	if (m_guardPage)  delete m_guardPage;
	if (m_portalPage) delete m_portalPage;
	m_patchPage  = NULL;
	m_msgPage    = NULL;
	m_attrPage   = NULL;
	m_sprPage    = NULL;
	m_roomPage   = NULL;
	m_guardPage  = NULL;
	m_portalPage = NULL;
	Game::deletePages();
}


void ManicGame::fillSprites(int room, SpriteList *s, int every)
{
        int n, m, step;

        step = every ? 0x20 : 0x100;

        for (n = 33280; n < 33536; n += step)
        {
                s->addSprite(n);
        }
	for (n = 0xB000; n < 0xFFFF; n += 1024)
	{
		for (m = 768; m < 1024; m += step)
		{
			s->addSprite(n + m);
		}
	}
}


jswByte *ManicGame::getRoom(int room)
{
	if (room < 0 || room > 19) return NULL;
	return m_mem->memoryAt(0xB000 + 1024 * room);	
}

int ManicGame::prevRoom(int room)
{
	if (room <= 0) return room;
	return --room;
}

int ManicGame::nextRoom(int room)
{
	if (room >= 19) return room;
	return ++room;
}

JswRoomList *ManicGame::getRoomList(void)
{
        int n;
        jswByte *b;
        char rname[40];

        JswRoomList *s = new JswRoomList();

        for (n = 0; n < 20; n++) 
        {
                b = getRoom(n);
                sprintf(rname, "%03d %-32.32s", n, b + 512);
                s->addString(rname);
        }
        return s;
}


bool ManicGame::featuresSettableByRoom()
{
	if (m_mem->peek(xltAddress(0x8770)) == 0x73)
	{
		return true;
	}
	return false;
}

bool ManicGame::roomHasFeature(jswByte room, jswByte feature)
{
/* See if my MM patch is installed */
	if (featuresSettableByRoom())
	{
		return (getRoom(room)[627] & feature);
	}
/* Use hardcoded room IDs */
	switch (feature)
	{
		case FEAT_EUGENE:
			return (m_mem->peek(xltAddress(0x8773)) == room);
		case FEAT_SKYLABS:
			return (m_mem->peek(xltAddress(0x877b)) == room);
		case FEAT_VGUARDS:
			// If have skylabs, can't have vertical guardians
			if (m_mem->peek(xltAddress(0x877b)) == room)
				return false;
			return (m_mem->peek(xltAddress(0x8783)) <= room);
		case FEAT_KONG:
			// If have skylabs, can't have Kong
			if (m_mem->peek(xltAddress(0x877b)) == room)
				return false;
			if (m_mem->peek(xltAddress(0x878B)) == room) 
				return true;
			if (m_mem->peek(xltAddress(0x8793)) == room) 
				return true;
			return false;
		case FEAT_SOLAR:
			// If have skylabs, can't have solar power
			if (m_mem->peek(xltAddress(0x877b)) == room)
				return false;
			return (m_mem->peek(xltAddress(0x879B)) == room);
	} 
	return false;
}



typedef Room *RoomPtr;

Room *ManicGame::newRoom(int r)
{
	if (r < 0 || r > 19) return NULL;
	return new ManicRoom(this, getRoom(r), r); 
}

Room *ManicGame::getRoomClass(int room)
{
	Room *r, **rc;
	int newsize, n;

	if (room < 0 || room >= 20) return NULL;

	if (room < m_roomCacheSize)
	{
		if (m_roomCache[room]) return m_roomCache[room];
		r = newRoom(room);
		m_roomCache[room] = r;
		return r;
	}

	r = newRoom(room);
	if (m_roomCacheSize == 0) newsize = 64;
	else			  newsize = 2 * m_roomCacheSize;

	while (newsize < room) newsize *= 2;

	rc = new RoomPtr[newsize];
	for (n = 0; n < newsize; n++) rc[n] = NULL;
	if (m_roomCache)
	{
		for (n = 0; n < m_roomCacheSize; n++) rc[n] = m_roomCache[n];
		delete m_roomCache;
	}
	m_roomCache = rc;
	m_roomCacheSize = newsize;
	rc[room] = r;
	return r;
}


char *ManicGame::getTitleMessage()
{
	return ((char *)m_mem->memoryAt(0x9D00));
}


jswByte *ManicGame::getRoomTitleAttrs(void)
{
	return getMem()->memoryAt(40704);
}

jswByte *ManicGame::getSprite(int id)
{
	int id0 = id & 0xFFFF;

	if (id0 >= 0x8000 && id0 < 0x8400)
	{
		jswByte *data = getRoom((id >> 16) & 0xFF);
		if (data) return (data + (id0 & 0x3FF));	
	}
	return Game::getSprite(id);
}

bool ManicGame::loadSprites(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent)
{
        xmlNodePtr cur;
	int memid, fileid, w, h, x, xpos, hexv;
	char *t, *t1, hex[3];
	int count = 0, nspr;

	/* Make a count of valid <sprite> nodes */
	for (cur = parent->xmlChildrenNode; cur; cur = cur->next)
	{
		if (cur->ns != ns) continue;
		if (strcmp((char *)cur->name, "sprite")) continue;
		char *s = (char *)xmlGetProp(cur, (xmlChar *)"base");
		if (!s) continue;
		xmlFree(s);
		s = (char *)xmlGetProp(cur, (xmlChar *)"w");
		if (!s) continue;
		w = atoi(s);
		xmlFree(s);
		if (!w) continue;	
		s = (char *)xmlGetProp(cur, (xmlChar *)"h");
		if (!s) continue;
		h = atoi(s);
		xmlFree(s);
		if (!h) continue;	
		t = (char *)xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
		if (!t) continue;
		xmlFree(t);
		++count;
	}
//	fprintf(stderr, "There are %d valid sprite nodes\n", count);
	if (!count) return false;

	if (m_sprites) delete [] m_sprites;
	m_sprites = new ImportedSprite[count];
	m_spriteCount = count;

	nspr = 0;
	for (cur = parent->xmlChildrenNode; cur; cur = cur->next)
	{
		if (cur->ns != ns) continue;
		if (strcmp((char *)cur->name, "sprite")) continue;
		char *s = (char *)xmlGetProp(cur, (xmlChar *)"base");
		if (!s) continue;
		fileid = 0;
		sscanf(s, "%x", &fileid);
		xmlFree(s);
		s = (char *)xmlGetProp(cur, (xmlChar *)"w");
		if (!s) continue;
		w = atoi(s);
		xmlFree(s);
		if (!w) continue;	
		s = (char *)xmlGetProp(cur, (xmlChar *)"h");
		if (!s) continue;
		h = atoi(s);
		xmlFree(s);
		if (!h) continue;	
		t = (char *)xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
		if (!t) continue;

		m_sprites[nspr].spriteID = fileid;
		t1 = t;
		x = 0;
		hex[0] = 0;
		xpos = 0;
		while (t1[0])
		{
			if (!isxdigit(t1[0]))
			{
				++t1;
				continue;
			}
			if (hex[0] == 0) 
			{
				hex[0] = *t1;
				hex[1] = 0;
				++t1;
				continue;
			}
			hex[1] = *t1;
			hex[2] = 0;
			++t1;
			hexv = 0;
			sscanf(hex, "%x", &hexv);
			hex[0] = 0;
			m_sprites[nspr].bits[x++] = hexv;
			xpos += 8;
			if (xpos >= 16) 
			{
/* Skip any pixels to the right of column 16 */
				for (; xpos < w; xpos += 8) 
				{
					while (!isxdigit(t1[0])) ++t1;
					++t1;
					while (!isxdigit(t1[0])) ++t1;
					++t1;
				}
			}
			if (x >= 32) break;
		}
		++nspr;
		xmlFree(t);
	}
	return true;
}



// Import sprites from file. As with JSW, this completely replaces the 
// version in Game.
bool ManicGame::importSprites(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent,
		int *roomMap)
{
	SpriteList list(getMem(), 0, 0, 50, 50);
	xmlNodePtr root, room, guardians, guardian, spriteref, cur;
	char *sframe, *sxref, *sg1, *sg2;
     	int g1, g2, frame, xref, x, y, y1;
	ImportedSprite pageSet[8];
	int base, swap;
	int n, m, found, match, count, roomNumber;
	jswByte *roomData;

	root = xmlDocGetRootElement(doc);
	if (!root) return false;
	if (!loadSprites(doc, ns, parent)) return true;	// Nothing to load

// Handle all nodes of form rooms/room/guardians/??/spriteref 
// and rooms/room/hints/hint/spriteref 
	for (room = root->xmlChildrenNode; room; room = room->next)
	{
		if (room->ns != ns) continue;
		if (strcmp((char *)room->name, "room")) continue;

// Only process rooms which are in the map (& hence will be imported) 
		char *num = (char *)xmlGetProp(room, (xmlChar *)"number");
		if (!num) continue;
		found = 0;
		match = atoi(num);
		xmlFree(num);
		for (n = 0; n < 256; n++)
		{
			if (roomMap[n] == match) 
			{ 
		/* Does mapped room actually exist, though? */
				if (getRoom(n)) 
				{
					roomNumber = n;
					found = 1; 
				}
				break; 
			}
		}
		if (!found) 
		{
//			printf("Room %d in file not mapped in database\n",
//					match);
			continue;
		}
		getRoomClass(roomNumber)->prepareForSpriteImport();
		roomData = getRoom(roomNumber);
// Room found. Now scan its spriterefs 
		for (guardians = room->xmlChildrenNode; guardians; 
			       guardians = guardians->next)	
		{
			if (guardians->ns != ns) continue;
			if (strcmp((char *)guardians->name, "guardians")) 
				continue;
			for (guardian = guardians->xmlChildrenNode;
				guardian;
				guardian = guardian->next)
			{
				if (guardian->ns != ns) continue;
				sg1 = (char *)xmlGetProp(guardian, 
						(xmlChar *)"framemask");
				sg2 = (char *)xmlGetProp(guardian, 
						(xmlChar *)"startframe");
				if (sg1) g1 = atoi(sg1); else g1 = 0;
				if (sg2) g2 = atoi(sg2); else g2 = 0;
				if (sg1) xmlFree(sg1);
				if (sg2) xmlFree(sg2);
				swap = 0;
				if (!strcmp((char *)(guardian->name), 
							"skylab"))
				{
					g1 = 7;
					base = 0;
				}
				else if (!strcmp((char *)(guardian->name),"vg"))
				{
					g1 = 3;
					base = 0;	
				}
				else if (!strcmp((char *)(guardian->name),"hg"))
				{
					if (g1 == 3) base = 4;
					else	     { base = 0; swap = 4; }
				}
				else	continue;
				for (x = 0; x < 8; x++) 
					pageSet[x].spriteID = -1;
				for (spriteref = guardian->xmlChildrenNode;
					spriteref;
					spriteref = spriteref->next)
				{
					if (spriteref->ns != ns) continue;
					if (strcmp((char *)spriteref->name, 
						"spriteref")) continue;
					sxref = (char *)xmlGetProp(
						spriteref, (xmlChar *)"xref");
					if (!sxref) continue;
					if (!sscanf(sxref, "%x", &xref))
					{
						xmlFree(sxref);
						continue;
					}
					xmlFree(sxref);
					sframe = (char *)xmlGetProp(
						spriteref, (xmlChar *)"frame");
					if (!sframe) continue;
					frame = atoi(sframe);
					xmlFree(sframe);
					pageSet[swap ^ ((frame & g1) | base)] = findSprite(xref);
				}
			}
		}
// Scan the hints and pick up any kongs.
		for (guardians = room->xmlChildrenNode; guardians; 
			       guardians = guardians->next)	
		{
			if (guardians->ns != ns) continue;
			if (strcmp((char *)guardians->name, "hints")) 
				continue;
			for (guardian = guardians->xmlChildrenNode;
				guardian;
				guardian = guardian->next)
			{
				if (guardian->ns != ns) continue;
				if (strcmp((char *)guardian->name, "hint")) continue;
				sg1 = (char *)xmlGetProp(guardian, 
						(xmlChar *)"type");
				sg2 = (char *)xmlGetProp(guardian, 
						(xmlChar *)"name");
				if (sg1 == NULL || sg2 == NULL ||
				    (sg1 && strcmp(sg1, "feature")) ||
				    (sg2 && strcmp(sg2, "kong")))
				{
					if (sg2) xmlFree(sg2);
					if (sg1) xmlFree(sg1);
					continue;
				}
				if (sg2) xmlFree(sg2);
				if (sg1) xmlFree(sg1);
				for (spriteref = guardian->xmlChildrenNode;
					spriteref;
					spriteref = spriteref->next)
				{
					if (spriteref->ns != ns) continue;
					if (strcmp((char *)spriteref->name, 
						"spriteref")) continue;
					sxref = (char *)xmlGetProp(
						spriteref, (xmlChar *)"xref");
					if (!sxref) continue;
					if (!sscanf(sxref, "%x", &xref))
					{
						xmlFree(sxref);
						continue;
					}
					xmlFree(sxref);
					sframe = (char *)xmlGetProp(
						spriteref, (xmlChar *)"frame");
					if (!sframe) continue;
					frame = atoi(sframe);
					xmlFree(sframe);
					pageSet[frame] = findSprite(xref);
				}
			}
		}
		// OK. pageSet should contain the sprite data for this room.
		for (n = 0; n < 8; n++)
		{
			memcpy(roomData + 0x300 + 32 * n, pageSet[n].bits, 32);
		}
	}
// Now deal with portals and eugenes.
	for (room = root->xmlChildrenNode; room; room = room->next)
	{
		if (room->ns != ns) continue;
		if (strcmp((char *)room->name, "room")) continue;

/* Only process rooms which are in the map (& hence will be imported) */
		char *num = (char *)xmlGetProp(room, (xmlChar *)"number");
		if (!num) continue;
		found = 0;
		match = atoi(num);
		xmlFree(num);
		for (n = 0; n < 256; n++)
		{
			if (roomMap[n] == match) 
			{ 
		/* Does mapped room actually exist, though? */
				if (getRoom(n)) found = 1; 
				roomNumber = n;
				break; 
			}
		}
		if (!found) 
		{
//			printf("Room %d in file not mapped in database\n",
//					match);
			continue;
		}
		roomData = getRoom(roomNumber);
		for (guardians = room->xmlChildrenNode; guardians; 
			       guardians = guardians->next)	
		{
			if (guardians->ns != ns) continue;
			if (!strcmp((char *)guardians->name, "portal"))
			{
				char *s = (char *)xmlGetProp(guardians, 
						(xmlChar *)"sprite");
                		if (!s) continue;
                		xref = 0;
                		sscanf(s, "%x", &xref);
				xmlFree(s);
				pageSet[0] = findSprite(xref);
				memcpy(roomData + 0x290, pageSet[0].bits, 32);
			}
			else if (!strcmp((char *)guardians->name, "hints")) 
			    for (guardian = guardians->xmlChildrenNode;
				guardian;
				guardian = guardian->next)
			{
				if (guardian->ns != ns) continue;
				if (strcmp((char *)guardian->name, "hint")) continue;
				sg1 = (char *)xmlGetProp(guardian, 
						(xmlChar *)"type");
				sg2 = (char *)xmlGetProp(guardian, 
						(xmlChar *)"name");
				if (sg1 == NULL || sg2 == NULL ||
				    (sg1 && strcmp(sg1, "feature")) ||
				    (sg2 && strcmp(sg2, "eugene")))
				{
					if (sg2) xmlFree(sg2);
					if (sg1) xmlFree(sg1);
					continue;
				}
				if (sg2) xmlFree(sg2);
				if (sg1) xmlFree(sg1);
				for (spriteref = guardian->xmlChildrenNode;
					spriteref;
					spriteref = spriteref->next)
				{
					if (spriteref->ns != ns) continue;
					if (strcmp((char *)spriteref->name, 
						"spriteref")) continue;
					sxref = (char *)xmlGetProp(
						spriteref, (xmlChar *)"xref");
					if (!sxref) continue;
					if (!sscanf(sxref, "%x", &xref))
					{
						xmlFree(sxref);
						continue;
					}
					xmlFree(sxref);
					sframe = (char *)xmlGetProp(
						spriteref, (xmlChar *)"frame");
					if (!sframe) continue;
					frame = atoi(sframe);
					xmlFree(sframe);
					pageSet[frame] = findSprite(xref);
				}
				memcpy(roomData + 0x2e0, pageSet[0].bits, 32);
			}
		}
	} 
	return true;
}

ImportedSprite ManicGame::findSprite(int xref)
{
	int n;
	ImportedSprite nil;

	for (n = 0; n < m_spriteCount; n++)
	{
		if (m_sprites[n].spriteID == xref) return m_sprites[n];
	}
	nil.spriteID = -1;
	return nil;
}

