/************************************************************************

    JSWED 2.3.0 - Editor for Jet Set Willy and derivatives

    Copyright (C) 2000-2001,2004-6  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*************************************************************************/

#include "jswed.hxx"
#include "room.hxx"
#include "manicroom.hxx"
#include "manicgame.hxx"

static jswByte query[] = { 0x00, 0x00, 0x3C, 0x42, 0x04, 0x08, 0x00, 0x08, 0x00 };


ManicRoom::ManicRoom(ManicGame *g, jswByte *data, int roomNo) : 
	Room(8, roomNo), m_data(data), m_game(g)
{
	prepareForSpriteImport();
}

int ManicRoom::getTitle(char *buf)
{
	memcpy(buf, m_data + 0x200, 32);
	buf[32] = 0;
        for (int n = 0; n < 32; n++)
        {
                if (buf[n] == ZXCH_COPY ) buf[n] = ZXK_COPY;
                if (buf[n] == ZXCH_POUND) buf[n] = ZXK_POUND;
        }
	return 32;
}

void ManicRoom::setTitle(const char *buf)
{
	char fmt[33];

	sprintf(fmt, "%-32.32s", buf);
	memcpy(m_data + 0x200, fmt, 32);
}


jswByte *ManicRoom::getCellPattern(int type)
{
	if (type == 8) /* item */
	{
		m_itemPattern[0] = ZX_WHITE;
		memcpy(m_itemPattern + 1, m_data + 692, 8);
		return m_itemPattern;
	}
	if (type < 0 | type > 7) 
	{
		query[0] = type & 0xFF;
		return query;
	}
	return m_data + 544 + 9 * type;
}


int ManicRoom::getDrawnCellAt(int x, int y)
{
	return m_data[y * 32 + x];
}

int ManicRoom::getCellAt(int x, int y)
{
	jswByte attrib = m_data[y * 32 + x];
	int n;

	for (n = 0; n < 8; n++)
	{
		if (m_data[544 + 9 * n] == attrib) return n;
	}
	query[0] = attrib;
	return 0x100 | attrib;
}

int ManicRoom::getConveyorDir()
{
	return m_data[623] & 1;
}


int ManicRoom::getConveyorY()
{
	return m_data[625] & 7;
}


void ManicRoom::drawOneOffs(VideoBitmap *bm, bool flashonly, bool draw)
{
        int len;
        int w=m_data[624]+(256*m_data[625]);
        int x =  (w & 0x1F);
        int y = ((w & 0xE0) >> 1) | ((w & 0x800) >> 4); //   ((w >> 1) & 0xF0);

        if (m_data[626] > 32) // Damaged room 
		m_data[626]=0;

	len = m_data[626];

	y = y >> 4;
        for (int n = 0; n < len; n++)
        {
                if (y >= 0 && y < 16 && x >= 0 && x < 32)
                {
			drawCell(bm, x, y, flashonly, draw);
		}
		x++;
	}
	Room::drawOneOffs(bm, flashonly, draw);
}


void ManicRoom::setCell(int x, int y, int type)
{
        if (x < 0 || x > 31 || y < 0 || y > 15) return;
	if (type < 0x100)
	{
		type &= 7;
		type = m_data[544 + 9 * type];
	}
	m_data[32 * y + x] = type & 0xFF;
}


int ManicRoom::getBackground()
{
	return (m_data[544] >> 3) & 0x0F;
}


int ManicRoom::getItemCoords(int *xs, int *ys, int *counts)
{
	int count = 0;
	int n, m, found;
	jswByte *ob;

	for (n = 0; n < 5; n++)
	{
		ob = m_data + 629 + 5 * n;
		if (ob[0] == 0xFF || ob[0] == 0) continue;

		xs[count] = ob[1] & 0x1F;
		ys[count] = ((ob[1] >> 5) & 7) | (ob[3] & 8);
		found = 0;
		for (m = 0; m < n; m++)
		{
			if (ys[m] == ys[count] && xs[m] == xs[count])
			{
				++counts[m];
				found = 1;
				break;
			}
		}
		if (!found) 
		{
			counts[count] = 1;
			++count;
		}
	}
	return count;
}


bool ManicRoom::getPortal(Portal &p)
{
	p.attribute = m_data[655];
	p.sprite    = m_data + 656;
        p.x         =   m_data[690] & 0x1F;
        p.y         = ((m_data[690] & 0xE0) >> 5) | (m_data[691] & 8) ;
	p.spriteID  = (m_roomNumber << 16) | 0x8000 | 656;

	p.targetRoom = m_roomNumber + 1;
	if (m_roomNumber == 19)  p.targetRoom = 0; 

	jswByte *data = m_game->getRoom(p.targetRoom);
	if (data)
	{
		p.targetX = m_data[620] & 0x1F;
		p.targetY = m_data[616] >> 4;
	}

	return true;
}

void ManicRoom::setPortal(Portal &p)
{
	m_data[655] = p.attribute;
	m_data[688] = (p.x) + ((p.y & 7) << 5);	
	m_data[689] = (m_data[689] & 0xFE) | ((p.y >> 3) & 1);
	m_data[690] = (p.x) + ((p.y & 7) << 5);	
	m_data[691] = (m_data[691] & 0xF7) | (p.y & 8);
	// (the sprite is done by ManicGame::importSprites() )
}

int ManicRoom::getCellBehaviour(int idx)
{
	switch(idx)
	{
		case 0: return CB_AIR; 
		case 1: 
		case 7: return CB_WATER;
		case 2: return CB_CRUMBLY;
		case 3: return CB_EARTH;
		case 4: if (getConveyorDir()) return CB_RCONV;
			return CB_LCONV;
		case 5: 
		case 6: return CB_FIRE;
	}
	return CB_WATER;
}


int ManicRoom::getCellTypeCount(bool includeSingles)
{
	return 8;
}

int ManicRoom::getRoomLength() const
{
	return 1024;
}


jswByte ManicRoom::getSolarAttr()
{
	if (m_game->roomHasFeature(m_roomNumber, FEAT_SOLAR))
	{
		int addr = m_game->xltAddress(0x8DA6);
		return m_game->getMem()->peek(addr);
		//return ZX_BRWHITE * 8 + ZX_YELLOW;	
	}
	else	return 0;
}

void ManicRoom::getSolarPos(int *x, int *y)
{
	// Peek the game engine. In Bug-Byte it's at 8D74h
	int addr = m_game->xltAddress(0x8D74);

	*x = m_game->getMem()->peek(addr) & 0x1F;
	*y = (m_game->getMem()->peek(addr) >> 5) | 
	     ((m_game->getMem()->peek(addr+1) & 1) << 3) ;
//	*y = 0;
//	*x = 23;
}

void ManicRoom::setSolarPos(int x, int y)
{
	int addr = m_game->xltAddress(0x8D74);
	x = (x & 31) | ((y & 7) << 5);
	y = ((y & 8) >> 3) | 0x5C;

	m_game->getMem()->poke(addr, x);	
	m_game->getMem()->poke(addr + 1, y);	
}


void ManicRoom::getCells(Cell *c)
{
	int n;
	jswByte *b = m_data + 544;
	for (n = 0; n < 8; n++)
	{
		memcpy(c[n].pattern, b, 9);
		b += 9;
		c[n].behaviour = getCellBehaviour(n);
	}
}


void ManicRoom::setCells(Cell *c)
{
	int n;
	jswByte *b = m_data + 544;
	for (n = 0; n < 8; n++)
	{
		memcpy(b, c[n].pattern, 9);
		b += 9;
	}
}




void ManicRoom::clear()
{
	int n;
	memset(m_data, 0, 1024);
	memset(m_data + 512, ' ', 32);
	for (n = 629; n <= 654; n += 5) m_data[n] = 0xFF;
	for (n = 702; n <= 730; n += 7) m_data[n] = 0xFF;
	for (n = 733; n <= 761; n += 7) m_data[n] = 0xFF;
}


extern void flip8(jswByte *data);

void ManicRoom::mirrorH()
{
	int x, y, n;
	jswByte bt;
	bool skylabs = m_game->roomHasFeature(m_roomNumber, FEAT_SKYLABS);
	bool vguards = m_game->roomHasFeature(m_roomNumber, FEAT_VGUARDS);

	for (y = 0; y < 16; y++)
		for (x = 0; x < 16; x++)
	{
		bt = m_data[y * 32 + (31-x)];
		m_data[y*32 + (31-x)] = m_data[y*32 + x];
		m_data[y*32 + x] = bt;
	}
	/* Willy's start position */
	x = 30 - (m_data[620] & 0x1F);
	m_data[620] = (m_data[620] & 0xE0) | x;
	m_data[618] ^= 1;

	/* Conveyor animation */
	m_data[623] ^= 1;
	x = (32 - (m_data[624] & 0x1F) - m_data[626]) & 0x1F;
	m_data[624] = (m_data[624] & 0xE0) | x;
	
	/* Mirror items */
	for (n = 630; n <= 650; n += 5)
	{
		x = 31 - (m_data[n] & 0x1F);
		m_data[n] = (m_data[n] & 0xE0) | x;
	}

	/* Mirror portal position */
	x = 30 - (m_data[688] & 0x1F);
	m_data[688] = (m_data[688] & 0xE0) | x;
	x = 30 - (m_data[690] & 0x1F);
	m_data[690] = (m_data[690] & 0xE0) | x;
	/* Mirror horizontal guardians */
	for (n = 702; n <= 723; n += 7)
	{
		x = 30 - (m_data[n+1] & 0x1F);
		m_data[n+1] = (m_data[n+1] & 0xE0) | x;
		m_data[n+4] = 7 - m_data[n+4];
		y = 30 - (m_data[n+5] & 0x1F);
		x = 30 - (m_data[n+6] & 0x1F);
		m_data[n+5] = (m_data[n+5] & 0xE0) | x;
		m_data[n+6] = (m_data[n+6] & 0xE0) | y;
	}
	/* Mirror vertical guardians */
	if (skylabs || vguards) 
	{
		for (n = 733; n <= 754; n += 7)
		{
			x = 31 - (m_data[n+3] & 0x1F);
			m_data[n+3] = (m_data[n+3] & 0xE0) | x;
		}
	}
/* Flip horizontal bitmaps */
	for (n = 545; n <= 608; n += 9)
	{
		flip8(m_data + n);
	}
	flip8(m_data + 692);
	createCellPatterns();
}



void ManicRoom::mirrorV()
{
	int x, y, n;

	for (y = 0; y < 8; y++) 
		for (x = 0; x < 32; x++)
	{
		jswByte *b = m_data +     y  * 32 + x;
		jswByte *c = m_data + (15-y) * 32 + x;
		jswByte a  = *b;
		*b = *c;
		*c = a;	
	}
	/* Mirror items */
	for (n = 630; n <= 650; n += 5)
	{
		y = ((m_data[n] & 0xE0) >> 5) | (m_data[n+2] & 8);
		y = 15 - y;
		m_data[n] = (m_data[n] & 0x1F) | ((y & 7) << 5) ;
		m_data[n+1] &= 0xFE;
		m_data[n+1] |= ((y & 8) >> 3);
		m_data[n+2] &= ~8;
		m_data[n+2] |= (y & 8);	
	}
	for (n = 545; n <= 608; n += 9)
	{
		for (x = 0; x < 4; x++)
		{
			y = m_data[7 + n - x];
			m_data[7 + n - x] = m_data[n + x];
			m_data[n + x] = y;	
		}
	}
	for (x = 0; x < 4; x++)
	{
		y = m_data[699 + n - x];
		m_data[699 + n - x] = m_data[692 + n];
		m_data[692 + n] = y;	
	}
	createCellPatterns();
}


void ManicRoom::getRoomMemMap(char *buf)
{
	bool vguards = m_game->roomHasFeature(m_roomNumber, FEAT_VGUARDS);

	buf[1024] = 0;
	memset(buf    , 'b', 512);
	memset(buf+512, 'x', 256);
	memset(buf+512, 't', 32);
	memset(buf+544, 'c', 72);
	memset(buf+656, 's', 32);
	memset(buf+702, 'g', 30);
	if (vguards)
	{
		memset(buf + 733, 'g', 28);
	}
	else
	{
		memset(buf + 736, 's', 32);
	}
	memset(buf + 768, 's', 256);
}

jswByte *ManicRoom::getExits()
{
	return NULL;
}

void ManicRoom::setBorder(int bdr)
{
	m_data[627] &= ~7;
	m_data[627] |= (bdr & 7);
}

jswByte ManicRoom::getBorder()
{
	return m_data[627] & 7;
}


void ManicRoom::clearItems()
{
	memset(m_data + 629, 0xFF, 25);
	for (int n = 1; n < 5; n++) m_data[629 + 5*n] = 0;
}

void ManicRoom::addItem(int x, int y)
{
        int n, m;

        for (n = 0; n < 5; n++)
        {
		if (m_data[n*5+629] == 0) break;
		if (m_data[n*5+629] == 0xFF) break;
	}
	if (n >= 5) return;
	// If this item was the last, move the 'last item' flag up one. 
	if (m_data[n*5+629] == 0xFF)
	{
		m_data[n*5+634] = 0xFF;
	}
	m_data[n*5+629] = (m_data[544] & 0xF8) | (3 + (n%4));
	m_data[n*5+630] = (x & 0x1F) | ((y << 5) & 0xE0);
	m_data[n*5+631] = ((y & 8) >> 3) | 0x5C;
	m_data[n*5+632] =  (y & 8)       | 0x60;
	m_data[n*5+633] = 0xFF;
}

void ManicRoom::setItemBitmap(jswByte *b)
{
	memcpy(m_data + 692, b, 8);
}

void ManicRoom::addRefs()
{
	Portal p;
	int n;

	// Portal sprites
	if (getPortal(p))
	{
		m_game->addRef(SPRITE(p.spriteID));
	}
	// Special sprites
	if (m_game->roomHasFeature(m_roomNumber, FEAT_EUGENE) ||
	    m_roomNumber < 3)
	{
		m_game->addRef(SPRITE((m_roomNumber << 16) | 0x8000 | 736));
	}
	// Guardian sprites
	if (m_game->roomHasFeature(m_roomNumber, FEAT_SKYLABS))
	{
		for (n = 0; n < 8; n++)
		{
			m_game->addRef(SPRITE(m_roomNumber << 16) | 0x8300 | (n << 5));
		}
	}
	else if (m_game->roomHasFeature(m_roomNumber, FEAT_KONG))
	{
		if (m_data[702] != 0xFF) for (n = 0; n < 4; n++)
		{
			m_game->addRef(SPRITE(m_roomNumber << 16) | 0x8300 | (n << 5));
		}
		for (n = 4; n < 8; n++)
		{
			m_game->addRef(SPRITE(m_roomNumber << 16) | 0x8300 | (n << 5));
		}
	}
	else if (m_game->roomHasFeature(m_roomNumber, FEAT_VGUARDS))
	{
		if (m_data[702] != 0xFF) for (n = 4; n < 8; n++)
		{
			m_game->addRef(SPRITE(m_roomNumber << 16) | 0x8300 | (n << 5));
		}
		if (m_data[733] != 0xFF) for (n = 0; n < 4; n++)
		{
			m_game->addRef(SPRITE(m_roomNumber << 16) | 0x8300 | (n << 5));
		}
	}
	else
	{
		if (m_data[702] != 0xFF) for (n = 0; n < 8; n++)
		{
			m_game->addRef(SPRITE(m_roomNumber << 16) | 0x8300 | (n << 5));
		}
	}
}

int ManicRoom::encodeCellNumber(int c)
{
	if (c >= 0 && c < 8) return m_data[544 + 9 * c];
	return c;
}

void ManicRoom::importCell(int x, int y, int attr, int type, int behaviour)
{
	m_data[y*32 + x] = attr;
}


bool ManicRoom::exportExtraCells(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent)
{
	jswByte *cp[8];
	jswByte pattern[9];
	int n, attr, used;

	/* Get the 8 cell patterns for this room */
	for (n = 0; n < 8; n++) cp[n] = getCellPattern(n);
	/* See which attributes are (a) nonstandard and (b) 
	 * used in this screen */
	for (attr = 0; attr < 256; attr++)
	{
		used = 0;
		/* Compare against the 8 we've already done */
		for (n = 0; n < 8; n++) if (cp[n][0] == attr) 
		{
			used = -1;
			break;
		}
		/* If used < 0, it's one of the ones that's been done */
		if (used < 0)  continue;
		/* Is it actually used onscreen? */
		for (n = 0; n < 512; n++)
		{
			if (m_data[n] == attr) used++;
		}
		if (used == 0) continue;
		/* Yes. So export it. */
		memcpy(pattern, cp[1], 9);	// Start off with water
		pattern[0] = attr;
/* Emulate genuine MM behaviour - including the Block Graphics bug - and 
 * find the pattern it would use to draw such a cell */
		for (n = 544; n < (1024-9); n++)
		{
			if (m_data[n] == attr) 
			{
				memcpy(pattern, m_data + n, 9);
				break;
			}
		}
		if (!exportCell(doc, ns, parent, attr, CB_WATER, pattern)) 
			return false;
	}
	return true;
}


bool ManicRoom::exportHints(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent)
{
	char buf[30];
	xmlNodePtr hint;
	int a;

	hint = xmlNewNode(ns, (xmlChar *)"hint");
	xmlSetProp(hint, (xmlChar *)"type", (xmlChar *)"nokamikaze");
	xmlSetProp(hint, (xmlChar *)"enabled", (xmlChar *)"1");
	xmlAddChild(parent, hint);
	// Conveyor animation is always present
	hint = xmlNewNode(ns, (xmlChar *)"hint");
	xmlSetProp(hint, (xmlChar *)"type", (xmlChar *)"conveyor");
	sprintf(buf, "%d", m_data[623]);
	xmlSetProp(hint, (xmlChar *)"direction", (xmlChar *)buf);
	sprintf(buf, "%d", m_data[624] & 31);
	xmlSetProp(hint, (xmlChar *)"x", (xmlChar *)buf);
	sprintf(buf, "%d", (m_data[624] >> 5) + (m_data[625] & 8));
	xmlSetProp(hint, (xmlChar *)"y", (xmlChar *)buf);
	sprintf(buf, "%d", m_data[625] & 7);
	xmlSetProp(hint, (xmlChar *)"animy", (xmlChar *)buf);
	sprintf(buf, "%d", m_data[626]);
	xmlSetProp(hint, (xmlChar *)"length", (xmlChar *)buf);
	xmlAddChild(parent, hint);
	// Solar power
	if ((a = getSolarAttr()))
	{
		int x,y;
		getSolarPos(&x, &y);

		hint = xmlNewNode(ns, (xmlChar *)"hint");
		xmlSetProp(hint, (xmlChar *)"type", (xmlChar *)"solar");
		sprintf(buf, "%d", x);
		xmlSetProp(hint, (xmlChar *)"x", (xmlChar *)buf);
		sprintf(buf, "%d", y);
		xmlSetProp(hint, (xmlChar *)"y", (xmlChar *)buf);
		sprintf(buf, "%d", a & 7) ;
		xmlSetProp(hint, (xmlChar *)"ink", (xmlChar *)buf);
		sprintf(buf, "%d", (a >> 3) & 7 );
		xmlSetProp(hint, (xmlChar *)"paper", (xmlChar *)buf);
		if (a & 0x40)
			xmlSetProp(hint, (xmlChar *)"bright", (xmlChar *)"1");
		if (a & 0x80)
			xmlSetProp(hint, (xmlChar *)"flash", (xmlChar *)"1");
		xmlAddChild(parent, hint);
	}
	// Air
	int air = (m_data[700] - 37) * 6;
	if      (m_data[701] &  4) air += 5;
	else if (m_data[701] &  8) air += 4;
	else if (m_data[701] & 16) air += 3;
	else if (m_data[701] & 32) air += 2;
	else if (m_data[701] & 64) air += 1;
	hint = xmlNewNode(ns, (xmlChar *)"hint");
	xmlSetProp(hint, (xmlChar *)"type", (xmlChar *)"air");
	sprintf(buf, "%d", air );
	xmlSetProp(hint, (xmlChar *)"amount", (xmlChar *)buf);
	xmlAddChild(parent, hint);
	// Special features
	if (m_game->roomHasFeature(m_roomNumber, FEAT_EUGENE))
	{
		hint = xmlNewNode(ns, (xmlChar *)"hint");
		xmlSetProp(hint, (xmlChar *)"type", (xmlChar *)"feature");
		xmlSetProp(hint, (xmlChar *)"name", (xmlChar *)"eugene");
		xmlAddChild(parent, hint);
		xmlNodePtr node = xmlNewNode(ns, (xmlChar *)"spriteref");
		xmlSetProp(node, (xmlChar *)"frame", (xmlChar *)"0");
		sprintf(buf, "%04x", 0x82E0 + (m_roomNumber << 16));
		xmlSetProp(node, (xmlChar *)"xref", (xmlChar *)buf);
		xmlAddChild(hint, node);
	}
	// This may not be necessary, as an importer could check the 
	// guardian table to see if there are skylabs.
	if (m_game->roomHasFeature(m_roomNumber, FEAT_SKYLABS))
	{
		hint = xmlNewNode(ns, (xmlChar *)"hint");
		xmlSetProp(hint, (xmlChar *)"type", (xmlChar *)"feature");
		xmlSetProp(hint, (xmlChar *)"name", (xmlChar *)"skylabs");
		xmlAddChild(parent, hint);
	}
	if (m_game->roomHasFeature(m_roomNumber, FEAT_VGUARDS))
	{
		hint = xmlNewNode(ns, (xmlChar *)"hint");
		xmlSetProp(hint, (xmlChar *)"type", (xmlChar *)"feature");
		xmlSetProp(hint, (xmlChar *)"name", (xmlChar *)"vguards");
		xmlAddChild(parent, hint);
	}
	/* Solar power we've already done */
	if (m_game->roomHasFeature(m_roomNumber, FEAT_KONG))
	{
		hint = xmlNewNode(ns, (xmlChar *)"hint");
		xmlSetProp(hint, (xmlChar *)"type", (xmlChar *)"feature");
		xmlSetProp(hint, (xmlChar *)"name", (xmlChar *)"kong");
		// Export the Kong sprite(s)
		for (int n = 0; n < 4; n++)
		{
			xmlNodePtr node = xmlNewNode(ns, (xmlChar *)"spriteref");
			sprintf(buf, "%d", n);
			xmlSetProp(node, (xmlChar *)"frame", (xmlChar *)buf);
			sprintf(buf, "%04x", n * 32 + 0x8300 + 
					(m_roomNumber << 16));
			xmlSetProp(node, (xmlChar *)"xref", (xmlChar *)buf);
			xmlAddChild(hint, node);
		}
		xmlAddChild(parent, hint);
	}
	/* Willy's starting position */
	hint = xmlNewNode(ns, (xmlChar *)"hint");
	xmlSetProp(hint, (xmlChar *)"type", (xmlChar *)"startpos");
	sprintf(buf, "%d", m_data[620] & 31);
	xmlSetProp(hint, (xmlChar *)"x", (xmlChar *)buf);
	sprintf(buf, "%d", (m_data[616] >> 4) & 15);
	xmlSetProp(hint, (xmlChar *)"y", (xmlChar *)buf);
	sprintf(buf, "%d", ((m_data[616] >> 1) & 7));
	xmlSetProp(hint, (xmlChar *)"ypix", (xmlChar *)buf);
	sprintf(buf, "%d", m_data[618]);
	xmlSetProp(hint, (xmlChar *)"direction", (xmlChar *)buf);
	sprintf(buf, "%d", m_data[617]);
	xmlSetProp(hint, (xmlChar *)"frame", (xmlChar *)buf);
	sprintf(buf, "%d", m_data[619]);
	xmlSetProp(hint, (xmlChar *)"jump", (xmlChar *)buf);
	xmlAddChild(parent, hint);
	return true;
}

bool ManicRoom::exportGuardians(xmlDocPtr doc, xmlNsPtr ns, 
		xmlNodePtr grandparent)
{
	xmlNodePtr parent = xmlNewNode(ns, (xmlChar *)"guardians");
	xmlSetProp(parent, (xmlChar *)"type", (xmlChar *)"jsw");
	bool skylabs = m_game->roomHasFeature(m_roomNumber, FEAT_SKYLABS);
	bool vguards = m_game->roomHasFeature(m_roomNumber, FEAT_VGUARDS);
	bool success;

	if (skylabs)      success = exportSGuardians(doc, ns, parent);
	else if (vguards) success = exportVGuardians(doc, ns, parent);
	else success = true;
	if (!success) return false;

	success = exportHGuardians(doc, ns, parent);
	xmlAddChild(grandparent, parent);
	return success;
}

bool ManicRoom::exportVGuardians(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent)
{
	int source = 0x2DD;
	int n, x, y, direction, startframe, limit1, limit2;
	char buf[20];
	xmlNodePtr node;

	for (n = 0; n < 4; n++)
	{
		if (m_data[source] == 0)
		{
			source += 7;
			continue;
		}
		if (m_data[source] == 0xFF) break;
		node = xmlNewNode(ns, (xmlChar *)"vg");

// Here's where we do the MM-to-JSW vertical guardian 
// conversion. Based on Andrew Broad's Java conversion function, described in
// <http://games.groups.yahoo.com/group/manicminerandjetsetwilly/message/4050>
// and used with permission.

		signed char velocity = m_data[source + 4];
		y = m_data[source + 2];
		limit1 = m_data[source+5];
		limit2 = m_data[source+6];
		if (velocity < 0) 
		{
			limit1 += (y - limit1) % velocity;
			limit2 -= (y - limit1) % velocity;
		}
		else
		{
			limit2 -= (limit2 - y)      % velocity;
			limit1 += (limit2 - limit1) % velocity;
		}
		if (limit2 == m_data[source+6]) limit2 -= abs(velocity);
		while (y > limit2)
		{
			if (velocity >= 0) velocity = -velocity;
			y += velocity;
		}
		if (y >= limit2 && velocity >= 0) velocity = -velocity;
		while (y < limit1)
		{
			if (velocity < 0) velocity = -velocity;
			y += velocity;
		}
		if (y <= limit1 && velocity < 0)
		{
			velocity = -velocity;
		}
		x = m_data[source + 3] & 0x1F;
		// OK. We now have y and the limits all nicely adjusted.	
		startframe = 0;	
		if (velocity < 0)
		     xmlSetProp(node, (xmlChar *)"direction", (xmlChar *)"0");
		else xmlSetProp(node, (xmlChar *)"direction", (xmlChar *)"1");
		sprintf(buf, "%d", m_data[source] & 7);
		xmlSetProp(node, (xmlChar *)"ink", (xmlChar *)buf);
		sprintf(buf, "%d", (m_data[source] >> 3) & 7);
		xmlSetProp(node, (xmlChar *)"paper", (xmlChar *)buf);
		if (m_data[source] & 0x40)
			xmlSetProp(node, (xmlChar *)"bright", (xmlChar *)"1");
		if (m_data[source] & 0x80)
			xmlSetProp(node, (xmlChar *)"flash", (xmlChar *)"1");
		xmlSetProp(node, (xmlChar *)"framemask", (xmlChar *)"3");
		sprintf(buf, "%d", x);
		xmlSetProp(node, (xmlChar *)"x", (xmlChar *)buf);
		sprintf(buf, "%d", startframe);
		xmlSetProp(node, (xmlChar *)"startframe", (xmlChar *)buf);
		sprintf(buf, "%d", y);
		xmlSetProp(node, (xmlChar *)"y", (xmlChar *)buf);
		sprintf(buf, "%d", 0xFF & (velocity * 2));
		xmlSetProp(node, (xmlChar *)"speed", (xmlChar *)buf);
		sprintf(buf, "%d", 2 * limit1);
		xmlSetProp(node, (xmlChar *)"limit1", (xmlChar *)buf);
		sprintf(buf, "%d", 2 * limit2);
		xmlSetProp(node, (xmlChar *)"limit2", (xmlChar *)buf);

		for (int frame = 0; frame < 8; frame++)
		{
			int xref = (frame << 5) & 0x60;
			xref += 0x8300;
			xmlNodePtr spr = xmlNewNode(ns, (xmlChar *)"spriteref");
			sprintf(buf, "%d", frame);
			xmlSetProp(spr, (xmlChar *)"frame", (xmlChar *)buf);
			sprintf(buf, "%06x", (m_roomNumber << 16) | xref);
			xmlSetProp(spr, (xmlChar *)"xref", (xmlChar *)buf);
			xmlAddChild(node, spr);
		}	
		xmlAddChild(parent, node);
		source += 7;
	}
	return true;
}



bool ManicRoom::exportSGuardians(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent)
{
	int source = 0x2DD;
	int n, x, y, direction, startframe, limit1, limit2;
	char buf[20];
	xmlNodePtr node;

	for (n = 0; n < 4; n++)
	{
		if (m_data[source] == 0)
		{
			source += 7;
			continue;
		}
		if (m_data[source] == 0xFF) break;
		node = xmlNewNode(ns, (xmlChar *)"skylab");
		x = m_data[source + 3] & 0x1F;
		startframe = 0;
		y = m_data[source + 2];
		signed char velocity = (signed char)(m_data[source + 4]);
		limit1 = m_data[source+5];
		limit2 = m_data[source+6];
		sprintf(buf, "%d", m_data[source] & 7);
		xmlSetProp(node, (xmlChar *)"ink", (xmlChar *)buf);
		sprintf(buf, "%d", (m_data[source] >> 3) & 7);
		xmlSetProp(node, (xmlChar *)"paper", (xmlChar *)buf);
		if (m_data[source] & 0x40)
			xmlSetProp(node, (xmlChar *)"bright", (xmlChar *)"1");
		if (m_data[source] & 0x80)
			xmlSetProp(node, (xmlChar *)"flash", (xmlChar *)"1");
		xmlSetProp(node, (xmlChar *)"framemask", (xmlChar *)"3");
		sprintf(buf, "%d", x);
		xmlSetProp(node, (xmlChar *)"x", (xmlChar *)buf);
		sprintf(buf, "%d", startframe);
		xmlSetProp(node, (xmlChar *)"startframe", (xmlChar *)buf);
		sprintf(buf, "%d", y);
		xmlSetProp(node, (xmlChar *)"y", (xmlChar *)buf);
		sprintf(buf, "%d", 0xFF & (velocity * 2));
		xmlSetProp(node, (xmlChar *)"speed", (xmlChar *)buf);
		sprintf(buf, "%d", 2 * limit1);
		xmlSetProp(node, (xmlChar *)"limit1", (xmlChar *)buf);
		sprintf(buf, "%d", 2 * limit2);
		xmlSetProp(node, (xmlChar *)"limit2", (xmlChar *)buf);

		for (int frame = 0; frame < 8; frame++)
		{
			int xref = (frame << 5) & 0xE0;
			xref += 0x8300;
			xmlNodePtr spr = xmlNewNode(ns, (xmlChar *)"spriteref");
			sprintf(buf, "%d", frame);
			xmlSetProp(spr, (xmlChar *)"frame", (xmlChar *)buf);
			sprintf(buf, "%06x", (m_roomNumber << 16) | xref);
			xmlSetProp(spr, (xmlChar *)"xref", (xmlChar *)buf);
			xmlAddChild(node, spr);
		}	
		xmlAddChild(parent, node);
		source += 7;
	}
	return true;
}

bool ManicRoom::exportHGuardians(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent)
{
	int source = 0x2BE;
	int n, x, direction, startframe, limit1, limit2;
	char buf[20];
	xmlNodePtr node;
	bool vguards = m_game->roomHasFeature(m_roomNumber, FEAT_VGUARDS);
	bool kong    = m_game->roomHasFeature(m_roomNumber, FEAT_KONG);

	for (n = 0; n < 5; n++)
	{
		if (m_data[source] == 0)
		{
			source += 7;
			continue;
		}
		if (m_data[source] == 0xFF) break;
		node = xmlNewNode(ns, (xmlChar *)"hg");

		x = m_data[source + 1] & 0x1F;
		limit1 = m_data[source+5] & 0x1F;
		limit2 = m_data[source+6] & 0x1F;
		startframe = 0;
		if (m_data[source + 4] >= 4) 
		{
			direction = 1;
			++x;
			if (x >= limit2)
			{
				direction = 0;
				x = limit2;
			}
		}
		else    direction = 0;
		// Auto-reverse if at RH end

		if (direction == 0)
		     xmlSetProp(node, (xmlChar *)"direction", (xmlChar *)"0");
		else xmlSetProp(node, (xmlChar *)"direction", (xmlChar *)"1");
		sprintf(buf, "%d", m_data[source] & 7);
		xmlSetProp(node, (xmlChar *)"ink", (xmlChar *)buf);
		sprintf(buf, "%d", (m_data[source] >> 3) & 7);
		xmlSetProp(node, (xmlChar *)"paper", (xmlChar *)buf);
		if (m_data[source] & 0x40)
			xmlSetProp(node, (xmlChar *)"bright", (xmlChar *)"1");
		if (kong || vguards)
		     xmlSetProp(node, (xmlChar *)"framemask", (xmlChar *)"3");
		else xmlSetProp(node, (xmlChar *)"framemask", (xmlChar *)"7");
		sprintf(buf, "%d", x);
		xmlSetProp(node, (xmlChar *)"x", (xmlChar *)buf);
		sprintf(buf, "%d", startframe);
		xmlSetProp(node, (xmlChar *)"startframe", (xmlChar *)buf);
		sprintf(buf, "%d", ((m_data[source+1] & 0xE0) >> 2) |
			           ((m_data[source+3] << 3) & 0x40));
		xmlSetProp(node, (xmlChar *)"y", (xmlChar *)buf);
		xmlSetProp(node, (xmlChar *)"speed", (xmlChar *)"0");
		if (m_data[source] & 0x80) limit2 |= 32;	// Slow
		sprintf(buf, "%d", limit1);
		xmlSetProp(node, (xmlChar *)"limit1", (xmlChar *)buf);
		sprintf(buf, "%d", limit2);
		xmlSetProp(node, (xmlChar *)"limit2", (xmlChar *)buf);

		for (int frame = 0; frame < 8; frame++)
		{
			int xref = (frame << 5) & 0xE0;
			if (kong || vguards) xref &= 0x60;
			xref ^= 0x80;
			xref += 0x8300;
			xmlNodePtr spr = xmlNewNode(ns, (xmlChar *)"spriteref");
			sprintf(buf, "%d", frame);
			xmlSetProp(spr, (xmlChar *)"frame", (xmlChar *)buf);
			sprintf(buf, "%06x", (m_roomNumber << 16) | xref);
			xmlSetProp(spr, (xmlChar *)"xref", (xmlChar *)buf);
			xmlAddChild(node, spr);
		}	
		xmlAddChild(parent, node);
		source += 7;
	}
	return true;
}


void ManicRoom::clearHints(void)
{
	int x, y;
	for (y = 0; y < 16; y++) for (x = 0; x < 32; x++)
	{
		if (m_data[y*32+x] == m_data[544])
		{
			m_data[623] = 2;
			m_data[624] = (x & 31) | ((y & 7) << 5);
			m_data[625] = (y & 8) | 0x70;
			m_data[626] = 1;
			x = 32;
			y = 16;
			break;
		}
	}
	m_data[627] &= (7 | FEAT_VGUARDS | FEAT_SKYLABS);
	// Reset special features. But leave VGUARDS and SKYLABS alone; the 
	// guardian import deals with them.
}


bool ManicRoom::importHints(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent)
{
	xmlNodePtr cur;
	char *attr;

	clearHints();
	for (cur = parent->xmlChildrenNode; cur; cur = cur->next)
	{
		if (cur->ns != ns) continue;
		if (strcmp((char *)cur->name, "hint")) continue;

		char *hintName = (char *)xmlGetProp(cur, (xmlChar *)"type");
		if (!hintName) continue;

		if (!strcmp(hintName, "conveyor"))
		{
			attr = (char *)xmlGetProp(cur, (xmlChar *)"direction");
			if (attr)
			{
				m_data[623] = atoi(attr);
				xmlFree(attr);
			}
			attr = (char *)xmlGetProp(cur, (xmlChar *)"x");
			if (attr)
			{
				m_data[624] &= 0xE0;
				m_data[624] |= atoi(attr) & 0x1F;
				xmlFree(attr);
			}
			attr = (char *)xmlGetProp(cur, (xmlChar *)"y");
			if (attr)
			{
				m_data[624] &= 0x1F;
				m_data[624] |= ((atoi(attr) & 7) << 5);
				m_data[625] &= 0xF7;
				m_data[625] |= (atoi(attr) & 8);
				xmlFree(attr);
			}
			attr = (char *)xmlGetProp(cur, (xmlChar *)"animy");
			if (attr)
			{
				m_data[625] &= 0xF8;
				m_data[625] |= (atoi(attr) & 7);
				xmlFree(attr);
			}
			attr = (char *)xmlGetProp(cur, (xmlChar *)"length");
			if (attr)
			{
				m_data[626] = atoi(attr);
				xmlFree(attr);
			}
		}
		if (!strcmp(hintName, "solar"))
		{
			if (m_game->featuresSettableByRoom())
				m_data[627] |= FEAT_SOLAR;
		}
		if (!strcmp(hintName, "air"))
		{
			attr = (char *)xmlGetProp(cur, (xmlChar *)"amount");
			if (attr && atoi(attr) < 162)
			{
				m_data[700] = (atoi(attr) / 6) + 37;
				switch(atoi(attr) % 6)
				{
					case 5: m_data[701] = 0xFC; break;
					case 4: m_data[701] = 0xF8; break;
					case 3: m_data[701] = 0xF0; break;
					case 2: m_data[701] = 0xE0; break;
					case 1: m_data[701] = 0xC0; break;
					case 0: m_data[701] = 0x80; break;
				}
			}
			if (attr) xmlFree(attr);
		}
		if (!strcmp(hintName, "feature") && m_game->featuresSettableByRoom())
		{
			attr = (char *)xmlGetProp(cur, (xmlChar *)"name");
			if (attr && !strcmp(attr, "eugene"))
			{
				m_data[627] |= FEAT_EUGENE;
			}
			if (attr && !strcmp(attr, "skylabs"))
			{
				m_data[627] |= FEAT_SKYLABS;
			}
			if (attr && !strcmp(attr, "vguards"))
			{
				m_data[627] |= FEAT_VGUARDS;
			}
			if (attr && !strcmp(attr, "kong"))
			{
				m_data[627] |= FEAT_KONG;
			}
			if (attr) xmlFree(attr);
		}
		if (!strcmp(hintName, "startpos"))
		{
			attr = (char *)xmlGetProp(cur, (xmlChar *)"x");
			if (attr)
			{
				m_data[620] &= 0xE0;
				m_data[620] |= atoi(attr);
				xmlFree(attr);
			}
			attr = (char *)xmlGetProp(cur, (xmlChar *)"y");
			if (attr)
			{
				m_data[620] &= 0x1F;
				m_data[620] |= (atoi(attr) & 7) << 5;
				m_data[621] &= 0xFE;
				m_data[621] |= (atoi(attr) & 8) >> 3;
				m_data[616] &= 0x0F;
				m_data[616] |= (atoi(attr) & 15) << 4;
				xmlFree(attr);
			}
			attr = (char *)xmlGetProp(cur, (xmlChar *)"ypix");
			if (attr)
			{
				m_data[616] &= 0xF1;
				m_data[616] |= (atoi(attr) & 7) << 1;
				xmlFree(attr);
			}
			attr = (char *)xmlGetProp(cur, (xmlChar *)"direction");
			if (attr)
			{
				m_data[618] = atoi(attr);
				xmlFree(attr);
			}
			attr = (char *)xmlGetProp(cur, (xmlChar *)"frame");
			if (attr)
			{
				m_data[617] = atoi(attr);
				xmlFree(attr);
			}
			attr = (char *)xmlGetProp(cur, (xmlChar *)"jump");
			if (attr)
			{
				m_data[619] = atoi(attr);
				xmlFree(attr);
			}
		}
		xmlFree(hintName);
	}
	return Room::importHints(doc, ns, parent);
}


void ManicRoom::prepareForSpriteImport(void)
{
	m_portalTaken = false;
	m_spritesTaken = false;
}
int ManicRoom::addLocalSprite(jswByte *bits)
{
	if (m_portalTaken) return -1;
	m_portalTaken = true;
	memcpy(m_data + 656, bits, 32);
	return (m_roomNumber << 16) | 0x8290;
}

int ManicRoom::addLocalSpritePage(jswByte *bits)
{
	if (m_spritesTaken) return -1;
	m_spritesTaken = true;
	memcpy(m_data + 768, bits, 32);
	return (m_roomNumber << 16) | 0x8300;
}


bool ManicRoom::importGuardians(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent)
{
	xmlNodePtr cur;
	jswByte *buf;
	int n, ng, hslot, vslot;
	char variant;
	int vgs_inited = 0;
	int bufmax, bufmax2, bufbase, tableBase;

	// Take off the vertical guardian and skylab flags. The
	// import will reinstate them if there are guardians of those types.
	m_data[627] &= ~(FEAT_VGUARDS | FEAT_SKYLABS);
	char *type = (char *)xmlGetProp(parent, (xmlChar *)"type");

	hslot = vslot = 0;
	bufbase = 0;
	// Zap HGs and the start of the VGs
	memset(m_data + 702, 0, 33);
	m_data[702] = 0xFF;
	m_data[733] = 0xFF;
	if (type && !strcmp(type, "jsw"))
	{
/* Import JSW guardians into MM; there is no separate MM guardian type. */
		for (cur = parent->xmlChildrenNode; cur; cur = cur->next)
		{
			if (cur->ns != ns) continue;
			if (!strcmp((char *)cur->name, "vg"))
			{
				if (!vgs_inited) memset(m_data + 733, 0, 29);
				if (m_game->featuresSettableByRoom())
					m_data[627] |= FEAT_VGUARDS;
				if (vslot < 4 && m_game->roomHasFeature(m_roomNumber, FEAT_VGUARDS))
				{
					vgs_inited = 1;
					importJswVGuard(doc, ns, cur, vslot++);
				}
			}
			if (!strcmp((char *)cur->name, "skylab"))
			{
				if (!vgs_inited) memset(m_data + 733, 0, 29);
				if (m_game->featuresSettableByRoom())
					m_data[627] |= FEAT_SKYLABS;
				if (vslot < 4 && m_game->roomHasFeature(m_roomNumber, FEAT_SKYLABS))
				{
					vgs_inited = 1;
					importJswSkylab(doc, ns, cur, vslot++);
				}
			}
			if (!strcmp((char *)cur->name, "hg"))
			{
				if (hslot < 4)
				{
					importJswHGuard(doc, ns, cur, hslot++);
				}
			}
		}
	}
	xmlFree(type);
	return true;
}

void ManicRoom::importJswHGuard(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent,
		int slot)
{
	jswByte jguard[9];
	jswByte buf[8];

	// Import the XML as a JSW guardian
	importJswGuardian(doc, ns, parent, jguard, m_game);

	buf[0] = (m_data[544] & 0x38);		// Room paper
	buf[0] |= jguard[1] & 7;		// ink
	buf[0] |= ((jguard[1] & 8) << 3);	// bright
	if (jguard[7] & 0x20) buf[0] |= 0x80;	// slow
	buf[1]  = jguard[2] & 0x1F;		// X
	buf[1] |= (jguard[3] & 0x70) << 1;	// Y low 3 bits
	buf[2] = 0x5C + ((jguard[3] >> 7) & 1);	// Y high bits
	buf[3] = 0x60 + ((jguard[3] >> 4) & 8);	// Y high bits
	buf[4] = (jguard[0] & 0x80) ? 0 : 7;	// Direction
	buf[5] = jguard[6] & 0x1F;		// Bound 1
	buf[6] = jguard[7] & 0x1F;		// Bound 2
	buf[5] |= (buf[1] & 0xE0);
	buf[6] |= (buf[1] & 0xE0);
	buf[7] = 0xff;
	memcpy(m_data + 702 + 7*slot, buf, 8);
}

void ManicRoom::importJswVGuard(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent,
		int slot)
{
	jswByte jguard[9];
	jswByte buf[8];

	// Import the XML as a JSW guardian
	importJswGuardian(doc, ns, parent, jguard, m_game);

	buf[0] = (m_data[544] & 0x38);		// Room paper
	buf[0] |= jguard[1] & 7;		// ink
	buf[0] |= ((jguard[1] & 8) << 3);	// bright
	buf[1] = (jguard[2] >> 5) & 3;		// first frame
	buf[2] = (jguard[3] >> 1) & 0x7F;	// Y-coord
	buf[3] = jguard[2] & 0x1F;		// X-coord
	buf[4] = ((signed char)jguard[4]) / 2;	// Velocity
	buf[5] = (jguard[6] >> 1) & 0x7F;	// Bound 1
	buf[6] = (jguard[7] >> 1) & 0x7F;	// Bound 1
	buf[7] = 0xff;
	memcpy(m_data + 733 + 7*slot, buf, 8);
}

void ManicRoom::importJswSkylab(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent,
		int slot)
{
	jswByte jguard[9];
	jswByte buf[8];

	// Import the XML as a JSW guardian
	importJswGuardian(doc, ns, parent, jguard, m_game);

	buf[0] = (m_data[544] & 0x38);		// Room paper
	buf[0] |= jguard[1] & 7;		// ink
	buf[0] |= ((jguard[1] & 8) << 3);	// bright
	buf[1] = 0;				// first frame
	buf[2] = (jguard[3] >> 1) & 0x7F;	// Y-coord
	buf[3] = jguard[2] & 0x1F;		// X-coord
	buf[4] = ((signed char)jguard[4]) / 2;	// Velocity
	buf[5] = (jguard[6] >> 1) & 0x7F;	// Bound 1
	buf[6] = (jguard[7] >> 1) & 0x7F;	// Bound 1
	buf[7] = 0xff;
	memcpy(m_data + 733 + 7*slot, buf, 8);
}

/* Given a list of cell definitions, set their numberInRoom elements to
 * the cell type within the room. This can also modify their behaviour
 * if it wants.
 *
 * If numberInRoom is set to a +ve value, the corresponding cell pattern 
 * in the room is set from the pattern of this cell.
 *
 * If numberInRoom is set to a -ve value, the cell pattern is ignored and 
 * pattern (-numberInRoom) is substituted. 
 */
void ManicRoom::normaliseCells(ImportCell *p, int count)
{
	int air = -1, water1 = -1, earth = -1, fire1 = -1, 
	    lconv = -1, rconv = -1, water2 = -1, fire2 = -1,
	    crumbly = -1;
	int n;

	for (n = 0; n < count; n++)
	{
		if (p[n].count == 0) continue;
		switch(p[n].behaviour)
		{
		    case CB_AIR:   if (air == -1)   air   = n; break;
		    case CB_WATER: if (water1 == -1) water1 = n; 
			           else if (water2 == -1) water2 = n;
				   break;
		    case CB_EARTH: if (earth == -1) earth = n; break;
		    case CB_FIRE:  if (fire1 == -1)  fire1  = n; 
				   else if (fire2 == -1) fire2 = n;
				   break;
		    case CB_LCONV: if (lconv == -1) lconv  = n; break;
		    case CB_RCONV: if (rconv == -1) rconv  = n; break;
		    case CB_CRUMBLY: if (crumbly == -1) crumbly = n; break;
		}
	}
	// Last chance for unsupported cell types - give them the spare water
	// cell if it's still there.
	for (n = 0; n < count; n++)
	{
		if (p[n].count == 0) continue;
		switch(p[n].behaviour)
		{
		    case CB_LRAMP:
	   	    case CB_RRAMP:
		    case CB_TRAMP:
		    case CB_TRAP: if (water1 == -1) water1 = n; 
			           else if (water2 == -1) water2 = n;
				   break;
		}
	}
	for (n = 0; n < count; n++)
	{
		p[n].behaviour = CB_WATER;
		p[n].numberInRoom = -1;
	}
	if (air != -1) 
	{ 
		p[air].behaviour = CB_AIR; 
		p[air].numberInRoom = 0; 
	}
	if (water1 != -1) 
	{ 
		p[water1].behaviour = CB_WATER; 
		p[water1].numberInRoom = 1; 
	}
	if (water2 != -1) 
	{ 
		p[water2].behaviour = CB_WATER; 
		p[water2].numberInRoom = 7; 
	}
	if (crumbly != -1) 
	{ 
		p[crumbly].behaviour = CB_CRUMBLY;
		p[crumbly].numberInRoom = 2; 
	}
	if (earth != -1) 
	{ 
		p[earth].behaviour = CB_EARTH; 
		p[earth].numberInRoom = 3; 
	}
	if (fire1 != -1) 
	{
	       	p[fire1].behaviour = CB_FIRE; 
		p[fire1].numberInRoom = 5; 
	}
	if (fire2 != -1) 
	{
	       	p[fire2].behaviour = CB_FIRE; 
		p[fire2].numberInRoom = 6; 
	}
	if (lconv != -1) 
	{ 
		p[lconv].behaviour = CB_LCONV; 
		p[lconv].numberInRoom = 4; 
	}
	else if (rconv != -1)
	{
		p[rconv].behaviour = CB_RCONV; 
		p[rconv].numberInRoom = 4; 
	}
}

const char *ManicRoom::supportedGuardianTypes()
{
	return "manic|jsw";
}
