/************************************************************************

    JSWED 2.3.0 - Editor for Jet Set Willy and derivatives

    Copyright (C) 2000-2001,2004  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*************************************************************************/

#include "jswed.hxx"
#include "progressform.hxx"

ProgressForm::ProgressForm(unsigned max) : VideoForm(3 * CHAR_W, 13 * CHAR_H, 
					 34 * CHAR_W, 5 * CHAR_H)
{
	m_bmpZero = NULL;
	m_bmp100  = NULL;
	m_cur     = 0;
	m_max     = max;
	m_xpercent= 0;
	m_message = "";
	m_bmBehind = newVideoBitmap(m_x, m_y, m_w, m_h);
	videoScreen->fillBox(m_x, m_y, m_w, m_h, ZX_WHITE);
	videoScreen->box(m_x, m_y, m_w, m_h, ZX_BLACK);
	redraw();
}

ProgressForm::~ProgressForm()
{
	if (m_bmBehind)
	{
		m_bmBehind->toScreen(m_x, m_y);
		delete m_bmBehind;
	}
}


void ProgressForm::redraw(void)
{
	char pbuf[34];
	int px, py;

	if (m_bmpZero) delete m_bmpZero;
	if (m_bmp100)  delete m_bmp100;

	float percent = (100.0 * m_cur) / (float)m_max;

	if (percent < 0.0)   percent = 0.0;
	if (percent > 100.0) percent = 100.0;
	sprintf(pbuf, "%d%%",(int)percent);

	m_bmpZero = newVideoBitmap(32 * CHAR_W, 4 + CHAR_H);
	m_bmp100  = newVideoBitmap(32 * CHAR_W, 4 + CHAR_H);

	px = ((32 - strlen(pbuf)) * CHAR_W) / 2;
	py = 2;

	m_bmpZero->getSurface()->fillBox(0, 0, 32 * CHAR_W, CHAR_H + 4, ZX_BRWHITE);
	m_bmpZero->getSurface()->box    (0, 0, 32 * CHAR_W, CHAR_H + 4, ZX_BLACK);
	m_bmpZero->getSurface()->drawText(px, py, pbuf, ZX_BLACK, ZX_BRWHITE);
	m_bmpZero->toSurface(m_bmp100, 0, 0);
	m_bmp100->getSurface()->fillReplace(0, 0, 32 * CHAR_W, CHAR_H + 4,
			ZX_BRWHITE, ZX_BRGREEN);

	px = m_xpercent;
	py = (32 * CHAR_W) - m_xpercent; 

	if (px)
	{
		m_bmp100->toScreen(m_x + CHAR_W, m_y + CHAR_H - 2,
				0, 0, px, 4 + CHAR_H);
	}
	if (py)
	{
		m_bmpZero->toScreen(m_x + CHAR_W + px, m_y + CHAR_H - 2,
				px, 0, py, 4 + CHAR_H);
	}
	sprintf(pbuf, "%-32.32s", m_message);
	videoScreen->drawText(m_x + CHAR_W, m_y + 3 * CHAR_H, pbuf, ZX_BLACK, 
			ZX_WHITE);
}


void ProgressForm::bar(const char *message, unsigned count)
{
	m_message = message;
	m_cur = count;
	float percent = (float)count / (float)m_max;
	m_xpercent = (int)(percent * 32 * CHAR_W);

	redraw();
}

void ProgressForm::inc(int n)
{
	m_cur += n;
	float percent = (float)m_cur / (float)m_max;
	int xpc = (int)(percent * 32 * CHAR_W);
	if (xpc != m_xpercent)
	{
		m_xpercent = xpc;
		redraw();
	}
}
