/************************************************************************

    JSWED 2.3.0 - Editor for Jet Set Willy and derivatives

    Copyright (C) 2000-2001,2004-6  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*************************************************************************/

#include "jswed.hxx"
#include "room.hxx"

#include "sun.xbm"

ImportCell::ImportCell()
{
	numberInFile = numberInRoom = count = 0;
}

Portal::Portal()
{
	x = y = 0;
	attribute = 0;
	sprite = NULL;
	spriteID = 0;
	targetX = 0;
	targetY = 0;
	targetRoom = 0;
}


bool Portal::operator == (Portal &v)
{
	if (attribute != v.attribute) return false;

	if (sprite && v.sprite) return !memcmp(sprite, v.sprite, 32);
	return false;
}


void Portal::operator = (Portal &v)
{
	x = v.x;
	y = v.y;
	attribute = v.attribute;
	sprite    = v.sprite;
}


bool Portal::operator != (Portal &v)
{
	if (attribute != v.attribute) return true;

	if (sprite && v.sprite) return memcmp(sprite, v.sprite, 32);
	return true;
}


typedef VideoBitmap *BitmapPointer;

Room::Room(int ncells, int roomNumber) : m_ncells(ncells), 
			m_roomNumber(roomNumber)
{
	int j;

	m_cellBitmaps = new BitmapPointer[ncells * 2 + 1];
	for (j = 0; j < 2 * ncells + 1; j++)
	{
		m_cellBitmaps[j] = NULL;
	}
	m_zxCellBitmaps = new jswByte[18 * ncells + 9];
	m_portalBitmap = NULL;
	m_solarBitmap = NULL;
	m_solarFlag = 0xFF;
}


Room::Room(Room &other)
{
	copyFrom(&other);
}


Room::~Room()
{
	for (int j = 0; j < 2 * m_ncells + 1; j++)
	{
		if (m_cellBitmaps[j]) delete m_cellBitmaps[j];
	}
	delete m_cellBitmaps;	
	delete m_zxCellBitmaps;
	if (m_portalBitmap) delete m_portalBitmap;
	if (m_solarBitmap) delete m_solarBitmap;
}



void Room::createCellPatterns()
{
	jswByte *bm;

	for (int j = 0; j < 2 * m_ncells + 1; j++)
	{
		bm = m_zxCellBitmaps + 9 * j;		
		jswByte *buf = getCellPattern(j/2);

// If the cell hasn't changed, don't recreate the bitmaps.
		if (m_cellBitmaps[j])
		{
			if (!memcmp(bm, buf, 9)) continue;
//printf("Regen %d: bm[0]=%02x buf[0]=%02x\n", j, bm[0], buf[0]);
			if (m_cellBitmaps[j]) delete m_cellBitmaps[j];
			m_cellBitmaps[j] = NULL;
		}
		if (buf) 
		{
			int ink   = buf[0] & 7;
			int paper = (buf[0] >> 3) & 7;

			if (buf[0] & 64)  
			{
				ink |= ZX_BRIGHT; 
				paper |= ZX_BRIGHT; 
			}
			m_cellBitmaps[j] = ::bitmapFromSprite(8, 8, buf+1, 1, ink, paper);
			memcpy(bm, buf, 9);
		}
	}
	Portal p;
      
	/* If there is no portal, delete any portal bitmap we have */
	if (!getPortal(p))
	{
		if (m_portalBitmap) delete m_portalBitmap;
		m_portalBitmap = NULL;
	}
	else
	{
		/* If the portal has changed since this was set up */
		if (p != m_portal
		|| (p.sprite && memcmp(m_portalSprite, p.sprite, 32)))
		{
			if (m_portalBitmap) delete m_portalBitmap;
			m_portal = p;
			/* Take a copy of the sprite */
			memcpy(m_portalSprite, p.sprite, 32);
			p.sprite = m_portalSprite;

			int ink   = m_portal.attribute & 7;
			int paper = (m_portal.attribute >> 3) & 7;

			if (m_portal.attribute & 64)  
			{
				ink |= ZX_BRIGHT; 
				paper |= ZX_BRIGHT; 
			}
			m_portalBitmap = ::bitmapFromSprite(16, 16, m_portal.sprite, 1, ink, paper);
		}
	}
	jswByte sa = getSolarAttr();
	/* If there is no solar power, delete the solar-power bitmap */
	{
		/* If the portal has changed since this was set up */
		if (sa != m_solarFlag)
		{
			if (m_solarBitmap) delete m_solarBitmap;
			int ink = sa ? ZX_BRORANGE: ZX_ORANGE;
		    
			m_solarBitmap = ::bitmapFromXbm(sun_width, sun_height,
					sun_bits, 0, ZX_BLACK, ink);	
		}
	}
}






void Room::drawCell(int x, int y, int type)
{
	if (type == CT_SOLAR)
	{
		if (!m_solarBitmap) createCellPatterns();
		if (m_solarBitmap) m_solarBitmap->toScreen(x, y);
		else videoScreen->fillBox(x, y, 16, 16, ZX_BLACK);	
		return;	
	}
	if (type == CT_ITEM)
	{
		int n = m_ncells * 2;

		m_cellBitmaps[n]->toScreen(x, y);
		return;
	}
	if (type < 0 || type >= m_ncells) return;

	if (!m_cellBitmaps[type * 2])
	{
		createCellPatterns();	
	}
	if (m_cellBitmaps[type * 2])
	{
		m_cellBitmaps[type*2]->toScreen(x, y);
	}
}	



void Room::drawCell(VideoBitmap *bm, int x, int y, bool flashonly, bool animated)
{
	int cell = getCellAt(x, y);
	jswByte *pattern = getCellPattern(cell);

	if (!pattern) return;
	/* "?" type cell. These are not cached, but generated on the fly. */
	if (cell >= m_ncells)
	{
		int ink   = pattern[0] & 7;
		int paper = (pattern[0] >> 3) & 7;

		if (pattern[0] & 64)  
		{
			ink |= ZX_BRIGHT; 
			paper |= ZX_BRIGHT; 
		}
		if (!flashonly || (pattern[0] & 0x80))
		{
			VideoBitmap *bm2 = ::bitmapFromSprite(8, 8, 
					pattern+1, 1, ink, paper);
			bm2->toSurface(bm, x * 16, y * 16);
			delete bm2;
		}
		return;
	}

	if (!flashonly || (pattern[0] & 0x80))
	{
		m_cellBitmaps[cell * 2 + (animated ? 1 : 0)]->toSurface(bm, x * 16, y * 16);
	}	
}


void Room::flashCells()
{
	int n, im, ink, paper;
	jswByte *bm;

	for (n = 0; n < 2*m_ncells; n++)
	{
		bm = m_zxCellBitmaps + 9 * n;
		m_cellBitmaps[n]->flash(bm[0]);
	}
	if (m_portalBitmap) m_portalBitmap->flash(m_portal.attribute);
	if (m_solarBitmap)  m_solarBitmap->flash(getSolarAttr());
}

/*
void Room::diag(const char *s, int l)
{
	fprintf(stderr, "::%s(%d) %p %p", s, l, this, m_cellBitmaps);
	if (m_cellBitmaps)
	{
		fprintf(stderr, " %p", m_cellBitmaps[0]);
	}
	fprintf(stderr, "\n");
}*/

#define CONVEYORANIM 4  /* 2 or even 1 look nicer, but 4 matches JSW best */

void Room::animateConveyor()
{
	int direction = getConveyorDir() & 1;
	int n; 
	int y = getConveyorY();

	for (n = 0; n < m_ncells; n++)
	{
		if (direction == 1)
		{
			m_cellBitmaps[n*2+1]->getSurface()->rotateR(0, 2*y,   16, 2, CONVEYORANIM);
			m_cellBitmaps[n*2+1]->getSurface()->rotateL(0, 2*y+4, 16, 2, CONVEYORANIM);
		}
		else
		{
			m_cellBitmaps[n*2+1]->getSurface()->rotateL(0, 2*y,   16, 2, CONVEYORANIM);
			m_cellBitmaps[n*2+1]->getSurface()->rotateR(0, 2*y+4, 16, 2, CONVEYORANIM);
		}
	}
}


void Room::drawOneOffs(VideoBitmap *bm, bool flashonly, bool draw)
{
	if (m_portalBitmap)
	{
		m_portalBitmap->toSurface(bm, 16 * m_portal.x, 16 * m_portal.y);
	}
	if (m_solarBitmap && getSolarAttr())
	{
		int x, y;
		x = y = -1;
		getSolarPos(&x, &y);
		if (x >=0 && y >= 0)
		{
			m_solarBitmap->toSurface(bm, 16 * x, 16 * y);
		}
	}
}


int Room::getRestartX()
{
	return -1;
}

int Room::getRestartY()
{
	return -1;
}

void Room::setRestartX(int x)
{

}

void Room::setRestartY(int y)
{

}


void Room::drawItems(VideoBitmap *bm, int ink)
{
	int xs[256];
	int ys[256];
	int counts[256];
	int n, count;
	char caption[3];

	count = getItemCoords(xs, ys, counts);
	for (n = 0; n < count; n++)
	{
		drawItem(bm, xs[n], ys[n], ink);
		if (counts[n] > 1)
		{
			if (counts[n] < 256)
			{
				sprintf(caption, "%x", counts[n]);
			}
			else
			{
				sprintf(caption, "??");
			}
			if (strlen(caption) == 1) bm->getSurface()->drawSmallText
				(16*xs[n] + 4, 16*ys[n] + 4, caption, 11 - ink, ZX_BLACK);
			else bm->getSurface()->drawSmallText
				(16*xs[n], 16*ys[n] + 4, caption, 11 - ink, ZX_BLACK);
		}
		++ink;
		if (ink == ZX_WHITE) ink = ZX_MAGENTA;
	}
}


void Room::drawItem(VideoBitmap *bm, int x, int y, int ink)
{
	int n = m_ncells * 2;

	m_cellBitmaps[n]->getSurface()->fillReplace(0,0,16, 16, ZX_WHITE, ink);
	m_cellBitmaps[n]->toSurface(bm, x * 16, y * 16);
	m_cellBitmaps[n]->getSurface()->fillReplace(0,0,16, 16, ink, ZX_WHITE);
}



jswByte *Room::getGuardianBuffer()
{
	return NULL;
}

void Room::deletePortal(void)
{

}

bool Room::getPortal(Portal &p)
{
	return false;
}

int Room::getMaxGuards()
{
	return 8;
}

void Room::setCellBehaviour(int idx, int cb)
{
}


void Room::getSolarPos(int *x, int *y)
{
}

void Room::setSolarPos(int x, int y)
{
}

jswByte Room::getSolarAttr()
{
	return 0;
}

void Room::setGuardianTable(int offset)
{
}

int Room::getDrawnCellAt(int x, int y)
{
	return getCellAt(x,y);
}


void Room::getSprites(int *buf, int granularity)
{
	*buf = 0;
}


void Room::setPortalPos(int x, int y)
{
	m_portal.x = x;
	m_portal.y = y;
}

void Room::copyFrom(Room *other)
{
	int j;

	m_ncells = other->m_ncells;
	m_zxCellBitmaps = new jswByte[18 * m_ncells + 9];
	memcpy(m_zxCellBitmaps, other->m_zxCellBitmaps, 18 * m_ncells + 9);
	m_portal = other->m_portal;
	memcpy(m_portalSprite, other->m_portalSprite, 32);
	m_solarFlag = other->m_solarFlag;
	m_portalBitmap = NULL;
	m_solarBitmap = NULL;
	m_cellBitmaps = new BitmapPointer[m_ncells * 2 + 1];
	for (j = 0; j < 2 * m_ncells + 1; j++)
	{
		m_cellBitmaps[j] = NULL;
	}
	memcpy(m_itemPattern, other->m_itemPattern, 9);
	m_roomNumber = other->m_roomNumber;	
}

bool Room::exportLayout(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent)
{
	char room_buf[90*16];
	char buf[20];
	int n, x, y;
	char *p;

	xmlNodePtr layout    = xmlNewNode(ns, (xmlChar *)"layout");
	xmlNodePtr cells     = xmlNewNode(ns, (xmlChar *)"cells");
	xmlNodePtr bits      = xmlNewNode(ns, (xmlChar *)"bits");
	sprintf(buf, "%d", getBorder());
	xmlSetProp(layout, (xmlChar *)"border", (xmlChar *)buf);
	for (n = 0; n < getCellTypeCount(true); n++)
	{
		int no = encodeCellNumber(n);
		int be = getCellBehaviour(n);
		jswByte *pattern = getCellPattern(n);

		if (!exportCell(doc, ns, cells, no, be, pattern)) return false;
	}
	if (!exportExtraCells(doc, ns, cells)) return false;
	strcpy(room_buf, "\n");
	xmlAddChild(layout, cells);
	for (y = 0; y < 16; y++)
	{
		for (x = 0; x < 32; x++)
		{
			sprintf(buf, "%02x", getDrawnCellAt(x, y));
			strcat(room_buf, buf);
		}
		strcat(room_buf, "\n");
	}
	xmlAddChild(bits, xmlNewText((xmlChar *)room_buf));
	xmlSetProp(bits, (xmlChar *)"width", (xmlChar *)"32");
	xmlSetProp(bits, (xmlChar *)"height", (xmlChar *)"16");
	xmlAddChild(layout, cells);
	xmlAddChild(layout, bits);
	xmlAddChild(parent, layout);
	return true;
}

bool Room::exportXML(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent)
{
	char buf[50];
	int n;
	char *ptr;

	xmlNodePtr room      = xmlNewNode(ns, (xmlChar *)"room");
	xmlNodePtr title     = xmlNewNode(ns, (xmlChar *)"title");
	xmlNodePtr hints     = xmlNewNode(ns, (xmlChar *)"hints");
	xmlNodePtr items     = xmlNewNode(ns, (xmlChar *)"items");
	xmlNodePtr portal    = xmlNewNode(ns, (xmlChar *)"portal");
	getTitle(buf);
	for (n = 0; buf[n]; n++)
	{
		if (buf[n] < 0 || buf[n] > 126) buf[n] = '?';
	}
	xmlNodePtr titletext = xmlNewText((xmlChar *)buf);
	xmlAddChild(title, titletext);
	sprintf(buf, "%d", m_roomNumber);	
	xmlSetProp(room, (xmlChar *)"number", (xmlChar *)buf);
	xmlAddChild(room, title);
	if (!exportLayout(doc, ns, room)) return false;
	if (getExits())
	{
		xmlNodePtr exits     = xmlNewNode(ns, (xmlChar *)"exits");
		if (!exportExits(doc, ns, exits)) return false;	
		xmlAddChild(room, exits);
	}
	if (!exportItems(doc, ns, items)) return false;
	xmlAddChild(room, items);
	if (!exportGuardians(doc, ns, room)) return false;
	if (!exportPortal(doc, ns, portal)) return false;
	if (!exportHints(doc, ns, hints)) return false;
	ptr = (char *)xmlGetProp(portal, (xmlChar *)"x");
	if (ptr)
	{
		xmlAddChild(room, portal);
		xmlFree(ptr);
	}
	if (hints->xmlChildrenNode) xmlAddChild(room, hints);
	xmlAddChild(parent, room);

	return true;
}

static const char *st_exnames[4] =
{
	"left", "right", "up", "down"
};

bool Room::exportExits(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr node)
{
	char buf[30];
	jswByte *ex = getExits();
	int n;

	for (n = 0; n < 4; n++)
	{
		sprintf(buf, "%d", ex[n]);
		xmlSetProp(node, (xmlChar *)st_exnames[n], (xmlChar *)buf);	
	}
	return true;
}

jswByte *Room::getItemBitmap()
{
	return getCellPattern(m_ncells) + 1;
}

bool Room::exportItems(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr node)
{
	int xs[256];
	int ys[256]; 
	int counts[256];
	int n, m, count;
	char buf[40];
	jswByte *bm = getItemBitmap();

	xmlNodePtr cell = xmlNewNode(ns, (xmlChar *)"cell");
	for (n = 0; n < 8; n++)
	{
		sprintf(buf + 2*n, "%02x", bm[n]);
	}
	xmlAddChild(cell, xmlNewText((xmlChar *)buf));
	xmlAddChild(node, cell);

	count = getItemCoords(xs, ys, counts);
	for (n = 0; n < count; n++)
	{
		for (m = 0; m < counts[n]; m++)
		{
			xmlNodePtr item = xmlNewNode(ns, (xmlChar *)"item");
			sprintf(buf, "%d", xs[n]);
			xmlSetProp(item, (xmlChar *)"x", (xmlChar *)buf);
			sprintf(buf, "%d", ys[n]);
			xmlSetProp(item, (xmlChar *)"y", (xmlChar *)buf);
			xmlAddChild(node, item);
		}
	}	
	return true;
}



bool Room::importExits(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr node, 
			int roomMap[256])
{
	int n, r;
	char *attr;
	jswByte *ex = getExits();
	if (!ex) return true;

	for (n = 0; n < 4; n++)
	{
		attr = (char *)xmlGetProp(node, (xmlChar *)st_exnames[n]);
		if (attr) 
		{
			for (r = 0; r < 256; r++) 
			{
				if (roomMap[r] == atoi(attr))
				{
					ex[n] = r;
				}
			}
			xmlFree(attr);
		}	
	}	
	return true;
}



bool Room::importTitle(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr node)
{
	char *t = (char *)xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
	if (t)	
	{	
		setTitle(t);
		xmlFree(t);
	}
	else	setTitle("");
	return true;
}


bool Room::importCells(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr node, 
		xmlNodePtr bits, ImportCell **parsedCells, int *ncells)
{
	ImportCell *cells;
	int n, count, ncell, rows;
	bool okay = true;
	xmlNodePtr cur;

	count = 0;
	for (cur = node->xmlChildrenNode; cur; cur = cur->next)
	{
		if (cur->ns != ns) continue;
		if (strcmp((char *)cur->name, "cell")) continue;
		++count;	
	}
	if (count == 0)
	{
		jswLog("No cell definitions in <cells>\n");
		return false;
	}
	
	initImportCells();
	cells   = new ImportCell[count];
	ncell   = 0;
	for (cur = node->xmlChildrenNode; cur; cur = cur->next)
	{
		if (cur->ns != ns) continue;
		if (strcmp((char *)cur->name, "cell")) continue;

		xmlChar *xcnumber    = xmlGetProp(cur, (xmlChar *)"number");
		xmlChar *xcbehaviour = xmlGetProp(cur, (xmlChar *)"behaviour");
		xmlChar *xcink       = xmlGetProp(cur, (xmlChar *)"ink");
		xmlChar *xcpaper     = xmlGetProp(cur, (xmlChar *)"paper");
		xmlChar *xcbright    = xmlGetProp(cur, (xmlChar *)"bright");
		xmlChar *xcflash     = xmlGetProp(cur, (xmlChar *)"flash");
		if (!xcbehaviour) xcbehaviour = (xmlChar *)xmlMemStrdup("water");
		if (!xcink)       xcink       = (xmlChar *)xmlMemStrdup("7");
		if (!xcpaper)     xcpaper     = (xmlChar *)xmlMemStrdup("7");
		if (!xcbright)    xcbright    = (xmlChar *)xmlMemStrdup("0");
		if (!xcflash)     xcflash     = (xmlChar *)xmlMemStrdup("0");
		cells[ncell].numberInRoom = ncell;
		cells[ncell].numberInFile = -1;
		if (xcnumber) 
		{
			sscanf((char *)xcnumber, "%x", 
					&cells[ncell].numberInFile);
		}
		cells[ncell].behaviour = CB_WATER;
		if (!strcmp((char *)xcbehaviour, "air")) 
			cells[ncell].behaviour = CB_AIR;
		else if (!strcmp((char *)xcbehaviour, "water")) 
			cells[ncell].behaviour = CB_WATER;
		else if (!strcmp((char *)xcbehaviour, "earth")) 
			cells[ncell].behaviour = CB_EARTH;
		else if (!strcmp((char *)xcbehaviour, "fire")) 
			cells[ncell].behaviour = CB_FIRE;
		else if (!strcmp((char *)xcbehaviour, "lramp")) 
			cells[ncell].behaviour = CB_LRAMP;
		else if (!strcmp((char *)xcbehaviour, "rramp")) 
			cells[ncell].behaviour = CB_RRAMP;
		else if (!strcmp((char *)xcbehaviour, "lconveyor")) 
			cells[ncell].behaviour = CB_LCONV;
		else if (!strcmp((char *)xcbehaviour, "rconveyor")) 
			cells[ncell].behaviour = CB_RCONV;
		else if (!strcmp((char *)xcbehaviour, "crumbly")) 
			cells[ncell].behaviour = CB_CRUMBLY;
		else if (!strcmp((char *)xcbehaviour, "trampoline")) 
			cells[ncell].behaviour = CB_TRAMP;
		else if (!strcmp((char *)xcbehaviour, "trap")) 
			cells[ncell].behaviour = CB_TRAP;

		cells[ncell].pattern[0] =  (atoi((char *)xcink  ) & 7)	|
					  ((atoi((char *)xcpaper) & 7) << 3)  |
					   (atoi((char *)xcbright) ? 0x40 : 0)|
					   (atoi((char *)xcflash)  ? 0x80 : 0);
		char *t = (char *)xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
		rows = strlen(t) / 2;
		if (rows > 8) rows = 8;
		for (n = 0; n < rows; n++)
		{
			char sc[3];

			int xv = -1;
			sprintf(sc, "%-2.2s", t + 2 * n);
			sscanf(sc, "%x", &xv);
			cells[ncell].pattern[n+1] = xv;
		}
		xmlFree(t);
		if (xcnumber) xmlFree(xcnumber);
		if (xcbehaviour) xmlFree(xcbehaviour);
		xmlFree(xcink);
		xmlFree(xcpaper);
		xmlFree(xcbright);
		xmlFree(xcflash);
		++ncell;
		if (ncell >= count) break;
	}
	// Calculate cell usage
	calcCellUsage(cells, count, doc, ns, bits);

	// Reduce the palette of cells to those actually available in the 
	// room.
	normaliseCells(cells, count);	
	for (n = 0; n < count; n++)
	{
		if (n >= m_ncells) break;
		jswByte *pat = getCellPattern(cells[n].numberInRoom);
		if (pat) memcpy(pat, cells[n].pattern, 9);
	}
	if (ncells)      *ncells = count;
	if (parsedCells) *parsedCells = cells;
	else	delete [] cells;
	return true;	
}

void Room::normaliseCells(ImportCell *p, int count)
{
	// Override in derived classes as required
}

bool Room::importLayout(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr node)
{
	ImportCell *cells = NULL;
	int ncells = 0;
	xmlNodePtr cur, bits;
	char *bdr = (char *)xmlGetProp(node, (xmlChar *)"border");

	if (bdr) 
	{
		setBorder(atoi(bdr));
		xmlFree(bdr);
	}
	bits = NULL;
	// Find the 'bits' node.
	for (cur = node->xmlChildrenNode; cur; cur = cur->next)
	{
		if (cur->ns != ns) continue;
		if (!strcmp((char *)cur->name, "bits"))
		{
			bits = cur;
			break;
		}
	}
	if (!bits)	// Room with no bitmap
	{
		return false;
	}

	for (cur = node->xmlChildrenNode; cur; cur = cur->next)
	{
		if (cur->ns != ns) continue;
		if (!strcmp((char *)cur->name, "cells"))
		{
			if (!importCells(doc, ns, cur, bits, &cells, &ncells))
			{
				if (cells) delete [] cells;
				return false;
			}
			break;
		}
	}
	if (!cells) return false; // No cells; hence no layout
// Load the <bits> 
	if (!importBits(doc, ns, bits, cells, ncells))
	{
		if (cells) delete [] cells;
		return false;
	}
	if (cells) delete [] cells;
	return true;
}
	



bool Room::importBits(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr node, 
		ImportCell *cells, int ncells)
{
	int width, height, x, y, n;

	xmlChar *w = xmlGetProp(node, (xmlChar *)"width");
	xmlChar *h = xmlGetProp(node, (xmlChar *)"height");
	xmlChar *a = xmlGetProp(node, (xmlChar *)"attrs");
	xmlChar *buf = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
	char *xs = (char *)buf;

	if (!w || !h || !buf) return false;
	width  = atoi((char *)w);
	height = atoi((char *)h);
	xmlFree(w);
	xmlFree(h);
	if (width == 0 || height == 0) 
	{
		xmlFree(buf);
		return false;
	}

	initImportBits();
	for (y = 0; y < height; y++)
		for (x = 0; x < width; x++)
		{
			char sc[3];

			if (y >= 16 || x >= 32) continue;

			if (!xs[0]) break;
			while (xs[0] && !isxdigit(xs[0])) ++xs;
			if (!xs[0]) break;
			sc[0] = *xs++;
			while (xs[0] && !isxdigit(xs[0])) ++xs;
			if (!xs[0]) break;
			sc[1] = *xs++;
			sc[2] = 0;
			int xv = -1;
			sscanf(sc, "%x", &xv);

			for (n = 0; n < ncells; n++) 
				if (cells[n].numberInFile == xv)
			{
				int nr = cells[n].numberInRoom;
				if (nr < 0) nr = -nr;
				importCell(x, y, cells[n].pattern[0], 
						nr, cells[n].behaviour);	
				break; // Don't do the same cell multiple times
			}

		}
	xmlFree(buf);
	return true;
}


void Room::calcCellUsage(ImportCell *cells, int ncells, 
		xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr node)
{
	int width, height, x, y, n;
	xmlChar *w = xmlGetProp(node, (xmlChar *)"width");
	xmlChar *h = xmlGetProp(node, (xmlChar *)"height");
	xmlChar *buf = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
	char *xs = (char *)buf;

	if (!w || !h || !buf) return;
	width  = atoi((char *)w);
	height = atoi((char *)h);
	xmlFree(w);
	xmlFree(h);
	if (width == 0 || height == 0) 
	{
		xmlFree(buf);
		return;
	}
	for (y = 0; y < height; y++)
	{
		for (x = 0; x < width; x++)
		{
			char sc[3];

			if (y >= 16 || x >= 32) continue;

			if (!xs[0]) break;
			while (xs[0] && !isxdigit(xs[0])) ++xs;
			if (!xs[0]) break;
			sc[0] = *xs++;
			while (xs[0] && !isxdigit(xs[0])) ++xs;
			if (!xs[0]) break;
			sc[1] = *xs++;
			sc[2] = 0;
			int xv = -1;
			sscanf(sc, "%x", &xv);

			for (n = 0; n < ncells; n++) 
				if (cells[n].numberInFile == xv)
			{
				++cells[n].count;
			}
		}
	}
	xmlFree(buf);
}

void Room::initImport()
{

}


void Room::initImportCells()
{

}



void Room::initImportBits()
{

}

bool Room::importItems(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr node)
{
	xmlNodePtr cur;
	int x, y, rows;
	jswByte buf[8];

	clearItems();
	for (cur = node->xmlChildrenNode; cur; cur = cur->next)
	{
		if (cur->ns != ns) continue;
		if (!strcmp((char *)cur->name, "cell"))
		{
			// XXX This will not work because in JSW48 the 
			// item bitmap is read-only.
			char *t = (char *)xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			jswByte *bm = getCellPattern(m_ncells) + 1;
			memcpy(buf, bm, 8);
			rows = strlen(t) / 2;
			if (rows > 8) rows = 8;
			for (int n = 0; n < rows; n++)
			{
				char sc[3];
	
				int xv = -1;
				sprintf(sc, "%-2.2s", t + 2 * n);
				sscanf(sc, "%x", &xv);
				buf[n] = xv;
			}
			setItemBitmap(buf);
			xmlFree(t);
		}
		if (!strcmp((char *)cur->name, "item"))
		{
			char *sx = (char *)xmlGetProp(cur, (xmlChar *)"x");	
			char *sy = (char *)xmlGetProp(cur, (xmlChar *)"y");	
			if (sx && sy) addItem(atoi(sx), atoi(sy));
			if (sx) xmlFree(sx);
			if (sy) xmlFree(sy);
		}
	}
	return true;
}


void Room::importCell(int x, int y, int attr, int type, int behaviour)
{
	setCell(x, y, type);
}



bool Room::importXML(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr node,
		int roomMap[256])
{
	bool okay = true;
	bool haveGuardians = false;
	bool haveHints = false;
	bool havePortal = false;
	bool haveItems = false;
	xmlNodePtr cur;

	if (strcmp((char *)node->name, "room") || node->ns != ns) return false;

	initImport();
	for (cur = node->xmlChildrenNode; cur; cur = cur->next)
	{
		if (cur->ns != ns) continue;
		if (!strcmp((char *)cur->name, "title"))
		{
			okay = importTitle(doc, ns, cur) && okay;
		}
		if (!strcmp((char *)cur->name, "hints"))
		{
			okay = importHints(doc, ns, cur) && okay;
			haveHints = true;
		}
		else if (!strcmp((char *)cur->name, "layout"))
		{
			okay = importLayout(doc, ns, cur) && okay;
		}
		else if (!strcmp((char *)cur->name, "exits"))
		{
			okay = importExits(doc, ns, cur, roomMap) && okay;
		}
		else if (!strcmp((char *)cur->name, "items"))
		{
			okay = importItems(doc, ns, cur) && okay;
			haveItems = true;
		}
		else if (!strcmp((char *)cur->name, "portal"))
		{
			okay = importPortal(doc, ns, cur, roomMap) && okay;
			havePortal = true;
		}
		else if (!strcmp((char *)cur->name, "guardians"))
		{
			// We'll come back to guardians later
		}
		else
		{
			jswLog("Room::ImportXML skipping unknown "
					"node %s\n", (char *)cur->name);
		}
	}
//
// Guardians are special, because there can be multiple alternative 
// representations in the one room. We find out which types this room
// supports, and search the file for them in priority order.
//
	const char *types = supportedGuardianTypes();
	if (types)
	{
// supportedGuardianTypes() returned a list of supported guardian types,
// in priority order, separated by | signs. 
		char *buf = new char[1 + strlen(types)];
		strcpy(buf, types);
		char *token;
		for (token = strtok(buf, "|"); token != NULL;
				token = strtok(NULL, "|"))
		{
			bool matched = false;
			for (cur = node->xmlChildrenNode; cur; cur = cur->next)
			{
				if (cur->ns != ns) continue;
				if (strcmp((char *)cur->name, "guardians")) 
					continue;
		
				char *type = (char *)xmlGetProp(cur, 
					(xmlChar *)"type");
				if (!type) continue;
				if (strcmp(type, token))
				{
					xmlFree(type);
					continue;
				}
				xmlFree(type);
				haveGuardians = true;
				matched = true;
				okay = importGuardians(doc, ns, cur) && okay;
				break;
			}
			if (matched) break;
		} 
		delete [] buf;
	}
	if (!haveGuardians)
	{
		jswByte *buf = getGuardianBuffer();
		if (buf) buf[0] = 0xFF;
	}
	if (!havePortal)
	{
		deletePortal();
	}
	if (!haveItems)
	{
		clearItems();
	}
	if (!haveHints)
	{
		clearHints();
	}
	return okay;
}

bool Room::exportPortal(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent)
{
	char buf[20];
	Portal p;
	if (!getPortal(p)) return true;

	sprintf(buf, "%d", p.x); 
	xmlSetProp(parent, (xmlChar *)"x", (xmlChar *)buf);
	sprintf(buf, "%d", p.y); 
	xmlSetProp(parent, (xmlChar *)"y", (xmlChar *)buf);

	sprintf(buf, "%d", p.attribute & 7);
	xmlSetProp(parent, (xmlChar *)"ink", (xmlChar *)buf);
	sprintf(buf, "%d", (p.attribute >> 3) & 7);
	xmlSetProp(parent, (xmlChar *)"paper", (xmlChar *)buf);
	if (p.attribute & 0x40) 
		xmlSetProp(parent, (xmlChar *)"bright", (xmlChar *)"1");
	if (p.attribute & 0x80) 
		xmlSetProp(parent, (xmlChar *)"flash", (xmlChar *)"1");

	sprintf(buf, "%d", p.targetRoom); 
	xmlSetProp(parent, (xmlChar *)"target", (xmlChar *)buf);
	sprintf(buf, "%d", p.targetX); 
	xmlSetProp(parent, (xmlChar *)"targetx", (xmlChar *)buf);
	sprintf(buf, "%d", p.targetY); 
	xmlSetProp(parent, (xmlChar *)"targety", (xmlChar *)buf);
	sprintf(buf, "%04x", p.spriteID); 
	xmlSetProp(parent, (xmlChar *)"sprite", (xmlChar *)buf);
	return true;
}


bool Room::importPortal(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent,
		int roomMap[256])
{
	Portal p;
	getPortal(p);
	char *prop;
	
	prop = (char *)xmlGetProp(parent, (xmlChar *)"x");
	if (prop) { p.x = atoi(prop); xmlFree(prop); }
	prop = (char *)xmlGetProp(parent, (xmlChar *)"y");
	if (prop) { p.y = atoi(prop); xmlFree(prop); }
//	fprintf(stderr, "Import portal x=%d y=%d\n", p.x, p.y);
	prop = (char *)xmlGetProp(parent, (xmlChar *)"ink");
	if (prop) 
	{
	      p.attribute = (p.attribute & 0xF8) | (atoi(prop) & 7);
	      xmlFree(prop);
	}
	prop = (char *)xmlGetProp(parent, (xmlChar *)"paper");
	if (prop) 
	{
	       p.attribute = (p.attribute & 0x47) | ((atoi(prop) & 7) << 3);
	       xmlFree(prop);
	}
	prop = (char *)xmlGetProp(parent, (xmlChar *)"bright");
	if (prop && atoi(prop)) p.attribute |= 0x40;
	if (prop) xmlFree(prop);
	prop = (char *)xmlGetProp(parent, (xmlChar *)"flash");
	if (prop && atoi(prop)) p.attribute |= 0x80;
	if (prop) xmlFree(prop);

	prop = (char *)xmlGetProp(parent, (xmlChar *)"target");
	if (prop) 
	{
		for (int r = 0; r < 256; r++) 
		{
			if (roomMap[r] == atoi(prop))
			{
				p.targetRoom = r;
			}
		}
		xmlFree(prop);	
	}	
	prop = (char *)xmlGetProp(parent, (xmlChar *)"targetx");
	if (prop) { p.targetX = atoi(prop); xmlFree(prop); }
	prop = (char *)xmlGetProp(parent, (xmlChar *)"targety");
	if (prop) { p.targetY = atoi(prop); xmlFree(prop); }
	prop = (char *)xmlGetProp(parent, (xmlChar *)"sprite");
	if (prop) { sscanf(prop, "%x", &p.spriteID); xmlFree(prop); }
// This needs to be done in the
// derived class's setPortal() method, because we have no
// m_game member.	
// 	p.spriteID = m_game->lookupSpriteRef(p.spriteID);

	setPortal(p);
	return true;
}

void Room::setPortal(Portal &p)
{

}



bool Room::exportHints(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent)
{
	char buf[20];
	int rx, ry;
	xmlNodePtr hint;

	rx = getRestartX();
	ry = getRestartY();
	if (rx >= 0 && ry >= 0)
	{
		hint = xmlNewNode(ns, (xmlChar *)"hint");
		xmlSetProp(hint, (xmlChar *)"type", (xmlChar *)"respawn");
		sprintf(buf, "%d", rx);
		xmlSetProp(hint, (xmlChar *)"x", (xmlChar *)buf);
		sprintf(buf, "%d", ry);
		xmlSetProp(hint, (xmlChar *)"y", (xmlChar *)buf);
		xmlAddChild(parent, hint);
	}
	return true;
}


bool Room::importHints(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent)
{
	xmlNodePtr cur;

	for (cur = parent->xmlChildrenNode; cur; cur = cur->next)
	{
		if (cur->ns != ns) continue;
		if (strcmp((char *)cur->name, "hint")) continue;

		char *hintName = (char *)xmlGetProp(cur, (xmlChar *)"type");
		if (!hintName) continue;
		if (!strcmp(hintName, "respawn"))
		{
			char *cx = (char *)xmlGetProp(cur, (xmlChar *)"x");
			char *cy = (char *)xmlGetProp(cur, (xmlChar *)"y");
			if (cx && cy)
			{
				setRestartX(atoi(cx));
				setRestartY(atoi(cx));
			}
			if (cx) xmlFree(cx);
			if (cy) xmlFree(cy);
		}
		xmlFree(hintName);
	}
	return true;
}


const char *Room::supportedGuardianTypes()
{
	return NULL;
}

bool Room::exportGuardians(xmlDocPtr doc, xmlNsPtr ns, 
		xmlNodePtr grandparent)
{
	return true;
}

bool Room::importGuardians(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent)
{
	return true;
}

static char *st_guardnodes[] =
{
	"blank",   "hg", "vg", "rope",  "arrow", "dgnw", "dgne", "vg",
	"special", "hg", "vg", "blank", "blank", "dgnw", "dgne", "vg"
};

static char *st_xguardnodes[] =
{
	"skylab",  "eugene",  "eugene",  "eugene", 
	"blank",   "lift",    "scenery", "blank", 
	"trigger", "switch",  "wall",    "stopper",
	"stopper", "trigger", "blank",   "blank",
};



bool Room::exportJswGuardian(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent,
					jswByte *buf, int tableBase)
{
	char data[40];
	int type = buf[0] & 0x0F;
	xmlNodePtr node;
	int address;

	if (type == 8) // JSW64 extended guardian
	{
		type = buf[0];
		node = xmlNewNode(ns, (xmlChar *)(st_xguardnodes[type/16]));
	}
	else
	{
		node = xmlNewNode(ns, (xmlChar *)(st_guardnodes[type]));
	}

	if (type == 7 || type == 9 || type == 13 || type == 14 || type == 15
	||  type == 0x38)
	{
		xmlSetProp(node, (xmlChar *)"multicolour", (xmlChar *)"1");
	}
	if (type == 0x28)
	{
		xmlSetProp(node, (xmlChar *)"multicolour", (xmlChar *)"2");
	}
	if (type == 0xD8)
	{
		xmlSetProp(node, (xmlChar *)"weak", (xmlChar *)"1");
	}
	if (type == 0xC8)
	{
		xmlSetProp(node, (xmlChar *)"type", (xmlChar *)"motion");
	}
	switch(type)
	{
		case 3: // Rope
			if (buf[0] & 64) 
				xmlSetProp(node, (xmlChar *)"stopped", 
						(xmlChar *)"1");
			if (buf[0] & 128) 
				xmlSetProp(node, (xmlChar *)"direction",
					(xmlChar *)"R");
			else	xmlSetProp(node, (xmlChar *)"direction",
					(xmlChar *)"L");

			sprintf(data, "%d", buf[1]);
			xmlSetProp(node, (xmlChar *)"pos", (xmlChar *)data);
			sprintf(data, "%d", buf[2] & 0x1F);
			xmlSetProp(node, (xmlChar *)"x", (xmlChar *)data);
			sprintf(data, "%d", buf[4]);
			xmlSetProp(node, (xmlChar *)"len", (xmlChar *)data);
			sprintf(data, "%d", buf[7]);
			xmlSetProp(node, (xmlChar *)"swing", (xmlChar *)data);
			break;
		case 4: // Arrow
			if (buf[0] & 128) 
				xmlSetProp(node, (xmlChar *)"direction",
					(xmlChar *)"R");
			else	xmlSetProp(node, (xmlChar *)"direction",
					(xmlChar *)"L");
			sprintf(data, "%d", buf[1]);
			xmlSetProp(node, (xmlChar *)"ink", (xmlChar *)data);
			sprintf(data, "%d", buf[2] / 2);
			xmlSetProp(node, (xmlChar *)"y", (xmlChar *)data);
			sprintf(data, "%d", buf[4]);
			xmlSetProp(node, (xmlChar *)"x", (xmlChar *)data);
			sprintf(data, "%02x", buf[3]);
			xmlSetProp(node, (xmlChar *)"middle", (xmlChar *)data);
			sprintf(data, "%02x", buf[6]);
			xmlSetProp(node, (xmlChar *)"edges", (xmlChar *)data);
			break;
		case 1: case 2: case 5: case 6: case 7:
		case 9: case 10: case 13: case 14: case 15:
		case 0x08: case 0x18: case 0x28: case 0x38:
		case 0x58: case 0x68: 
		case 0x98: // switch.
		// Byte 0
			sprintf(data, "%d", (buf[0] >> 5));
			xmlSetProp(node, (xmlChar *)"frame", (xmlChar *)data);
			if (buf[0] & 16) xmlSetProp(node, (xmlChar *)"fastanim",
				(xmlChar *)"1");
			if ((buf[0] & 0x0F) != 8)
			{
				if (buf[0] & 0x80) xmlSetProp(node, 
					(xmlChar *)"direction", (xmlChar *)"1");
				else xmlSetProp(node, (xmlChar *)"direction", 
						(xmlChar *)"0");
			}
		// Byte 1
			sprintf(data, "%d", buf[1] & 7);
			xmlSetProp(node, (xmlChar *)"ink", (xmlChar *)data);
			if (buf[1] & 8) xmlSetProp(node, (xmlChar *)"bright",
				(xmlChar *)"1");
			sprintf(data, "%d", buf[1] >> 5);
			xmlSetProp(node, (xmlChar *)"framemask", (xmlChar *)data);
			sprintf(data, "%d", buf[2] & 31);
			xmlSetProp(node, (xmlChar *)"x", (xmlChar *)data);
			sprintf(data, "%d", buf[2] >> 5);
			xmlSetProp(node, (xmlChar *)"startframe", (xmlChar *)data);
			sprintf(data, "%d", buf[3] / 2);
			xmlSetProp(node, (xmlChar *)"y", (xmlChar *)data);
			sprintf(data, "%d", buf[4]);
			xmlSetProp(node, (xmlChar *)"speed", (xmlChar *)data);
			sprintf(data, "%d", buf[6]);
			xmlSetProp(node, (xmlChar *)"limit1", (xmlChar *)data);
			sprintf(data, "%d", buf[7]);
			xmlSetProp(node, (xmlChar *)"limit2", (xmlChar *)data);

// Reference the sprites separately as child nodes
// <spriteref frame="0" xref="be00" />
// <spriteref frame="1" xref="be20" />
// etc.
			for (int frame = 0; frame < 8; frame++)
			{
				int xref = (frame << 5) & 0xE0;
		/* Skylabs use all frames, so don't limit.
		 * Other x8 guardians use 1 frame; this should work as 
		 * normal. */
				if (type != 8)
				{
					xref &= (buf[1] & 0xE0);
					xref |= (buf[2] & 0xE0);
				}
				xref += 256 * buf[5];
				xmlNodePtr spr = xmlNewNode(ns, (xmlChar *)"spriteref");
				sprintf(data, "%d", frame);
				xmlSetProp(spr, (xmlChar *)"frame", (xmlChar *)data); 	
				sprintf(data, "%04x", xref);
				xmlSetProp(spr, (xmlChar *)"xref", (xmlChar *)data); 
				xmlAddChild(node, spr);
			}
	//		sprintf(data, "%02x00", buf[5]);
	//		xmlSetProp(node, (xmlChar *)"sprite", (xmlChar *)data);
			break;
		case 0x88: // trigger.
		case 0xD8: // weak trigger.
			address = buf[1] + 256 * buf[2];
			// XXX Should check room size rather than fixed 
			// limits
			if (address < 0x8000 || address >= 0x8400)
			{
				sprintf(data, "%04x", address);
				xmlSetProp(node, (xmlChar *)"address", 
						(xmlChar *)data); 
			}
			else
			{
				sprintf(data, "%d", address - tableBase);
				xmlSetProp(node, (xmlChar *)"relative", 
						(xmlChar *)data); 
			}
			data[0] = 0;
			for (int b = 3; b < 8; b++)
			{	
				sprintf(data+strlen(data), "%02x", buf[b]);
			}
			xmlSetProp(node, (xmlChar *)"newvalues", 
						(xmlChar *)data);
			break;
		case 0xA8: // opening wall
			address = buf[1] + 256 * buf[2];
			// XXX Should check room size rather than fixed 
			// limits
			if (address < 0x8000 || address >= 0x8400)
			{
				sprintf(data, "%04x", address);
				xmlSetProp(node, (xmlChar *)"address", 
						(xmlChar *)data); 
			}
			else
			{
				sprintf(data, "%d", address - tableBase);
				xmlSetProp(node, (xmlChar *)"relative", 
						(xmlChar *)data); 
			}
			sprintf(data, "%d", buf[3]);
			xmlSetProp(node, (xmlChar *)"height", 
						(xmlChar *)data);
			sprintf(data, "%d", buf[6]);
			xmlSetProp(node, (xmlChar *)"x", (xmlChar *)data);
			sprintf(data, "%d", buf[7]);
			xmlSetProp(node, (xmlChar *)"y", (xmlChar *)data);
			break;
		case 0xB8: // stopper. No parameters.
		case 0xC8: // stopper. No parameters.
			break;

	}	
	xmlAddChild(parent, node);
	return true;
}

bool Room::importJswGuardian(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr cur,
					jswByte *buf, Game *game, int tableBase)
{
	unsigned char guard[8];
	char *prop;

	memset(guard, 0, sizeof(guard));

	if (!strcmp((char *)cur->name, "vg")     || 
	    !strcmp((char *)cur->name, "hg")     || 
	    !strcmp((char *)cur->name, "dgnw")   || 
	    !strcmp((char *)cur->name, "eugene") || 
	    !strcmp((char *)cur->name, "skylab") || 
	    !strcmp((char *)cur->name, "lift")   || 
	    !strcmp((char *)cur->name, "switch") || 
	    !strcmp((char *)cur->name, "scenery")) 
	{
// Decode byte 0
		if      (!strcmp((char *)cur->name, "vg"))      guard[0] = 2;
		else if (!strcmp((char *)cur->name, "hg"))      guard[0] = 1;
		else if (!strcmp((char *)cur->name, "dgnw"))    guard[0] = 5;
		else if (!strcmp((char *)cur->name, "dgnw"))    guard[0] = 6;
		else if (!strcmp((char *)cur->name, "skylab"))  guard[0] = 8;
		else if (!strcmp((char *)cur->name, "eugene"))  guard[0] = 0x18;
		else if (!strcmp((char *)cur->name, "lift"))    guard[0] = 0x58;
		else if (!strcmp((char *)cur->name, "scenery")) guard[0] = 0x68;
		else if (!strcmp((char *)cur->name, "switch"))  guard[0] = 0x98;

		prop = (char *)xmlGetProp(cur, (xmlChar *)"multicolour");
		if (prop && atoi(prop))
		{
			if (guard[0] == 1) guard[0] = 9;
			if (guard[0] == 2) guard[0] = 7;
			if (guard[0] == 5) guard[0] = 13;
			if (guard[0] == 6) guard[0] = 14;
			if (guard[0] == 0x18) guard[0] += 0x10 * atoi(prop);
		}
		if (prop) xmlFree(prop);
		prop = (char *)xmlGetProp(cur, (xmlChar *)"direction");
		if (prop && atoi(prop) && (guard[0] & 0x0F) != 8)
			guard[0] |= 0x80;
		if (prop) xmlFree(prop);
		prop = (char *)xmlGetProp(cur, (xmlChar *)"frame");
		if (prop) { guard[0] |= (atoi(prop) & 7) << 5; xmlFree(prop); }
		prop = (char *)xmlGetProp(cur, (xmlChar *)"fastanim");
		if (prop && atoi(prop) && ((guard[0] & 7) != 7)) guard[0] |= 16;
		if (prop) xmlFree(prop);
// Decode byte 1	
		prop = (char *)xmlGetProp(cur, (xmlChar *)"ink");
		if (prop) { guard[1] |= (atoi(prop) & 7); xmlFree(prop); }
		prop = (char *)xmlGetProp(cur, (xmlChar *)"bright");
		if (prop && atoi(prop)) guard[1] |= 8;
		if (prop) xmlFree(prop);
		prop = (char *)xmlGetProp(cur, (xmlChar *)"framemask");
		if (prop) { guard[1] |= ((atoi(prop) & 7) << 5); xmlFree(prop); }
// Decode byte 2
		prop = (char *)xmlGetProp(cur, (xmlChar *)"x");
		if (prop) { guard[2] |= (atoi(prop) & 31); xmlFree(prop); }
		prop = (char *)xmlGetProp(cur, (xmlChar *)"startframe");
		if (prop) { guard[2] |= ((atoi(prop) & 7) << 5); xmlFree(prop); }
// Decode byte 3
		prop = (char *)xmlGetProp(cur, (xmlChar *)"y");
		if (prop) { guard[3] |= (atoi(prop) * 2); xmlFree(prop); }
// Decode byte 4
		prop = (char *)xmlGetProp(cur, (xmlChar *)"speed");
		if (prop) { guard[4] |= atoi(prop); xmlFree(prop); }
// Decode byte 5 - sprite page
		xmlNodePtr spriteref;
		for (spriteref = cur->xmlChildrenNode; spriteref != NULL;
			spriteref = spriteref->next)
		{
			if (spriteref->ns != ns) continue;
			if (strcmp((char *)spriteref->name, "spriteref"))
				continue;
			prop = (char *)xmlGetProp(spriteref, (xmlChar *)"xref");
			if (prop)
			{
				int sprPage;
				sscanf(prop, "%x", &sprPage);	
		// Map file ID to game sprite ID
				sprPage = game->lookupSpriteRef(sprPage );
				guard[5] = sprPage / 256;
				xmlFree(prop);
				break;
			}
		}

// Decode bytes 6,7 
		prop = (char *)xmlGetProp(cur, (xmlChar *)"limit1");
		if (prop) { guard[6] = atoi(prop); xmlFree(prop); }
		prop = (char *)xmlGetProp(cur, (xmlChar *)"limit2");
		if (prop) { guard[7] = atoi(prop); xmlFree(prop); }
		memcpy(buf, guard, 8);
		return true;
	}
	if (!strcmp((char *)cur->name, "rope"))
	{
		guard[0] = 3;

		prop = (char *)xmlGetProp(cur, (xmlChar *)"stopped"); 
		if (prop && atoi(prop)) guard[0] |= 64;
		if (prop) xmlFree(prop);

		prop = (char *)xmlGetProp(cur, (xmlChar *)"direction");
		if (prop && prop[0] == 'R') guard[0] |= 128;
		if (prop) xmlFree(prop);
		
		prop = (char *)xmlGetProp(cur, (xmlChar *)"pos");
		if (prop) { guard[1] = atoi(prop); xmlFree(prop); }
		prop = (char *)xmlGetProp(cur, (xmlChar *)"x");
		if (prop) { guard[2] = atoi(prop) & 0x1F; xmlFree(prop); }
		prop = (char *)xmlGetProp(cur, (xmlChar *)"len");
		if (prop) { guard[4] = atoi(prop); xmlFree(prop); }
		prop = (char *)xmlGetProp(cur, (xmlChar *)"swing");
		if (prop) { guard[7] = atoi(prop); xmlFree(prop); }
		memcpy(buf, guard, 8);
		return true;
	}
	if (!strcmp((char *)cur->name, "arrow"))
	{
		guard[0] = 4;
		guard[1] = 7;
		guard[3] = 0xFF;
		int v;

		prop = (char *)xmlGetProp(cur, (xmlChar *)"direction");
		if (prop && prop[0] == 'R') guard[0] |= 128;
		if (prop) xmlFree(prop);

		prop = (char *)xmlGetProp(cur, (xmlChar *)"ink");
		if (prop) { guard[1] = atoi(prop); xmlFree(prop); }
		prop = (char *)xmlGetProp(cur, (xmlChar *)"y");
		if (prop) { guard[2] = 2*atoi(prop); xmlFree(prop); }
		prop = (char *)xmlGetProp(cur, (xmlChar *)"x");
		if (prop) { guard[4] = atoi(prop); xmlFree(prop); }

		prop = (char *)xmlGetProp(cur, (xmlChar *)"middle");
		if (prop && sscanf(prop, "%x", &v)) guard[3] = v;
		if (prop) xmlFree(prop);
		prop = (char *)xmlGetProp(cur, (xmlChar *)"edges");
		if (prop && sscanf(prop, "%x", &v)) guard[6] = v;
		if (prop) xmlFree(prop);
		memcpy(buf, guard, 8);
		return true;
	}
	if (!strcmp((char *)cur->name, "trigger"))
	{
		int addr = tableBase;
		char bt[3];
		int v;

		prop = (char *)xmlGetProp(cur, (xmlChar *)"weak");
		if (prop && atoi(prop)) guard[0] = 0xD8;
		else			guard[0] = 0x88;
		if (prop) xmlFree(prop);
		prop = (char *)xmlGetProp(cur, (xmlChar *)"relative");
		if (prop) { addr = atoi(prop) + tableBase; xmlFree(prop); }
		prop = (char *)xmlGetProp(cur, (xmlChar *)"address");
		if (prop && sscanf(prop, "%x", &v)) addr = v;
		if (prop) xmlFree(prop);
		guard[1] = addr & 0xFF;
		guard[2] = addr >> 8;
		prop = (char *)xmlGetProp(cur, (xmlChar *)"newvalues");
		void *p = prop;
		if (prop) for (int n = 0; n < 5; n++)
		{
			strncpy(bt, prop, 2);
			bt[2] = 0;
			if (prop[0] == 0) break;
			if (sscanf(bt, "%x", &v)) guard[n+3] = v;
			prop += strlen(bt);
		} 
		if (p) xmlFree(p);
		memcpy(buf, guard, 8);
		return true;
	}
	if (!strcmp((char *)cur->name, "wall"))
	{
		int addr = tableBase;
		int v;

		guard[0] = 0xA8;
		prop = (char *)xmlGetProp(cur, (xmlChar *)"relative");
		if (prop) addr = atoi(prop) + tableBase;
		if (prop) xmlFree(prop);
		prop = (char *)xmlGetProp(cur, (xmlChar *)"address");
		if (prop && sscanf(prop, "%x", &v)) addr = v;
		if (prop) xmlFree(prop);
		guard[1] = addr & 0xFF;
		guard[2] = addr >> 8;
		prop = (char *)xmlGetProp(cur, (xmlChar *)"height");
		if (prop && atoi(prop)) guard[3] = atoi(prop);
		if (prop) xmlFree(prop);
		guard[4] = 1;
		guard[5] = 0;
		prop = (char *)xmlGetProp(cur, (xmlChar *)"x");
		if (prop)  { guard[6] = atoi(prop); xmlFree(prop); }
		prop = (char *)xmlGetProp(cur, (xmlChar *)"y");
		if (prop)  { guard[7] = atoi(prop); xmlFree(prop); }
		memcpy(buf, guard, 8);
		return true;
	}
	if (!strcmp((char *)cur->name, "stopper"))
	{
		prop = (char *)xmlGetProp(cur, (xmlChar *)"type");
		if (prop && !strcmp(prop, "motion")) guard[0] = 0xC8;
		else guard[0] = 0xB8;
		if (prop) xmlFree(prop);
		memset(guard+1, 0, 7);
		memcpy(buf, guard, 8);
		return true;
	}
	memset(guard, 0, 8);	// Blank (unknown type) 
	return true;
}


void Room::clearHints()
{

}


int Room::encodeCellNumber(int c)
{
	return c;
}


bool Room::exportExtraCells(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent)
{
	return true;
}

bool Room::exportCell(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent, 
		int number, int behaviour, jswByte *pattern)
{
	char buf[20];
	char *p;

	xmlNodePtr cell = xmlNewNode(ns, (xmlChar *)"cell");
	sprintf(buf, "%02x", number);
	xmlSetProp(cell, (xmlChar *)"number", (xmlChar *)buf);

	switch(behaviour)
	{
		case CB_AIR:   p = "air"; break;
		case CB_WATER: p = "water"; break;
		case CB_EARTH: p = "earth"; break;
		case CB_FIRE:  p = "fire"; break;
		case CB_LRAMP: p = "lramp"; break;
		case CB_RRAMP: p = "rramp"; break;
		case CB_LCONV: p = "lconveyor"; break;
		case CB_RCONV: p = "rconveyor"; break;
		case CB_CRUMBLY: p = "crumbly"; break;
		case CB_TRAMP: p = "trampoline"; break;
		case CB_TRAP: p = "trap"; break;
		default: p = "unknown"; break;
	}
	xmlSetProp(cell, (xmlChar *)"behaviour", (xmlChar *)p);
	sprintf(buf, "%d", pattern[0] & 7);
	xmlSetProp(cell, (xmlChar *)"ink", (xmlChar *)buf);
	sprintf(buf, "%d", (pattern[0] >> 3) & 7);
	xmlSetProp(cell, (xmlChar *)"paper", (xmlChar *)buf);
	if (pattern[0] & 0x40) 
		xmlSetProp(cell, (xmlChar *)"bright", (xmlChar *)"1");
	if (pattern[0] & 0x80) 
		xmlSetProp(cell, (xmlChar *)"flash", (xmlChar *)"1");
	for (int pt = 0; pt < 8; pt++)
	{
		sprintf(buf + 2*pt, "%02x", pattern[pt+1]);
	}
	xmlAddChild(cell, xmlNewText((xmlChar *)buf));
	xmlAddChild(parent, cell);	
	return true;
}


void Room::prepareForSpriteImport(void)
{
}

int Room::addLocalSprite(jswByte *bits)
{
	return -1;
}

int Room::addLocalSpritePage(jswByte *bits)
{
	return -1;
}

