/************************************************************************

    JSWED 2.3.0 - Editor for Jet Set Willy and derivatives

    Copyright (C) 2000-2001, 2004-6  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*************************************************************************/

#include "jswed.hxx"
#include "jswgame.hxx"
#include "room.hxx"
#include "j48roomnav.hxx"
#include "roomrange.hxx"
#include "jumproom.hxx"


int RoomImport::doExport(Game *g)
{
	int room, from, upto;

	JswRoomList *s = g->getRoomList();
	RoomRangeForm rf(1, s);

	int rv = rf.doModal();
	if (rv >= ZXE_QUIT) 
	{
		delete s;
		return rv;
	}
	if (rv != ZXE_OK) 
	{
		delete s;
		return ZXE_CONTINUE;
	}

	xmlDocPtr doc = xmlNewDoc((xmlChar *)"1.0");
	if (!doc)
	{
		delete s;
		return alert("Error", "Cancel", "Failed to create XML document object.");
	}
	xmlNsPtr ns; 
	xmlNodePtr root;
	root = xmlNewNode(NULL, (xmlChar *)"rooms");
	xmlDocSetRootElement(doc, root);
	ns = xmlNewNs(root, NAMESPACE, NULL);

	from = rf.getFrom();
	upto = rf.getUpto();
	g->clearRefCounts();
	delete s;
	for (room = from; room <= upto; room++)
	{
		Room *rm = g->getRoomClass(room);
		if (!rm) continue;
		rm->exportXML(doc, ns, root);
		rm->addRefs();
	}
	g->exportReferencedSprites(doc, ns, root, 16, 16);
        char filename[PATH_MAX];
        JswFileForm ff;

	filename[0] = 0;
	rv = ff.doModal(filename);
	if (rv != ZXE_OK)
	{
		xmlFreeDoc(doc);
		if (rv >= ZXE_QUIT) return rv;
		return ZXE_CONTINUE;
	}
	if (xmlSaveFile(filename, doc) < 0)
	{
		xmlFreeDoc(doc);
		return alert("File was not saved.", "Cancel");
	}
	xmlFreeDoc(doc);
	return ZXE_CONTINUE;
}


int RoomImport::doImport(Game *g)
{
        char filename[PATH_MAX];
        JswFileForm ff;
	xmlDocPtr doc;
	xmlNsPtr ns;
	xmlNodePtr cur, child;
	int rv;
	int from, upto, room;
	int map[256];

	filename[0] = 0;
	rv = ff.doModal(filename);
	if (rv != ZXE_OK)
	{
		if (rv >= ZXE_QUIT) return rv;
		return ZXE_CONTINUE;
	}
	doc = xmlParseFile(filename);
	if (doc == NULL)
	{
		return alert("Couldn't load XML", "Cancel",
				"Unable to load and parse the file:\n%s",
				filename);
	}
	cur = xmlDocGetRootElement(doc);
	if (cur == NULL)
	{
		xmlFreeDoc(doc);
		return alert("Couldn't load XML", "Cancel",
				"XML document is empty");
	}
	ns = xmlSearchNsByHref(doc, cur, NAMESPACE);
	if (ns == NULL)         // No namespace
	{
		xmlFreeDoc(doc);
		return alert("Couldn't load XML", "Cancel",
				"No JSWED namespace found in the file.");
	}
	if (strcmp((char *)cur->name, "rooms")) // Wrong root
	{
		xmlFreeDoc(doc);
		return alert("Couldn't load XML", "Cancel",
				"First node is not <rooms>.");
	}
	int roomCount = 0;
	// OK. File loaded. See how many rooms there are.
	for (child = cur->xmlChildrenNode; child; child = child->next)
	{
		if (child->ns != ns) continue;
		if (!strcmp((char *)child->name, "room")) ++roomCount;
	}
	if (roomCount == 0)
	{
		xmlFreeDoc(doc);
		return alert("Couldn't load XML", "Cancel",
				"No rooms in the file.");
	}
	JswRoomList *s = g->getRoomList();
	RoomRangeForm rf(0, s, roomCount);
	rv = rf.doModal();

	if (rv != ZXE_OK)
	{
		delete s;
		xmlFreeDoc(doc);
		if (rv >= ZXE_QUIT) return rv;
		return ZXE_CONTINUE;
	}
	from = rf.getFrom();
	upto = rf.getUpto();
	delete s;
	child = cur->xmlChildrenNode; 
/* Populate the map of rooms */
	for (room = 0; room < 256; room++) map[room] = -1;
	for (room = from; child != NULL; child = child->next)
	{
		if (child->ns != ns) continue;
		if (strcmp((char *)child->name, "room")) continue;
		char *num = (char *)xmlGetProp(child, (xmlChar *)"number");

		if (num) 
		{
			map[room++] = atoi(num);
			xmlFree(num);
		}
	}
/* Build a reference count of guardians and sprites, in the rooms that are
 * not being used. */
	g->clearRefCounts();
	for (room = 0; room < 256; room++) if (map[room] == -1)
	{
		Room *rm = g->getRoomClass(room);
		if (rm) rm->addRefs();
	}
	g->addFixedRefs();
/* Import sprites and map them */
	g->importSprites(doc, ns, cur, map);

//	g->dumpRefs();
	child = cur->xmlChildrenNode; 
	for (room = from; child != NULL; child = child->next)
	{
		if (child->ns != ns) continue;
		if (strcmp((char *)child->name, "room")) continue;

		Room *rm = g->getRoomClass(room);
		if (!rm) continue;
		rm->importXML(doc, ns, child, map);
		child = child->next;
		++room;
		if (room > upto) break;
	}
	xmlFreeDoc(doc);

	return ZXE_CONTINUE;
}

