/************************************************************************

    JSWED 2.3.0 - Editor for Jet Set Willy and derivatives

    Copyright (C) 2000-2001,2004  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*************************************************************************/

#include "jswed.hxx"
#include "jswroomlist.hxx"
#include "roomrange.hxx"

RoomRangeForm::RoomRangeForm (int mode, JswRoomList *list, int count) :
	VideoForm(6 * CHAR_W, 9 * CHAR_H, 38 * CHAR_W, 13 * CHAR_H)
{
	m_mode = mode;
	m_list = list;
	m_count = count;
        m_ok = new VideoButton(ZXE_OK, m_x + CHAR_W, m_y + m_h - 2 * CHAR_H,
				                        "  OK  ");
        m_cancel = new VideoButton(ZXE_CANCEL, m_x + m_w - 7 * CHAR_W, 
		                        m_y + m_h - 2 * CHAR_H, "Cancel");
        m_first = new VideoButton(-1, m_x + CHAR_SW, m_y + 3 * CHAR_SH," From ");
        m_last  = new VideoButton(-1, m_x + CHAR_SW, m_y + 5 * CHAR_H, "  To  ");

	m_first->setListener(this);
	m_last->setListener(this);
	m_from = 0;
	if (m_mode)	m_upto = list->getItemCount() - 1;
	else		m_upto = m_count - 1;	
}


RoomRangeForm::~RoomRangeForm()
{
	delete m_ok;
	delete m_cancel;
	delete m_first;
	delete m_last;
}


int RoomRangeForm::doModal()
{
	return VideoForm::doModal();
}


void RoomRangeForm::redraw(void)
{
	VideoForm::redraw();
	if (m_mode) redrawBorder("Select rooms to export");
	else	    redrawBorder("Select first room to import");

	while (m_upto > 0 && !m_list->getItem(m_upto))
	{
		--m_upto;
	}

	videoScreen->drawText(m_x + CHAR_W, m_y + 6 * CHAR_SH,
			m_list->getItem(m_from), ZX_BLACK, ZX_WHITE);
	videoScreen->drawText(m_x + CHAR_W, m_y + 13 * CHAR_SH,
			m_list->getItem(m_upto), ZX_BLACK, ZX_WHITE);
}


int RoomRangeForm::onButtonSelect(VideoButton *b)
{
	if (b == m_first || b == m_last)
	{
		JswRoomList l2;
		l2.copyFrom(m_list);
		if (b == m_first) 
		{
			if (m_mode) l2.setTitle("First room to export");
			else	    l2.setTitle("First room to import");
			l2.setSelected(m_from);
		}
		else
		{	
			if (m_mode) l2.setTitle("Last room to export");
			else	    l2.setTitle("Last room to import");
			l2.setSelected(m_upto);
		}
		int rv = l2.doModal();
		if (rv >= ZXE_QUIT) return rv;
		if (rv == ZXE_OK)
		{
			if (b == m_first) m_from = l2.getSelected();
			else		  m_upto = l2.getSelected();
			if (!m_mode) m_upto = m_from + m_count - 1;
			redraw();
		}
		return ZXE_CONTINUE;
	}
	return ZXE_CONTINUE;
}


void RoomRangeForm::onSelect(ScrollingList *sender, int nSel)
{

}

void RoomRangeForm::hideChildren(void)
{
	VideoForm::hideChildren();
	removeChild(m_first);
	if (m_mode) removeChild(m_last);
	removeChild(m_ok);
	removeChild(m_cancel);
}


void RoomRangeForm::showChildren(int redraw)
{
	addChild(m_first);
	if (m_mode) addChild(m_last);
	addChild(m_ok);
	addChild(m_cancel);
	VideoForm::showChildren(redraw);
}


