/************************************************************************

    JSWED 2.3.0 - Editor for Jet Set Willy and derivatives

    Copyright (C) 2000-2001,2004  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*************************************************************************/


#include "jswed.hxx"

static int read3Dos(const char *path, const char *name, int minLen, int addr, jswByte *loadAt, int maxLen = -1);
static int write3Dos(const char *path, const char *name, const jswByte *header, const jswByte *data);
static int write3Dos(const char *path, const char *name, int len, int addr, const jswByte *data);
static void constructHeader(jswByte *header, int t, int a, int b);

#define TUNMAX 0xAD6

static const jswByte snaTemplate[] =
{
        0x3f,           /* 0: I */
        0x58, 0x27,     /* 1: HL' */
        0x9b, 0x36,     /* 3: DE' */
        0xfb, 0xfe,     /* 5: BC' */
	0xA3, 0x05,	/* 7: AF' */
        0x33, 0x86,     /* 9: HL */
        0x08, 0x58,     /* B: DE */
        0x04, 0x69,     /* D: BC */
        0x3a, 0x5c,     /* F: IY */
        0x74, 0x84,     /* 11:IX */
        0x00,           /* 13:IFF2 */
        0x73,           /* 14:R */
        0x0a, 0x00,     /* 15:AF */
        0xfe, 0x5b,     /* 17:SP */
        0x01,           /* 19:IM 1 */
        0x00,           /* 1A:Border */
};

static const jswByte z80Map[] = { 10, /* I */
                                  19, 20, 17, 18, 15, 16, 22, 21, /* Alt regs*/
                                   4,  5, 13, 14,  2,  3, 23, 24,
                                  25, 26,                         /* Main regs*/
                                  27, 11,  1,  0,  8,  9, 29, 12 };

static const int pages48[] = { 8, 4, 5, -1 };
static const int pages128[] = { 3, 4, 5, 6, 7, 8, 9, 10, -1};
static const int addrs128[] = { 0xC000, 0, 0x8000, 0,0, 0x4000, 0, 0 };

static const int banks48[]  = { 5, 2, 0 };
static const int banks128[] = { 0, 1, 2, 3, 4, 5, 6, 7 };
static const int addrs48[]  = { 0x4000, 0x8000, 0xC000 };


#define VAR -1
#define LTR 'j'

/* "Universal" JSW loader. This works on cassette or +3 disc systems, 
 * and for JSW64 or JSW128 games. */
static signed short jsw_univ_loader[] = 
{
0x50,0x4c,0x55,0x53,0x33,0x44,0x4f,0x53,
0x1a,0x01,0x00,0x45,0x04,0x00,0x00,0x00,
0xc5,0x03,0x00,0x00,0xbf,0x03,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x4b,
/* End of +3DOS header */
0x00,0x0a,0x68,0x00,0xea,0x4a,0x65,0x74,
0x2d,0x53,0x65,0x74,0x20,0x57,0x69,0x6c,
0x6c,0x79,0x20,0x36,0x34,0x20,0x2d,0x20,
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x42,0x41,0x53,0x49,0x43,0x20,0x6c,0x6f,
0x61,0x64,0x65,0x72,0x20,0x20,0x20,0x20,
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x20,0x20,0x20, 'U', 'n', 'i', 'v', 'e',
 'r', 's', 'a', 'l', ' ', 'v', 'e', 'r',
 's', 'i', 'o', 'n',0x20,0x20,0x20,0x20,
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x20,0x20,0x20,0x4a,0x43,0x45,0x20,0x32,
0x30,0x30,0x34,0x0d,

/* 20 CLEAR 24999 */
0x00,0x14,0x0d,0x00,0xfd,0x32,0x34,0x39,0x39,0x39,0x0e,0x00,
0x00,0xa7,0x61,0x00,0x0d,

/* 29 was REM LOAD "nocache.bin" CODE -- now spare. */
0x00,0x1d,0x84,0x00,
0xea,0x20,0x20,0x20,0x20,0x20,0x20,0x20,	
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,	/* 0x10 */
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,	/* 0x20 */
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,	/* 0x30 */
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,	/* 0x40 */
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,	/* 0x50 */
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,	/* 0x60 */
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,	/* 0x70 */
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,	/* 0x80 */
0x20,0x20,0x20,0x0d,

0x00,0x1e,0x1a,0x00,0xef,0x22,0x68,	/* 30 load "hiload.j?7" */
0x69,0x6c,0x6f,0x61,0x64,0x2e, LTR,VAR,
0x37,0x22,0xaf,0x32,0x35,0x30,0x30,0x30,	/* code 25000 */
0x0e,0x00,0x00,0xa8,0x61,0x00,0x0d,

/* 40 LET x=USR 32121 */
	0x00,0x28,0x10,0x00,
	0xf1,0x58,0x3d,0xc0,	/* 16 bytes */
	0x33,0x32,0x31,0x32,
	0x31,0x0E,0x00,0x00,
	0x79,0x7d,0x00,0x0d,
/* Load all the rest */
0x00,0x32,0x3e,0x00,0xef,0x22,0x72,0x6f,0x6f,
0x6d,0x73,0x2e, LTR,VAR,0x36,0x22,0xaf,
0x33,0x32,0x37,0x36,0x38,0x0e,0x00,0x00,
0x00,0x80,0x00,0x3a,0xf4,0x33,0x32,0x30,
0x30,0x30,0x0e,0x00,0x00,0x00,0x7d,0x00,
0x2c,0x36,0x0e,0x00,0x00,0x06,0x00,0x00,
0x3a,0xf1,0x58,0x3d,0xc0,0x33,0x32,0x30,
0x30,0x31,0x0e,0x00,0x00,0x01,0x7d,0x00,
0x0d,0x00,0x3c,0x3e,0x00,0xef,0x22,0x72,
0x6f,0x6f,0x6d,0x73,0x2e, LTR,VAR,0x34,
0x22,0xaf,0x33,0x32,0x37,0x36,0x38,0x0e,
0x00,0x00,0x00,0x80,0x00,0x3a,0xf4,0x33,
0x32,0x30,0x30,0x30,0x0e,0x00,0x00,0x00,
0x7d,0x00,0x2c,0x34,0x0e,0x00,0x00,0x04,
0x00,0x00,0x3a,0xf1,0x58,0x3d,0xc0,0x33,
0x32,0x30,0x30,0x31,0x0e,0x00,0x00,0x01,
0x7d,0x00,0x0d,0x00,0x46,0x3e,0x00,0xef,
0x22,0x72,0x6f,0x6f,0x6d,0x73,0x2e, LTR,
VAR,0x33,0x22,0xaf,0x33,0x32,0x37,0x36,
0x38,0x0e,0x00,0x00,0x00,0x80,0x00,0x3a,
0xf4,0x33,0x32,0x30,0x30,0x30,0x0e,0x00,
0x00,0x00,0x7d,0x00,0x2c,0x33,0x0e,0x00,
0x00,0x03,0x00,0x00,0x3a,0xf1,0x58,0x3d,
0xc0,0x33,0x32,0x30,0x30,0x31,0x0e,0x00,
0x00,0x01,0x7d,0x00,0x0d,0x00,0x50,0x3e,
0x00,0xef,0x22,0x72,0x6f,0x6f,0x6d,0x73,
0x2e, LTR,VAR,0x31,0x22,0xaf,0x33,0x32,
0x37,0x36,0x38,0x0e,0x00,0x00,0x00,0x80,
0x00,0x3a,0xf4,0x33,0x32,0x30,0x30,0x30,
0x0e,0x00,0x00,0x00,0x7d,0x00,0x2c,0x31,
0x0e,0x00,0x00,0x01,0x00,0x00,0x3a,0xf1,
0x58,0x3d,0xc0,0x33,0x32,0x30,0x30,0x31,
0x0e,0x00,0x00,0x01,0x7d,0x00,0x0d,0x00,
0x5a,0x54,0x00,0xef,0x22,0x74,0x69,0x74,
0x6c,0x65,0x2e, LTR,VAR,0x37,0x22,0xaf,
0x33,0x32,0x37,0x36,0x38,0x0e,0x00,0x00,
0x00,0x80,0x00,0x3a,0xf4,0x33,0x32,0x30,
0x30,0x30,0x0e,0x00,0x00,0x00,0x7d,0x00,
0x2c,0x37,0x0e,0x00,0x00,0x07,0x00,0x00,
0x3a,0xf4,0x33,0x32,0x30,0x33,0x30,0x0e,
0x00,0x00,0x1e,0x7d,0x00,0x2c,0x32,0x37,
0x0e,0x00,0x00,0x1b,0x00,0x00,0x3a,0xf1,
0x78,0x3d,0xc0,0x33,0x32,0x30,0x30,0x31,
0x0e,0x00,0x00,0x01,0x7d,0x00,0x0d,0x00,
0x64,0x19,0x00,0xef,0x22,0x72,0x74,0x69,
0x6d,0x65,0x2e, LTR,VAR,0x32,0x22,0xaf,
0x33,0x32,0x37,0x36,0x38,0x0e,0x00,0x00,
0x00,0x80,0x00,0x0d,
0x00,0x78,0x19,0x00,0xef,0x20,
0x22, 'm', 'a', 'i', 'n', '.', 't', 'u',
 'n',0x22,0xaf,0x36,0x33,0x32,0x33,0x32,
0x0e,0x00,0x00,0x00,0xf7,0x00,0x0d,0x00,
0x82,0x1b,0x00,0xef,0x22, 'c', 'h', 'e',
 'a', 't', '.', 't', 'u', 'n',0x22,0x20,
0x20,0xaf,0x36,0x33,0x37,0x34,0x36,0x0e,
0x00,0x00,0x02,0xf9,0x00,0x0d,0x00,0x8c,
0x18,0x00,0xef,0x22,0x74,0x75,0x6e,0x65,
0x2e,0x62,0x69,0x6e,0x22,0xaf,0x36,0x34,
0x30,0x30,0x30,0x0e,0x00,0x00,0x00,0xfa,
0x00,0x0d,0x00,0x96,0x7d,0x00,0xea,0x50,
0x41,0x55,0x53,0x45,0x20,0x32,0x30,0x30,
0x3a,0x20,0x52,0x45,0x4d,0x20,0x20,0x20,
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x20,0x20,0x4c,0x69,0x6e,0x65,0x20,0x31,
0x35,0x30,0x20,0x63,0x61,0x6e,0x20,0x62,
0x65,0x20,0x64,0x65,0x6c,0x65,0x74,0x65,
0x64,0x20,0x69,0x66,0x20,0x20,0x6e,0x6f,
0x74,0x20,0x6c,0x6f,0x61,0x64,0x69,0x6e,
0x67,0x20,0x66,0x72,0x6f,0x6d,0x20,0x2b,
0x33,0x20,0x64,0x69,0x73,0x63,0x20,0x20,
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x20,0x20,0x0d,
/* Load title.tun at 16384 */
0x00,0x9f,0x1a,0x00,0xef,0x20,0x22, 't', 'i', 't', 'l', 'e',	/* Tune filenames */
 '.', 't', 'u', 'n',0x22,0xaf,0x31,0x36,
0x33,0x38,0x34,0x0e,0x00,0x00,0x00,0x40,
0x00,0x0d,
/* 160 LET x=USR 32039 */
0x00,0xa0,0x10,0x00,0xf1,0x58,0x3d,0xc0,0x33,0x32,0x30,0x33,0x39,
0x0e,0x00,0x00,0x27,0x7d,0x00,0x0d,0x78,
0x00,0x00,0xfd,0x7f,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
};


/*
// 25 April 2004: This is actually the JSW64 tape loader, but will do 
//               just as well for JSW128. Fields containing VAR will be
//               replaced with the game variant letter ('s' for JSW128).
//
//               The same thing has been done for the +3 disc loader.
//
static signed short jsw_tape_loader[] = 
{
0x50,0x4c,0x55,0x53,0x33,0x44,0x4f,0x53,
0x1a,0x01,0x00,0x45,0x04,0x00,0x00,0x00,
0xc5,0x03,0x00,0x00,0xbf,0x03,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x4b,
0x00,0x0a,0x68,0x00,0xea,0x4a,0x65,0x74,
0x2d,0x53,0x65,0x74,0x20,0x57,0x69,0x6c,
0x6c,0x79,0x20,0x36,0x34,0x20,0x2d,0x20,
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x42,0x41,0x53,0x49,0x43,0x20,0x6c,0x6f,
0x61,0x64,0x65,0x72,0x20,0x20,0x20,0x20,
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x20,0x20,0x20,0x43,0x61,0x73,0x73,0x65,
0x74,0x74,0x65,0x20,0x76,0x65,0x72,0x73,
0x69,0x6f,0x6e,0x20,0x20,0x20,0x20,0x20,
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x20,0x20,0x20,0x4a,0x43,0x45,0x20,0x32,
0x30,0x30,0x34,0x0d,0x00,0x14,0x0d,0x00,
0xfd,0x33,0x31,0x39,0x39,0x39,0x0e,0x00,
0x00,0xff,0x7c,0x00,0x0d,0x00,0x1e,0x98,
0x00,0xea,0x4c,0x4f,0x41,0x44,0x20,0x22,
0x4e,0x4f,0x43,0x41,0x43,0x48,0x45,0x2e,
0x42,0x49,0x4e,0x22,0x20,0x43,0x4f,0x44,
0x45,0x20,0x33,0x32,0x30,0x30,0x30,0x3a,
0x20,0x4c,0x45,0x54,0x20,0x58,0x3d,0x20,
0x55,0x53,0x52,0x20,0x33,0x32,0x30,0x30,
0x30,0x3a,0x20,0x52,0x45,0x4d,0x20,0x20,
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x20,0x20,0x20,0x20,0x4c,0x69,0x6e,0x65,
0x20,0x33,0x30,0x20,0x64,0x69,0x73,0x61,
0x62,0x6c,0x65,0x64,0x2c,0x20,0x6e,0x6f,
0x74,0x20,0x20,0x20,0x20,0x20,0x20,0x75,
0x73,0x69,0x6e,0x67,0x20,0x2b,0x33,0x20,
0x6f,0x72,0x20,0x2b,0x32,0x41,0x20,0x64,
0x69,0x73,0x63,0x20,0x20,0x20,0x20,0x20,
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x0d,0x00,0x28,0x1a,0x00,0xef,0x22,0x68,
0x69,0x6c,0x6f,0x61,0x64,0x2e,0x62,0x69,
0x6e,0x22,0xaf,0x33,0x32,0x30,0x30,0x30,
0x0e,0x00,0x00,0x00,0x7d,0x00,0x0d,0x00,
0x32,0x3e,0x00,0xef,0x22,0x72,0x6f,0x6f,
0x6d,0x73,0x2e,0x6a,VAR,0x36,0x22,0xaf,
0x33,0x32,0x37,0x36,0x38,0x0e,0x00,0x00,
0x00,0x80,0x00,0x3a,0xf4,0x33,0x32,0x30,
0x30,0x30,0x0e,0x00,0x00,0x00,0x7d,0x00,
0x2c,0x36,0x0e,0x00,0x00,0x06,0x00,0x00,
0x3a,0xf1,0x58,0x3d,0xc0,0x33,0x32,0x30,
0x30,0x31,0x0e,0x00,0x00,0x01,0x7d,0x00,
0x0d,0x00,0x3c,0x3e,0x00,0xef,0x22,0x72,
0x6f,0x6f,0x6d,0x73,0x2e,0x6a,VAR,0x34,
0x22,0xaf,0x33,0x32,0x37,0x36,0x38,0x0e,
0x00,0x00,0x00,0x80,0x00,0x3a,0xf4,0x33,
0x32,0x30,0x30,0x30,0x0e,0x00,0x00,0x00,
0x7d,0x00,0x2c,0x34,0x0e,0x00,0x00,0x04,
0x00,0x00,0x3a,0xf1,0x58,0x3d,0xc0,0x33,
0x32,0x30,0x30,0x31,0x0e,0x00,0x00,0x01,
0x7d,0x00,0x0d,0x00,0x46,0x3e,0x00,0xef,
0x22,0x72,0x6f,0x6f,0x6d,0x73,0x2e,0x6a,
VAR,0x33,0x22,0xaf,0x33,0x32,0x37,0x36,
0x38,0x0e,0x00,0x00,0x00,0x80,0x00,0x3a,
0xf4,0x33,0x32,0x30,0x30,0x30,0x0e,0x00,
0x00,0x00,0x7d,0x00,0x2c,0x33,0x0e,0x00,
0x00,0x03,0x00,0x00,0x3a,0xf1,0x58,0x3d,
0xc0,0x33,0x32,0x30,0x30,0x31,0x0e,0x00,
0x00,0x01,0x7d,0x00,0x0d,0x00,0x50,0x3e,
0x00,0xef,0x22,0x72,0x6f,0x6f,0x6d,0x73,
0x2e,0x6a,VAR,0x31,0x22,0xaf,0x33,0x32,
0x37,0x36,0x38,0x0e,0x00,0x00,0x00,0x80,
0x00,0x3a,0xf4,0x33,0x32,0x30,0x30,0x30,
0x0e,0x00,0x00,0x00,0x7d,0x00,0x2c,0x31,
0x0e,0x00,0x00,0x01,0x00,0x00,0x3a,0xf1,
0x58,0x3d,0xc0,0x33,0x32,0x30,0x30,0x31,
0x0e,0x00,0x00,0x01,0x7d,0x00,0x0d,0x00,
0x5a,0x54,0x00,0xef,0x22,0x74,0x69,0x74,
0x6c,0x65,0x2e,0x6a,VAR,0x37,0x22,0xaf,
0x33,0x32,0x37,0x36,0x38,0x0e,0x00,0x00,
0x00,0x80,0x00,0x3a,0xf4,0x33,0x32,0x30,
0x30,0x30,0x0e,0x00,0x00,0x00,0x7d,0x00,
0x2c,0x37,0x0e,0x00,0x00,0x07,0x00,0x00,
0x3a,0xf4,0x33,0x32,0x30,0x33,0x30,0x0e,
0x00,0x00,0x1e,0x7d,0x00,0x2c,0x32,0x37,
0x0e,0x00,0x00,0x1b,0x00,0x00,0x3a,0xf1,
0x78,0x3d,0xc0,0x33,0x32,0x30,0x30,0x31,
0x0e,0x00,0x00,0x01,0x7d,0x00,0x0d,0x00,
0x64,0x19,0x00,0xef,0x22,0x72,0x74,0x69,
0x6d,0x65,0x2e,0x6a,VAR,0x32,0x22,0xaf,
0x33,0x32,0x37,0x36,0x38,0x0e,0x00,0x00,
0x00,0x80,0x00,0x0d,0x00,0x6e,0x1a,0x00,
//
// Make tune names generic. 
//   sonata.tun -> title.tun
//   grace.tun  -> main.tun
//   poacher.tun-> cheat.tun
//
0xef,0x20,0x22, 't', 'i', 't', 'l', 'e', //0xef,0x22, 's', 'o', 'n', 'a', 't', 'a',
 '.', 't', 'u', 'n',0x22,0xaf,0x36,0x30,
0x34,0x35,0x38,0x0e,0x00,0x00,0x2a,0xec,
0x00,0x0d,0x00,0x78,0x19,0x00,0xef,0x20, //0x00,0x0d,0x00,0x78,0x19,0x00,0xef,0x22,
0x22, 'm', 'a', 'i', 'n', '.', 't', 'u', //'g', 'r', 'a', 'c', 'e', '.', 't', 'u',
 'n',0x22,0xaf,0x36,0x33,0x32,0x33,0x32,
0x0e,0x00,0x00,0x00,0xf7,0x00,0x0d,0x00,
0x82,0x1b,0x00,0xef,0x22, 'c', 'h', 'e', //0x82,0x1b,0x00,0xef,0x22, 'p', 'o', 'a',
 'a', 't', '.', 't', 'u', 'n',0x22,0x20, //'c', 'h', 'e', 'r', '.', 't', 'u', 'n',
0x20,0xaf,0x36,0x33,0x37,0x34,0x36,0x0e, //0x22,0xaf,0x36,0x33,0x37,0x34,0x36,0x0e,
0x00,0x00,0x02,0xf9,0x00,0x0d,0x00,0x8c,
0x18,0x00,0xef,0x22,0x74,0x75,0x6e,0x65,
0x2e,0x62,0x69,0x6e,0x22,0xaf,0x36,0x34,
0x30,0x30,0x30,0x0e,0x00,0x00,0x00,0xfa,
0x00,0x0d,0x00,0x96,0x7d,0x00,0xea,0x50,
0x41,0x55,0x53,0x45,0x20,0x32,0x30,0x30,
0x3a,0x20,0x52,0x45,0x4d,0x20,0x20,0x20,
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x20,0x20,0x4c,0x69,0x6e,0x65,0x20,0x31,
0x35,0x30,0x20,0x63,0x61,0x6e,0x20,0x62,
0x65,0x20,0x64,0x65,0x6c,0x65,0x74,0x65,
0x64,0x20,0x69,0x66,0x20,0x20,0x6e,0x6f,
0x74,0x20,0x6c,0x6f,0x61,0x64,0x69,0x6e,
0x67,0x20,0x66,0x72,0x6f,0x6d,0x20,0x2b,
0x33,0x20,0x64,0x69,0x73,0x63,0x20,0x20,
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x20,0x20,0x0d,0x00,0xa0,0x10,0x00,0xf1,
0x58,0x3d,0xc0,0x33,0x33,0x37,0x39,0x32,
0x0e,0x00,0x00,0x00,0x84,0x00,0x0d,0x78,
0x00,0x00,0xfd,0x7f,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
};

static signed short jsw_disk_loader[] = 
{
0x50,0x4c,0x55,0x53,0x33,0x44,0x4f,0x53,
0x1a,0x01,0x00,0x3a,0x04,0x00,0x00,0x00,
0xba,0x03,0x00,0x00,0xb4,0x03,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x2a,
0x00,0x0a,0x68,0x00,0xea,0x4a,0x65,0x74,
0x2d,0x53,0x65,0x74,0x20,0x57,0x69,0x6c,
0x6c,0x79,0x20,0x36,0x34,0x20,0x2d,0x20,
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x42,0x41,0x53,0x49,0x43,0x20,0x6c,0x6f,
0x61,0x64,0x65,0x72,0x20,0x20,0x20,0x20,
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x20,0x20,0x20,0x44,0x69,0x73,0x6b,0x65,
0x74,0x74,0x65,0x20,0x76,0x65,0x72,0x73,
0x69,0x6f,0x6e,0x20,0x20,0x20,0x20,0x20,
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x20,0x20,0x20,0x4a,0x43,0x45,0x20,0x32,
0x30,0x30,0x34,0x0d,0x00,0x14,0x0d,0x00,
0xfd,0x33,0x31,0x39,0x39,0x39,0x0e,0x00,
0x00,0xff,0x7c,0x00,0x0d,0x00,0x1e,0x8f,
0x00,0xef,0x22,0x4e,0x4f,0x43,0x41,0x43,
0x48,0x45,0x2e,0x42,0x49,0x4e,0x22,0xaf,
0x33,0x32,0x30,0x30,0x30,0x0e,0x00,0x00,
0x00,0x7d,0x00,0x3a,0xf1,0x58,0x3d,0xc0,
0x33,0x32,0x30,0x30,0x30,0x0e,0x00,0x00,
0x00,0x7d,0x00,0x3a,0xea,0x20,0x20,0x20,
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x20,0x20,0x4c,0x69,0x6e,0x65,0x20,0x33,
0x30,0x20,0x73,0x68,0x6f,0x75,0x6c,0x64,
0x20,0x62,0x65,0x20,0x64,0x69,0x73,0x61,
0x62,0x6c,0x65,0x64,0x20,0x69,0x66,0x20,
0x75,0x73,0x69,0x6e,0x67,0x20,0x2b,0x33,
0x20,0x6f,0x72,0x20,0x2b,0x32,0x41,0x20,
0x64,0x69,0x73,0x63,0x20,0x20,0x20,0x20,
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x0d,
0x00,0x28,0x1a,0x00,0xef,0x22,0x68,0x69,
0x6c,0x6f,0x61,0x64,0x2e,0x62,0x69,0x6e,
0x22,0xaf,0x33,0x32,0x30,0x30,0x30,0x0e,
0x00,0x00,0x00,0x7d,0x00,0x0d,0x00,0x32,
0x3e,0x00,0xef,0x22,0x72,0x6f,0x6f,0x6d,
0x73,0x2e,0x6a, VAR,0x36,0x22,0xaf,0x33,
0x32,0x37,0x36,0x38,0x0e,0x00,0x00,0x00,
0x80,0x00,0x3a,0xf4,0x33,0x32,0x30,0x30,
0x30,0x0e,0x00,0x00,0x00,0x7d,0x00,0x2c,
0x36,0x0e,0x00,0x00,0x06,0x00,0x00,0x3a,
0xf1,0x58,0x3d,0xc0,0x33,0x32,0x30,0x30,
0x31,0x0e,0x00,0x00,0x01,0x7d,0x00,0x0d,
0x00,0x3c,0x3e,0x00,0xef,0x22,0x72,0x6f,
0x6f,0x6d,0x73,0x2e,0x6a, VAR,0x34,0x22,
0xaf,0x33,0x32,0x37,0x36,0x38,0x0e,0x00,
0x00,0x00,0x80,0x00,0x3a,0xf4,0x33,0x32,
0x30,0x30,0x30,0x0e,0x00,0x00,0x00,0x7d,
0x00,0x2c,0x34,0x0e,0x00,0x00,0x04,0x00,
0x00,0x3a,0xf1,0x58,0x3d,0xc0,0x33,0x32,
0x30,0x30,0x31,0x0e,0x00,0x00,0x01,0x7d,
0x00,0x0d,0x00,0x46,0x3e,0x00,0xef,0x22,
0x72,0x6f,0x6f,0x6d,0x73,0x2e,0x6a, VAR,
0x33,0x22,0xaf,0x33,0x32,0x37,0x36,0x38,
0x0e,0x00,0x00,0x00,0x80,0x00,0x3a,0xf4,
0x33,0x32,0x30,0x30,0x30,0x0e,0x00,0x00,
0x00,0x7d,0x00,0x2c,0x33,0x0e,0x00,0x00,
0x03,0x00,0x00,0x3a,0xf1,0x58,0x3d,0xc0,
0x33,0x32,0x30,0x30,0x31,0x0e,0x00,0x00,
0x01,0x7d,0x00,0x0d,0x00,0x50,0x3e,0x00,
0xef,0x22,0x72,0x6f,0x6f,0x6d,0x73,0x2e,
0x6a, VAR,0x31,0x22,0xaf,0x33,0x32,0x37,
0x36,0x38,0x0e,0x00,0x00,0x00,0x80,0x00,
0x3a,0xf4,0x33,0x32,0x30,0x30,0x30,0x0e,
0x00,0x00,0x00,0x7d,0x00,0x2c,0x31,0x0e,
0x00,0x00,0x01,0x00,0x00,0x3a,0xf1,0x58,
0x3d,0xc0,0x33,0x32,0x30,0x30,0x31,0x0e,
0x00,0x00,0x01,0x7d,0x00,0x0d,0x00,0x5a,
0x54,0x00,0xef,0x22,0x74,0x69,0x74,0x6c,
0x65,0x2e,0x6a, VAR,0x37,0x22,0xaf,0x33,
0x32,0x37,0x36,0x38,0x0e,0x00,0x00,0x00,
0x80,0x00,0x3a,0xf4,0x33,0x32,0x30,0x30,
0x30,0x0e,0x00,0x00,0x00,0x7d,0x00,0x2c,
0x37,0x0e,0x00,0x00,0x07,0x00,0x00,0x3a,
0xf4,0x33,0x32,0x30,0x33,0x30,0x0e,0x00,
0x00,0x1e,0x7d,0x00,0x2c,0x32,0x37,0x0e,
0x00,0x00,0x1b,0x00,0x00,0x3a,0xf1,0x78,
0x3d,0xc0,0x33,0x32,0x30,0x30,0x31,0x0e,
0x00,0x00,0x01,0x7d,0x00,0x0d,0x00,0x64,
0x19,0x00,0xef,0x22,0x72,0x74,0x69,0x6d,
0x65,0x2e,0x6a, VAR,0x32,0x22,0xaf,0x33,
0x32,0x37,0x36,0x38,0x0e,0x00,0x00,0x00,
0x80,0x00,0x0d,0x00,0x6e,0x1a,0x00,0xef,
0x20,0x22,0x74,0x69,0x74,0x6c,0x65,0x2e,
0x74,0x75,0x6e,0x22,0xaf,0x36,0x30,0x34,
0x35,0x38,0x0e,0x00,0x00,0x2a,0xec,0x00,
0x0d,0x00,0x78,0x19,0x00,0xef,0x20,0x22,
0x6d,0x61,0x69,0x6e,0x2e,0x74,0x75,0x6e,
0x22,0xaf,0x36,0x33,0x32,0x33,0x32,0x0e,
0x00,0x00,0x00,0xf7,0x00,0x0d,0x00,0x82,
0x1b,0x00,0xef,0x22,0x63,0x68,0x65,0x61,
0x74,0x2e,0x74,0x75,0x6e,0x22,0x20,0x20,
0xaf,0x36,0x33,0x37,0x34,0x36,0x0e,0x00,
0x00,0x02,0xf9,0x00,0x0d,0x00,0x8c,0x18,
0x00,0xef,0x22,0x74,0x75,0x6e,0x65,0x2e,
0x62,0x69,0x6e,0x22,0xaf,0x36,0x34,0x30,
0x30,0x30,0x0e,0x00,0x00,0x00,0xfa,0x00,
0x0d,0x00,0x96,0x7b,0x00,0xf2,0x32,0x30,
0x30,0x0e,0x00,0x00,0xc8,0x00,0x00,0x3a,
0xea,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x4c,0x69,0x6e,0x65,0x20,0x31,0x35,0x30,
0x20,0x63,0x61,0x6e,0x20,0x62,0x65,0x20,
0x64,0x65,0x6c,0x65,0x74,0x65,0x64,0x20,
0x69,0x66,0x20,0x20,0x6e,0x6f,0x74,0x20,
0x6c,0x6f,0x61,0x64,0x69,0x6e,0x67,0x20,
0x66,0x72,0x6f,0x6d,0x20,0x2b,0x33,0x20,
0x64,0x69,0x73,0x63,0x20,0x20,0x20,0x20,
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x0d,
0x00,0xa0,0x10,0x00,0xf1,0x58,0x3d,0xc0,
0x33,0x33,0x37,0x39,0x32,0x0e,0x00,0x00,
0x00,0x84,0x00,0x0d,0x78,0x00,0x00,0xfd,
0x7f,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
}; 
*/

/*
static const jswByte old_jsw128_ldr[] = {
	0x50, 0x4c, 0x55, 0x53, 0x33, 0x44, 0x4f, 0x53,
	0x1a, 0x01, 0x00, 0x18, 0x04, 0x00, 0x00, 0x00,
	0x98, 0x03, 0x00, 0x00, 0x98, 0x03, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xca,
	0x00, 0x0a, 0x68, 0x00, 0xea, 0x4a, 0x65, 0x74,
	0x2d, 0x53, 0x65, 0x74, 0x20, 0x57, 0x69, 0x6c,
	0x6c, 0x79, 0x20, 0x31, 0x32, 0x38, 0x20, 0x2d,
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x42, 0x41, 0x53, 0x49, 0x43, 0x20, 0x6c, 0x6f,
	0x61, 0x64, 0x65, 0x72, 0x20, 0x20, 0x20, 0x20,
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x20, 0x20, 0x20, 0x43, 0x61, 0x73, 0x73, 0x65,
	0x74, 0x74, 0x65, 0x20, 0x76, 0x65, 0x72, 0x73,
	0x69, 0x6f, 0x6e, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x20, 0x20, 0x20, 0x4a, 0x43, 0x45, 0x20, 0x31,
	0x39, 0x39, 0x36, 0x0d, 0x00, 0x14, 0x0d, 0x00,
	0xfd, 0x33, 0x31, 0x39, 0x39, 0x39, 0x0e, 0x00,
	0x00, 0xff, 0x7c, 0x00, 0x0d, 0x00, 0x1e, 0x98,
	0x00, 0xea, 0x4c, 0x4f, 0x41, 0x44, 0x20, 0x22,
	0x4e, 0x4f, 0x43, 0x41, 0x43, 0x48, 0x45, 0x2e,
	0x42, 0x49, 0x4e, 0x22, 0x20, 0x43, 0x4f, 0x44,
	0x45, 0x20, 0x33, 0x32, 0x30, 0x30, 0x30, 0x3a,
	0x20, 0x4c, 0x45, 0x54, 0x20, 0x58, 0x3d, 0x20,
	0x55, 0x53, 0x52, 0x20, 0x33, 0x32, 0x30, 0x30,
	0x30, 0x3a, 0x20, 0x52, 0x45, 0x4d, 0x20, 0x20,
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x20, 0x20, 0x20, 0x20, 0x4c, 0x69, 0x6e, 0x65,
	0x20, 0x33, 0x30, 0x20, 0x64, 0x69, 0x73, 0x61,
	0x62, 0x6c, 0x65, 0x64, 0x2c, 0x20, 0x6e, 0x6f,
	0x74, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x75,
	0x73, 0x69, 0x6e, 0x67, 0x20, 0x2b, 0x33, 0x20,
	0x6f, 0x72, 0x20, 0x2b, 0x32, 0x41, 0x20, 0x64,
	0x69, 0x73, 0x63, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x0d, 0x00, 0x28, 0x1a, 0x00, 0xef, 0x22, 0x68,
	0x69, 0x6c, 0x6f, 0x61, 0x64, 0x2e, 0x62, 0x69,
	0x6e, 0x22, 0xaf, 0x33, 0x32, 0x30, 0x30, 0x30,
	0x0e, 0x00, 0x00, 0x00, 0x7d, 0x00, 0x0d, 0x00,
	0x2a, 0x3e, 0x00, 0xef, 0x22, 0x72, 0x6f, 0x6f,
	0x6d, 0x73, 0x2e, 0x6a, 0x73, 0x36, 0x22, 0xaf,
	0x33, 0x32, 0x37, 0x36, 0x38, 0x0e, 0x00, 0x00,
	0x00, 0x80, 0x00, 0x3a, 0xf4, 0x33, 0x32, 0x30,
	0x30, 0x30, 0x0e, 0x00, 0x00, 0x00, 0x7d, 0x00,
	0x2c, 0x36, 0x0e, 0x00, 0x00, 0x06, 0x00, 0x00,
	0x3a, 0xf1, 0x58, 0x3d, 0xc0, 0x33, 0x32, 0x30,
	0x30, 0x31, 0x0e, 0x00, 0x00, 0x01, 0x7d, 0x00,
	0x0d, 0x00, 0x2c, 0x3e, 0x00, 0xef, 0x22, 0x72,
	0x6f, 0x6f, 0x6d, 0x73, 0x2e, 0x6a, 0x73, 0x34,
	0x22, 0xaf, 0x33, 0x32, 0x37, 0x36, 0x38, 0x0e,
	0x00, 0x00, 0x00, 0x80, 0x00, 0x3a, 0xf4, 0x33,
	0x32, 0x30, 0x30, 0x30, 0x0e, 0x00, 0x00, 0x00,
	0x7d, 0x00, 0x2c, 0x34, 0x0e, 0x00, 0x00, 0x04,
	0x00, 0x00, 0x3a, 0xf1, 0x58, 0x3d, 0xc0, 0x33,
	0x32, 0x30, 0x30, 0x31, 0x0e, 0x00, 0x00, 0x01,
	0x7d, 0x00, 0x0d, 0x00, 0x32, 0x3e, 0x00, 0xef,
	0x22, 0x72, 0x6f, 0x6f, 0x6d, 0x73, 0x2e, 0x6a,
	0x73, 0x33, 0x22, 0xaf, 0x33, 0x32, 0x37, 0x36,
	0x38, 0x0e, 0x00, 0x00, 0x00, 0x80, 0x00, 0x3a,
	0xf4, 0x33, 0x32, 0x30, 0x30, 0x30, 0x0e, 0x00,
	0x00, 0x00, 0x7d, 0x00, 0x2c, 0x33, 0x0e, 0x00,
	0x00, 0x03, 0x00, 0x00, 0x3a, 0xf1, 0x58, 0x3d,
	0xc0, 0x33, 0x32, 0x30, 0x30, 0x31, 0x0e, 0x00,
	0x00, 0x01, 0x7d, 0x00, 0x0d, 0x00, 0x3c, 0x3e,
	0x00, 0xef, 0x22, 0x72, 0x6f, 0x6f, 0x6d, 0x73,
	0x2e, 0x6a, 0x73, 0x31, 0x22, 0xaf, 0x33, 0x32,
	0x37, 0x36, 0x38, 0x0e, 0x00, 0x00, 0x00, 0x80,
	0x00, 0x3a, 0xf4, 0x33, 0x32, 0x30, 0x30, 0x30,
	0x0e, 0x00, 0x00, 0x00, 0x7d, 0x00, 0x2c, 0x31,
	0x0e, 0x00, 0x00, 0x01, 0x00, 0x00, 0x3a, 0xf1,
	0x58, 0x3d, 0xc0, 0x33, 0x32, 0x30, 0x30, 0x31,
	0x0e, 0x00, 0x00, 0x01, 0x7d, 0x00, 0x0d, 0x00,
	0x46, 0x54, 0x00, 0xef, 0x22, 0x74, 0x69, 0x74,
	0x6c, 0x65, 0x2e, 0x6a, 0x73, 0x37, 0x22, 0xaf,
	0x33, 0x32, 0x37, 0x36, 0x38, 0x0e, 0x00, 0x00,
	0x00, 0x80, 0x00, 0x3a, 0xf4, 0x33, 0x32, 0x30,
	0x30, 0x30, 0x0e, 0x00, 0x00, 0x00, 0x7d, 0x00,
	0x2c, 0x37, 0x0e, 0x00, 0x00, 0x07, 0x00, 0x00,
	0x3a, 0xf4, 0x33, 0x32, 0x30, 0x33, 0x30, 0x0e,
	0x00, 0x00, 0x1e, 0x7d, 0x00, 0x2c, 0x32, 0x37,
	0x0e, 0x00, 0x00, 0x1b, 0x00, 0x00, 0x3a, 0xf1,
	0x78, 0x3d, 0xc0, 0x33, 0x32, 0x30, 0x30, 0x31,
	0x0e, 0x00, 0x00, 0x01, 0x7d, 0x00, 0x0d, 0x00,
	0x50, 0x19, 0x00, 0xef, 0x22, 0x72, 0x74, 0x69,
	0x6d, 0x65, 0x2e, 0x6a, 0x73, 0x32, 0x22, 0xaf,
	0x33, 0x32, 0x37, 0x36, 0x38, 0x0e, 0x00, 0x00,
	0x00, 0x80, 0x00, 0x0d, 0x00, 0x51, 0x4b, 0x00,
	0xef, 0x22, 0x73, 0x6f, 0x6e, 0x61, 0x74, 0x61,
	0x2e, 0x74, 0x75, 0x6e, 0x22, 0xaf, 0x36, 0x30,
	0x34, 0x35, 0x38, 0x0e, 0x00, 0x00, 0x2a, 0xec,
	0x00, 0x3a, 0xef, 0x22, 0x67, 0x72, 0x61, 0x63,
	0x65, 0x2e, 0x74, 0x75, 0x6e, 0x22, 0xaf, 0x36,
	0x33, 0x32, 0x33, 0x32, 0x0e, 0x00, 0x00, 0x00,
	0xf7, 0x00, 0x3a, 0xef, 0x22, 0x74, 0x75, 0x6e,
	0x65, 0x2e, 0x62, 0x69, 0x6e, 0x22, 0xaf, 0x36,
	0x34, 0x30, 0x30, 0x30, 0x0e, 0x00, 0x00, 0x00,
	0xfa, 0x00, 0x0d, 0x00, 0x5a, 0x7d, 0x00, 0xea,
	0x50, 0x41, 0x55, 0x53, 0x45, 0x20, 0x32, 0x30,
	0x30, 0x3a, 0x20, 0x52, 0x45, 0x4d, 0x20, 0x20,
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x20, 0x20, 0x20, 0x4c, 0x69, 0x6e, 0x65, 0x20,
	0x39, 0x30, 0x20, 0x63, 0x61, 0x6e, 0x20, 0x62,
	0x65, 0x20, 0x64, 0x65, 0x6c, 0x65, 0x74, 0x65,
	0x64, 0x20, 0x69, 0x66, 0x20, 0x20, 0x6e, 0x6f,
	0x74, 0x20, 0x6c, 0x6f, 0x61, 0x64, 0x69, 0x6e,
	0x67, 0x20, 0x66, 0x72, 0x6f, 0x6d, 0x20, 0x2b,
	0x33, 0x20, 0x64, 0x69, 0x73, 0x63, 0x20, 0x20,
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x20, 0x20, 0x20, 0x0d, 0x00, 0x64, 0x10, 0x00,
	0xf1, 0x58, 0x3d, 0xc0, 0x33, 0x33, 0x37, 0x39,
	0x32, 0x0e, 0x00, 0x00, 0x00, 0x84, 0x00, 0x0d,
};


static const jswByte jswp3_ldr[] = 
{
	0x50, 0x4c, 0x55, 0x53, 0x33, 0x44, 0x4f, 0x53,
	0x1a, 0x01, 0x00, 0xf7, 0x03, 0x00, 0x00, 0x00,
	0x77, 0x03, 0x00, 0x00, 0x71, 0x03, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x60,
	0x00, 0x0a, 0x4e, 0x00, 0xea, 0x4a, 0x65, 0x74,
	0x2d, 0x53, 0x65, 0x74, 0x20, 0x57, 0x69, 0x6c,
	0x6c, 0x79, 0x20, 0x31, 0x32, 0x38, 0x20, 0x2d,
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x42, 0x41, 0x53, 0x49, 0x43, 0x20, 0x6c, 0x6f,
	0x61, 0x64, 0x65, 0x72, 0x20, 0x20, 0x20, 0x20,
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x20, 0x4a, 0x43, 0x45, 0x20, 0x31, 0x39, 0x39,
	0x36, 0x0d, 0x00, 0x14, 0x0d, 0x00, 0xfd, 0x33,
	0x31, 0x39, 0x39, 0x39, 0x0e, 0x00, 0x00, 0xff,
	0x7c, 0x00, 0x0d, 0x00, 0x1e, 0x8b, 0x00, 0xef,
	0x22, 0x4e, 0x4f, 0x43, 0x41, 0x43, 0x48, 0x45,
	0x2e, 0x42, 0x49, 0x4e, 0x22, 0xaf, 0x33, 0x32,
	0x30, 0x30, 0x30, 0x0e, 0x00, 0x00, 0x00, 0x7d,
	0x00, 0x3a, 0xf1, 0x58, 0x3d, 0xc0, 0x33, 0x32,
	0x30, 0x30, 0x30, 0x0e, 0x00, 0x00, 0x00, 0x7d,
	0x00, 0x3a, 0xea, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x52, 0x65, 0x6d, 0x6f, 0x76, 0x65, 0x20, 0x6c,
	0x69, 0x6e, 0x65, 0x20, 0x33, 0x30, 0x20, 0x69,
	0x66, 0x20, 0x6e, 0x6f, 0x74, 0x20, 0x20, 0x20,
	0x20, 0x20, 0x20, 0x6f, 0x6e, 0x20, 0x61, 0x20,
	0x2b, 0x33, 0x20, 0x6f, 0x72, 0x20, 0x2b, 0x32,
	0x41, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x20, 0x0d, 0x00, 0x28, 0x1a, 0x00, 0xef, 0x22,
	0x48, 0x49, 0x4c, 0x4f, 0x41, 0x44, 0x2e, 0x42,
	0x49, 0x4e, 0x22, 0xaf, 0x33, 0x32, 0x30, 0x30,
	0x30, 0x0e, 0x00, 0x00, 0x00, 0x7d, 0x00, 0x0d,
	0x00, 0x32, 0x3e, 0x00, 0xef, 0x22, 0x52, 0x4f,
	0x4f, 0x4d, 0x53, 0x2e, 0x4a, 0x53, 0x36, 0x22,
	0xaf, 0x33, 0x32, 0x37, 0x36, 0x38, 0x0e, 0x00,
	0x00, 0x00, 0x80, 0x00, 0x3a, 0xf4, 0x33, 0x32,
	0x30, 0x30, 0x30, 0x0e, 0x00, 0x00, 0x00, 0x7d,
	0x00, 0x2c, 0x36, 0x0e, 0x00, 0x00, 0x06, 0x00,
	0x00, 0x3a, 0xf1, 0x58, 0x3d, 0xc0, 0x33, 0x32,
	0x30, 0x30, 0x31, 0x0e, 0x00, 0x00, 0x01, 0x7d,
	0x00, 0x0d, 0x00, 0x3c, 0x3e, 0x00, 0xef, 0x22,
	0x52, 0x4f, 0x4f, 0x4d, 0x53, 0x2e, 0x4a, 0x53,
	0x34, 0x22, 0xaf, 0x33, 0x32, 0x37, 0x36, 0x38,
	0x0e, 0x00, 0x00, 0x00, 0x80, 0x00, 0x3a, 0xf4,
	0x33, 0x32, 0x30, 0x30, 0x30, 0x0e, 0x00, 0x00,
	0x00, 0x7d, 0x00, 0x2c, 0x34, 0x0e, 0x00, 0x00,
	0x04, 0x00, 0x00, 0x3a, 0xf1, 0x58, 0x3d, 0xc0,
	0x33, 0x32, 0x30, 0x30, 0x31, 0x0e, 0x00, 0x00,
	0x01, 0x7d, 0x00, 0x0d, 0x00, 0x46, 0x3e, 0x00,
	0xef, 0x22, 0x52, 0x4f, 0x4f, 0x4d, 0x53, 0x2e,
	0x4a, 0x53, 0x33, 0x22, 0xaf, 0x33, 0x32, 0x37,
	0x36, 0x38, 0x0e, 0x00, 0x00, 0x00, 0x80, 0x00,
	0x3a, 0xf4, 0x33, 0x32, 0x30, 0x30, 0x30, 0x0e,
	0x00, 0x00, 0x00, 0x7d, 0x00, 0x2c, 0x33, 0x0e,
	0x00, 0x00, 0x03, 0x00, 0x00, 0x3a, 0xf1, 0x58,
	0x3d, 0xc0, 0x33, 0x32, 0x30, 0x30, 0x31, 0x0e,
	0x00, 0x00, 0x01, 0x7d, 0x00, 0x0d, 0x00, 0x50,
	0x3e, 0x00, 0xef, 0x22, 0x52, 0x4f, 0x4f, 0x4d,
	0x53, 0x2e, 0x4a, 0x53, 0x31, 0x22, 0xaf, 0x33,
	0x32, 0x37, 0x36, 0x38, 0x0e, 0x00, 0x00, 0x00,
	0x80, 0x00, 0x3a, 0xf4, 0x33, 0x32, 0x30, 0x30,
	0x30, 0x0e, 0x00, 0x00, 0x00, 0x7d, 0x00, 0x2c,
	0x31, 0x0e, 0x00, 0x00, 0x01, 0x00, 0x00, 0x3a,
	0xf1, 0x58, 0x3d, 0xc0, 0x33, 0x32, 0x30, 0x30,
	0x31, 0x0e, 0x00, 0x00, 0x01, 0x7d, 0x00, 0x0d,
	0x00, 0x5a, 0x54, 0x00, 0xef, 0x22, 0x54, 0x49,
	0x54, 0x4c, 0x45, 0x2e, 0x4a, 0x53, 0x37, 0x22,
	0xaf, 0x33, 0x32, 0x37, 0x36, 0x38, 0x0e, 0x00,
	0x00, 0x00, 0x80, 0x00, 0x3a, 0xf4, 0x33, 0x32,
	0x30, 0x30, 0x30, 0x0e, 0x00, 0x00, 0x00, 0x7d,
	0x00, 0x2c, 0x37, 0x0e, 0x00, 0x00, 0x07, 0x00,
	0x00, 0x3a, 0xf4, 0x33, 0x32, 0x30, 0x33, 0x30,
	0x0e, 0x00, 0x00, 0x1e, 0x7d, 0x00, 0x2c, 0x32,
	0x37, 0x0e, 0x00, 0x00, 0x1b, 0x00, 0x00, 0x3a,
	0xf1, 0x78, 0x3d, 0xc0, 0x33, 0x32, 0x30, 0x30,
	0x31, 0x0e, 0x00, 0x00, 0x01, 0x7d, 0x00, 0x0d,
	0x00, 0x64, 0x19, 0x00, 0xef, 0x22, 0x52, 0x54,
	0x49, 0x4d, 0x45, 0x2e, 0x4a, 0x53, 0x32, 0x22,
	0xaf, 0x33, 0x32, 0x37, 0x36, 0x38, 0x0e, 0x00,
	0x00, 0x00, 0x80, 0x00, 0x0d, 0x00, 0x6e, 0x4b,
	0x00, 0xef, 0x22, 0x73, 0x6f, 0x6e, 0x61, 0x74,
	0x61, 0x2e, 0x74, 0x75, 0x6e, 0x22, 0xaf, 0x36,
	0x30, 0x34, 0x35, 0x38, 0x0e, 0x00, 0x00, 0x2a,
	0xec, 0x00, 0x3a, 0xef, 0x22, 0x67, 0x72, 0x61,
	0x63, 0x65, 0x2e, 0x74, 0x75, 0x6e, 0x22, 0xaf,
	0x36, 0x33, 0x32, 0x33, 0x32, 0x0e, 0x00, 0x00,
	0x00, 0xf7, 0x00, 0x3a, 0xef, 0x22, 0x74, 0x75,
	0x6e, 0x65, 0x2e, 0x62, 0x69, 0x6e, 0x22, 0xaf,
	0x36, 0x34, 0x30, 0x30, 0x30, 0x0e, 0x00, 0x00,
	0x00, 0xfa, 0x00, 0x0d, 0x00, 0x78, 0x7d, 0x00,
	0xf2, 0x32, 0x30, 0x30, 0x0e, 0x00, 0x00, 0xc8,
	0x00, 0x00, 0x3a, 0xea, 0x20, 0x20, 0x20, 0x20,
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x20, 0x20, 0x20, 0x20, 0x4c, 0x69, 0x6e, 0x65,
	0x20, 0x31, 0x32, 0x30, 0x20, 0x63, 0x61, 0x6e,
	0x20, 0x62, 0x65, 0x20, 0x64, 0x65, 0x6c, 0x65,
	0x74, 0x65, 0x64, 0x20, 0x69, 0x66, 0x20, 0x6e,
	0x6f, 0x74, 0x20, 0x6c, 0x6f, 0x61, 0x64, 0x69,
	0x6e, 0x67, 0x20, 0x66, 0x72, 0x6f, 0x6d, 0x20,
	0x2b, 0x33, 0x20, 0x64, 0x69, 0x73, 0x63, 0x20,
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x20, 0x20, 0x20, 0x20, 0x0d, 0x00, 0x82, 0x10,
	0x00, 0xf1, 0x58, 0x3d, 0xc0, 0x33, 0x33, 0x37,
	0x39, 0x32, 0x0e, 0x00, 0x00, 0x00, 0x84, 0x00,
	0x0d, 0x78, 0x00, 0x00, 0xfd, 0x7f, 0x00, 0xe5,
	0xe5, 0xe5, 0xe5, 0xe5, 0xe5, 0xe5, 0xe5, 0xe5,
};
*/


static const jswByte jsw48_ldr[] = 
{
	0x50, 0x4c, 0x55, 0x53, 0x33, 0x44, 0x4f, 0x53,
	0x1a, 0x01, 0x00, 0xac, 0x00, 0x00, 0x00, 0x00,
	0x2c, 0x00, 0x00, 0x00, 0x2c, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x7c,
	0x00, 0x0a, 0x0d, 0x00, 0xfd, 0x33, 0x32, 0x37,
	0x36, 0x37, 0x0e, 0x00, 0x00, 0xff, 0x7f, 0x00,
	0x0d, 0x00, 0x14, 0x05, 0x00, 0xef, 0x22, 0x22,
	0xaf, 0x0d, 0x00, 0x1e, 0x0e, 0x00, 0xf5, 0xc0,
	0x33, 0x33, 0x37, 0x39, 0x32, 0x0e, 0x00, 0x00,
	0x00, 0x84, 0x00, 0x0d, 
};



static const jswByte hiload_head[128] = 
{
	 'P',  'L',  'U',  'S',  '3',  'D',  'O',  'S',
	0x1A, 0x01, 0x00, 0x2D, 0x1E, 0x00, 0x00, 0x03, /* 10 */
	0xAD, 0x1D, 0x00, 0x00, 0x04, 0x08, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 20 */
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 30 */
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 40 */
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 50 */
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 60 */
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 70 */
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x9C, /* 80 */
};

#define ATTR 9	/* Attribute to use for loading screen */

static const jswByte hiload64[] = 
{
0x00,0x3a,0x5c,0x5b,0x01,0xfd,0x7f,0x5f,
0xe6,0xf8,0x57,0x3a,0x00,0x7d,0xe6,0x07,
0xb2,0xf3,0xed,0x79,0xc5,0xd5,0x21,0x00,
0x80,0x11,0x00,0xc0,0x01,0x00,0x40,0xed,
0xb0,0xd1,0xc1,0xed,0x59,0xfb,0xc9,0xcd,
0xd2,0x7d,0x20,0x09,0xcd,0xae,0x7d,0xcd,
0x9c,0x01,0xcd,0xc0,0x7d,0xf3,0x2a,0x3d,
0x5c,0x23,0x36,0x13,0x2b,0x36,0x03,0x2b,
0x36,0x1b,0x2b,0x36,0x76,0x2b,0x36,0x00,
0x2b,0x36,0x51,0xf9,0xfd,0xcb,0x01,0xa6,
0x3a,0x5c,0x5b,0x01,0xfd,0x7f,0x5f,0xf6,
0x07,0xf3,0xed,0x79,0xc5,0xd5,0x21,0xa8,
0x61,0x11,0x00,0xdb,0x01,0x2a,0x1a,0xed,
0xb0,0x21,0x00,0x40,0x01,0xd6,0x0a,0xed,
0xb0,0xd1,0xc1,0xed,0x59,0xfb,0xc3,0x00,
0x84,0xcd,0xd2,0x7d,0x20,0x0f,0xcd,0xae,
0x7d,0x11,0x00,0x00,0x21,0x80,0x00,0xcd,
0x3f,0x01,0xcd,0xc0,0x7d,0x3e,ATTR,0x21,
0x00,0x58,0x11,0x01,0x58,0x01,0xff,0x02,
0x77,0xed,0xb0,0x32,0x8d,0x5c,0x32,0x8f,
0x5c,0xee,0x07,0x32,0x48,0x5c,0x0f,0x0f,
0x0f,0xe6,0x07,0xd3,0xfe,0xc9,0x3a,0x5c,
0x5b,0x01,0xfd,0x7f,0xcb,0xa7,0xf6,0x07,
0xf3,0x32,0x5c,0x5b,0xed,0x79,0xfb,0xc9,
0x3a,0x5c,0x5b,0x01,0xfd,0x7f,0xe6,0xf8,
0xcb,0xe7,0xf3,0x32,0x5c,0x5b,0xed,0x79,
0xfb,0xc9,0xe5,0xcd,0xe5,0x7d,0x7c,0xe1,
0xfe,0x02,0xc8,0xfe,0x01,0x11,0x3f,0x7f,
0xc2,0x45,0x7e,0xa7,0xc9,0xf5,0x21,0xff,
0xff,0x3a,0x52,0x0b,0xfe,0xc3,0x28,0x22,
0xfe,0xd6,0x28,0x02,0xf1,0xc9,0x3a,0xaa,
0x04,0x2c,0xfe,0xcd,0x28,0xf6,0x2c,0x2c,
0xfe,0x11,0x28,0xf0,0x2c,0xfe,0x3e,0x28,
0xeb,0x2c,0xfe,0xca,0x28,0xe6,0x2e,0xff,
0xf1,0xc9,0x3a,0x53,0x0b,0xfe,0x9f,0x28,
0x19,0xfe,0x7e,0x20,0xd7,0x26,0x02,0x3a,
0x4c,0x00,0x2e,0x01,0xfe,0x38,0x28,0xcc,
0xfe,0x3b,0x2e,0x04,0x28,0xc6,0x2e,0xff,
0xf1,0xc9,0x26,0x01,0x3a,0x40,0x15,0xfe,
0x41,0x2e,0x01,0x28,0xb7,0x2d,0xfe,0x53,
0x28,0xb2,0x2d,0xf1,0xc9,0xed,0x53,0x3d,
0x7f,0xfd,0x21,0x3a,0x5c,0xfb,0xcd,0xe5,
0x7d,0x7c,0xb7,0x28,0x51,0x3d,0xc2,0xe4,
0x7e,0xfd,0xcb,0x01,0x66,0x28,0x47,0x2a,
0x8b,0x5b,0x22,0x3b,0x7f,0x21,0x97,0x03,
0x22,0x39,0x7f,0x21,0x74,0x7e,0x22,0x8b,
0x5b,0xc3,0x1d,0x5b,0x2a,0x3b,0x7f,0x22,
0x8b,0x5b,0xed,0x7b,0xb4,0x5c,0x33,0x21,
0xff,0x5b,0x22,0x81,0x5b,0x76,0xfd,0xcb,
0x01,0xae,0x21,0x66,0x5b,0xcb,0x56,0xca,
0x16,0x7f,0xcd,0x45,0x1f,0xdd,0x2a,0x83,
0x5b,0x01,0x14,0x00,0xdd,0x09,0xcd,0x56,
0x1d,0xcd,0x20,0x1f,0x18,0x70,0x76,0xfd,
0xcb,0x01,0xae,0xfd,0xcb,0x30,0x4e,0xc4,
0xcd,0x0e,0x21,0x00,0x00,0xfd,0x74,0x37,
0xfd,0x74,0x26,0x22,0x0b,0x5c,0x21,0x01,
0x00,0x22,0x16,0x5c,0xcd,0xb0,0x16,0xfd,
0xcb,0x37,0xae,0xcd,0x6e,0x0d,0xfd,0xcb,
0x02,0xee,0xed,0x5b,0x3d,0x7f,0x1a,0xe6,
0x7f,0xcd,0x3b,0x0c,0x1a,0x13,0x87,0x30,
0xf5,0xc3,0x49,0x13,0xfd,0xcb,0x01,0x66,
0x28,0xbc,0x2a,0x6c,0x5b,0x22,0x3b,0x7f,
0x11,0x61,0x00,0x19,0x22,0x39,0x7f,0x21,
0x00,0x7f,0x22,0x6c,0x5b,0xc3,0x3a,0x5b,
0x2a,0x3b,0x7f,0x22,0x6c,0x5b,0xed,0x7b,
0xb4,0x5c,0x33,0x21,0xff,0x5b,0x22,0x6a,
0x5b,0x76,0xfd,0xcb,0x01,0xae,0x21,0x00,
0x00,0xfd,0x74,0x37,0xfd,0x74,0x26,0x22,
0x0b,0x5c,0x23,0x22,0x16,0x5c,0xef,0xb0,
0x16,0xfd,0xcb,0x37,0xae,0xef,0x6e,0x0d,
0xfd,0xcb,0x02,0xee,0xed,0x5b,0x3d,0x7f,
0xc3,0x00,0x00,0x00,0x00,0x00,0x00,0x34,
0x20,0x52,0x65,0x71,0x75,0x69,0x72,0x65,
0x73,0x20,0x31,0x32,0x38,0x6b,0x20,0x6d,
0x65,0x6d,0x6f,0x72,0xf9,
};

/* Old hiload 
static const jswByte hiload_ldr[] =
{
	0x50, 0x4c, 0x55, 0x53, 0x33, 0x44, 0x4f, 0x53,
	0x1a, 0x01, 0x00, 0xa8, 0x00, 0x00, 0x00, 0x03,
	0x28, 0x00, 0x00, 0x7d, 0x00, 0x80, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x48,
	0x00, 0x3a, 0x5c, 0x5b, 0x01, 0xfd, 0x7f, 0x5f,
	0xe6, 0xf8, 0x57, 0x3a, 0x00, 0x7d, 0xe6, 0x07,
	0xb2, 0xf3, 0xed, 0x79, 0xc5, 0xd5, 0x21, 0x00,
	0x80, 0x11, 0x00, 0xc0, 0x01, 0x00, 0x40, 0xed,
	0xb0, 0xd1, 0xc1, 0xed, 0x59, 0xfb, 0xc9, 0x00,
};


static const jswByte nocache_ldr[] =
{
	0x50, 0x4c, 0x55, 0x53, 0x33, 0x44, 0x4f, 0x53,
	0x1a, 0x01, 0x00, 0xa8, 0x00, 0x00, 0x00, 0x03,
	0x28, 0x00, 0x00, 0x7d, 0xf5, 0x80, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x3d,
	0x3a, 0x5c, 0x5b, 0x01, 0xfd, 0x7f, 0xf5, 0xc5,
	0xcb, 0xa7, 0xf6, 0x07, 0xf3, 0x32, 0x5c, 0x5b,
	0xed, 0x79, 0xfb, 0x11, 0x00, 0x00, 0x21, 0x80,
	0x00, 0xcd, 0x3f, 0x01, 0xc1, 0xf1, 0xf3, 0x32,
	0x5c, 0x5b, 0xed, 0x79, 0xfb, 0xc9, 0x00, 0x00,
	0x6f, 0x72, 0x79, 0x21, 0x00, 0x0d, 0x46, 0x6f,
	0x75, 0x6e, 0x64, 0x20, 0x00, 0x52, 0x65, 0x63,
	0x6f, 0x72, 0x64, 0x20, 0x30, 0x30, 0x0d, 0x00,
	0x0d, 0x46, 0x69, 0x6c, 0x65, 0x20, 0x65, 0x78,
	0x69, 0x73, 0x74, 0x73, 0x20, 0x44, 0x65, 0x6c,
	0x65, 0x74, 0x65, 0x28, 0x59, 0x2f, 0x4e, 0x29,
	0x3f, 0x00, 0x0d, 0x46, 0x69, 0x6c, 0x65, 0x20,
	0x41, 0x62, 0x73, 0x65, 0x6e, 0x74, 0x00, 0x0d,
	0x42, 0x61, 0x64, 0x20, 0x4e, 0x61, 0x6d, 0x65,
	0x00, 0x0d, 0x50, 0x72, 0x6f, 0x74, 0x65, 0x63,
	0x74, 0x65, 0x64, 0x00, 0x49, 0x6e, 0x63, 0x6c,
};
*/


/* This structure describes a file that JSWED may find on tape, and where 
 * to load it. Used when loading JSW128 and JSW64 tape files. */
typedef struct valid_bank
{
	char *filename;
	unsigned short address;
	unsigned char  bank;
} VALID_BANK;

/* JSWED 2.1.0: Create a data structure to store the valid filenames on tape
 * (for JSW128/JSW64) and where we load them */
static VALID_BANK validBanks[] =
{
	{ "HILOAD.JS7", 0xDB00, 7 },
	{ "HILOAD.JV7", 0xDB00, 7 },
	{ "HILOAD.JW7", 0xDB00, 7 },
	{ "HILOAD.JX7", 0xDB00, 7 },
	{ "HILOAD.JY7", 0xDB00, 7 },
	{ "HILOAD.JZ7", 0xDB00, 7 },
	{ "HILOAD.J[7", 0xDB00, 7 },
	{ "ROOMS.JS6 ", 0xC000, 6 },
	{ "ROOMS.JS4 ", 0xC000, 4 },
	{ "ROOMS.JS3 ", 0xC000, 3 },
	{ "ROOMS.JS1 ", 0xC000, 1 },
	{ "TITLE.JS7 ", 0xC000, 7 },
	{ "RTIME.JS2 ", 0x8000, 0 },
	{ "ROOMS.JV6 ", 0xC000, 6 },
	{ "ROOMS.JV4 ", 0xC000, 4 },
	{ "ROOMS.JV3 ", 0xC000, 3 },
	{ "ROOMS.JV1 ", 0xC000, 1 },
	{ "TITLE.JV7 ", 0xC000, 7 },
	{ "RTIME.JV2 ", 0x8000, 0 },
	{ "ROOMS.JW6 ", 0xC000, 6 },
	{ "ROOMS.JW4 ", 0xC000, 4 },
	{ "ROOMS.JW3 ", 0xC000, 3 },
	{ "ROOMS.JW1 ", 0xC000, 1 },
	{ "TITLE.JW7 ", 0xC000, 7 },
	{ "RTIME.JW2 ", 0x8000, 0 },
	{ "ROOMS.JX6 ", 0xC000, 6 },
	{ "ROOMS.JX4 ", 0xC000, 4 },
	{ "ROOMS.JX3 ", 0xC000, 3 },
	{ "ROOMS.JX1 ", 0xC000, 1 },
	{ "TITLE.JX7 ", 0xC000, 7 },
	{ "RTIME.JX2 ", 0x8000, 0 },
	{ "ROOMS.JY6 ", 0xC000, 6 },
	{ "ROOMS.JY4 ", 0xC000, 4 },
	{ "ROOMS.JY3 ", 0xC000, 3 },
	{ "ROOMS.JY1 ", 0xC000, 1 },
	{ "TITLE.JY7 ", 0xC000, 7 },
	{ "RTIME.JY2 ", 0x8000, 0 },
	{ "ROOMS.JZ6 ", 0xC000, 6 },
	{ "ROOMS.JZ4 ", 0xC000, 4 },
	{ "ROOMS.JZ3 ", 0xC000, 3 },
	{ "ROOMS.JZ1 ", 0xC000, 1 },
	{ "TITLE.JZ7 ", 0xC000, 7 },
	{ "RTIME.JZ2 ", 0x8000, 0 },
	{ "ROOMS.J[6 ", 0xC000, 6 },
	{ "ROOMS.J[4 ", 0xC000, 4 },
	{ "ROOMS.J[3 ", 0xC000, 3 },
	{ "ROOMS.J[1 ", 0xC000, 1 },
	{ "TITLE.J[7 ", 0xC000, 7 },
	{ "RTIME.J[2 ", 0x8000, 0 },
	{ "TITLE.TUN ", 0xF52A, 7 },
	{ "SONATA.TUN", 0xF52A, 7 },
	{ "GRACE.TUN ", 0xF700, 0 },
	{ "MAIN.TUN  ", 0xF700, 0 },
	{ "CHEAT.TUN ", 0xF902, 0 },
	{ "TUNE.BIN  ", 0xFA00, 0 },
	{ NULL, 0, 0 }
};


static void createLoader(jswByte *target, char c)
{
	int n;
	for (n = 0; n < sizeof(jsw_univ_loader) / sizeof(jsw_univ_loader[0]); n++)
	{
		if (jsw_univ_loader[n] == VAR)	target[n] = c;
		else				target[n] = jsw_univ_loader[n];	
	}
}


static void createHiload(jswByte *target, SpectrumMemory *mem)
{
	memcpy(target,  hiload_head, 128);
	memcpy(target + 128, mem->memoryAt(0xDB00, 7), 7000);
	memcpy(target + 7128, hiload64, sizeof(hiload64));
}




static int tapnamecmp(char *s, char *t)
{
	for (int n = 0; n < 10; n++)
	{
		if (tolower(s[n]) != tolower(t[n])) return -1;
	}
	return 0;
}



static int retError(char *s)
{
	char buf[41];
	sprintf(buf, "%-40.40s", s);

	videoScreen->drawText(0, VIDEO_H - CHAR_H, buf,
                      ZX_BLACK, ZX_WHITE);
        return videoWaitForKey(ZXE_CANCEL);
}

int TapeHeader::read(FILE *fp)
{
	jswByte dlen[2];  
	int datalen;
	char maybe_hdr[19];
	
	if (fread(dlen, 1, 2, fp) < 2) 
	{
		return ZXE_CANCEL;	// End of .TAP file
	}
	datalen = (dlen[0] & 0xFF) + 256 * (dlen[1] & 0xFF);
	if (datalen != 0x13)
	{
		fseek(fp, datalen, SEEK_CUR);
		return ZXE_CONTINUE;
	}
	if (fread(maybe_hdr, 1, 19, fp) < 19)
		return retError("R TAP loading error");
	if (maybe_hdr[0] != 0) return ZXE_CONTINUE;	
	if (maybe_hdr[1] >  3) return ZXE_CONTINUE;
	memcpy(m_data, maybe_hdr + 1, 17);
	return ZXE_OK;
}

int TapeHeader::readData(FILE *fp, jswByte *buf)
{
        jswByte dlen[2];
        int datalen, ch;

        if (fread(dlen, 1, 2, fp) < 2)
		return retError("R TAP loading error");
        datalen = (dlen[0] & 0xFF) + 256 * (dlen[1] & 0xFF);
	if (datalen - 2 != getSize())
	{
		return retError("R TAP loading error");
	}	
	ch = fgetc(fp);	// Block type
	if (ch != 0xFF)
	{
		ungetc(ch, fp);
        	return retError("R TAP loading error");
	}
	if (fread(buf, 1, datalen-2, fp) < (unsigned int)datalen-2)
	{
		return retError("R TAP loading error");
	}
	fgetc(fp);	// Checksum, ignored	
	return ZXE_OK;
}


void TapeHeader::showMe(int y)
{
	char str[40];

	switch (m_data[0])
	{
		case 0: strcpy(str, "Program: ");         break;
		case 1: strcpy(str, "Number array: ");    break;
		case 2: strcpy(str, "Character array: "); break;
		case 3: strcpy(str, "Bytes: ");           break;
	}
	str[10 + strlen(str)] = 0;
	memcpy(str + strlen(str), m_data + 1, 10);
	videoScreen->drawText(0, y, str, ZX_BLACK, ZX_WHITE);	
}



SpectrumMemory::SpectrumMemory()
{
	m_is128 = 0;
	m_type = 0;
	zeroise();
	clearDirty();
}

SpectrumMemory::SpectrumMemory(const char *path, FILE *fp)
{
	m_is128 = 0;
	m_type = 0;
	zeroise();
	fromDisc(path, fp);
	clearDirty();
}

SpectrumMemory::SpectrumMemory(SpectrumMemory *mem, int is128)
{
	m_type = mem->m_type;
	zeroise();

	if (is128< 0) m_is128 = mem->m_is128;
	else	      m_is128 = is128;

	copyFrom(mem);
	clearDirty();
}

void SpectrumMemory::zeroise()
{
	memset(m_rom,      0, sizeof(m_rom));
	memset(m_oldrom,   0, sizeof(m_oldrom));
	memset(m_ram48,    0, sizeof(m_ram48));
	memset(m_oldram48, 0, sizeof(m_oldram48));
	memset(m_ram1,     0, sizeof(m_ram1));
	memset(m_oldram1,  0, sizeof(m_oldram1));
	memset(m_ram3,     0, sizeof(m_ram3));
	memset(m_oldram3,  0, sizeof(m_oldram3));
	memset(m_ram4,     0, sizeof(m_ram4));
	memset(m_oldram4,  0, sizeof(m_oldram4));
	memset(m_ram6,     0, sizeof(m_ram6));
	memset(m_oldram6,  0, sizeof(m_oldram6));
	memset(m_ram7,     0, sizeof(m_ram7));
	memset(m_oldram7,  0, sizeof(m_oldram7));
}

void SpectrumMemory::copyFrom(SpectrumMemory *mem)
{
	memcpy(m_ram48,   mem->m_ram48,   49152);
	memcpy(m_snahdr,  mem->m_snahdr,     27);

	if (m_is128 && mem->m_is128)
	{
		memcpy(m_ram1, mem->m_ram1, 16384);
                memcpy(m_ram3, mem->m_ram3, 16384);
                memcpy(m_ram4, mem->m_ram4, 16384);
                memcpy(m_ram6, mem->m_ram6, 16384);
                memcpy(m_ram7, mem->m_ram7, 16384);
	}
	else if (m_is128)
	{
		memset(m_ram1, 0, 16384);
                memset(m_ram3, 0, 16384);
                memset(m_ram4, 0, 16384);
                memset(m_ram6, 0, 16384);
                memset(m_ram7, 0, 16384);
	}	
	clearDirty();
}



SpectrumMemory::~SpectrumMemory()
{

}

void SpectrumMemory::setDirty()
{
	m_forceDirty = true;
}

void SpectrumMemory::clearDirty()
{
	m_forceDirty = false;
	memcpy(m_oldram48, m_ram48,   49152);

	if (m_is128)
	{
		memcpy(m_oldram1, m_ram1, 16384);
                memcpy(m_oldram3, m_ram3, 16384);
                memcpy(m_oldram4, m_ram4, 16384);
                memcpy(m_oldram6, m_ram6, 16384);
                memcpy(m_oldram7, m_ram7, 16384);
	}
}


bool SpectrumMemory::isDirty()
{
	if (m_forceDirty) return true;

	if (memcmp(m_oldram48, m_ram48,   49152)) return true;

	if (m_is128)
	{
		if (memcmp(m_oldram1, m_ram1, 16384)) return true;
		if (memcmp(m_oldram3, m_ram3, 16384)) return true;
		if (memcmp(m_oldram4, m_ram4, 16384)) return true;
		if (memcmp(m_oldram6, m_ram6, 16384)) return true;
		if (memcmp(m_oldram7, m_ram7, 16384)) return true;
	}
	return false;
}


int SpectrumMemory::fromDisc(const char *path, FILE *fp)
{
	// Determine file size, by seeking to the end and back.
	long p0, p1;
	jswByte hdr[128];

	p0 = ftell(fp);
	fseek(fp, 0, SEEK_END);
	p1 = ftell(fp);
	fseek(fp, p0, SEEK_SET);

	int rv = fread(hdr, 1, sizeof(hdr), fp);
	if (rv < (int)sizeof(hdr)) memset(hdr + rv, 0x1a, sizeof(hdr) - rv);
	
	if (isHeaderPlus3Dos(hdr)) // Raw binary, or ZXT file
	{
		if (memcmp(hdr + 15, "TAPEFILE", 8)) m_type = 'O';	
		else				     m_type = 'T';
	}
	// SZX
	else if (!memcmp(hdr, "ZXST", 4))	     m_type = 'Y';
	else	// SNA, TAP or Z80
	{
		// Undo header read
		fseek(fp, p0, SEEK_SET);	
	
		// Guess type	
		if (hdr[0] == 0x13 && hdr[1] == 0x00 && hdr[2] == 0x00) 
			m_type = 'T';
		else if ((p1 - p0) == 49179 ||  (p1 - p0) == 49280) 
			m_type = 'N';
		else if ((p1 - p0) == 131103) 
			m_type = 'n';
		else m_type = 'Z';
	}
	switch(m_type)
	{
		case 'O': return loadPlus3(path, fp, hdr);
		case 'T': return loadTap(path, fp);
		case 'N': return loadSna(path, fp);
		case 'n': return loadSna128(path, fp);
		case 'Y': return loadSZX(path, fp);
		case 'Z': return loadZ80(path, fp);
	}
	return ZXE_CANCEL;
}


int SpectrumMemory::loadSna(const char *path, FILE *fp)
{
	if (fread(m_snahdr, 1, sizeof(m_snahdr), fp) < sizeof(m_snahdr)) 
		return ZXE_CANCEL;
	if (fread(m_ram48,  1, sizeof(m_ram48),  fp) < sizeof(m_ram48))
		return ZXE_CANCEL;

	m_snahdr[23] = 0xFE;
	m_snahdr[24] = 0x5B;

	m_is128	= 0;
	return ZXE_OK;
}



int SpectrumMemory::loadSna128(const char *path, FILE *fp)
{
	jswByte dummy[4];

	if (fread(m_snahdr, 1, sizeof(m_snahdr), fp) < sizeof(m_snahdr)) 
		return ZXE_CANCEL;
	if (fread(m_ram48,  1, sizeof(m_ram48),  fp) < sizeof(m_ram48))
		return ZXE_CANCEL;
        if (fread(dummy, 1, sizeof(dummy), fp) < sizeof(dummy))
                return ZXE_CANCEL;
        if (fread(memoryAt(0xC000, 1), 1, 0x4000, fp) < 0x4000)
                return ZXE_CANCEL;
        if (fread(memoryAt(0xC000, 3), 1, 0x4000, fp) < 0x4000)
                return ZXE_CANCEL;
        if (fread(memoryAt(0xC000, 4), 1, 0x4000, fp) < 0x4000)
                return ZXE_CANCEL;
        if (fread(memoryAt(0xC000, 6), 1, 0x4000, fp) < 0x4000)
                return ZXE_CANCEL;
        if (fread(memoryAt(0xC000, 7), 1, 0x4000, fp) < 0x4000)
                return ZXE_CANCEL;

	m_snahdr[23] = 0xFE;
	m_snahdr[24] = 0x5B;

	m_type = 'N';
	m_is128	= 1;
	return ZXE_OK;
}


// fp is file containing the BASIC loader.
int SpectrumMemory::loadPlus3(const char *pathname, FILE *fp, jswByte *hdr)
{
	char filename[8][20];
	char variant;

	memcpy(m_snahdr, snaTemplate, 27);

       if ((hdr[15] == 3 && hdr[16] == 0    && hdr[17] == 0x80) ||
           (hdr[15] == 3 && hdr[16] == 0xFF && hdr[17] == 0x7F))
        {
	        if (fread(memoryAt(0x8000),  1, 
			sizeof(m_ram48),  fp) < 0x7FFF) return ZXE_CANCEL;
		
		memset(m_ram48, 0, 0x4000);
		m_is128 = 0;
		return ZXE_OK;
        }
/* JSW128 pre HL9 */
	if ( (hdr[15] == 0 && hdr[16] == 0x98 && hdr[17] == 0x03) ||
	     (hdr[15] == 0 && hdr[16] == 0x77 && hdr[17] == 0x03))
	{
                 if (read3Dos(pathname,"rooms.js6", 0x4000,0xC000,memoryAt(0xC000, 6))
                 &&  read3Dos(pathname,"rooms.js4", 0x4000,0xC000,memoryAt(0xC000, 4))
                 &&  read3Dos(pathname,"rooms.js3", 0x4000,0xC000,memoryAt(0xC000, 3))
                 &&  read3Dos(pathname,"rooms.js1", 0x4000,0xC000,memoryAt(0xC000, 1))
                 &&  read3Dos(pathname,"title.js7" ,0x1B00,0xC000,memoryAt(0xC000, 7))
                 &&  read3Dos(pathname,"rtime.js2", 0x6C80,0x8000,memoryAt(0x8000))
                 &&  read3Dos(pathname,"sonata.tun",0x06AE,0x752A,memoryAt(60458))
                 &&  read3Dos(pathname,"grace.tun", 0x0100,0xF700,memoryAt(0xF700))
                 &&  read3Dos(pathname,"tune.bin",  0x0600,0xFA00,memoryAt(0xFA00 )))

		m_is128 = 1;
		return ZXE_OK;  
	}
/* JSW128 HL9+, and JSW64 */
	if ((hdr[15] == 0 && hdr[16] == 0xBA && hdr[17] == 0x03) || 
	    (hdr[15] == 0 && hdr[16] == 0xC5 && hdr[17] == 0x03))
	{
		unsigned ldrLen = hdr[17] * 256 + hdr[16];
		int n;
		jswByte *loader = new jswByte[ldrLen];
//
// To distinguish between JSW128 and the 6 variants of JSW64, pull in the
// loader and extract a variant number.
//
		if (fread(loader, 1, ldrLen, fp) < ldrLen) 
		{
			videoScreen->drawText(0, VIDEO_H - CHAR_H, "Can't load the BASIC loader",
                                ZX_BLACK, ZX_WHITE);
			delete loader;
			return ZXE_CANCEL;
		}
// Find a reference to a variant-dependent filename.
		for (n = 0; n < ldrLen - 10; n++)
		{
			if (!memcmp(loader + n, "rooms.", 6)) break;
		}
		variant = loader[n + 7];
		sprintf(filename[0], "rooms.j%c6", variant);	
		sprintf(filename[1], "rooms.j%c4", variant);	
		sprintf(filename[2], "rooms.j%c3", variant);	
		sprintf(filename[3], "rooms.j%c1", variant);	
		sprintf(filename[4], "title.j%c7", variant);	
		sprintf(filename[5], "rtime.j%c2", variant);	
		sprintf(filename[6], "hiload.j%c7", variant);	

                 if (read3Dos(pathname,filename[0], 0x4000,0xC000,memoryAt(0xC000, 6))
                 &&  read3Dos(pathname,filename[1], 0x4000,0xC000,memoryAt(0xC000, 4))
                 &&  read3Dos(pathname,filename[2], 0x4000,0xC000,memoryAt(0xC000, 3))
                 &&  read3Dos(pathname,filename[3], 0x4000,0xC000,memoryAt(0xC000, 1))
                 &&  read3Dos(pathname,filename[4] ,0x1B00,0xC000,memoryAt(0xC000, 7))
                 &&  read3Dos(pathname,filename[5], 0x8000,0x8000,memoryAt(0x8000))
                 &&  read3Dos(pathname,"title.tun", 0x06AE,0x752A,memoryAt(0xF52A, 7), 0xAD6)
                 &&  read3Dos(pathname,"main.tun",  0x0100,0xF700,memoryAt(0xF700))
                 &&  read3Dos(pathname,"cheat.tun", 0x00FE,0xF902,memoryAt(0xF902))
                 &&  read3Dos(pathname,"tune.bin",  0x0600,0xFA00,memoryAt(0xFA00))
		 &&  read3Dos(pathname,filename[6], 7000,  0xDB00,memoryAt(0xDB00, 7))
		 )
		{
			m_is128 = 1;
			return ZXE_OK;  
		}
	}
	videoScreen->drawText(0, VIDEO_H - CHAR_H, "Can't identify this file",
                                ZX_BLACK, ZX_WHITE);
	return ZXE_CANCEL;
}

#define TRUNCATED ("8 End of file (truncated Z80 file)")

#define REFILLBUF \
                for (n = 0; n < 4; n++)				\
                {						\
                        ch = fgetc(fp);				\
                        if (ch == EOF) return retError TRUNCATED;	\
                        z80buf[n] = ch;				\
                }


int SpectrumMemory::loadZ80(const char *path, FILE *fp)
{
	int z80base;
	jswByte zhdr1[30], z80buf[6];

        memcpy(m_snahdr, snaTemplate, 27);

	if (fread(zhdr1, 1, 30, fp) < 30) return retError TRUNCATED;

	if (zhdr1[6] || zhdr1[7])	// v1 Z80 
	{
                int n, addr;
		int ch;
                jswByte o,rlelen,p;

		m_is128 = 0;
		REFILLBUF
		addr = 0x4000;
		if (zhdr1[12] & 0x20)	// Compressed?
		{
			while (addr < 0x10000)
			{
  				if ((z80buf[0]==0xED) 
  				&&  (z80buf[1]==0xED))  // RLE signature
	    			{
					rlelen=z80buf[2];
					p=z80buf[3];

					for(o=rlelen;o>0;o--)
					{
						poke(addr++, p);
					}
					REFILLBUF
    				}
  				else
    				{
					poke(addr++, z80buf[0]);
    					for(n=0;n<3;n++)
    						z80buf[n]=z80buf[n+1];  // Do 1 byte 
					ch = fgetc(fp);
					if (ch == EOF) return retError(TRUNCATED);
					z80buf[3] = ch;
    				}
  			}
/*
			if (z80buf[0] || z80buf[3] || (z80buf[1]!=0xED) || 
			    (z80buf[2]!=0xED))
				fprintf(stderr,"Warning: End-of-Z80-file signature not found.\n"); */
		}
		else  // Not compressed 
		{
			if (fread(m_ram48, 1, 49152, fp) < 49152) 
				return retError(TRUNCATED);
		}
	}
	else // Later .Z80 than v1 
	{
		int ch, hdrlen, z80ver;

		hdrlen = fgetc(fp);
		ch     = fgetc(fp);
		if (hdrlen == EOF || ch == EOF) return retError(TRUNCATED);
			
		hdrlen += (ch << 8);
		switch(hdrlen)
		{	
			case 23: // v2 Z80 
			z80ver=2;
			break;
			
			default:
/*
			fprintf(stderr,"Warning: Z80 file of unknown type "
                                       "(header length %d).\nI hope "
                                       "it's similar to v3.x...\n",hdrlen);
*/
			case 54: // v3 Z80 
			z80ver=3;
			break;
		}
		// Read PC and hardware type 
		for (int n = 0; n < 3; n++)
		{
			ch = fgetc(fp); if (ch == EOF) return retError(TRUNCATED);
		}
		if ((z80ver == 2) && (ch > 1)) ++ch; // Translate to Z80v3 num.
/*		if (ch == 3)
			fprintf(stderr,"Warning: This file uses the SamRam.\n");
*/		if (ch > 3) m_is128 = 1;	// 128k
		else	    m_is128 = 0;

		hdrlen -= 3;	// Skip remainder of Z80 header
		while (hdrlen--)
		{
			ch = fgetc(fp);
			if (ch == EOF) return retError(TRUNCATED); 
		}
		z80base = ftell(fp);
		if (m_is128)
		{
			if (loadZ80Bank(8, 0x4000, 0, fp, z80base) || 
			    loadZ80Bank(5, 0x8000, 0, fp, z80base) ||
			    loadZ80Bank(3, 0xC000, 0, fp, z80base) || 
                            loadZ80Bank(4, 0xC000, 1, fp, z80base) || 
                            loadZ80Bank(6, 0xC000, 3, fp, z80base) || 
                            loadZ80Bank(7, 0xC000, 4, fp, z80base) || 
                            loadZ80Bank(9, 0xC000, 6, fp, z80base) || 
			    loadZ80Bank(10,0xC000, 7, fp, z80base)) return retError(TRUNCATED);
		}
		else
		{
                        if (loadZ80Bank(8, 0x4000, 0, fp, z80base) ||
                            loadZ80Bank(4, 0x8000, 0, fp, z80base) ||
                            loadZ80Bank(5, 0xC000, 0, fp, z80base)) return retError(TRUNCATED);
		}	
	}			
	return ZXE_OK;
}


#undef REFILLBUF
#define REFILLBUF \
                for (n = 0; n < 4; n++)                         \
                {                                               \
                        ch = fgetc(fp);                         \
                        if (ch == EOF) return -1;               \
                        z80buf[n] = ch;                         \
                }


int SpectrumMemory::loadZ80Bank(int z80Id, int zxAddr, int zxBank, FILE *fp,
				int z80base)
{
	int m1, n, p, ch;
	jswByte o, rlelen, z80buf[4];
	jswByte *dest = memoryAt(zxAddr, zxBank);

	if (z80Seek(z80Id, fp,z80base)) return -1;

	REFILLBUF

        for(m1 = 0; m1< 0x4000;)   /* m=Spectrum address */
        {
                if ((z80buf[0]==0xED)
                &&  (z80buf[1]==0xED))  /* RLE signature */
                {
                        rlelen=z80buf[2];
                        p=z80buf[3];

                        for(o=rlelen;o>0;o--) dest[m1++] = p;
			if (m1 == 0x4000) return 0;
                        REFILLBUF
                }
                else
                {
                        dest[m1++]=z80buf[0];
                        for(n=0;n<3;n++)
                                z80buf[n]=z80buf[n+1];  /* Do 1 byte */
                        ch = fgetc(fp);
			if (ch == EOF) return -1;
			z80buf[3] = ch;
                }
        }
	return 0;
}

int SpectrumMemory::z80Seek(int bank,FILE *fp, int z80base)
{
        unsigned long zpos=z80base,banklen;
	int ll, lh, bn;

        while (!feof(fp))
        {
                if(fseek(fp,zpos,SEEK_SET)<0) return -1;
		ll = fgetc(fp);
		lh = fgetc(fp);
		bn = fgetc(fp);
		if (ll == EOF || lh == EOF || bn == EOF) return -1;
                banklen = (ll + 256 * lh);
                if (bank == bn) return 0;
                zpos += (banklen + 3);
        }
        return -1;
}
			


int SpectrumMemory::loadTap(const char *path, FILE *fp)
{
	TapeHeader th;
	int rv;
	int y = 0;
	int n;

        memcpy(m_snahdr, snaTemplate, 27);
	
	do
	{
		rv = th.read(fp);
		if (rv >= ZXE_CANCEL) break;
		if (rv == ZXE_CONTINUE) continue;
		th.showMe(y); 
		y += CHAR_H; 
		if (y + 3 * CHAR_H > VIDEO_H) y = 0; 
		if (th.getType() != 3)
		{
			rv = ZXE_CONTINUE;
			continue;
		}
		// We have the header of a code file

		// Check for JSW128 tape blocks, by name. We ignore the
		// loader completely.
		char *s = th.getName();

		// Check for JSW2 by name
		if (!tapnamecmp(s, "jsw\252      "))
		{
			rv = th.readData(fp, memoryAt(0x4000));
			if (rv == ZXE_OK) rv = ZXE_CONTINUE;
			continue;
		}
		if (!tapnamecmp(s, "JS1       "))
		{
			printf("Loading %s at %04x len %04x\n",
				s, th.getAddress(), th.getSize());
			rv = th.readData(fp, memoryAt(th.getAddress()));
			if (rv == ZXE_OK) rv = ZXE_CONTINUE;
			continue;
		}
		if (!tapnamecmp(s, "JS2       "))
		{
			printf("Loading %s at %04x len %04x\n",
				s, th.getAddress(), th.getSize());
			rv = th.readData(fp, memoryAt(th.getAddress()));
			continue;	
		}

		// JSW48 has a single 32k block. But new JSW128/JSW64 runtimes
		// are also a single 32k block. We distinguish between them 
		// by checking for a "RTIME" name.
		if ((tapnamecmp(s, "RTIME.JS2 ")) &&
		   (tapnamecmp(s, "RTIME.JV2 ")) &&
		   (tapnamecmp(s, "RTIME.JW2 ")) &&
		   (tapnamecmp(s, "RTIME.JX2 ")) &&
		   (tapnamecmp(s, "RTIME.JY2 ")) &&
		   (tapnamecmp(s, "RTIME.JZ2 ")) &&
		   (tapnamecmp(s, "RTIME.J[2 ")) &&
		   (th.getSize() == 0x8000 || th.getSize() == 0x7FFF)) // JSW48
		{
			rv = th.readData(fp, memoryAt(0x8000));
			continue;	
		}
		// v2.1.0: Use an array of valid name structures rather than
		// hardcoding each case.
		for (n = 0; validBanks[n].filename; n++)
		{
			if (!tapnamecmp(s, validBanks[n].filename)) 
			{
				m_is128 = 1;
				rv = th.readData(fp, memoryAt(validBanks[n].address, validBanks[n].bank));
// In HL8 and below, TUNE.BIN is the last file, so stop when we read it.
				if (peek(0x8C4B) == 0x1E)
				{
					if (rv == ZXE_OK && (tapnamecmp(s, "TUNE.BIN  ")))
						rv = ZXE_CONTINUE;		
				}
// Otherwise, the last file is TITLE.TUN
				else
				{
					if (rv == ZXE_OK && (tapnamecmp(s, "TITLE.TUN ")))
						rv = ZXE_CONTINUE;		
				}
				break;
			}
		}
		if (!validBanks[n].filename) // not matched
		{
			rv = ZXE_CONTINUE;
		}
	} while (rv == ZXE_CONTINUE);

	if (rv == ZXE_CANCEL)
	{
		return retError("8 End of file (no JSW data found)");
	}
// JSW128 hacklevels lower than 9 load the tune at 0EC2Ah, bank 0, so copy
// the definition in there. 
	if (m_is128 && peek(0x8C4B) == 0x1E)
	{
		memcpy(memoryAt(0xEC2A), 
			memoryAt(0xF52A, 7), 2774);

	}
        return rv;
}





int SpectrumMemory::toDisc(const char *path, FILE *fp)
{
	int rv = toDiscInternal(path, fp);
	if (rv == ZXE_OK) clearDirty();
	return rv;
}

int SpectrumMemory::toDiscInternal(const char *path, FILE *fp)
{
	int err;
	VideoMenu vm("Save as", "+3DOS", "TAP", "SNA", "SZX", "Z80", NULL);
	
	do
	{
	        switch(m_type)
       		{
               		case 'O': return savePlus3(path, fp);
                	case 'T': return saveTap(path, fp);
                	case 'N': return saveSna(path, fp);
                	case 'Y': return saveSZX(path, fp);
                	case 'Z': return saveZ80(path, fp);
		}
		err = vm.doModal();
		if (err != ZXE_OK) return err;
		switch(vm.getSelected())
		{
			case 0: m_type = 'O'; break;
			case 1: m_type = 'T'; break;
			case 2: m_type = 'N'; break;
			case 3: m_type = 'Y'; break;
			case 4: m_type = 'Z'; break;
		}

        } while(1);
 
	return ZXE_CANCEL;
}



int SpectrumMemory::savePlus3(const char *path, FILE *fp)
{
	jswByte header[128];
	char hiload7[20];
	char room6[20];
	char room4[20];
	char room3[20];
	char room1[20];
	char title7[20];
	char rtime2[20];
	char variant;

	if (m_is128)
	{
		const char *ldrname = path;
		char *s;
		jswByte jswp3_ldr[sizeof(jsw_univ_loader) / sizeof(jsw_univ_loader[0])];
		jswByte hiload[7725];
		switch(peek(0x85C9))
		{
			case 'V': case 'W': case 'X': case 'Y': case 'Z': 
			case '[':
				variant = tolower(peek(0x85C9));
				break;
			default:
				variant = 's';
				break;
		}
		
		createLoader(jswp3_ldr, variant);
		createHiload(hiload, this);

		s = strrchr(ldrname, '/'); if (s) ldrname = s+1;
#ifdef _WIN32
		s = strrchr(ldrname, '\\'); if (s) ldrname = s+1;
		s = strrchr(ldrname, ':'); if (s) ldrname = s+1;
#endif	
	
		sprintf(room6, "rooms.j%c6", variant);
		sprintf(room4, "rooms.j%c4", variant);
		sprintf(room3, "rooms.j%c3", variant);
		sprintf(room1, "rooms.j%c1", variant);
		sprintf(rtime2, "rtime.j%c2", variant);
		sprintf(title7, "title.j%c7", variant);
		sprintf(hiload7, "hiload.j%c7", variant);

                if (write3Dos(path,ldrname,jswp3_ldr, jswp3_ldr + 128)
                 && write3Dos(path,hiload7,hiload, hiload + 128)
// Now part of hiload.bin
//		 && write3Dos(path,"nocache.bin",nocache_ldr, nocache_ldr+128)
                 && write3Dos(path, room6,0x4000,0xC000,memoryAt(0xC000, 6))
                 && write3Dos(path, room4,0x4000,0xC000,memoryAt(0xC000, 4))
                 && write3Dos(path, room3,0x4000,0xC000,memoryAt(0xC000, 3))
                 && write3Dos(path, room1,0x4000,0xC000,memoryAt(0xC000, 1))
                 && write3Dos(path, title7, 0x1B00,0xC000,memoryAt(0xC000, 7))
                 && write3Dos(path, rtime2, 0x8000,0x8000,memoryAt(0x8000))
                 && write3Dos(path,"title.tun",0x0AD6,0x752A,memoryAt(0xF52A, 7))
                 && write3Dos(path,"main.tun", 0x0100,0xF700,memoryAt(0xF700))
                 && write3Dos(path,"cheat.tun",0x00FE,0xF902,memoryAt(0xF902))
                 && write3Dos(path,"tune.bin", 0x0600,0xFA00,memoryAt(0xFA00)))
			return ZXE_OK;
		return ZXE_CANCEL;
	}
	else
	{
		constructHeader(header, 3, 0x8000, 0x8000);
		if (fwrite(header, 1, sizeof(header), fp) < sizeof(header))
			return ZXE_CANCEL;
                if (fwrite(memoryAt(0x8000),  1, 0x8000,  fp) < 0x8000) 
			return ZXE_CANCEL;
	}
        return ZXE_OK;
}


/* Based in the routing in m3conv4 by Ian Cull */
static int writeZ80Rle(jswByte id, jswByte *page, FILE *outfp)
{
	/* Note the big buffer. Pathological RLEing can occur! */
	jswByte *source, *dest, buf[32768], byte;
        unsigned n,length,tel;

	memset(buf, 0, sizeof(buf));	// To shut valgrind up
/* Compress the 16K block of memory */
	byte = 0;	
        length = 0x4000;
        source = page;
        dest = buf;
        tel = 0;
        do
        {
//		fprintf(stderr, "id=%d source=%p length=%d\n", 
//				id, source, length);
                if (tel==0)
                {
                        byte=*source++;
                        tel=1;
                        length--;
                }
                else if ((tel<255)&&(byte==*source)&&(length!=0))
                {
                        source++;
			tel++;
                        length--;
                }
                else if ((tel>=5)||((byte==0xED)&&(tel>=2)))
                {
                        *dest++ = 0xED;
                        *dest++ = 0xED;
                        *dest++ = (jswByte)tel;
                        *dest++ = byte;
                        tel=0;
                }
                else
                {
                        for (n=0;n<tel;n++) *dest++ = byte;
                        tel=0;
                        if ((byte==0xED)&&(length!=0))
                        {
                                *dest++ = *source++;
                                length--;
                        }
                }
        } while ((length!=0)||(tel!=0));

/* Calculate new length of compressed block */

        length = dest-buf;

/* Write block header information */
        fputc((jswByte)(length & 0xFF),outfp);             /* length */
        fputc((jswByte)(length >> 8),outfp);
        fputc((jswByte)id,outfp);                    /* page number */

/* Write the block */
//	fprintf(stderr, "-- id=%d buf=%p length=%d\n", 
//				id, buf, length);
	if (fwrite(buf, 1, length, outfp) < length)
	{
		return retError("Disc full");
	}
	return ZXE_OK;
}

int SpectrumMemory::saveZ80(const char *path, FILE *fp)
{
	int n;
	const int *pages, *banks, *addrs;
	jswByte z80Header[55];

	memset(z80Header, 0, sizeof(z80Header));
	/* Generate Z80 header */
	for (n = 0; n < (int)sizeof(z80Map); n++) 
		z80Header[z80Map[n]] = m_snahdr[n]; 
	z80Header[28] = z80Header[27];
	z80Header[6]  = z80Header[7] = 0;	/* v2 file */
	z80Header[30] = 23;			/* v2 file */

	/* Pop PC manually */
	n = z80Header[8] + 256 * z80Header[9];
	z80Header[32] = peek(n++);
	z80Header[33] = peek(n++);

	z80Header[9] = n / 256;
	z80Header[8] = n % 256;
	z80Header[34] = is128() ? 3 : 0;
	z80Header[35] = 0x10;	/* OUT 7FFD */

	/* Set up the AY registers as BASIC leaves them - all zeroes 
	 * except for number 7 */
	z80Header[46] = 0xFF;

        if (fwrite(z80Header, 1, sizeof(z80Header), fp) < sizeof(z80Header))
                return retError("Disc full");
	/* Now write the Z80 file itself */
	if (is128()) { pages = pages128; banks = banks128; addrs = addrs128; }
	else 	     { pages = pages48;  banks = banks48;  addrs = addrs48;  }
	for (n = 0; pages[n] >= 0; n++)
	{
		jswByte *page = memoryAt(addrs[n], banks[n]);

		int rv = writeZ80Rle(pages[n], page, fp);	
		if (rv != ZXE_OK) return rv;
	}
        return ZXE_OK;
}


static int read3Dos(const char *path, const char *name, int minLen, int addr, 
		jswByte *loadAt, int maxLen)
{
        char fbuf[PATH_MAX];
        strcpy(fbuf, path);
	jswByte header[128], n, sum;
	int hlen, haddr;

	if (maxLen < 0) maxLen = minLen;
        char *s = strrchr(fbuf, '/');
        if (s) *s = 0;
#ifdef _WIN32
        else
        {
                s = strrchr(fbuf, '\\');
                if (s) *s = 0;
                else { s = strrchr(fbuf, ':'); if (s) s[1] = 0; }
        }
#endif
	// No path separator at all. So there's no path, just filenames.
	if (!s) strcpy(fbuf, "");
	else	strcat(fbuf, "/");
        strcat(fbuf, name);

//	videoScreen->drawSmallText(0, 0, fbuf, ZX_BLACK, ZX_WHITE);

        FILE *fp = fopen(fbuf, "rb");
	if (!fp)
	{
		alert("Open failed", "Cancel", "Cannot open file:\n%s", name);
		return 0;
	}
	if (fread(header, 1, 128, fp) < 128)
	{
               	fclose(fp);
		alert("Open failed", "Cancel", "File is not in "
				"+3DOS format:\n%s", name);
                return 0;
	}
	for (n = sum = 0; n < 127; n++) sum += header[n];
	hlen  = header[16] + 256 * header[17];
	haddr = header[18] + 256 * header[19];
	if (header[n] != sum || memcmp(header, "PLUS3DOS\032", 9) ||
	    header[15] != 3 || hlen < minLen)
	{
                fclose(fp);
		alert("Open failed", "Cancel", "Unexpected +3DOS header in "
				"file:\n%s", name);
                return 0;
	} 
        if (hlen < maxLen) maxLen = hlen;
        if (fread(loadAt, 1, maxLen, fp) < (unsigned)minLen)
	{
		fclose(fp);
		alert("Read failed", "Cancel", "Unexpected end of "
				"file:\n%s", name);
		return 0;
	}
	fclose(fp);
	return 1;
}


static int write3Dos(const char *path, const char *name, const jswByte *header,
			const jswByte *data)
{
	char fbuf[PATH_MAX];
	//char mbuf[50];
	int datlen = header[11] + 256 * header[12] + 65536 * header[13];
	FILE *fp;
	
	datlen -= 128;

	strcpy(fbuf, path);

	char *s = strrchr(fbuf, '/'); 
	if (s) *s = 0;
#ifdef _WIN32
	else 
	{
		s = strrchr(fbuf, '\\');
		if (s) *s = 0;
		else { s = strrchr(fbuf, ':'); if (s) s[1] = 0; } 
	}
#endif
        // No path separator at all. So there's no path, just filenames.
	if (!s) strcpy(fbuf, "");
        else    strcat(fbuf, "/");
 
	strcat(fbuf, name);

//	videoScreen->drawSmallText(0, 0, fbuf, ZX_BLACK, ZX_WHITE);

/* Skip the overwriting check and assume the user knows what he/she is doing...
	fp = fopen(fbuf, "rb");
	if (fp)
	{
		fclose(fp);
		sprintf(mbuf, "Overwrite %s? ", name);
		VideoMenu vm(mbuf, "Yes", "Cancel save", NULL);
		int rv = vm.doModal();
		if (rv >= ZXE_CANCEL) return 0;
		if (vm.getSelected() != 0) return 0;
	} */
	fp = fopen(fbuf, "wb");
	if (!fp) return 0;

	if (fwrite(header, 1, 128, fp) < 128) return 0;
	if (fwrite(data,   1, datlen, fp) < (unsigned)datlen) return 0;
	fclose(fp);
	return 1;
}

static int write3Dos(const char *path, const char *name, int len, int addr,
                        const jswByte *data)
{
        jswByte header[128];

        constructHeader(header, 3, len, addr);

        return write3Dos(path, name, header, data);
}


static int tapWrite(FILE *fp, char *name, const jswByte *header, 
			const jswByte *data)
{
	jswByte tapHeader[25];
	jswByte csum;
	unsigned n, len;

	memset(tapHeader, 0, sizeof(tapHeader));	
	tapHeader[0] = 0x13;
	tapHeader[3] = header[15];
	memcpy(tapHeader + 4,  name, 10);
	memcpy(tapHeader + 14, header + 16, 6);
	
	csum = 0;
	for (n = 2; n < 20; n++) csum ^= tapHeader[n];
	tapHeader[n] = csum;
	if (fwrite(tapHeader, 1, 21, fp) < 21) return 0;
	
	len = header[16] + 256 * header[17];
	csum = 0xFF;
	for (n = 0; n < len; n++) csum ^= data[n];
	
	tapHeader[0] = ((len + 2)     ) & 0xFF;
	tapHeader[1] = ((len + 2) >> 8) & 0xFF;
	tapHeader[2] = 0xFF;
	if (fwrite(tapHeader, 1, 3, fp) < 3) return 0;
	if (fwrite(data,      1, len, fp) < len) return 0;
	if (fwrite(&csum,     1, 1, fp) < 1) return 0;
	return 1;
}

static void constructHeader(jswByte *header, int type, int len, int addr)
{
        jswByte csum;
        int n;

        memcpy(header, hiload_head, 128);
        header[11] = ((len + 128)      ) & 0xFF;
        header[12] = ((len + 128) >>  8) & 0xFF;
        header[13] = ((len + 128) >> 16) & 0xFF;
        header[14] = ((len + 128) >> 24) & 0xFF;
        header[15] = type;
        header[16] = len & 0xFF;
        header[17] = (len >> 8) & 0xFF;
	header[18] = addr & 0xFF;
	header[19] = (addr >> 8) & 0xFF;

	csum = 0;	
	for (n = 0; n < 127; n++) csum += header[n];
	header[n] = csum;
}

static int tapWrite(FILE *fp, char *name, int len, int addr, 
			const jswByte *data)
{
	jswByte header[128];
	
	constructHeader(header, 3, len, addr);

	return tapWrite(fp, name, header, data);
}

int SpectrumMemory::saveTap(const char *path, FILE *fp)
{
	char variant;
	char hiload7[20];
	char room6[20];
	char room4[20];
	char room3[20];
	char room1[20];
	char title7[20];
	char rtime2[20];

	if (is128())
	{
		jswByte hiload[7725];
		jswByte jsw128_ldr[sizeof(jsw_univ_loader) / sizeof(jsw_univ_loader[0])];
	
		switch(peek(0x85C9))
		{
			case 'V': case 'W': case 'X': case 'Y': case 'Z': 
			case '[':
				variant = tolower(peek(0x85C9));
				break;
			default:
				variant = 's';
				break;
		}
		createLoader(jsw128_ldr, variant);
		createHiload(hiload, this);

		sprintf(room6,   "rooms.j%c6 ", variant);
		sprintf(room4,   "rooms.j%c4 ", variant);
		sprintf(room3,   "rooms.j%c3 ", variant);
		sprintf(room1,   "rooms.j%c1 ", variant);
		sprintf(rtime2,  "rtime.j%c2 ", variant);
		sprintf(title7,  "title.j%c7 ", variant);
		sprintf(hiload7, "hiload.j%c7", variant);
		if (tapWrite(fp,"JetSet.128",jsw128_ldr, jsw128_ldr + 128)
		 && tapWrite(fp, hiload7, hiload, hiload + 128)
		 && tapWrite(fp, room6,  0x4000,0xC000,memoryAt(0xC000, 6))
                 && tapWrite(fp, room4,  0x4000,0xC000,memoryAt(0xC000, 4))
                 && tapWrite(fp, room3,  0x4000,0xC000,memoryAt(0xC000, 3))
                 && tapWrite(fp, room1,  0x4000,0xC000,memoryAt(0xC000, 1))
		 && tapWrite(fp, title7, 0x1B00,0xC000,memoryAt(0xC000, 7))
		 && tapWrite(fp, rtime2, 0x8000,0x8000,memoryAt(0x8000))
		 && tapWrite(fp,"main.tun  ",0x0100,0xF700,memoryAt(0xF700))
		 && tapWrite(fp,"cheat.tun ",0x00FE,0xF902,memoryAt(0xF902))
		 && tapWrite(fp,"tune.bin  ",0x0600,0xFA00,memoryAt(0xFA00 ))
		 && tapWrite(fp,"title.tun ",0x0AD6,0x752A,memoryAt(0xF52A,7)))
			return ZXE_OK;
		else return retError("Disc full");
	}
	else
	{
		if (tapWrite(fp, "JetSet    ",jsw48_ldr, jsw48_ldr+128)
		&&  tapWrite(fp, "jsw.js2   ",0x8000,0x8000,memoryAt(0x8000)))
			return ZXE_OK; 
		else return retError("Disc full");	
	}
	
        return ZXE_CANCEL;
}


int SpectrumMemory::saveSna(const char *path, FILE *fp)
{
	jswByte theSna[49179];

	memcpy(theSna, m_snahdr, 27);
	memcpy(theSna + 27, memoryAt(0x4000), 0xC000);

	if (is128())
	{
		if (fwrite(theSna, 1, 49179, fp) < 49179              ||
		    fwrite(theSna + 0x1C19, 1,2, fp) < 2              ||
		    fputc(0x10, fp) == EOF                            ||
		    fputc(0x00, fp) == EOF                            ||
		    fwrite(memoryAt(0xC000, 1),1,0x4000, fp) < 0x4000 ||
		    fwrite(memoryAt(0xC000, 3),1,0x4000, fp) < 0x4000 ||
                    fwrite(memoryAt(0xC000, 4),1,0x4000, fp) < 0x4000 ||
                    fwrite(memoryAt(0xC000, 6),1,0x4000, fp) < 0x4000 ||
                    fwrite(memoryAt(0xC000, 7),1,0x4000, fp) < 0x4000)
			return retError("Disc full");	
	}
	else if (fwrite(theSna, 1, 49179, fp) < 49179) 
		return retError("Disc full");
	return ZXE_OK;
}


void SpectrumMemory::writeMemoryAt(int address, int bank, jswByte *src, int len)
{
	jswByte *dest = memoryAt(address, bank);
	memcpy(dest, src, len);
}

jswByte *SpectrumMemory::memoryAt(int address, int bank)
{
	switch(bank)
	{
		case 1: return m_ram1 + (address & 0x3FFF);
		case 2: return m_ram48 + 0x4000 + (address & 0x3FFF);
		case 3: return m_ram3 + (address & 0x3FFF);
		case 4: return m_ram4 + (address & 0x3FFF);
		case 5: return m_ram48 + (address & 0x3FFF);
		case 6: return m_ram6 + (address & 0x3FFF);
		case 7: return m_ram7 + (address & 0x3FFF);
	}
	if (address < 0x4000) return m_ram48;
	return m_ram48 + (address - 0x4000);
}



jswByte SpectrumMemory::peek(int address, int bank)
{
	if (address < 0x4000 || address > 0xFFFF) return 0;
	return *memoryAt(address, bank);
}


void SpectrumMemory::poke(int address, int value, int bank)
{
	jswByte *pmem;

	if (address < 0x4000 || address > 0xFFFF) return;
	pmem = memoryAt(address,bank);
	*pmem = (value & 0xFF);
}


jswByte SpectrumMemory::peekSnaHeader(int address)
{
	if (address >= 0 && address < sizeof(m_snahdr)) return m_snahdr[address];
	return 0xFF;
}


void SpectrumMemory::pokeSnaHeader(int address, int value)
{
	if (address >= 0 && address < sizeof(m_snahdr)) m_snahdr[address] = value;
}


int SpectrumMemory::loadSZX(const char *path, FILE *fp)
{
	return alert("SZX support not yet present", "Cancel");
}


int SpectrumMemory::saveSZX(const char *path, FILE *fp)
{
	return alert("SZX support not yet present", "Cancel");
}

