/************************************************************************

    JSWED 2.3.0 - Editor for Jet Set Willy and derivatives

    Copyright (C) 2005  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*************************************************************************/

#define CHANNELS 3

#define TUNE_W   249
#define TUNE_X   250
#define TUNE_V   251
#define TUNEDJNZ 252
#define TUNERET  253
#define TUNECALL 254
#define TUNESTOP 255

class NoteSequence;

struct Note128
{
	Note128();
	Note128(unsigned t, unsigned p1 = 0);
	Note128(NoteSequence &s);
	~Note128();

	void clear();
	inline bool isNote() const { return (m_type < TUNE_W); }

	unsigned m_type;
	unsigned m_duration;
	unsigned m_effect;
	unsigned m_repeats;
	NoteSequence *m_body;
	unsigned m_offset;

	bool operator == (Note128 &other);
	bool operator != (Note128 &other);
public:
	void copy(Note128 *other);
	void dump(FILE *fp);
	unsigned memLength(void);
	unsigned store(SpectrumMemory *mem, unsigned base, unsigned addr);
};


class NoteSequence
{
	Note128 *m_data;
	int m_count;
	int m_max;

	NoteSequence(Note128 *d, int c);

public:
	NoteSequence();
	NoteSequence(int len);
	~NoteSequence();

	void assertValid(void);
	inline void clear() { m_count = 0; }
	inline int length() const { return m_count; }
	inline const Note128 *getData() { return m_data; }

	void append(Note128 datum);
	void append(const NoteSequence &other);
	bool equals(NoteSequence *other);
	bool operator == (NoteSequence &other);
public:
	void subset(NoteSequence *s, int first, int count = -1);
	unsigned memLength();
	unsigned store(SpectrumMemory *mem, unsigned base, unsigned addr);

	int getHcf(int soFar);

	void dump(FILE *fp);
};




class Tune128
{
private:
	NoteSequence m_data[CHANNELS + 1];
	void load(int channel, SpectrumMemory *mem, unsigned base, int term, unsigned top);
	int store(int channel, SpectrumMemory *mem, unsigned base, unsigned addr);
	void compress(NoteSequence &src, NoteSequence &dest);

public:
	Tune128();
	~Tune128();

	void getChannel(NoteSequence *s, int ch, int starttime = 0);

	void append(int channel, Note128 n);

	void load(SpectrumMemory *mem, int addr);
	int  store(SpectrumMemory *mem, int addr, int len);
	int	getTimeBase();

	Tune128 & operator =(Tune128 &other);
};

